/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989,1988 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:$
 *
 * 23-Nov-91  Bob Wheeler (bobw) at Carnegie-Mellon University
 *	Created.
 */
/*
 *	File: app.c
 * 	Author: Bob Wheeler, Carnegie-Mellon University
 *	Date:	11/91
 *
 *	Assembler pre-processor for pa-risc assembly files
 */

#include <stdio.h>
#include <errno.h>
#include <strings.h>
#include <sys/param.h>

#define BUFSIZE 4096

main(argc, argv)
	int argc;
	char **argv;
{
	FILE *infile = (FILE *)0;
	FILE *outfile = (FILE *)0;
	char buffer[BUFSIZ];
	int lineno = 0;
	int templine = 0;
	char tempname[MAXPATHLEN+1];
	char filename[MAXPATHLEN+1];
	char infilename[MAXPATHLEN+1];

	argv++;
	argc--;

	while(argc > 0) {
		if (!strcmp(*argv,"-")) {
			if (!infile)
				infile = stdin;
			else if (!outfile)
				outfile = stdout;
			else
				usage();
		} else if (!infile)
			if (!(infile = fopen(*argv, "r"))) {
				perror("can't open input file");
				exit(1);
			} else
				(void) strcpy(infilename, *argv);
		else if (!outfile) 
			if (!(outfile = fopen(*argv, "w"))) {
				perror("can't open output file");
				exit(1);
			}
		argv++;
		argc--;
	}

	if (!infile)
		infile = stdin;

	if (!outfile)
		outfile = stdout;

	(void) sprintf(filename, "\"%s\"", 
		       infile == stdin ? "stdin" : infilename);

	while(fgets(buffer, BUFSIZE, infile)) {
		char *b;
		int comment_seen = 0;
		lineno++;

		if (buffer[0] == '#' && 
		    sscanf(buffer, "# %d %s", &templine, tempname) == 2) {
			lineno = templine - 1;
			(void) strcpy(filename, tempname);
			(void) fputs(buffer, outfile);
			continue;
		}

		for (b = buffer; *b; b++) {
			if (*b == '!' && !comment_seen) {
				(void) fprintf(outfile,
					       "\n# %d %s\n",lineno,filename);
				continue;
			}

			if (*b == ';')
				comment_seen++;

			(void) fputc(*b, outfile);
		}
	}

	(void) fclose(infile);
	(void) fclose(outfile);
}

usage()
{
	(void) fprintf(stderr, "usage: app [infile] [outfile]\n");
}


			
