/* Definitions of target machine for GNU compiler, for the HP Spectrum.
   Copyright (C) 1988 Free Software Foundation, Inc.
   Contributed by Michael Tiemann (tiemann@mcc.com).

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* RCS $Header: tm-hp9k8.h,v 1.6 91/09/24 16:20:40 bob Exp $ */

/* Note that some other tm- files include this one and then override
   many of the definitions that relate to assembler syntax.  */

/* This is a complete and utter lie, but I'd like to be able to
   look at the output of the C compiler notes stuff in the
   assembler.  */
#define DBX_DEBUGGING_INFO

/* Names to predefine in the preprocessor for this target machine.  */

/* Brain-dead loader */

#define LINK_SPEC "-u main"

#if 0
#define CPP_PREDEFINES "-Dhp9000s800 -D__hp9000s800 -Dhp9k8 -DPWB -Dhpux -Dunix -D_HPUX_SOURCE"
#else
#define CPP_PREDEFINES "-Dhp9000s800 -D__hp9000s800 -Dhp9k8 -D_HPUX_SOURCE -Dhp9000 -Dhp800 -Dspectrum -Dpa_risc -DPA_RISC -DHP800 -DREVARGV -DMACH -DBYTE_MSF -DBIT_MSF -DCMUCS -DCMU"
#endif

/* Print subsidiary information on the compiler version in use.  */

#define TARGET_VERSION fprintf (stderr, " (hp9000s800)");

/* Run-time compilation parameters selecting different hardware subsets.

   On the the hp9k800, we don't yet need any.  */

extern int target_flags;

/* Macro to define tables used to set the flags.
   This is a list in braces of pairs in braces,
   each pair being { "NAME", VALUE }
   where VALUE is the bits to set or minus the bits to clear.
   An empty string NAME is used to identify the default VALUE.  */

#define TARGET_SWITCHES  \
  {{ "", TARGET_DEFAULT}}

#define TARGET_DEFAULT 1

/* target machine storage layout */

/* Define this if most significant bit is lowest numbered
   in instructions that operate on numbered bit-fields.  */
#define BITS_BIG_ENDIAN

/* Define this if most significant byte of a word is the lowest numbered.  */
/* That is true on the hp9k8.  */
#define BYTES_BIG_ENDIAN

/* Define this if most significant word of a multiword number is numbered.  */
/* For the hp9k800 we can decide arbitrarily
   since there are no machine instructions for them.  */
#define WORDS_BIG_ENDIAN

/* number of bits in an addressible storage unit */
#define BITS_PER_UNIT 8

/* Width in bits of a "word", which is the contents of a machine register.
   Note that this is not necessarily the width of data type `int';
   if using 16-bit ints on a 68000, this would still be 32.
   But on a machine with 16-bit registers, this would be 16.  */
#define BITS_PER_WORD 32

/* Width of a word, in units (bytes).  */
#define UNITS_PER_WORD 4

/* Width in bits of a pointer.
   See also the macro `Pmode' defined below.  */
#define POINTER_SIZE 32

/* Allocation boundary (in *bits*) for storing pointers in memory.  */
#define POINTER_BOUNDARY 32

/* Allocation boundary (in *bits*) for storing arguments in argument list.  */
#define PARM_BOUNDARY 32

/* Largest alignment required for any stack parameter, in bits.
   Don't define this if it is equal to PARM_BOUNDRY */
#define MAX_PARM_BOUNDARY 64

/* Boundary (in *bits*) on which stack pointer should be aligned.  */
#define STACK_BOUNDARY 64

/* Allocation boundary (in *bits*) for the code of a function.  */
#define FUNCTION_BOUNDARY 32

/* Alignment of field after `int : 0' in a structure.  */
#define EMPTY_FIELD_BOUNDARY 32

/* No data type wants to be aligned rounder than this.  */
#define BIGGEST_ALIGNMENT 64

/* Get around hp-ux assembler bug. */
#define CONSTANT_ALIGNMENT(CODE, TYPEALIGN) \
  ((TYPEALIGN) < 32 ? 32 : (TYPEALIGN))

/* Define this if move instructions will actually fail to work
   when given unaligned data.  */
#define STRICT_ALIGNMENT

/* Every structure or union's size must be a multiple of 2 bytes.  */

#define STRUCTURE_SIZE_BOUNDARY 16

/* Generate calls to memcpy, memcmp and memset.  */
#define TARGET_MEM_FUNCTIONS

/* Standard register usage.  */

/* Number of actual hardware registers.
   The hardware registers are assigned numbers for the compiler
   from 0 to just below FIRST_PSEUDO_REGISTER.
   All registers that the compiler knows about must be given numbers,
   even those that are not normally considered general registers.

   the hp9k800 has 32 fullword registers and 16 floating point registers.
   16 additional registers are reserved.  */

#define FIRST_PSEUDO_REGISTER 64

/* 1 for registers that have pervasive standard uses
   and are not available for the register allocator.

   On the hp9k800, these are:
   Reg 0	= 0 (hardware).
   Reg 1	= ADDIL target/Temporary (hardware).
   Reg 2	= Return Pointer
   Reg 3	= Unused
   Reg 4	= Frame Pointer (Gnu)
   Reg 5-18	= Preserved Registers
   Reg 19-22	= Temporary Registers
   Reg 23-26	= Temporary/Parameter Registers
   Reg 27	= Global Data Pointer (hp)
   Reg 28	= Temporary/???/Return Value register
   Reg 29	= Temporary/Static Chain/Return Value register
   Reg 30	= stack pointer
   Reg 31	= Temporary/Millicode Return Pointer (hp)

   Freg 0-3	= Status Registers
   Freg 4-7	= Arguments/Return Value
   Freg 8-11	= Temporary Registers
   Freg 12-15	= Preserved Registers

   Freg 16-31	= Reserved

*/

#define FIXED_REGISTERS  \
 {1, 1, 1, 1, 1, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 1, 0, 0, 1, 1, \
  1, 1, 1, 1, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0}

/* 1 for registers not available across function calls.
   These must include the FIXED_REGISTERS and also any
   registers that can be used without being saved.
   The latter must include the registers where values are returned
   and the register where structure-value addresses are passed.
   Aside from that, you can include as many other registers as you like.  */
#define CALL_USED_REGISTERS  \
 {1, 1, 1, 1, 1, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 1, 1, 1, 1, 1, \
  1, 1, 1, 1, 1, 1, 1, 1, \
  1, 1, 1, 1, 1, 1, 1, 1, \
  1, 1, 1, 1, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 0, 0, 0}

/* Allocated the call used registers first.  This should minimize
   the number of registers that need to be saved (as call used
   registers will generally not be allocated across a call).

   It is possible that it would be wise to allocate the floating point
   registers before the regular ones, but I doubt it matters.  Same
   comment for parameters versus normal.  */

#define REG_ALLOC_ORDER			\
 {19, 20, 21, 22, 23, 24, 25, 26,	\
  27, 28, 29, 30, 31, 40, 41, 42,	\
  43, 36, 37, 38, 39,  5,  6,  7,	\
   8,  9, 10, 11, 12, 13, 14, 15,	\
  16, 17, 18, 44, 45, 46, 47,  1,	\
   2,  3,  4, 32, 33, 34, 35,  0}


/* Return number of consecutive hard regs needed starting at reg REGNO
   to hold something of mode MODE.
   This is ordinarily the length in words of a value of mode MODE
   but can be less for certain modes in special long registers.

   On the hp9k800, ordinary registers hold 32 bits worth;
   The floating point registers are 64 bits wide.  */
#define HARD_REGNO_NREGS(REGNO, MODE)   \
  ((REGNO)<32 ? ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD) : 1)

/* Value is 1 if hard register REGNO can hold a value of machine-mode MODE.
   On the hp9k800, the cpu registers can hold any mode.  We
   force this to be an even register is it cannot hold the full mode.  */
#define HARD_REGNO_MODE_OK(REGNO, MODE) \
  ((REGNO) < 32 ? ((GET_MODE_SIZE (MODE) <= 4) ? 1 : ((REGNO) & 1) == 0) : 1)

/* Value is 1 if it is a good idea to tie two pseudo registers
   when one has mode MODE1 and one has mode MODE2.
   If HARD_REGNO_MODE_OK could produce different values for MODE1 and MODE2,
   for any hard reg, then this must be 0 for correct output.  */
#define MODES_TIEABLE_P(MODE1, MODE2) 1

/* Specify the registers used for certain standard purposes.
   The values of these macros are register numbers.  */

/* the hp9k800 pc isn't overloaded on a register that the compiler knows about.  */
/* #define PC_REGNUM  */

/* Register to use for pushing function arguments.  */
#define STACK_POINTER_REGNUM 30

/* Base register for access to local variables of the function.  */
#define FRAME_POINTER_REGNUM 4

/* Value should be nonzero if functions must have frame pointers.
   Zero means the frame pointer need not be set up (and parms
   may be accessed via the stack pointer) in functions that seem suitable.
   This is computed in `reload', in reload1.c.  */
#define FRAME_POINTER_REQUIRED 0

/* Base register for access to arguments of the function.  */
#define ARG_POINTER_REGNUM 4

/* Register in which static-chain is passed to a function.  */
/* ??? */
#define STATIC_CHAIN_REGNUM 29

/* Register in which address to store a structure value
   is passed to a function.  */
#define STRUCT_VALUE_REGNUM 28

/* Define the classes of registers for register constraints in the
   machine description.  Also define ranges of constants.

   One of the classes must always be named ALL_REGS and include all hard regs.
   If there is more than one class, another class must be named NO_REGS
   and contain no registers.

   The name GENERAL_REGS must be the name of a class (or an alias for
   another name such as ALL_REGS).  This is the class of registers
   that is allowed by "g" or "r" in a register constraint.
   Also, registers outside this class are allocated only when
   instructions express preferences for them.

   The classes must be numbered in nondecreasing order; that is,
   a larger-numbered class must never be contained completely
   in a smaller-numbered class.

   For any two classes, it is very desirable that there be another
   class that represents their union.  */

  /* The hp9k800 has two kinds of registers, hence four classes.  */

enum reg_class { NO_REGS, GENERAL_REGS, FP_REGS, ALL_REGS, LIM_REG_CLASSES };

#define N_REG_CLASSES (int) LIM_REG_CLASSES

/* Give names of register classes as strings for dump file.   */

#define REG_CLASS_NAMES {"NO_REGS", "GENERAL_REGS", "FP_REGS", "ALL_REGS" }

/* Define which registers fit in which classes.
   This is an initializer for a vector of HARD_REG_SET
   of length N_REG_CLASSES.  */

#define REG_CLASS_CONTENTS {{0,0}, {-1,0}, {0,-1}, {-1,-1}}

/* The same information, inverted:
   Return the class number of the smallest class containing
   reg number REGNO.  This could be a conditional expression
   or could index an array.  */

#define REGNO_REG_CLASS(REGNO) ((REGNO)<32 ? GENERAL_REGS : FP_REGS)

/* The class value for index registers, and the one for base regs.  */
#define INDEX_REG_CLASS GENERAL_REGS
#define BASE_REG_CLASS GENERAL_REGS

/* Get reg_class from a letter such as appears in the machine description.  */

#define REG_CLASS_FROM_LETTER(C) \
  ((C)=='r' ? GENERAL_REGS : ((C)=='f' ? FP_REGS : NO_REGS))

/* The letters I, J, K, L and M in a register constraint string
   can be used to stand for particular ranges of immediate operands.
   This macro defines what the ranges are.
   C is the letter, and VALUE is a constant value.
   Return 1 if VALUE is in the range specified by C.

HP9000/800 immediate field sizes:
  5 bits: scalar/floating short loads + stores; deposit; conditional branch
  11 bits: arithmetic immediate, compare immediate
  14 bits: loads and stores; load offset
  21 bits: load and add immediate long
  (there are also 13-bit and 26-bit immediates but only in system instructions)

   `I' is used for the 11 bit constants.
   `J' is used for the 14 bit constants.
   `K' is used for the 21 bit constants.
   `L' is used for the 5 bit constants.
   `M' is used for 0.  */

#define CONST_OK_FOR_LETTER_P(VALUE, C)  \
  ((C) == 'I' ? (unsigned) ((VALUE) + 0x400) < 0x800		\
   : (C) == 'J' ? (unsigned) ((VALUE) + 0x2000) < 0x4000	\
   : (C) == 'K' ? (VALUE) & 0x3fff == 0				\
   : (C) == 'L' ? (unsigned) ((VALUE) + 0x10) < 0x20		\
   : (C) == 'M' ? (VALUE) == 0					\
   : 0)

/* Similar, but for floating constants, and defining letters G and H.
   Here VALUE is the CONST_DOUBLE rtx itself.  */

#define CONST_DOUBLE_OK_FOR_LETTER_P(VALUE, C)  \
  ((C) == 'G' && XINT (VALUE, 0) == 0 && XINT (VALUE, 1) == 0)

/* Given an rtx X being reloaded into a reg required to be
   in class CLASS, return the class of reg to actually use.
   In general this is just CLASS; but on some machines
   in some cases it is preferable to use a more restrictive class.  */
#define PREFERRED_RELOAD_CLASS(X,CLASS) (CLASS)

/* Return the maximum number of consecutive registers
   needed to represent mode MODE in a register of class CLASS.  */
#define CLASS_MAX_NREGS(CLASS, MODE)	\
  ((CLASS) == FP_REGS ? 1 : ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD))

/* Stack layout; function entry, exit and calling.  */

/* Define this if pushing a word on the stack
   makes the stack pointer a smaller address.  */
/* #define STACK_GROWS_DOWNWARD */

/* Believe it or not.  */
#define ARGS_GROW_DOWNWARD

/* Define this if the nominal address of the stack frame
   is at the high-address end of the local variables;
   that is, each additional local variable allocated
   goes at a more negative offset in the frame.  */
/* #define FRAME_GROWS_DOWNWARD */

/* Offset within stack frame to start allocating local variables at.
   If FRAME_GROWS_DOWNWARD, this is the offset to the END of the
   first local allocated.  Otherwise, it is the offset to the BEGINNING
   of the first local allocated.  */
#define STARTING_FRAME_OFFSET 8

/* If we generate an insn to push BYTES bytes,
   this says how many the stack pointer really advances by.
   On the hp9k800, don't define this because there are no push insns.  */
/*  #define PUSH_ROUNDING(BYTES) */

/* If BYTES is the size of arguments for a function call,
   return the size of the argument block (which is BYTES suitably rounded).
   Define this only on machines where the entire call block is allocated
   before the args are stored into it.  */

#define ROUND_CALL_BLOCK_SIZE(BYTES)  \
   (((BYTES) + 7) & ~7)

/* Offset of first parameter from the argument pointer register value.
   This value will be negated because the arguments grow down.
   Also note that on STACK_GROWS_UPWARD machines (such as this one)
   this is the distance from the frame pointer to the end of the first
   argument, not it's beginning.  To get the real offset of the first
   argument, the size of the argument must be added. */

/* #define FIRST_PARM_OFFSET(FNDECL) 36  */
#define FIRST_PARM_OFFSET(FNDECL) 32 

/* Absolute value of offset from top-of-stack address to location to store the
   function parameter if it can't go in a register.
   Addresses for following parameters are computed relative to this one.  */
/* #define FIRST_PARM_CALLER_OFFSET(FNDECL) 36 */
#define FIRST_PARM_CALLER_OFFSET(FNDECL) 32 


/* When a parameter is passed in a register, stack space is still
   allocated for it.  */
#define REG_PARM_STACK_SPACE

/* The weird HPPA calling conventions require a minimum of 48 bytes on 
   the stack */
#define STACK_ARGS_ADJUST(SIZE)						\
{									\
  if (SIZE.var)								\
    abort(); /*don't know how to do this yet */				\
  else if (SIZE.constant < 48)						\
    SIZE.constant = 48;							\
}
  
/* Value is 1 if returning from a function call automatically
   pops the arguments described by the number-of-args field in the call.
   FUNTYPE is the data type of the function (as a tree),
   or for a library call it is an identifier node for the subroutine name.  */

#define RETURN_POPS_ARGS(FUNTYPE) 0

/* Define how to find the value returned by a function.
   VALTYPE is the data type of the value (as a tree).
   If the precise function being called is known, FUNC is its FUNCTION_DECL;
   otherwise, FUNC is 0.  */

/* On the hp9k800 the value is found in register(s) 28(-29), unless
   the mode is SF or DF. Then the value is returned in fr4 (36) */


#define FUNCTION_VALUE(VALTYPE, FUNC)  \
  gen_rtx (REG, TYPE_MODE (VALTYPE), (TYPE_MODE (VALTYPE) == SFmode ||\
				      TYPE_MODE (VALTYPE) == DFmode\
				      ? 36 : 28))

#define FUNCTION_OUTGOING_VALUE(VALTYPE, FUNC)  \
  FUNCTION_VALUE(VALTYPE, FUNC)

/* Define how to find the value returned by a library function
   assuming the value has mode MODE.  */

#define LIBCALL_VALUE(MODE) \
  gen_rtx (REG, MODE, MODE == SFmode || MODE == DFmode ? 36 : 28)

/* 1 if N is a possible register number for a function value
   as seen by the caller.  */

#define FUNCTION_VALUE_REGNO_P(N) ((N) == 28 || (N) == 36)

/* 1 if N is a possible register number for function argument passing.  */

#define FUNCTION_ARG_REGNO_P(N) (((N) >= 23 && (N) <= 26) || \
				 ((N) >= 36 || (N) <= 39))

/* Define a data type for recording info about an argument list
   during the scan of that argument list.  This data type should
   hold all necessary information about the function itself
   and about the args processed so far, enough to enable macros
   such as FUNCTION_ARG to determine where the next arg should go.

   On the hp9k800, this is a single integer, which is a number of words
   of arguments scanned so far (including the invisible argument,
   if any, which holds the structure-value-address).
   Thus 4 or more means all following args should go on the stack.  */

#define CUMULATIVE_ARGS int

/* Initialize a variable CUM of type CUMULATIVE_ARGS
   for a call to a function whose data type is FNTYPE.
   For a library call, FNTYPE is 0.
*/

/* A dreadful hack to allow output_call_preamble to put the correct 
   description in the assembly code output. hp_arg_modes is an array
   of ints to avoid warnings caused by include file order. It has to
   be organized as a stack to handle nested calls. Unfortunately,
   these macros get called to set up the parameters to a function as
   well as to pass parameters. Thus, every call to
   INIT_CUMULATIVE_ARGS is not matched by a call to output_call_preamble ().
   We try clean up in output_call_preamble, but that never gets called
   if a function doesn't call anything .... so, in FUNCTION_EPILOGUE,
   reset hp_call_level to -1 (All this doesn't matter too much, just
   to prevent memory leaks).
 */

void init_hp_arg_modes ();
extern int (*hp_arg_modes)[4];
extern int hp_call_level;

#define INIT_CUMULATIVE_ARGS(CUM,FNTYPE) (init_hp_arg_modes (),(CUM) = 0)

/* Figure out the size in words of the function argument. */

#define FUNCTION_ARG_SIZE(MODE, TYPE)	\
  ((((MODE) != BLKmode ? GET_MODE_SIZE (MODE) : int_size_in_bytes (TYPE))+3)/4)

/* Arguments which must have double-word alignment may need to have some
   padding on the stack.  FUNCTION_ARG_ADJUST bumps the register pointer
   the appropriate amount, and returns the number of bytes of extra padding
   needed.  It is used in parallel with FUNCTION_ARG_ADVANCE.

   This is nice but it never gets used anywhere.

#define FUNCTION_ARG_ADJUST(CUM, MODE, TYPE, NAMED)			\
  ((((CUM) & 01) && (TYPE) != 0 && TYPE_ALIGN (TYPE) > BITS_PER_WORD)	\
   ? ((CUM)++,4) : 0)
*/

/* Update the data in CUM to advance over an argument
   of mode MODE and data type TYPE.
   (TYPE is null for libcalls where that information may not be available.)  */

#define FUNCTION_ARG_ADVANCE(CUM, MODE, TYPE, NAMED)			\
    (((((CUM) & 01) && (TYPE) != 0 && TYPE_ALIGN (TYPE) > BITS_PER_WORD)\
      && (CUM)++), ((CUM) < 4 ?						\
		    hp_arg_modes[hp_call_level][(CUM)] = (int)(MODE) \
		    : 0), \
     (CUM) += FUNCTION_ARG_SIZE(MODE, TYPE))

/* Determine where to put an argument to a function.
   Value is zero to push the argument on the stack,
   or a hard register in which to store the argument.

   MODE is the argument's machine mode.
   TYPE is the data type of the argument (as a tree).
    This is null for libcalls where that information may
    not be available.
   CUM is a variable of type CUMULATIVE_ARGS which gives info about
    the preceding args and about the function being called.
   NAMED is nonzero if this argument is a named parameter
    (otherwise it is an extra parameter matching an ellipsis).  */

/* On the hp9k800 the first four words of args are normally in registers
   and the rest are pushed.  But any arg that won't entirely fit in regs
   is pushed.  */

#define FUNCTION_ARG(CUM, MODE, TYPE, NAMED)		      		\
  (4 >= ((CUM) + FUNCTION_ARG_SIZE ((MODE), (TYPE)))			\
   ? gen_rtx (REG, (MODE), (MODE) == SFmode ? (36 + (CUM)) :    	\
	      ((MODE) == DFmode ? ((CUM) ? 39 : 37) : 			\
	       (27 - (CUM) - FUNCTION_ARG_SIZE ((MODE), (TYPE))))) 	\
   : 0)

/* Define where a function finds its arguments.
   This would be different from FUNCTION_ARG if we had register windows.  */

#define FUNCTION_INCOMING_ARG(CUM, MODE, TYPE, NAMED)	\
  FUNCTION_ARG (CUM, MODE, TYPE, NAMED)

/* For an arg passed partly in registers and partly in memory,
   this is the number of registers used.
   For args passed entirely in registers or entirely in memory, zero.  */

#define FUNCTION_ARG_PARTIAL_NREGS(CUM, MODE, TYPE, NAMED) 0

/* This macro generates the assembly code for function entry.
   FILE is a stdio stream to output the code to.
   SIZE is an int: how many units of temporary storage to allocate.
   Refer to the array `regs_ever_live' to determine which registers
   to save; `regs_ever_live[I]' is nonzero if register number I
   is ever used in the function.  This macro is responsible for
   knowing which registers should not be saved even if used.  */

#define FUNCTION_PROLOGUE(FILE, SIZE) function_prologue (FILE, SIZE)

/* Output assembler code to FILE to increment profiler label # LABELNO
   for profiling a function entry.  */

#define FUNCTION_PROFILER(FILE, LABELNO)  \
   abort ();

/* EXIT_IGNORE_STACK should be nonzero if, when returning from a function,
   the stack pointer does not matter.  The value is tested only in
   functions that have frame pointers.
   No definition is equivalent to always zero.  */

extern int may_call_alloca;
extern int current_function_pretend_args_size;

#define EXIT_IGNORE_STACK	\
 (get_frame_size () != 0	\
  || may_call_alloca || current_function_pretend_args_size)

/* This macro generates the assembly code for function exit,
   on machines that need it.  If FUNCTION_EPILOGUE is not defined
   then individual return instructions are generated for each
   return statement.  Args are same as for FUNCTION_PROLOGUE.

   The function epilogue should not depend on the current stack pointer!
   It should use the frame pointer only.  This is mandatory because
   of alloca; we also take advantage of it to omit stack adjustments
   before returning.  */

#define FUNCTION_EPILOGUE(FILE, SIZE) \
{hp_call_level = -1;  function_epilogue (FILE, SIZE);}

/* If the memory address ADDR is relative to the frame pointer,
   correct it to be relative to the stack pointer instead.
   This is for when we don't use a frame pointer.
   ADDR should be a variable name.  */

#define FIX_FRAME_POINTER_ADDRESS(ADDR,DEPTH)  \
{ int offset = -1;							\
  rtx regs = stack_pointer_rtx;						\
  if (ADDR == frame_pointer_rtx)					\
    offset = 0;								\
  else if (GET_CODE (ADDR) == PLUS && XEXP (ADDR, 0) == frame_pointer_rtx \
	   && GET_CODE (XEXP (ADDR, 1)) == CONST_INT)			\
    offset = INTVAL (XEXP (ADDR, 1));					\
  else if (GET_CODE (ADDR) == PLUS && XEXP (ADDR, 0) == frame_pointer_rtx) \
    { rtx other_reg = XEXP (ADDR, 1);					\
      offset = 0;							\
      regs = gen_rtx (PLUS, Pmode, stack_pointer_rtx, other_reg); }	\
  else if (GET_CODE (ADDR) == PLUS && XEXP (ADDR, 1) == frame_pointer_rtx) \
    { rtx other_reg = XEXP (ADDR, 0);					\
      offset = 0;							\
      regs = gen_rtx (PLUS, Pmode, stack_pointer_rtx, other_reg); }	\
  if (offset >= 0)							\
    { int regno;							\
      extern char call_used_regs[];					\
      for (regno = 2; regno < FRAME_POINTER_REGNUM; regno++)		\
	if (regs_ever_live[regno] && ! call_used_regs[regno])		\
	  offset += 4;							\
      offset -= 4;							\
      ADDR = plus_constant (regs, offset + (DEPTH)); } }


/* Addressing modes, and classification of registers for them.  */

#define HAVE_POST_INCREMENT
#define HAVE_POST_DECREMENT

#define HAVE_PRE_DECREMENT
#define HAVE_PRE_INCREMENT

/* Macros to check register numbers against specific register classes.  */

/* These assume that REGNO is a hard or pseudo reg number.
   They give nonzero only if REGNO is a hard reg of the suitable class
   or a pseudo reg currently allocated to a suitable hard reg.
   Since they use reg_renumber, they are safe only once reg_renumber
   has been allocated, which happens in local-alloc.c.  */

#define REGNO_OK_FOR_INDEX_P(REGNO) \
  ((REGNO) < 32 || (unsigned) reg_renumber[REGNO] < 32)
#define REGNO_OK_FOR_BASE_P(REGNO)  \
  ((REGNO) < 32 || (unsigned) reg_renumber[REGNO] < 32)
#define REGNO_OK_FOR_FP_P(REGNO) \
(((REGNO) ^ 0x20) < 32 || (unsigned) (reg_renumber[REGNO] ^ 0x20) < 32)

/* Now macros that check whether X is a register and also,
   strictly, whether it is in a specified class.

   These macros are specific to the the hp9k800, and may be used only
   in code for printing assembler insns and in conditions for
   define_optimization.  */

/* 1 if X is an fp register.  */

#define FP_REG_P(X) (REG_P (X) && REGNO_OK_FOR_FP_P (REGNO (X)))

/* Maximum number of registers that can appear in a valid memory address.  */

#define MAX_REGS_PER_ADDRESS 2

/* Recognize any constant value that is a valid address.  */

#define CONSTANT_ADDRESS_P(X)  CONSTANT_P (X)

/* Nonzero if the constant value X is a legitimate general operand.
   It is given that X satisfies CONSTANT_P or is a CONST_DOUBLE.  */

/*#define LEGITIMATE_CONSTANT_P(X) (1)*/
#define LEGITIMATE_CONSTANT_P(X)		\
 (GET_CODE (X) != CONST_DOUBLE)

/* The macros REG_OK_FOR..._P assume that the arg is a REG rtx
   and check its validity for a certain class.
   We have two alternate definitions for each of them.
   The usual definition accepts all pseudo regs; the other rejects
   them unless they have been allocated suitable hard regs.
   The symbol REG_OK_STRICT causes the latter definition to be used.

   Most source files want to accept pseudo regs in the hope that
   they will get allocated to the class that the insn wants them to be in.
   Source files for reload pass need to be strict.
   After reload, it makes no difference, since pseudo regs have
   been eliminated by then.  */

#ifndef REG_OK_STRICT

/* Nonzero if X is a hard reg that can be used as an index
   or if it is a pseudo reg.  */
#define REG_OK_FOR_INDEX_P(X) ((unsigned) REGNO (X) - 32 >= 32)
/* Nonzero if X is a hard reg that can be used as a base reg
   or if it is a pseudo reg.  */
#define REG_OK_FOR_BASE_P(X) ((unsigned) REGNO (X) - 32 >= 32)

#else

/* Nonzero if X is a hard reg that can be used as an index.  */
#define REG_OK_FOR_INDEX_P(X) REGNO_OK_FOR_INDEX_P (REGNO (X))
/* Nonzero if X is a hard reg that can be used as a base reg.  */
#define REG_OK_FOR_BASE_P(X) REGNO_OK_FOR_BASE_P (REGNO (X))

#endif

/* GO_IF_LEGITIMATE_ADDRESS recognizes an RTL expression
   that is a valid memory address for an instruction.
   The MODE argument is the machine mode for the MEM expression
   that wants to use this address.

   On the hp9k800, the actual legitimate addresses must be
   REG+REG, REG+(REG*SCALE) or REG+SMALLINT.
   But we can treat a SYMBOL_REF as legitimate if it is part of this
   function's constant-pool, because such addresses can actually
   be output as REG+SMALLINT.  */

#define VAL_5_BITS_P(X) ((unsigned)(X) + 0x10 < 0x20)
#define INT_5_BITS(X) VAL_5_BITS_P (INTVAL (X))

#define VAL_11_BITS_P(X) ((unsigned)(X) + 0x400 < 0x800)
#define INT_11_BITS(X) VAL_11_BITS_P (INTVAL (X))

#define VAL_14_BITS_P(X) ((unsigned)(X) + 0x2000 < 0x4000)
#define INT_14_BITS(X) VAL_14_BITS_P (INTVAL (X))

#define FITS_14_BITS(X)	\
   (GET_CODE (X) == CONST_INT && INT_14_BITS (X))

#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR)  \
{							\
  if (REG_P (X) && REG_OK_FOR_BASE_P (X)		\
      || ((GET_CODE (X) == PRE_DEC || GET_CODE (X) == POST_DEC		\
	   || GET_CODE (X) == PRE_INC || GET_CODE (X) == POST_INC)	\
	   && REG_P (XEXP (X, 0))			\
	   && REG_OK_FOR_BASE_P (XEXP (X, 0))))		\
    goto ADDR;						\
  else if (GET_CODE (X) == PLUS)			\
    {							\
      rtx base = 0, index;				\
      if (REG_P (XEXP (X, 0))				\
	  && REG_OK_FOR_BASE_P (XEXP (X, 0)))		\
	base = XEXP (X, 0), index = XEXP (X, 1);	\
      else if (REG_P (XEXP (X, 1))			\
	       && REG_OK_FOR_BASE_P (XEXP (X, 1)))	\
	base = XEXP (X, 1), index = XEXP (X, 0);	\
      if (base != 0)					\
	if (GET_CODE (index) == CONST_INT		\
	    && INT_14_BITS (index))			\
	  goto ADDR;					\
	/* Next two disabled because of asymmetrical architecture.  */	\
	else if (0 && REG_P (index) && REG_OK_FOR_INDEX_P (index))	\
	  goto ADDR;					\
	else if (0 && GET_CODE (index) == MULT)		\
	  { rtx xfoo0 = XEXP (index, 0), xfoo1 = XEXP (index, 1);\
	    if (GET_CODE (xfoo0) == CONST_INT		\
		&& INTVAL (xfoo0) == GET_MODE_SIZE (MODE)\
		&& REG_P (xfoo1)			\
		&& REG_OK_FOR_INDEX_P (xfoo1)		\
		||					\
		GET_CODE (xfoo1) == CONST_INT		\
		&& INTVAL (xfoo1) == GET_MODE_SIZE (MODE)\
		&& REG_P (xfoo0)			\
		&& REG_OK_FOR_INDEX_P (xfoo0))		\
	      goto ADDR;				\
	  }						\
    }							\
}

/* Try machine-dependent ways of modifying an illegitimate address
   to be legitimate.  If we find one, return the new, valid address.
   This macro is used in only one place: `memory_address' in explow.c.

   OLDX is the address as it was before break_out_memory_refs was called.
   In some cases it is useful to look at this to decide what needs to be done.

   MODE and WIN are passed so that this macro can use
   GO_IF_LEGITIMATE_ADDRESS.

   It is always safe for this macro to do nothing.  It exists to recognize
   opportunities to optimize the output.  */

/* On the hp9k800, change REG+N into REG+REG, and REG+(X*Y) into REG+REG.  */

#define LEGITIMIZE_ADDRESS(X,OLDX,MODE,WIN)	\
{ if (GET_CODE (X) == PLUS && CONSTANT_ADDRESS_P (XEXP (X, 1)))	\
    (X) = gen_rtx (PLUS, SImode, XEXP (X, 0),			\
		   copy_to_mode_reg (SImode, XEXP (X, 1)));	\
  if (GET_CODE (X) == PLUS && CONSTANT_ADDRESS_P (XEXP (X, 0)))	\
    (X) = gen_rtx (PLUS, SImode, XEXP (X, 1),			\
		   copy_to_mode_reg (SImode, XEXP (X, 0)));	\
  if (GET_CODE (X) == PLUS && GET_CODE (XEXP (X, 0)) == MULT)	\
    (X) = gen_rtx (PLUS, SImode, XEXP (X, 1),			\
		   force_operand (XEXP (X, 0), 0));		\
  if (GET_CODE (X) == PLUS && GET_CODE (XEXP (X, 1)) == MULT)	\
    (X) = gen_rtx (PLUS, SImode, XEXP (X, 0),			\
		   force_operand (XEXP (X, 1), 0));		\
  if (memory_address_p (MODE, X))				\
    goto WIN; }

/* Go to LABEL if ADDR (a legitimate address expression)
   has an effect that depends on the machine mode it is used for.  */

#define GO_IF_MODE_DEPENDENT_ADDRESS(ADDR,LABEL)	\
  if (GET_CODE (ADDR) == PRE_DEC	\
      || GET_CODE (ADDR) == POST_DEC	\
      || GET_CODE (ADDR) == PRE_INC	\
      || GET_CODE (ADDR) == POST_INC)	\
    goto LABEL

/* Specify the machine mode that this machine uses
   for the index in the tablejump instruction.  */
#define CASE_VECTOR_MODE SImode

/* Define this if the tablejump instruction expects the table
   to contain offsets from the address of the table.
   Do not define this if the table should contain absolute addresses.  */
/* #define CASE_VECTOR_PC_RELATIVE */

/* Specify the tree operation to be used to convert reals to integers.  */
#define IMPLICIT_FIX_EXPR FIX_ROUND_EXPR

/* This is the kind of divide that is easiest to do in the general case.  */
#define EASY_DIV_EXPR TRUNC_DIV_EXPR

/* Define this as 1 if `char' should by default be signed; else as 0.  */
#define DEFAULT_SIGNED_CHAR 1

/* Max number of bytes we can move from memory to memory
   in one reasonably fast instruction.  */
#define MOVE_MAX 4

/* Nonzero if access to memory by bytes is slow and undesirable.  */
#define SLOW_BYTE_ACCESS 0

/* Do not break .stabs pseudos into continuations.  */
#define DBX_CONTIN_LENGTH 0

/* Value is 1 if truncating an integer of INPREC bits to OUTPREC bits
   is done just by pretending it is already truncated.  */
#define TRULY_NOOP_TRUNCATION(OUTPREC, INPREC) 1

/* We assume that the store-condition-codes instructions store 0 for false
   and some other value for true.  This is the value stored for true.  */

#define STORE_FLAG_VALUE 1

/* When a prototype says `char' or `short', really pass an `int'.  */
#define PROMOTE_PROTOTYPES

/* Specify the machine mode that pointers have.
   After generation of rtl, the compiler makes no further distinction
   between pointers and any other objects of this machine mode.  */
#define Pmode SImode

/* A function address in a call instruction
   is a byte address (for indexing purposes)
   so give the MEM rtx a byte's mode.  */
#define FUNCTION_MODE SImode

/* Define this if addresses of constant functions
   shouldn't be put through pseudo regs where they can be cse'd.
   Desirable on machines where ordinary constants are expensive
   but a CALL with constant address is cheap.  */
#define NO_FUNCTION_CSE

/* Compute the cost of computing a constant rtl expression RTX
   whose rtx-code is CODE.  The body of this macro is a portion
   of a switch statement.  If the code is computed here,
   return it with a return statement.  Otherwise, break from the switch.  */

#define CONST_COSTS(RTX,CODE) \
  case CONST_INT:						\
    if (INTVAL (RTX) == 0) return 0;				\
    if (INT_14_BITS (RTX)) return 1;				\
  case CONST:							\
  case LABEL_REF:						\
  case SYMBOL_REF:						\
    return 2;							\
  case CONST_DOUBLE:						\
    return 4;

/* Tell emit-rtl.c how to initialize special values on a per-function basis.  */
extern int optimize;
extern struct rtx_def *cc0_reg_rtx;

typedef struct { struct rtx_def *reg1;
		 struct rtx_def *cc0_src; } cc_status_mdep;
#define CC_STATUS_MDEP cc_status_mdep

/* Tell final.c how to eliminate redundant test instructions.  */

/* Here we define machine-dependent flags and fields in cc_status
   (see `conditions.h').  */

/* Nonzero if the results of the previous comparison are
   in the floating point condition code register.  */

#define CC_IN_FCCR 04000

/* Nonzero if the results of the previous comparison are
   int the coprocessor's condition code register.  */

#define CC_IN_CCCR 010000

/* Nonzero if we know the value of %hi(%g1).  */
#define CC_KNOW_HI_REG1 0100000

#define CC_STATUS_MDEP_INIT (cc_status.mdep.reg1 = 0)

/* Store in cc_status the expressions
   that the condition codes will describe
   after execution of an instruction whose pattern is EXP.
   Do not alter them if the instruction would not alter the cc's.  */

#define NOTICE_UPDATE_CC(EXP, INSN) \
{ if (GET_CODE (EXP) == SET)					\
    { if (GET_CODE (SET_DEST (EXP)) == CC0)			\
	{ cc_status.flags = 0;					\
	  cc_status.value1 = SET_DEST (EXP);			\
	  cc_status.value2 = SET_SRC (EXP);			\
	}							\
      else if (GET_CODE (SET_DEST (EXP)) == REG)		\
	{ if ((cc_status.value1					\
	       && reg_overlap_mentioned_p (SET_DEST (EXP), cc_status.value1))) \
	    cc_status.value1 = 0;				\
	  if ((cc_status.value2					\
	      && reg_overlap_mentioned_p (SET_DEST (EXP), cc_status.value2))) \
	    cc_status.value2 = 0;				\
	}							\
      else if (GET_CODE (SET_DEST (EXP)) == MEM)		\
	{ rtx x = cc_status.mdep.reg1; int know = cc_status.flags & CC_KNOW_HI_REG1;	\
	  CC_STATUS_INIT;					\
	  if (x && know)					\
	    { cc_status.mdep.reg1 = x; cc_status.flags |= CC_KNOW_HI_REG1; }	\
	}							\
    }								\
  else if (GET_CODE (EXP) == PARALLEL				\
	   && GET_CODE (XVECEXP (EXP, 0, 0)) == SET)		\
    { if (GET_CODE (SET_DEST (XVECEXP (EXP, 0, 0))) == CC0)	\
	{ cc_status.flags = 0;					\
	  cc_status.value1 = SET_DEST (XVECEXP (EXP, 0, 0));	\
	  cc_status.value2 = SET_SRC (XVECEXP (EXP, 0, 0));	\
	}							\
      else if (GET_CODE (SET_DEST (XVECEXP (EXP, 0, 0))) == REG) \
	{ if ((cc_status.value1					\
	       && reg_overlap_mentioned_p (SET_DEST (XVECEXP (EXP, 0, 0)), cc_status.value1))) \
	    cc_status.value1 = 0;				\
	  if ((cc_status.value2					\
	       && reg_overlap_mentioned_p (SET_DEST (XVECEXP (EXP, 0, 0)), cc_status.value2))) \
	    cc_status.value2 = 0;				\
	}							\
      else if (GET_CODE (SET_DEST (XVECEXP (EXP, 0, 0))) == MEM) \
	{ rtx x = cc_status.mdep.reg1; int know = cc_status.flags & CC_KNOW_HI_REG1;	\
	  CC_STATUS_INIT;					\
	  if (x && know)					\
	    { cc_status.mdep.reg1 = x; cc_status.flags |= CC_KNOW_HI_REG1; }	\
	}							\
    }								\
  else if (GET_CODE (EXP) == PARALLEL)				\
  /* insn-peep has changed this insn beyond recognition
     by NOTICE_UPDATE_CC.  However, we know it is either
     a call or a branch with a delay slot filled, so we can
     give up on knowing condition codes in any case.  */	\
    { CC_STATUS_INIT; }						\
  else if (GET_CODE (EXP) == CALL)				\
    { /* all bets are off */ CC_STATUS_INIT; }			\
}

/* Control the assembler format that we output.  */

/* Output at beginning of assembler file.  */

#define ASM_FILE_START(FILE) do { fprintf (FILE, "\t.SPACE $PRIVATE$\n\
\t.SUBSPA $DATA$,QUAD=1,ALIGN=8,ACCESS=31\n\
\t.SPACE $TEXT$\n\
\t.SUBSPA $LIT$,QUAD=0,ALIGN=8,ACCESS=44\n\
\t.SUBSPA $CODE$,QUAD=0,ALIGN=8,ACCESS=44,CODE_ONLY\n\
\t.IMPORT $global$,DATA\n\
\t.IMPORT $$dyncall,MILLICODE\n");} while (0)


/* Output to assembler file text saying following lines
   may contain character constants, extra white space, comments, etc.  */

#define ASM_APP_ON ""

/* Output to assembler file text saying following lines
   no longer contain unusual constructs.  */

#define ASM_APP_OFF ""

/* We don't yet know how to identify GCC to HP series 800.  */
#define ASM_IDENTIFY_GCC(FILE) fprintf (FILE, "; gcc_compiled.:\n")

/* Output before code.  */

#define TEXT_SECTION_ASM_OP "\t.SPACE $TEXT$\n\t.SUBSPA $CODE$\n"

/* Output before writable data.  */

#define DATA_SECTION_ASM_OP "\t.SPACE $PRIVATE$\n\t.SUBSPA $DATA$\n"

/* How to refer to registers in assembler output.
   This sequence is indexed by compiler's hard-register-number (see above).  */

#define REGISTER_NAMES \
{"0", "1", "2", "3", "4", "5", "6", "7", "8", "9",		\
 "10", "11", "12", "13", "14", "15", "16", "17", "18", "19",	\
 "20", "21", "22", "23", "24", "25", "26", "27", "28", "29",	\
 "30", "31",							\
 "0", "1", "2", "3", "4", "5", "6", "7",\
 "8", "9", "10", "11", "12", "13", "14", "15",}

/* How to renumber registers for dbx and gdb.  */

#define DBX_REGISTER_NUMBER(REGNO) (REGNO)

/* This is how to output the definition of a user-level label named NAME,
   such as the label on a static function or variable NAME.  */

#define ASM_OUTPUT_LABEL(FILE, NAME)	\
  do { assemble_name (FILE, NAME); fputc ('\n', FILE); } while (0)

/* This is how to output a command to make the user-level label named NAME
   defined for reference from other files.  */

#define ASM_OUTPUT_EXTERNAL(FILE, DECL, NAME)	\
  do { fputs ("\t.IMPORT ", FILE);				\
	 assemble_name (FILE, NAME);				\
       if (TREE_CODE (DECL) == VAR_DECL && ! TREE_READONLY (DECL))     	\
	 fputs (",DATA\n", FILE);				\
       else							\
	 fputs (",CODE\n", FILE);				\
     } while (0)

/* The bogus HP assembler requires ALL external references to be 
   "imported", even library calls. They look a bit different, so
   here's this macro. */

#define ASM_OUTPUT_EXTERNAL_LIBCALL(FILE, RTL) \
  do { fputs ("\t.IMPORT ", FILE);					\
       assemble_name (FILE, XSTR ((RTL), 0));		       		\
       fputs (",CODE\n", FILE);						\
     } while (0)

#define ASM_GLOBALIZE_LABEL(FILE, NAME, DECL)				\
  do { fputs ("\t.EXPORT ", FILE); assemble_name (FILE, NAME);		\
       if (TREE_CODE (DECL) == FUNCTION_DECL) 				\
	 {								\
	   tree fntype = DECL_RESULT_TYPE (DECL); 			\
	   tree tree_type = TREE_TYPE (DECL);				\
	   tree parm;							\
	   int i;							\
	   fputs (",PRIV_LEV=3", FILE);					\
	   for (parm = DECL_ARGUMENTS (DECL), i = 0; parm && i < 4;	\
		parm = TREE_CHAIN (parm), i++)				\
	     {								\
	       if (TYPE_MODE (DECL_ARG_TYPE (parm)) == SFmode)		\
		 fprintf (FILE, ",ARGW%d=FR", i);			\
	       else if (TYPE_MODE (DECL_ARG_TYPE (parm)) == DFmode)	\
		 {							\
		   if (i == 0 || i == 2)				\
		     {							\
		       fprintf (FILE, ",ARGW%d=FU", i++);		\
		       fprintf (FILE, ",ARGW%d=FR", i);			\
		     }							\
		   else if (i == 1)					\
		     {							\
		       fprintf (FILE, ",ARGW%d=FU", ++i);		\
		       fprintf (FILE, ",ARGW%d=FR", ++i);		\
		     }							\
		 }							\
	       else							\
		 fprintf (FILE, ",ARGW%d=GR", i);			\
	     }								\
	     /* anonymous args */					\
	     if (TYPE_ARG_TYPES (tree_type) != 0			\
		 && (TREE_VALUE (tree_last (TYPE_ARG_TYPES (tree_type)))\
		     != void_type_node))				\
	       {							\
		 for (; i < 4; i++)					\
		   fprintf (FILE, ",ARGW%d=GR", i);			\
	       }							\
	     if (TYPE_MODE (fntype) == DFmode)				\
	       fprintf (FILE, ",RTNVAL=FR");				\
	     else if (TYPE_MODE (fntype) == SFmode)			\
	       fprintf (FILE, ",RTNVAL=FU");				\
	     else if (fntype != void_type_node)				\
	       fprintf (FILE, ",RTNVAL=GR");				\
	 }								\
 fputs ("\n", FILE);} while (0)

/* This is how to output a reference to a user-level label named NAME.
   `assemble_name' uses this.  */

#define ASM_OUTPUT_LABELREF(FILE,NAME)	\
  fprintf (FILE, "%s", NAME)

/* This is how to output an internal numbered label where
   PREFIX is the class of label and NUM is the number within the class.  */

#define ASM_OUTPUT_INTERNAL_LABEL(FILE,PREFIX,NUM)	\
  fprintf (FILE, "%s$%04d\n", PREFIX, NUM)

/* This is how to store into the string LABEL
   the symbol_ref name of an internal numbered label where
   PREFIX is the class of label and NUM is the number within the class.
   This is suitable for output with `assemble_name'.  */

#define ASM_GENERATE_INTERNAL_LABEL(LABEL,PREFIX,NUM)	\
  sprintf (LABEL, "*%s$%04d", PREFIX, NUM)

/* This is how to output an assembler line defining a `double' constant.  */

#define ASM_OUTPUT_DOUBLE(FILE,VALUE)  \
  do { union { double d; int i[2];} __u;	\
    __u.d = (VALUE);				\
    fprintf (FILE, "\t; .double %.20e\n\t.word %d ; = 0x%x\n\t.word %d ; = 0x%x\n",	\
	     __u.d, __u.i[0], __u.i[0], __u.i[1], __u.i[1]);	\
  } while (0)

/* This is how to output an assembler line defining a `float' constant.  */

#define ASM_OUTPUT_FLOAT(FILE,VALUE)  \
  do { union { float f; int i;} __u;		\
    __u.f = (VALUE);				\
    fprintf (FILE, "\t; .float %.12e\n\t.word %d ; = 0x%x\n", __u.f, __u.i, __u.i); \
  } while (0)

/* This is how to output an assembler line defining an `int' constant.  */

#define ASM_OUTPUT_INT(FILE,VALUE)  \
( fprintf (FILE, "\t.word "),			\
  output_addr_const (FILE, (VALUE)),		\
  fprintf (FILE, "\n"))

/* Likewise for `short' and `char' constants.  */

#define ASM_OUTPUT_SHORT(FILE,VALUE)  \
( fprintf (FILE, "\t.half "),			\
  output_addr_const (FILE, (VALUE)),		\
  fprintf (FILE, "\n"))

#define ASM_OUTPUT_CHAR(FILE,VALUE)  \
( fprintf (FILE, "\t.byte "),			\
  output_addr_const (FILE, (VALUE)),		\
  fprintf (FILE, "\n"))

/* This is how to output an assembler line for a numeric constant byte.  */

#define ASM_OUTPUT_BYTE(FILE,VALUE)  \
  fprintf (FILE, "\t.byte 0x%x\n", (VALUE))

#define ASM_OUTPUT_ASCII(FILE, P, SIZE)  \
  output_ascii (FILE, P, SIZE)

#define ASM_OUTPUT_REG_PUSH(FILE,REGNO)  \
  fprintf (FILE, "\tstws,mb %s,4(0,30)\n", reg_names[REGNO])

/* This is how to output an insn to pop a register from the stack.
   It need not be very fast code.  */

#define ASM_OUTPUT_REG_POP(FILE,REGNO)  \
  fprintf (FILE, "\tldws,ma -4(0,30),%s\n", reg_names[REGNO])

/* This is how to output an element of a case-vector that is absolute.
   Note that this method makes filling these branch delay slots
   virtually impossible.  */

#define ASM_OUTPUT_ADDR_VEC_ELT(FILE, VALUE)  \
  fprintf (FILE, "\tb L$%04d\n\tnop\n", VALUE)

/* This is how to output an element of a case-vector that is relative.
   (the hp9k800 does not use such vectors,
   but we must define this macro anyway.)  */

#define ASM_OUTPUT_ADDR_DIFF_ELT(FILE, VALUE, REL)  \
  fprintf (FILE, "\tword L%d-L%d\n", VALUE, REL)

/* This is how to output an assembler line
   that says to advance the location counter
   to a multiple of 2**LOG bytes.  */

#define ASM_OUTPUT_ALIGN(FILE,LOG)	\
    fprintf (FILE, "\t.align %d\n", (1<<(LOG)))

#define ASM_OUTPUT_SKIP(FILE,SIZE)  \
  fprintf (FILE, "\t.blockz %d\n", (SIZE))

/* This says how to output an assembler line
   to define a global common symbol.  */

#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE, ROUNDED)  \
( data_section (),					\
  assemble_name ((FILE), (NAME)),			\
  fputs ("\t.comm ", (FILE)),				\
  fprintf ((FILE), "%d\n", (ROUNDED)))

/* This says how to output an assembler line
   to define a local common symbol.  */

#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE, ROUNDED)  \
( data_section (),					\
  fprintf ((FILE), "\t.align %d\n", (SIZE) <= 4 ? 4 : 8),	\
  assemble_name ((FILE), (NAME)),				\
  fprintf ((FILE), "\n\t.blockz %d\n", (ROUNDED)))

/* Store in OUTPUT a string (made with alloca) containing
   an assembler-name for a local static variable named NAME.
   LABELNO is an integer which is different for each call.  */

#define ASM_FORMAT_PRIVATE_NAME(OUTPUT, NAME, LABELNO)	\
( (OUTPUT) = (char *) alloca (strlen ((NAME)) + 12),	\
  sprintf ((OUTPUT), "%s___%d", (NAME), (LABELNO)))

/* Define the parentheses used to group arithmetic operations
   in assembler code.  */

#define ASM_OPEN_PAREN "("
#define ASM_CLOSE_PAREN ")"

/* Define results of standard character escape sequences.  */
#define TARGET_BELL 007
#define TARGET_BS 010
#define TARGET_TAB 011
#define TARGET_NEWLINE 012
#define TARGET_VT 013
#define TARGET_FF 014
#define TARGET_CR 015

/* Print operand X (an rtx) in assembler syntax to file FILE.
   CODE is a letter or dot (`z' in `%z0') or 0 if no letter was specified.
   For `%' followed by punctuation, CODE is the punctuation and X is null.

   On the hp9k800, the CODE can be `r', meaning this is a register-only operand
   and an immediate zero should be represented as `r0'.  */

#define PRINT_OPERAND(FILE, X, CODE)  \
{ if ((CODE) == 'O')							\
    switch (GET_CODE (X))						\
      {									\
      case PLUS:							\
	fprintf (FILE, "add%s",						\
		 GET_CODE (XEXP (X, 1)) == CONST_INT ? "i" : ""); break;\
      case MINUS:							\
	fprintf (FILE, "sub%s",						\
		 GET_CODE (XEXP (X, 0)) == CONST_INT ? "i" : ""); break;\
      case AND:								\
	fprintf (FILE, "and%s",						\
		 GET_CODE (XEXP (X, 1)) == NOT ? "cm" : ""); break;	\
      case IOR:								\
	fprintf (FILE, "or"); break;					\
      case XOR:								\
	fprintf (FILE, "xor"); break;					\
      case ASHIFT:							\
	fprintf (FILE, "sh%dadd", INTVAL (XEXP (X, 1))); break;		\
      /* Too lazy to handle bitfield conditions yet.  */		\
      default:								\
	printf ("Can't grok '%c' operator:\n", CODE);			\
	debug_rtx (X);							\
	abort ();							\
      }									\
  else if ((CODE) == 'C')						\
    switch (GET_CODE (X))						\
      {									\
      case EQ:								\
	fprintf (FILE, "="); break;					\
      case NE:								\
	fprintf (FILE, "<>"); break;					\
      case GT:								\
	fprintf (FILE, ">"); break;					\
      case GE:								\
	fprintf (FILE, ">="); break;					\
      case GEU:								\
	fprintf (FILE, ">>="); break;					\
      case GTU:								\
	fprintf (FILE, ">>"); break;					\
      case LT:								\
	fprintf (FILE, "<"); break;					\
      case LE:								\
	fprintf (FILE, "<="); break;					\
      case LEU:								\
	fprintf (FILE, "<<="); break;					\
      case LTU:								\
	fprintf (FILE, "<<"); break;					\
      default:								\
	printf ("Can't grok '%c' operator:\n", CODE);			\
	debug_rtx (X);							\
	abort ();							\
      }									\
  else if ((CODE) == 'N')						\
    switch (GET_CODE (X))						\
      {									\
      case EQ:								\
	fprintf (FILE, "<>"); break;					\
      case NE:								\
	fprintf (FILE, "="); break;					\
      case GT:								\
	fprintf (FILE, "<="); break;					\
      case GE:								\
	fprintf (FILE, "<"); break;					\
      case GEU:								\
	fprintf (FILE, "<<"); break;					\
      case GTU:								\
	fprintf (FILE, "<<="); break;					\
      case LT:								\
	fprintf (FILE, ">="); break;					\
      case LE:								\
	fprintf (FILE, ">"); break;					\
      case LEU:								\
	fprintf (FILE, ">>"); break;					\
      case LTU:								\
	fprintf (FILE, ">>="); break;					\
      default:								\
	printf ("Can't grok '%c' operator:\n", CODE);			\
	debug_rtx (X);							\
	abort ();							\
      }									\
  else if ((CODE) == 'M')						\
    switch (GET_CODE (XEXP (X, 0)))					\
      {									\
      case PRE_DEC:							\
      case PRE_INC:							\
	fprintf (FILE, "s,mb");						\
	break;								\
      case POST_DEC:							\
      case POST_INC:							\
	fprintf (FILE, "s,ma");						\
	break;								\
      default:								\
	break;								\
      }									\
  else if ((CODE) == 'G')						\
    {									\
      if (GET_CODE (X) == SYMBOL_REF && RTX_UNCHANGING_P (X))		\
	assemble_name (FILE, XSTR (X, 0));				\
      else if (GET_CODE (X) == SYMBOL_REF)			       	\
	{ assemble_name (FILE, XSTR (X, 0)); fprintf (FILE, "-$global$"); }\
      else if (GET_CODE (X) == CONST)					\
	{ char *sep = "";						\
	  assemble_name (FILE, XSTR (XEXP (XEXP (X, 0), 0), 0));	\
	  if (GET_CODE (XEXP (XEXP (X, 0), 1)) == CONST_INT)		\
	    {								\
	      if (INTVAL (XEXP (XEXP (X, 0), 1)) >= 0)			\
		sep = GET_CODE (XEXP (X, 0)) == PLUS ? "+" : "-";	\
	    }								\
	  else sep = GET_CODE (XEXP (X, 0)) == PLUS ? "+" : "-";	\
	  fprintf (FILE, "-$global$%s",	sep);				\
	  output_addr_const (FILE, XEXP (XEXP (X, 0), 1));		\
        }								\
      else abort ();							\
    }									\
  else if (GET_CODE (X) == REG)						\
    fprintf (FILE, "%s", reg_names [REGNO (X)]);			\
  else if (GET_CODE (X) == MEM)						\
    {									\
      int size = GET_MODE_SIZE (GET_MODE (X));				\
      rtx base = XEXP (XEXP (X, 0), 0);					\
      switch (GET_CODE (XEXP (X, 0)))					\
	{								\
	case PRE_DEC:							\
	case POST_DEC:							\
	  fprintf (FILE, "-%d(0,%s)", size, reg_names [REGNO (base)]);	\
	  break;							\
	case PRE_INC:							\
	case POST_INC:							\
	  fprintf (FILE, "%d(0,%s)", size, reg_names [REGNO (base)]);	\
	  break;							\
	default:							\
	  output_address (XEXP (X, 0));					\
	  break;							\
	}								\
    }									\
  else if (GET_CODE (X) == CONST_DOUBLE && GET_MODE (X) == SFmode)	\
    { union { double d; int i[2]; } u;					\
      union { float f; int i; } u1;					\
      u.i[0] = XINT (X, 0); u.i[1] = XINT (X, 1);			\
      u1.f = u.d;							\
      if (CODE == 'f')							\
	fprintf (FILE, "0r%.9g", u1.f);					\
      else								\
	fprintf (FILE, "0x%x", u1.i); }					\
  else if (GET_CODE (X) == CONST_DOUBLE && GET_MODE (X) != DImode)	\
    { union { double d; int i[2]; } u;					\
      u.i[0] = XINT (X, 0); u.i[1] = XINT (X, 1);			\
      fprintf (FILE, "0r%.20g", u.d); }					\
  else if ((CODE) == 'r' && (X) == const0_rtx)				\
    fprintf (FILE, "0");						\
  else if (GET_CODE (X) == CONST_INT)					\
    fprintf (FILE, "%d", INTVAL (X));					\
  else { output_addr_const (FILE, X); }}

/* Print a memory address as an operand to reference that memory location.  */

#define PRINT_OPERAND_ADDRESS(FILE, ADDR)  \
{ register rtx addr = ADDR;						\
  register rtx base;							\
  int offset;								\
  switch (GET_CODE (addr))						\
    {									\
    case REG:								\
      fprintf (FILE, "0(0,%s)", reg_names [REGNO (addr)]);		\
      break;								\
    case PLUS:								\
      if (GET_CODE (XEXP (addr, 0)) == CONST_INT)			\
	offset = INTVAL (XEXP (addr, 0)), base = XEXP (addr, 1);	\
      else if (GET_CODE (XEXP (addr, 1)) == CONST_INT)			\
	offset = INTVAL (XEXP (addr, 1)), base = XEXP (addr, 0);	\
      else								\
	abort ();							\
      fprintf (FILE, "%d(0,%s)", offset, reg_names [REGNO (base)]);	\
      break;								\
    default:								\
      output_addr_const (FILE, addr);					\
    }}


/* Delayed brancing instructions.  */
/* #define HAVE_DELAYED_BRANCH */

/* This is true, but I don't fully understand the code yet, so ... */
/* #define HAVE_ANNULLED_BRANCH  */

/* Define DBR_SLOTS_AFTER to be the number of slots availible
   after INSN.  Non-zero values for this macro deliniate delayed
   branch blocks.  */

extern struct rtx_def *find_call_subrtx ();

/* On the Spectrum any branch, jump, or call gives 1 slot; the rest
   give 0.  Returns don't exist (they are jumps to the exit code), so
   they aren't worried about. case statements are special; control
   doesn't always flow through the last branch in them, so it can't
   have any delay slots.  This is the garbage at the end of the macro.
   We also can't give any delay slots to calls to a location specified
   by a register; see the machine description for why.  */

#define DBR_SLOTS_AFTER(INSN)					\
  ((GET_CODE (INSN) == CALL_INSN				\
    && GET_CODE (PATTERN (INSN)) != SEQUENCE			\
    && GET_CODE (XEXP (find_call_subrtx (PATTERN (INSN)), 0)) != REG)\
   || (GET_CODE (INSN) == JUMP_INSN				\
       && GET_CODE (PATTERN (INSN)) != RETURN			\
       && GET_CODE (PATTERN (INSN)) != SEQUENCE			\
       && !(GET_CODE (PATTERN (INSN)) == PARALLEL		\
	    && XVECLEN (PATTERN (INSN), 0) == 2			\
	    && GET_CODE (XVECEXP (PATTERN (INSN), 0, 0)) == SET	\
	    && GET_CODE (SET_DEST (XVECEXP (PATTERN (INSN), 0, 0))) == PC \
	    && GET_CODE (XVECEXP (PATTERN (INSN), 0, 1)) == USE)))

/* Anything that is really only one asm instruction can go in a
   delay slot.  Just about everything is a SET.  The following would
   allow patterns in which each of the source and the dest are single
   insn moves, but both together aren't.  But since one end of
   everything on this machine must be a register, this shouldn't be a
   problem.  */

#define DBR_INSN_ELIGIBLE_P(INSN, DINSN) \
  (GET_CODE (PATTERN (INSN)) == SET	 \
   && strict_single_insn_op_p (SET_DEST (PATTERN (INSN)), VOIDmode)  \
   && strict_single_insn_op_p (SET_SRC (PATTERN (INSN)), VOIDmode))


/* All eligible instructions have one delay slot.  */
#define DBR_INSN_SLOTS(INSN) 1

/* This just for convenience in reading the asm language output.  */
#define DBR_OUTPUT_SEQEND(FILE) \
  { output_asm_insn ("\t\t; Delay end", 0); }
