/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	asm_macro.h,v $
 * Revision 2.2  91/07/09  23:16:23  danner
 * 	Initial Checkin. Currently unused. In fact, I'm not sure where
 * 	 this file came from.
 * 	[91/07/08            danner]
 * 
 */

/************************************************************************
 * Various compiler macros used for speed and efficiency.               *
 ************************************************************************/

/*
 * PSR_TYPE is the type of the Process Status Register.
 */
typedef unsigned long psr_type;

/*
 * disable_interrupts_return_psr()
 *
 *    The INTERRUPT_DISABLE bit is set in the PSR and the *PREVIOUS*
 *    PSR is returned.  Intended to be used with set_psr() [below] as in:
 *
 *	{
 *	    psr_type psr;
 *	        .
 *	        .
 *	    psr = disable_interrupts_return_psr();
 *	        .
 *   		SHORT [time-wise] CRITICAL SECTION HERE
 *	        .
 *	    set_psr(psr);
 *	        .
 *	        .
 */
#define disable_interrupts_return_psr()				\
	({							\
		psr_type __foo1_;				\
		psr_type __foo2_;				\
		asm("ldcr %0, cr1" : "=r" (__foo1_));		\
		asm("set  %0, %1, 1<1>" : "=r" (__foo2_) : "r" (__foo1_)); \
		asm("stcr %0, cr1" :: "r" (__foo2_));		\
		__foo1_;					\
	})

/*
 * set_psr(psr_type value)
 * Sets the psr. See comments above.
 */
#define set_psr(PSR)						\
	{							\
		psr_type __foo1_ = (psr_type)(PSR);		\
		asm("stcr %0, cr1" :: "r" (__foo1_));		\
	}
