/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	autoconf.c,v $
 * Revision 2.6  92/05/21  17:17:36  jfriedl
 * 	     Cleaned up and de-linted.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.5  92/05/04  11:26:17  danner
 * 	Lance support. 
 * 	[92/05/03  16:45:17  danner]
 * 
 * Revision 2.4  92/02/18  18:00:26  elf
 * 	Include updates.
 * 	[92/01/20            danner]
 * 
 * Revision 2.3  91/08/24  12:01:23  af
 * 	pcb_zone_init -> pcb_module_init and now is called by MI code.
 * 	[91/08/14            rvb]
 * 	Modifed slave_machine_configure to call cpu_configuration_print.
 * 	[91/07/18            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:45:50  danner
 * 	pcb_zone_init -> pcb_module_init and now is called by MI code.
 * 	[91/08/14            rvb]
 * 	Modifed slave_machine_configure to call cpu_configuration_print.
 * 	[91/07/18            danner]
 * 
 * Revision 2.2  91/07/09  23:16:34  danner
 * 	cputypes.h -> platforms.h
 * 	[91/06/24            danner]
 * 
 * 	Butchered for 3.0
 * 	[91/05/06            danner]
 * 
 * Revision 2.2  91/04/05  14:01:06  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * Copyright (c) 1982,1986 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 *	@(#)autoconf.c	7.1 (Berkeley) 6/6/86
 */

/*
 * Setup the system to run on the current machine.
 *
 * Configure() is called at boot time and initializes the uba and mba
 * device tables and the memory controller monitoring.  Available
 * devices are determined (from possibilities mentioned in ioconf.c),
 * and the drivers are initialized.
 */

#include <cpus.h>
#include <platforms.h>

#include <sys/reboot.h>

#include <luna88k/board.h>
#include <motorola/m88k/m88100/m882xx.h>
#include <luna88k/cpu_number.h> 
#include <sys/types.h>

#include <mach/machine.h>
#include <vm/vm_kern.h>
#include <mach/vm_param.h>

#include <chips/busses.h>

extern struct bus_driver spc_driver;
extern struct bus_device se_driver;

extern void se_intr();
extern void spc_intr();

struct bus_ctlr bus_master_init[] = {
/* driver, name, unit, intr, addr+am, phys, adaptor, alive, flags */
{ &spc_driver, "spc", 0, spc_intr, 0x0, 0, 0, '?', 0, 0,} ,
  0
};

struct bus_device bus_device_init[] = {
{ &spc_driver, "rz", 0, 0, 0x0, 0, 0, '?', 0, 0, 0, 0, },
{ &spc_driver, "rz", 1, 0, 0x0, 0, 0, '?', 0, 0, 1, 0, },
{ &spc_driver, "rz", 2, 0, 0x0, 0, 0, '?', 0, 0, 2, 0, },
{ &spc_driver, "rz", 3, 0, 0x0, 0, 0, '?', 0, 0, 3, 0, },
{ &spc_driver, "rz", 4, 0, 0x0, 0, 0, '?', 0, 0, 4, 0, },
{ &spc_driver, "rz", 5, 0, 0x0, 0, 0, '?', 0, 0, 5, 0, },
{ &spc_driver, "rz", 6, 0, 0x0, 0, 0, '?', 0, 0, 6, 0, },
{ &spc_driver, "rz", 7, 0, 0x0, 0, 0, '?', 0, 0, 7, 0, },
{ &spc_driver, "tz", 0, 0, 0x0, 0, 0, '?', 0, 0, 0, 0, },
{ &spc_driver, "tz", 1, 0, 0x0, 0, 0, '?', 0, 0, 1, 0, },
{ &spc_driver, "tz", 2, 0, 0x0, 0, 0, '?', 0, 0, 2, 0, },
{ &spc_driver, "tz", 3, 0, 0x0, 0, 0, '?', 0, 0, 3, 0, },
{ &spc_driver, "tz", 4, 0, 0x0, 0, 0, '?', 0, 0, 4, 0, },
{ &spc_driver, "tz", 5, 0, 0x0, 0, 0, '?', 0, 0, 5, 0, },
{ &spc_driver, "tz", 6, 0, 0x0, 0, 0, '?', 0, 0, 6, 0, },
{ &spc_driver, "tz", 7, 0, 0x0, 0, 0, '?', 0, 0, 7, 0, },
/* lance */
{ &se_driver,  "am", 0, se_intr, 0x0, 0, 0, '?', 0, -1, -1, 0, },
  0
};

/*
 * The following several variables are related to
 * the configuration process, and are used in initializing
 * the machine.
 */
#ifdef unused
u_short	*caddr;		/* address being probed */
int	devtype;
int	nexnum;		/* current nexus number */
int	nsbi;		/* current sbi number */
#endif
int	cold = 1;	/* if 1, still working on cold-start */
int	dkn;		/* number of iostat dk numbers assigned so far */

extern int loadpt;

extern char *bdevname[];	/* see ../conf/param.c */
extern int nbdevnames;		/* see ../conf/param.c */
extern char *clu_name_list[];	/* see ../conf/param.c */


#ifdef WANT_PROTOTYPES

 /* chips/busses.c */
 extern int configure_bus_device(char*,caddr_t,caddr_t,u_long,char*);
 extern int configure_bus_master(char*,caddr_t,caddr_t,int,char*);

 /* chips/lance.c */
 extern void se_intr(int,int);

 /* device/subrs.c */
 extern unsigned int strlen(char*);

 /* kern/printf.c */
 extern void printf(char*,...);

 /* luna88k/locore/locore.c */
 extern void raw_putstr(char*);

 /* luna88k/locore/locore_asm_routines.c */
 extern void enable_interrupt(void);
 extern void set_cpu_number(unsigned);

 /* luna88k/machdep.c */
 extern char* strarg(char*);
 extern unsigned int spl0(void);
 extern unsigned int spl2(void);
 extern unsigned int splhigh(void);
 extern void cpu_configuration_print(int);
 extern void level3_intr(void);

 /* luna88k/pmap.c */
 extern void pmap_set_kernel_mmu(void);

 /* luna88k/softint.c */
 extern void softcall_init(void);

 /* luna88kdev/uart.c */
 extern void utinit(void);

 /* motorola/m88k/m88100/cmmu.c */
 extern unsigned int cmmu_cpu_number(void);
 extern void cmmu_flush_cache(caddr_t,int);
 extern void cmmu_flush_tlb(unsigned int,caddr_t,int);

 /* scsi/adapters/scsi_89352_hdw.c */
 extern void spc_intr(int);

 /* forwards */
 static void set_ether_addr(void);
 void machine_init(void);
 void slave_config(void);
 void slave_machine_init(void);
 void slave_pre_main(void);
#endif /* ifdef WANT_PROTOTYPES */


/*
 * Determine mass storage and memory configuration for a machine.
 * Get cpu type, and then switch out to machine specific procedures
 * which will probe adaptors to see what is out there.
 */
void machine_init()
{
    int i;
    extern int	cpu_sets[];
    extern void level3_intr();
    extern void (*scb_level3[])();

    /*
     * Copy the cpu existing flag from cpu_sets.
     */
    for (i = 0; i < NCPUS; i++)
	machine_slot[i].is_cpu = cpu_sets[i] ? TRUE : FALSE;

    machine_slot[master_cpu].is_cpu = TRUE;
    machine_slot[master_cpu].running = TRUE;
    machine_slot[master_cpu].cpu_type = CPU_TYPE_MC88000;
    machine_slot[master_cpu].cpu_subtype = CPU_SUBTYPE_LUNA88K;

    softcall_init();		/* initialize softcall */
    utinit();

    scb_level3[0]  = level3_intr;

    spl2();
    enable_interrupt();
    /* Probe the spc based scsi busses */
    configure_bus_master("spc", (caddr_t)SCSI_ADDR, 0, 0, "On Board SCSI");
    set_ether_addr();
    configure_bus_device("am", (caddr_t)TRI_PORT_RAM, (caddr_t)TRI_PORT_RAM,
		         0, "On board LANCE");

#if 0
    coswinit();
    cninit();
    nvprobe();
    lcslave();
#endif

    spl0();

#if 0
    lbfind();
    setroot();
    /*
     * Configure swap area and related system
     * parameter based on device(s) used.
     */
    swapconf();
#endif
    cold = 0;

    cpu_configuration_print(TRUE);
}

volatile int scpus = 0;		/* number of slave cpus */

/*
 * Slave CPU pre-main routine.
 * Determine CPU number and set it.
 *
 * Running on an interrupt stack here; do nothing fancy.
 */
void slave_pre_main()
{
    set_cpu_number(cmmu_cpu_number());	/* Determine cpu number by CMMU */
    splhigh();
    enable_interrupt();
    pmap_set_kernel_mmu();      /* Go virtual */
    cmmu_flush_cache(0, FLUSH_ALL);
    cmmu_flush_tlb(1, (caddr_t)-1, 0);
    cmmu_flush_tlb(0, (caddr_t)-1, 0);
}

void slave_machine_init()
{
    int i, ncpus = 0;

    /*
     *	Slave MMU start
     */
    slave_config();
    scpus++;

    /* interlock until all slaves are up */
    for (i = 0; i < NCPUS; i++)
	if (machine_slot[i].is_cpu)
	    ncpus++;
    ncpus--; /* discount one for the master cpu not being a slave */

    /* see machdep.c (start_other_cpus) for why this wait is here */
    while(scpus != ncpus)
	;
#ifdef JEFF_DEBUG
    cmmu_flush_cache(0, FLUSH_ALL);
    cmmu_flush_tlb(1, (caddr_t)-1, 0);
    cmmu_flush_tlb(0, (caddr_t)-1, 0);
    {
	char *msg[] = {
	    "[slave cpu 0 continuing]\n\r",
	    "[slave cpu 1 continuing]\n\r",
	    "[slave cpu 2 continuing]\n\r",
	    "[slave cpu 3 continuing]\n\r",
	};
	raw_putstr(msg[cpu_number()]);
    }
#endif
}

void slave_config()
{
    
    register int mycpu = cpu_number();

    machine_slot[mycpu].is_cpu = TRUE;
    machine_slot[mycpu].running = TRUE;
    machine_slot[mycpu].cpu_type = CPU_TYPE_MC88000;
    machine_slot[mycpu].cpu_subtype = CPU_SUBTYPE_LUNA88K;
    
    cpu_configuration_print(FALSE);
}

/* Fish out our ethernet address */

unsigned int ether_addr[6];

static void set_ether_addr()
{
    register int i;
    register unsigned char *src, c;
    extern char *strarg();

    if (src = strarg("ENADDR"), src == NULL)
	printf("set_etheraddr: Ethernet address not found\n");
    else if(strlen(src) != 12)
        printf("set_etheraddr: The length of address must be 12.\n");
    else
    {
	for(i = 0; i < 6; i++)
	{
	    c = *src++;
	    c = c <= '9' ? c - '0' : (c&0xF) + 9;
	    c*=16;
	    c += *src <= '9' ? *src - '0' : (*src&0xF) + 9;
	    src++;
	    ether_addr[i] = c;
	}
    }
}
