/* 
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon 
 * the rights to redistribute these changes.
 */

/*
 * HISTORY
 * $Log:	conf.c,v $
 * Revision 2.6  92/05/04  11:26:22  danner
 * 	Added ethernet.
 * 	[92/04/21            danner]
 * 
 * Revision 2.5  92/01/03  20:40:29  dbg
 * 	Added devinfo routine to scsi to accomodate MI change
 * 	that screwed up extra large writes.
 * 	[91/12/26  11:10:01  af]
 * 
 * Revision 2.4  91/08/28  11:14:59  jsb
 * 	Added entries for new dev_info field.
 * 	[91/08/27  17:58:24  jsb]
 * 
 * Revision 2.3  91/07/11  11:00:41  danner
 * 	Copyright Fixes
 * 
 * Revision 2.2  91/07/09  23:16:51  danner
 * 	Initial 3.0 Checkin
 * 	[91/06/26  11:58:45  danner]
 * 
 * Revision 2.2  90/12/04  14:49:59  jsb
 * 	First checkin.
 * 	[90/12/03  21:47:04  jsb]
 * 
 */

#include <device/conf.h>

extern int cnopen(), cnclose(), cnread(), cnwrite(), cngetstat(),
  cnsetstat(), cnmmap(), cnportdeath();

#define   cnname "console"

/* Mapped time */

extern int timeopen(),timeclose(),timemmap();
#define timename	"time"

/* Scsi driver */

extern int rz_open(), rz_close(), rz_read(), rz_write();
int     rz_get_status(), rz_set_status(), rz_devinfo();
#define rz_name "sd"

/* ethernet (named am for 2.5 compat) */

int	se_open(), se_output(), se_get_status(), se_set_status(),
	se_setinput(), se_restart();
#define se_name "am"

int	SE_open(), SE_close(), SE_mmap(), SE_portdeath(),
	SE_get_status(), SE_set_status();
#define SE_name "SE"


/*
 * List of devices - console must be at slot 0
 */
struct dev_ops	dev_name_list[] =
{
	/*name,		open,		close,		read,
	  write,	getstat,	setstat,	mmap,
	  async_in,	reset,		port_death,	subdev,
	  dev_info */

	{ cnname,	cnopen,		cnclose,	cnread,
	  cnwrite,	cngetstat,	cnsetstat,	cnmmap,
	  nodev,	nulldev,	cnportdeath,	0,
	  nodev },

	{ timename,	timeopen,	timeclose,	nulldev,
	  nulldev,	nulldev,	nulldev,	timemmap,
	  nodev,	nulldev,	nulldev,	0,
	  nodev },

	{ rz_name,	rz_open,	rz_close,	rz_read,
	  rz_write,	rz_get_status,	rz_set_status,	nodev,
	  nodev,	nulldev,	nulldev,	8,
	  rz_devinfo },

#if 0
	{ SE_name,	SE_open,	SE_close,	nulldev,
	  nulldev,	SE_get_status,	SE_set_status,	SE_mmap,
	  nodev,	nulldev,	SE_portdeath,	0,
	  nodev },
#endif
	{ se_name,	se_open,	nulldev,	nodev,
	  se_output,	se_get_status,	se_set_status,	nodev,
	  se_setinput,	se_restart,	nulldev,	0,
	  nodev },
};
int	dev_name_count = sizeof(dev_name_list)/sizeof(dev_name_list[0]);

/*
 * Indirect list.
 */
struct dev_indirect dev_indirect_list[] = {

	/* console */
	{ "console",	&dev_name_list[0],	0 }
};
int	dev_indirect_count = sizeof(dev_indirect_list)
				/ sizeof(dev_indirect_list[0]);



