/*
 * Mach Operating System
 * Copyright (c) 1992-1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	db_interface.c,v $
 * Revision 2.9  92/08/03  18:13:43  jfriedl
 * 	added watchpoint support.
 * 	[92/07/31            jfriedl]
 * 
 * 	Test for pmap_initialized in db_{read,write}_bytes. [danner]
 * 	Fixed luna_db_print_frame for data exceptions [jfriedl]
 * 
 * Revision 2.8  92/05/21  17:18:10  jfriedl
 * 	Cleaned up and de-linted.
 * 	Made CHAR args to db_putc() and luna_db_set_palette() UNSIGNED.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.6.1.1  92/04/13  10:03:16  danner
 * 	Fixed typo into dm{a,t,d}2 printing.
 * 	[92/04/12            danner]
 * 
 * Revision 2.6  92/04/01  10:54:28  rpd
 * 	upgraded luna_db_where to use db_find_xtrn_task_sym_and_offset.
 * 	[92/03/21            danner]
 * 	Replaced fixed length wait with db_wait variable.
 * 	[92/03/16  17:18:45  danner]
 * 
 * Revision 2.5  92/03/03  15:38:16  rpd
 * 	Fixed TASK_NULL handling bug in db_task_va_to_phys.
 * 	[92/03/02            danner]
 * 
 * Revision 2.4  92/02/18  18:00:43  elf
 * 	Upgraded to db_task_trap.
 * 	[92/02/14            danner]
 * 
 * 	New functions to support updated ddb.
 * 	[92/02/12            danner]
 * 
 * 	Made db_read_bytes, db_write_bytes multi address space capable.
 * 	[92/02/10            danner]
 * 	Decreased maximum wait by factor of 8.
 * 	[92/02/02            danner]
 * 	Include changes.
 * 	[92/01/20            danner]
 * 
 * Revision 2.3.1.1  91/09/19  13:48:13  danner
 * 	Upgraded luna_db_print_frame to surpress printing of the data
 * 	 unit and fp control registers if the vector indicates is a trap
 * 	 (as opposed to an interrupt or exception).
 * 	[91/09/18            danner]
 * 
 * 	Changed to always print mask in 8 hex digits.
 * 	[91/09/17            danner]
 * 
 * 	Added >> 8 to print of mask, because mask is << 8 before it is
 * 	 saved in pre_ext_int.
 * 	[91/09/17            danner]
 * 
 * 	Added print of interrupt mask.
 * 	[91/08/26            danner]
 * 
 * Revision 2.3  91/08/24  12:05:26  af
 * 	Changes from jfriedl, including 2.5 compatability.
 * 
 * 	Added volatile declarations to compensate for dropping
 * 	 volatile-global flag from gcc.
 * 	[91/07/14            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:46:09  danner
 * 	Changes from jfriedl, including 2.5 compatability.
 * 
 * 	Added volatile declarations to compensate for dropping
 * 	 volatile-global flag from gcc.
 * 	[91/07/14            danner]
 * 
 * Revision 2.2  91/07/09  23:17:05  danner
 * 	Picked up include of assym.s
 * 	[91/07/09            danner]
 * 
 * 	Added printing of epsr, vector to frame printing routine. Disabled
 * 	      frame crawling support, since locore support is gone.
 * 	[91/07/08            danner]
 * 
 * 	Changed db_interface to use ENTRY_ASM in gimmeabreak.
 * 	[91/07/08            danner]
 * 	Conditionalized db_read_bytes, db_write_bytes to deal with
 * 	 unitialized pmap subsystem.
 * 	[91/05/13            danner]
 * 
 * Revision 2.4  91/05/07  11:10:44  rvb
 * 	The debugger will now be entered along the boot path if
 * 	 dipswitch three is set. Support added for the kdb_entry
 * 	 breakpoint. 
 * 	[91/05/05            danner]
 * 
 * 	Changed a call of simple_unlock to db_simple_lock. Added machine
 * 	 dependent command "clearscrn" to the debugger so it is usable
 * 	 when droped into from X.
 * 	[91/04/14            danner]
 * 
 * 	Simplified the trap/break path interface; modified
 * 	 luna_db_grab_console to clear the screen on entry.
 * 	[91/04/12            danner]
 * 
 * 	Made the print of the cpu entering the debugger as master NOISY.
 * 	[91/04/12            danner]
 * 
 * Revision 2.3  91/04/10  16:05:40  mbj
 * 	Luna debugger interface changes.
 * 
 * Revision 2.2  91/04/05  14:01:47  mbj
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * 	Added machine specific commands for ddb, and supporting technology.
 * 	[91/03/13            danner]
 * 
 * 	Added current_map function.
 * 	[91/02/06            danner]
 * 
 * 	Created
 * 	[91/01/08            danner]
 * 
 */

/*
 * luna interface to ddb debugger
 */

#include <cpus.h>

#include <machine/db_machdep.h>
#include "assym.s" /* for EF_NREGS */
#include <motorola/m88k/m88100/m88100.h>
#include <motorola/m88k/m88100/m882xx.h>
#include <luna88k/board.h>
#include <luna88k/rom1ep.h>
#include <luna88k/pio.h>
#include <luna88k/locore.h>

#include <mach/machine.h>

#include <kern/thread.h>
#include <kern/task.h>
#include <kern/queue.h>

#include <vm/vm_kern.h>
#include <vm/vm_map.h>

#include <ddb/db_command.h>
#include <ddb/db_sym.h>
#include <ddb/db_run.h>

#ifndef MACH_KERNEL
#include <sys/user.h>
#include <sys/proc.h>
#endif
typedef db_regs_t *db_regs_p;

extern int  rawmem_size;
extern int  cmmu_cpu_lock;
/* state pointers for the exception frame */

db_regs_p db_cur_exc_frame[NCPUS] = {0,0,0,0};
db_regs_p db_master_frame = 0;

simple_lock_data_t ddb_entry_lock;
simple_lock_data_t ddb_softcall_lock;
simple_lock_data_t ddb_print_lock;

int ddb_master = -1;
int ddb_depth  = 0;
volatile int db_barrier[3]={0,0,0};
volatile int db_frozen[NCPUS] = {0,0,0,0};
int bwait=0;
int db_noisy = 0;
int db_console_grabbed;
int db_was_console_on;

static void luna_db_sysreset();
static void luna_db_poweroff();
static void luna_db_registers();
static void luna_db_rom_abort();
static void luna_db_where();
#ifndef MACH_KERNEL
void luna_db_task();
#endif
static void luna_db_freeze_cpu();
static void luna_db_thaw_cpu();
static void luna_db_status_cpu();
static void luna_db_noise();
static void luna_db_clear_screen();
static void luna_db_frame_search();
static void luna_db_print_frame();
static void luna_db_iflush();
static void luna_db_dflush();
static void luna_db_peek();

struct db_command luna_cpu_cmds[] =
{
    {"freeze",    luna_db_freeze_cpu,0, 0},
    {"thaw",      luna_db_thaw_cpu,  0, 0},
    {"status",    luna_db_status_cpu, 0, 0},
    {(char  *) 0,}
};

struct db_command luna_cache_cmds[] =
{
    { "iflush",    luna_db_iflush, 0, 0},
    { "dflush",    luna_db_dflush, 0, 0},
    { "peek",      luna_db_peek, 0, 0},
    { (char *) 0,}
};

struct db_command db_machine_cmds[] =
{
    {"registers",	luna_db_registers,	0, 0},
    {"sysreset",	luna_db_sysreset,  	0, 0},
    {"poweroff",	luna_db_poweroff,	0, 0},
    {"abort",		luna_db_rom_abort,	0, 0},
    {"rom",		luna_db_rom_abort,	0, 0},
    {"where",		luna_db_where,		0, 0},
#ifndef MACH_KERNEL
    {"task",		luna_db_task,		0, 0},
#endif
    {"noise",		luna_db_noise,		0, 0},
    {"clearscrn",	luna_db_clear_screen,	0, 0},
    {"cpu",		0,			0, luna_cpu_cmds},
    {"searchFrame",	luna_db_frame_search,	0, 0},
    {"frame",		luna_db_print_frame,	0, 0},
    {"cache",		0,			0, luna_cache_cmds},
    {(char  *) 0,}
};


void kdb_init()
{
    simple_lock_init(&ddb_entry_lock);
    simple_lock_init(&ddb_softcall_lock);
    simple_lock_init(&ddb_print_lock);
#ifdef DB_MACHINE_COMMANDS
    db_machine_commands_install(db_machine_cmds);
#endif
    ddb_init();

    printf("ddb enabled\n");
    /* if dipswitch 3 is set, trap into the debugger via gimmeabreak */
    if (diptest(3))
	gimmeabreak();
}

/* forward declarations */
static void luna_db_str();
static void luna_db_str1();
static void luna_db_str2();

/*
  when the below routine is entered interrupts should be on
  but spl should be high
*/

static void luna_db_freeze(trap_type)
int trap_type;
{
    int cpu = cpu_number();

    while (db_frozen[cpu])
    {
	/* if we are frozen just chill here until db_barrier[0] opens,
	 then drop into the debugger as a slave. */
	NOISY2(luna_db_str1("(cpu %d) freezing\n",cpu);)
	while(!db_barrier[0])
	    ;
	luna_db_trap(trap_type,0,1);
    }
}

/* fielded a non maskable interrupt - from ext_int (machdep.c) */
int ddb_nmi_trap(level,eframe)
int level;
db_regs_t *eframe;
{
    db_regs_t *oldframe;
    int cpu = cpu_number();

    NOISY(luna_db_str1("kernel: nmi interrupt cpu %d\n",cpu);)
    oldframe = db_cur_exc_frame[cpu];
    db_cur_exc_frame[cpu] = eframe;
    luna_db_trap(T_KDB_ENTRY,0,0);
    luna_db_freeze(T_KDB_ENTRY);
    db_cur_exc_frame[cpu] = oldframe;

    return 0;
}

/*
 * when the below routine is entered interrupts should be on
 * but spl should be high
 */

/* softcall entry - always a slave */
int ddb_softcall_trap(level,eframe)
int level;
db_regs_t *eframe;
{
    db_regs_t *oldframe;
    int cpu = cpu_number();

    if (!db_simple_lock_held(&ddb_softcall_lock))
    {
	/* the softcall lock is not held; this means
	   this  call is spurious; just return */
	NOISY2(luna_db_str1("kernel: spurious softcall on cpu %d\n",cpu);)
    }
    else
    {
	NOISY(luna_db_str1("kernel: softcall interrupt cpu %d\n",cpu);)
	oldframe = db_cur_exc_frame[cpu];
	db_cur_exc_frame[cpu] = eframe;
	luna_db_trap(T_KDB_ENTRY,0,1);
	luna_db_freeze(T_KDB_ENTRY);
	db_cur_exc_frame[cpu] = oldframe;
    }
    return 0;
}

/*
 * When the below routine is entered interrupts should be on
 * but spl should be high
 *
 * The following routine is for breakpoint and watchpoint entry.
 * When a processor hits this, it needs to do a software interrupt
 * (softcall) to halt the other processors. But since multiple threads
 * could have hit breakpoints, protect the softcall by a lock. So if
 * a processor hits here and is unable to grab the lock, that means another
 * processor has hit a breakpoint/watchpoint and already emitted the
 * softcall (or is in the process of doing so).
 *
 * In the case that the softcall lock is already held, the processor
 * should drop into the kernel as slave; when it recieves the softcall
 * interrupt latter, it will notice that the softcall lock is not held
 * and just ignore the interrupt and continue. (See above).
 *
 */

/* breakpoint/watchpoint entry */
int ddb_break_trap(type,eframe)
int type;
db_regs_t *eframe;
{
    db_regs_t *oldframe;
    int cpu = cpu_number();

    NOISY(luna_db_str1("kernel: break/watch/ cpu %d\n",cpu);)
    oldframe = db_cur_exc_frame[cpu];
    db_cur_exc_frame[cpu] = eframe;
    if (db_simple_lock_try(&ddb_softcall_lock))
    {
	/* we got the lock - signal other cpus */
#if 0
	luna_db_str2("cpu %d breakpoint trap, frame = 0x%x\n", cpu,(int)eframe);
#endif
	luna_db_signal_cpus(cpu); /* halt other cpus */
	luna_db_trap(type,0,0);
	db_simple_unlock(&ddb_softcall_lock);
    }
    else
	luna_db_trap(type,0,1); /* enter as slave */

    luna_db_freeze(type);
    db_cur_exc_frame[cpu] = oldframe;
    if (type == T_KDB_BREAK)
    {
	/* back up an instruction and retry the instruction at the
	   breakpoint address */
	eframe->sfip = eframe->snip;
	eframe->snip = eframe->sxip;
    }

    return 0;
}

/* gimmeabreak - drop execute the ENTRY trap */
void gimmeabreak()
{
    asm (ENTRY_ASM); /* entry trap */
    /* ends up at routine below */
}

/* enter at splhigh */

int ddb_entry_trap(level,eframe)
int level;
db_regs_t *eframe;
{
    db_regs_t *oldframe;
    int cpu = cpu_number();

    NOISY(luna_db_str1("kernel: entry trap cpu %d\n",cpu);)
    oldframe = db_cur_exc_frame[cpu];
    db_cur_exc_frame[cpu] = eframe;
    if (db_simple_lock_try(&ddb_softcall_lock))
    {
	/* we got the lock - signal other cpus */
	luna_db_signal_cpus(cpu); /* halt other cpus */
	luna_db_trap(T_KDB_ENTRY,0,0);
	db_simple_unlock(&ddb_softcall_lock);
    }
    else
	luna_db_trap(T_KDB_ENTRY,0,1); /* enter as slave */

    luna_db_freeze(T_KDB_ENTRY);
    db_cur_exc_frame[cpu] = oldframe;
    /* unlike break traps, we need not restart the instruction */

    return 0;
}

/*
 * when the below routine is entered interrupts should be on
 * but spl should be high
 */

/* error trap - unreturnable */
void ddb_error_trap(error,eframe)
char *error;
db_regs_t *eframe;
{
    db_regs_t *oldframe;
    int cpu = cpu_number();

    luna_db_str2("KERNEL:  terminal error %s (cpu %d)\n",(int)error,cpu);
    luna_db_str("KERNEL:  Exiting debugger will cause abort to rom\n");
    luna_db_str1("at 0x%x ", eframe->sxip & ~3);
    luna_db_str2("dmt0 0x%x dma0 0x%x", eframe->dmt0, eframe->dma0);
    luna_db_pause(1000000);
    oldframe = db_cur_exc_frame[cpu];
    db_cur_exc_frame[cpu] = eframe;
    if (db_simple_lock_try(&ddb_softcall_lock))
    {
	luna_db_signal_cpus(cpu); /* halt other cpus */
	luna_db_trap(T_KDB_BREAK,0,0);
	luna_db_freeze(T_KDB_BREAK);
	while(1)
	    call_rom_abort(2,eframe);
    }
    else
    {
	luna_db_str1("(cpu %d) with ERROR entering debugger as slave;\n",cpu);
	luna_db_trap(T_KDB_BREAK,0,1);
	luna_db_freeze(T_KDB_BREAK);
	while(1)
	    call_rom_abort(2,eframe);
    }
}


extern int pmap_initialized;
/*
 * Translate from virtual address va in task task
 * to physical address pa. Return 1 if succesfull
 * translation, 0 if not
 */

static vm_offset_t db_task_va_to_phys(va, task, pa)
vm_offset_t va;
task_t task;
vm_offset_t *pa;
{
    pmap_t pm;
    vm_offset_t p;

    if (task == TASK_NULL)
    {
	if (badaddr((void*)va, 1))
	    return 0;
	if (!pmap_initialized)
	    *pa = va;
	else
	{
	    p = pmap_extract_unlocked(kernel_pmap,va);
	    *pa = p ? p : va;
	}
	return 1;
    }

    /* check map */
    if (badwordaddr(task->map))
	return 0;

    /* check pmap sanity */
    if (badwordaddr(task->map->pmap))
	return 0;

    pm = task->map->pmap;

    if (pm == kernel_pmap)
    {
	if (badaddr((void*)va, 1))
	    return 0;
	p = pmap_extract_unlocked(pm,va);
	*pa = p ? p : va;
	return 1;
    }

    p = pmap_extract_unlocked(pm,va);
    if (p == 0 && va != 0) /* pmap_extract_unlocked failed */
	return 0;
    *pa = p;
    return 1;
}

/*
 * read some bytes for the debugger
 */

void db_read_bytes(addr,size,data,task)
char *addr;
unsigned int size;
char *data;
task_t	task;
{
    char *paddr;

    while(size--)
    {
	if (!db_task_va_to_phys((vm_offset_t)addr,task,(vm_offset_t*)&paddr))
	    db_error("Translation failure");

      if (paddr < rawmem_size && pmap_initialized)
	paddr += VEQR_ADDR; /* set pointer into physical mapping region */ 

	if (!badaddr(paddr, 1))
	    *data++ = *paddr;
	else
	    db_error("Invalid virtual address or page not resident");
	addr++;
    }
}

/*
 * write some bytes for the debugger
 */

void db_write_bytes(addr,size,data,task)
char *addr;
unsigned int size;
char *data;
task_t task;
{
    char *paddr;

    while(size--)
    {
	if (!db_task_va_to_phys((vm_offset_t)addr,task,(vm_offset_t*)&paddr))
	    db_error("Invalid virtual address or page not resident");

      if (paddr < rawmem_size && pmap_initialized)
	paddr += VEQR_ADDR; /* set pointer into physical mapping region */ 

	if (!badaddr(paddr, 1))
	    *paddr = *data++;
	else
	    db_error("Invalid virtual address or page not resident");
	addr++;
    }
}

/* to print a character to the console */
void db_putc(c)
unsigned c;
{
    while (mayput(c & 0xff) == -1)
	;
    if (c == '\n')
	db_putc('\r');
}

/* to peek at the console; returns -1 if no character is there */
int db_getc()
{
    return mayget();
}

int db_wait = 8;

static void luna_db_trap(type, code, soft)
int type, code, soft;
{
    int cpu = cpu_number();

    if (db_spl() != 0x03ffff00 /* splhigh */)
	luna_db_str2("WARNING: (cpu %d) spl is not high in luna_db_trap (spl=%x)\n", cpu,db_spl());

    if (db_are_interrupts_disabled())
	luna_db_str1("WARNING: (cpu %d) entered debugger with interrupts disabled\n", cpu);

    db_simple_lock(&ddb_entry_lock);

    if (ddb_master == -1 && !soft)
    {
	ddb_master = cpu;
	ddb_depth = 1;
    }
    else
	if (ddb_master == cpu)

    ddb_depth++;
    db_simple_unlock(&ddb_entry_lock);

    NOISY2(luna_db_str1(cpu==ddb_master ?
	"(cpu %d) master\n" :  "(cpu %d) slave\n",cpu);)

    if (ddb_master == cpu) /* self is master */
    {
	int oldbwait = bwait; /* record current barrier status */
	int i;
	int ncpus=0;
	int delay = 2;
	db_regs_t *oframe;

	db_console_grabbed = FALSE;

	for (ncpus=0, i=0; i<NCPUS; i++)
	    if (machine_slot[i].is_cpu)
		ncpus++;

	ncpus--;
	NOISY(luna_db_str1("%d slaves expected\n",ncpus);)
	bwait = 0;      /* set counter to zero */
	db_barrier[0] = 1; /* let the slaves through to register */

	i = 0;
	while(bwait != ncpus && i<db_wait)
	{
	    luna_db_pause(delay);
	    delay *= 2;
	    i++;
	}

	NOISY(luna_db_str2("%d slaves responded (count=%d) \n", bwait,i);)
	if (bwait != ncpus)
	    luna_db_str1("WARNING:{%d}  not all cpus responded\n",cpu);

	/* relock the first barrier */
	db_barrier[0] = 0;

	NOISY(luna_db_str1("cpu %d entering ddb as master\n",cpu);)
	oframe = db_master_frame;
	db_master_frame = db_cur_exc_frame[cpu];

	/* reset the frozen masks to what they should be */
	luna_db_release();
	db_task_trap(type,code,USERMODE(db_master_frame->epsr));
	/* If any of the single step flags are set,
	   we should freeze all the other cpus temporarily */
	luna_db_hold_down();

	db_master_frame = oframe;
	db_simple_lock(&ddb_entry_lock);
	NOISY2(luna_db_str("master leaving luna_db_trap\n");)
	if (--ddb_depth == 0)
	    ddb_master = -1;
	db_simple_unlock(&ddb_entry_lock);
	/* release the first barrier */

	/* lock the third, unlock the second let the slaves count clear */
	db_barrier[2] = 0; /* order is important */
	db_barrier[1] = 1;
	for (i=0, delay = 2; bwait && i<24; i++)
	{
	    luna_db_pause(delay);
	    delay *= 2;
	}

	if (bwait!=0)
	    luna_db_str1("barrier wait %d after long wait; cont. anyway\n",bwait);

	bwait = oldbwait;
	db_barrier[1] = 0; /* relock the second */
	db_barrier[2] = 1; /* unlock the third */

	luna_db_release_console();
    }
    else
    {
	/* we are a slave. register our status */
	while(!db_barrier[0])
		;
	db_simple_lock(&ddb_entry_lock);
	bwait++;
	db_simple_unlock(&ddb_entry_lock);
	while(!db_barrier[1])
		;
	db_simple_lock(&ddb_entry_lock);
	bwait--;
	db_simple_unlock(&ddb_entry_lock);
	while(!db_barrier[2])
		;
    }
}

/*
 * pause for 2*ticks many cycles
 */
static void luna_db_pause(ticks)
unsigned volatile ticks;
{
    while (ticks)
	ticks -= 1;
    return;
}

/*
 * send soft interrupt to every other cpu
 * besides self and frozen cpus.
 *
 * Not sure what spl status one should have when
 * this is invoked.
 */

static void luna_db_signal_cpus(self)
int self;
{
    int i;

    for (i = 0; i < NCPUS; i++)
    {
	if (i != self && machine_slot[i].is_cpu && db_frozen[i]==0)
	{
	    /* not self, unfrozen */
	    RVPtr->EPsoft_intr_rom[i] = LUNA_DDB_SOFTCALL;
	    switch(i)
	    {
	      case 0:
		*((volatile int *) SOFT_INT0) = 0;
		break;
	      case 1:
		*((volatile int *) SOFT_INT1) = 0;
		break;
	      case 2:
		*((volatile int *) SOFT_INT2) = 0;
		break;
	      case 3:
		*((volatile int *) SOFT_INT3) = 0;
		break;
	    }
	}
    }
}

static int luna_db_print_lock_acquire()
{
    int count = 0;
    int got;
    /* try to get the print lock, but give up after 5 megacycles -
       about a second */

    while (!(got = db_simple_lock_try(&ddb_print_lock)) && count < 5*1024*1024)
    {
	while (db_simple_lock_held(&ddb_print_lock) && count < 5*1024*1024)
	    count++;
	count++;
    }

    return got;
}

static int luna_db_print_lock_release(held)
int held;
{
    if (held)
	db_simple_unlock(&ddb_print_lock);
    return 0;
}

static void luna_db_str(str)
char *str;
{
    int acq = luna_db_print_lock_acquire();

    db_printf(str);
    luna_db_print_lock_release(acq);
}

static void luna_db_str1(str,arg1)
char *str;
int arg1;
{
    int acq = luna_db_print_lock_acquire();

    db_printf(str, arg1);
    luna_db_print_lock_release(acq);
}

static void luna_db_str2(str,arg1,arg2)
char *str;
int arg1;
int arg2;
{
    int acq = luna_db_print_lock_acquire();

    db_printf(str, arg1, arg2);
    luna_db_print_lock_release(acq);
}

/*
 * When bm_set_palette_colors() is used to set the foreground and
 * background colors, it can return the old values of those colors.
 * However, only for 8-plane boards. Therefore, we'll stuff some valid
 * values in here now so that we'll have something from which to restore
 * the palette for one and four plane boards.
 *
 * The default X colormap defaults to have white and black preallocated
 * to indicies #0 and #1, so we'll pre-load ddb_saved_palettte with
 * white and black.  The 8-plane board will overwrite these, but they'll
 * be here for when the 4-plane board needs to restore the palette....
 */
static unsigned char ddb_saved_palette[2][3] =
	{ /*WHITE*/0xff, 0xff, 0xff,	/*BLACK*/0x00, 0x00, 0x00 };


extern void bm_align_screen();
extern void bm_set_direct_framebuffer_access();
extern void bm_set_palette_colors();

/* get a hold of the console - via rom call (see rom1ep.h) */
static int luna_db_grab_console()
{
    int acq = luna_db_print_lock_acquire();
    int i;

    bm_set_direct_framebuffer_access();
    i = CONSON();
    if (!i)
    {
        bm_set_palette_colors(/*put previous values in */ddb_saved_palette,
		/* foreground R,G,B */ 9, 7, 15,	/* purple */
		/* background R G B */ 0, 0, 0); 	/* black  */
	bm_align_screen();
    }

    luna_db_print_lock_release(acq);
    db_was_console_on = i;
    return i;
}

static void luna_db_release_console()
{
    if (db_console_grabbed && !db_was_console_on)
    {
	CONSOFF();
        bm_set_palette_colors(/*put previous values in */0,
		ddb_saved_palette[0][0],
		ddb_saved_palette[0][1],
		ddb_saved_palette[0][2],
		ddb_saved_palette[1][0],
		ddb_saved_palette[1][1],
		ddb_saved_palette[1][2]);
	bm_align_screen();
    }
}

void db_printing()
{
    if (db_console_grabbed == FALSE)
    {
	db_console_grabbed = TRUE;
	luna_db_grab_console();
    }
}

static void luna_db_clear_screen()
{
    /* this clears the screen of garbage, then redraws the console text */
    CONSOFF();
    CONSON();
    bm_align_screen();
}

/* machine dependent debugger commands */
static void luna_db_rom_abort()
{
    call_rom_abort(2,db_cur_exc_frame[cpu_number()]);
}

static void luna_db_sysreset()
{
    db_printf("<sysreset>\n");
    *(volatile int *)RESET_CPU_ALL = 0;
}

static void luna_db_poweroff()
{
    POWERDOWN();
}

extern unsigned int db_maxoff;

/*
 * If have_addr is true,  addr is cpu number of frame
 * to dump. Otherwise use master cpu. addr of 5 is all
 */

static void luna_db_registers(addr,have_addr,count,modif)
int addr;
boolean_t have_addr;
int count;
char *modif;
{
    int cpu = ddb_master;
    struct m88100_saved_state *s;

    if (have_addr && addr == 5)
    {
	int i;
	for (i=0; i<NCPUS; i++)
	    if (machine_slot[i].is_cpu)
	luna_db_registers(i,TRUE,0,0);
	return;
    }

    if (have_addr)
	if (addr >= 0 && addr < NCPUS && machine_slot[addr].is_cpu)
	    cpu = addr;
	else
	    db_printf("bad cpu number %d passed to luna show reg., assuming %d instead\n", addr,cpu);

    s = db_cur_exc_frame[cpu];
    luna_db_print_frame(s, TRUE);
}

/*
 *
 * If you really feel like understanding the following procedure and
 * macros, see pages 6-22 to 6-30 (Section 6.7.3) of
 *
 * MC881000 RISC Microprocessor User's Manual Second Edition
 * (Motorola Order: MC88100UM/AD REV 1)
 *
 * and ERRATA-5 (6-23, 6-24, 6-24) of
 *
 * Errata to MC88100 User's Manual Second Edition MC88100UM/AD Rev 1
 * (Oct 2, 1990)
 * (Motorola Order: MC88100UMAD/AD)
 *
 */

/* macros for decoding dmt registers */

#define XMEM(x)  ((x) & (1<<12))
#define XMEM_MODE(x) ((((x)>>2 & 0xf) == 0xf) ? "" : ".bu")
#define MODE(x) ((x)>>2 & 0xf)
#define DOUB(x) ((x) & (1<<13))
#define SIGN(x) ((x) & (1<<6))
#define DAS(x) (((x) & (1<<14)) ? "" : ".usr")
#define REG(x) (((x)>>7) & 0x1f)
#define STORE(x) ((x) & 0x2)

/*
 * return 1 if the printing of the next stage should be surpressed
 */

static int luna_dmx_print(t,d,a,no)
unsigned t,d,a,no;
{
    static unsigned addr_mod[16] = { 0, 0, 1, 0, 2, 0, 0, 0,
				     3, 2, 0, 0, 0, 0, 0, 0};
    static char *mode[16]  = { "?", ".b", ".b", ".h", ".b", "?", "?", "?",
			      ".b", ".h", "?" , "?" , "?" , "?", "?", ""};
    static unsigned mask[16] = { 0,           0xff,        0xff00,     0xffff,
				 0xff0000,    0,           0,          0,
				 0xff000000U, 0xffff0000U, 0,          0,
				 0,           0,           0,    0xffffffffU};
    static unsigned shift[16] = { 0,  0, 8, 0, 16, 0, 0, 0,
				 24, 16, 0, 0,  0, 0, 0, 0};
    int reg = REG(t);

    if (XMEM(t))
    {
	db_printf("xmem%s%s r%d(0x%x) <-> mem(0x%x),",
	    XMEM_MODE(t), DAS(t), reg,
	    (((t)>>2 & 0xf) == 0xf) ? d : (d & 0xff), a );
	return 1;
    }
    else
    {
	if (MODE(t) == 0xf)
	{
	    /* full or double word */
	    if (STORE(t))
		if (DOUB(t) && no == 2)
		    db_printf("st.d%s -> mem(0x%x) (** restart sxip **)",
			DAS(t), a);
		else
		    db_printf("st%s (0x%x) -> mem(0x%x)", DAS(t), d, a);
	    else /* load */
		if (DOUB(t) && no == 2)
		    db_printf("ld.d%s r%d <- mem(0x%x), r%d <- mem(0x%x)",
			DAS(t), reg, a, reg+1, a+4);
		else
		    db_printf("ld%s r%d <- mem(0x%x)",  DAS(t), reg, a);
	}
	else
	{
	    /* fractional word - check if load or store */
	    a += addr_mod[MODE(t)];
	    if (STORE(t))
		db_printf("st%s%s (0x%x) -> mem(0x%x)", mode[MODE(t)], DAS(t),
		(d & mask[MODE(t)]) >> shift[MODE(t)], a);
	    else
		db_printf("ld%s%s%s r%d <- mem(0x%x)",
		mode[MODE(t)], SIGN(t) ? "" : "u", DAS(t), reg, a);
	}
    }
    return 0;
}

static void luna_db_print_frame(s, have_addr)
struct m88100_saved_state *s;
boolean_t have_addr;
{
    char *name;
    db_expr_t offset;
    int surpress1 = 0, surpress2 = 0;

    if (!frame_is_sane(s))  /* see db_trace.c */
    {
	db_printf("Bad address or insane frame given as argument\n");
	return;
    }

    if (!have_addr)
	return;

#define R(i) s->r[i]
#define IPMASK(x) ((x) &  ~(3))
    db_printf("R00-05: 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x\n",
	R(0),R(1),R(2),R(3),R(4),R(5));
    db_printf("R06-11: 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x\n",
	R(6),R(7),R(8),R(9),R(10),R(11));
    db_printf("R12-17: 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x\n",
	R(12),R(13),R(14),R(15),R(16),R(17));
    db_printf("R18-23: 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x\n",
	R(18),R(19),R(20),R(21),R(22),R(23));
    db_printf("R24-29: 0x%08x  0x%08x  0x%08x  0x%08x  0x%08x  0x%08x\n",
	R(24),R(25),R(26),R(27),R(28),R(29));
    db_printf("R30-31: 0x%08x  0x%08x\n",R(30),R(31));

    db_printf("sxip: 0x%08x ",s->sxip);
    db_find_xtrn_sym_and_offset((db_addr_t) IPMASK(s->sxip),&name,&offset);
    if (name!= 0 && (unsigned)offset <= db_maxoff)
	db_printf("%s+0x%08x",name,(unsigned)offset);
    db_printf("\n");
    if (s->snip != s->sxip+4)
    {
	db_printf("snip: 0x%08x ",s->snip);
	db_find_xtrn_sym_and_offset((db_addr_t) IPMASK(s->snip),&name,&offset);
	if (name!= 0 && (unsigned)offset <= db_maxoff)
	    db_printf("%s+0x%08x",name,(unsigned)offset);
	db_printf("\n");
    }
    if (s->sfip != s->snip+4)
    {
	db_printf("sfip: 0x%08x ",s->sfip);
	db_find_xtrn_sym_and_offset((db_addr_t) IPMASK(s->sfip),&name,&offset);
	if (name!= 0 && (unsigned)offset <= db_maxoff)
	    db_printf("%s+0x%08x",name,(unsigned)offset);
	db_printf("\n");
    }

    db_printf("vector: 0x%02x                    interrupt mask: 0x%08x\n",
	s->vector, s->mask >> 8);
    db_printf("epsr: 0x%08x                current thread: 0x%x\n",
	s->epsr, current_thread());

    /*
     * If the vector indicates trap, instead of an exception or
     * interrupt, skip the check of dmt and fp regs.
     *
     * Interrupt and exceptions are vectored at 0-10 and 114-127.
     */

    if (!((0 <= s->vector && s->vector <= 10)
       || (114 <= s->vector && s->vector <= 127)))
    {
	db_printf("\n\n");
	return;
    }

    if (s->vector == /*data*/3 || s->dmt0 & 1)
    {
	db_printf("dmt,d,a0: 0x%08x  0x%08x  0x%08x ",s->dmt0,s->dmd0,s->dma0);
	db_find_xtrn_sym_and_offset((db_addr_t) s->dma0,&name,&offset);
	if (name!= 0 && (unsigned)offset <= db_maxoff)
	    db_printf("%s+0x%08x",name,(unsigned)offset);
	db_printf("\n          ");
	surpress1 = luna_dmx_print(s->dmt0|0x01, s->dmd0, s->dma0, 0);
	db_printf("\n");

	if ((s->dmt1 & 1) && (!surpress1))
	{
	    db_printf("dmt,d,a1: 0x%08x  0x%08x  0x%08x ",s->dmt1, s->dmd1,s->dma1);
	    db_find_xtrn_sym_and_offset((db_addr_t) s->dma1,&name,&offset);
	    if (name!= 0 && (unsigned)offset <= db_maxoff)
		db_printf("%s+0x%08x",name,(unsigned)offset);
	    db_printf("\n          ");
	    surpress2 = luna_dmx_print(s->dmt1, s->dmd1, s->dma1, 1);
	    db_printf("\n");

	    if ((s->dmt2 & 1) && (!surpress2))
	    {
		db_printf("dmt,d,a2: 0x%08x  0x%08x  0x%08x ",s->dmt2, s->dmd2,s->dma2);
		db_find_xtrn_sym_and_offset((db_addr_t) s->dma2,&name,&offset);
		if (name!= 0 && (unsigned)offset <= db_maxoff)
		    db_printf("%s+0x%08x",name,(unsigned)offset);
		db_printf("\n          ");
		(void) luna_dmx_print(s->dmt2, s->dmd2, s->dma2, 2);
		db_printf("\n");
	    }
	}
    }

    if (s->fpecr & 255) /* floating point error occured */
    {
	db_printf("fpecr: 0x%08x fpsr: 0x%08x fpcr: 0x%08x\n",
	    s->fpecr,s->fpsr,s->fpcr);
	db_printf("fcr1-4: 0x%08x  0x%08x  0x%08x  0x%08x\n",
	    s->fphs1, s->fpls1, s->fphs2, s->fpls2);
	db_printf("fcr5-8: 0x%08x  0x%08x  0x%08x  0x%08x\n",
	    s->fppt,  s->fprh,  s->fprl,  s->fpit);
    }
    db_printf("\n\n");
}

/* display where all the cpus are stopped at */
static void luna_db_where()
{
    int i;
    struct m88100_saved_state *s;
    char *name;
    int *offset;

    for (i=0; i<NCPUS; i++)
	if (machine_slot[i].is_cpu && (s = db_cur_exc_frame[i]))
	{
	    int l = luna_pc(s); /* clear low bits */

	    db_find_xtrn_task_sym_and_offset((db_addr_t) l,&name,
	        (db_addr_t*)&offset,
		USERMODE(s->epsr) ? active_threads[i]->task : kernel_task);
	    if (name!=0 && offset <= db_maxoff)
		db_printf("cpu %d: stopped at 0x%x  (%s+0x%x)\n",
			i, l, name, offset);
	    else
		db_printf("cpu %d: stopped at 0x%x\n", i, l);
	}
}

#ifndef MACH_KERNEL
/* display a particular task, including 2.5 info */
void luna_db_task(addr,have_addr)
struct task *addr;
boolean_t have_addr;
{
  thread_t thread;

  if (!have_addr)
    {
      db_printf("task command needs task pointer as argument\n");
      return;
    }

  db_printf("task 0x%x:  pid = %d, name = %s ",
	    addr->u_address->uu_procp->p_pid,
	    addr, addr->u_address->uu_comm);

  if (addr->thread_count == 0)
    db_printf("no threads\n");
  else
    if (addr->thread_count > 1)
      db_printf("%d threads\n", addr->thread_count);

  queue_iterate(&addr->thread_list, thread, thread_t, thread_list)
    db_printf("thread 0x%x\n ", thread);

  return;
}
#endif /* not defined  MACH_KERNEL */

/*
 * Walk back a stack, looking for exception frames.
 * These frames are recognized by the routine frame_is_sane. Frames
 * only start with zero, so we only call frame_is_sane if the
 * current address contains zero.
 *
 * If addr is given, it is assumed to an address on the stack to be
 * searched. Otherwise, r31 of the current cpu is used.
 */
static void luna_db_frame_search(addr,have_addr)
int addr;
boolean_t have_addr;
{
#if 1
    db_printf("sorry, frame search currently disabled.\n");
#else
    if (have_addr)
	addr &= ~3; /* round to word */
    else
	addr = (DDB_REGS -> r[31]);

    /* walk back up stack until 8k boundry, looking for 0 */
    while (addr & ((8*1024)-1))
    {
	int i;
	db_read_bytes(addr,4,&i,kernel_task);
	if (i == 0 && frame_is_sane(i))
	    db_printf("frame found at 0x%x\n", i);
	addr += 4;
    }

    db_printf("(Walked back until 0x%x)\n",addr);
#endif
}

int luna_pc(regs)
struct m88100_saved_state *regs;
{
    if (regs->sxip & 2) /* is valid */
	return regs->sxip & ~3;
    if (regs->snip & 2)
	return regs->snip & ~3;
    if (regs->sfip & 2)
	return regs->sfip & ~3;
    /* we are in trouble - none of the program counters are valid */
    return 0;
}

/* flush icache of cpu number addr. If addr unspecified use, self */
static void luna_db_iflush(addr, have_addr)
int addr;
boolean_t have_addr;
{
    if (have_addr)
    {
	if (!(addr >= 0 && addr < NCPUS && machine_slot[addr].is_cpu))
	{
	    db_printf("bad cpu number %d specified\n",addr);
	    return;
	}
    }
    else
	addr = cpu_number();

    if (cmmu_cpu_lock)
    {
	db_printf("cmmu_cpu_lock held; clear to force this operation\n");
	return;
    }

    cmmu_remote_set(addr, CMMU_SCR, 0, CMMU_FLUSH_CACHE_CBI_ALL);
}

/* flush dcache of cpu number addr. If addr unspecified use, self */
static void luna_db_dflush(addr, have_addr)
int addr;
boolean_t have_addr;
{
    if (have_addr)
    {
	if (!(addr >= 0 && addr < NCPUS && machine_slot[addr].is_cpu))
	{
	    db_printf("bad cpu number %d specified\n",addr);
	    return;
	}
    }
    else
	addr = cpu_number();

    if (cmmu_cpu_lock)
    {
	db_printf("cmmu_cpu_lock held; clear to force this operation\n");
	return;
    }

    cmmu_remote_set(addr, CMMU_SCR, 1, CMMU_FLUSH_CACHE_CBI_ALL);
}

/* probe my cache */
static void luna_db_peek(addr, have_addr)
int addr;
boolean_t have_addr;
{
    int pa12;
    int cpu = cpu_number();
    int valmask;

    if (!have_addr)
    {
	db_printf("peek requires requires address\n");
	return;
    }

    if (cmmu_cpu_lock)
    {
	db_printf("cmmu_cpu_lock held; clear to force this operation\n");
	return;
    }

    pa12 = addr & ~((1<<12) -1);

    /* probe dcache */
    cmmu_remote_set(cpu, CMMU_SAR, 1, addr);

    valmask = cmmu_remote_get(cpu, CMMU_CSSP, 1);
    printf("dcache valmask 0x%x\n", (unsigned)valmask);
    printf("dcache tag ports 0x%x 0x%x 0x%x 0x%x\n",
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP0, 1),
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP1, 1),
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP2, 1),
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP3, 1));

#if 0
    if ( (pa12 == cmmu_remote_get(cpu, CMMU_CTP0, 1) &&
	(valmask & 0x3000 != 0x3000)) ||
	(pa12 == cmmu_remote_get(cpu, CMMU_CTP1, 1) &&
	(valmask & 0xb000 != 0xb000)) ||
	(pa12 == cmmu_remote_get(cpu, CMMU_CTP2, 1) &&
	(valmask & 0x30000 != 0x30000)) ||
	(pa12 == cmmu_remote_get(cpu, CMMU_CTP3, 1) &&
	(valmask & 0xb0000 != 0xb0000)) )
    {
	db_printf("line held in dcache\n");
    }
#endif

    /* probe icache */
    cmmu_remote_set(cpu, CMMU_SAR, 0, addr);

    valmask = cmmu_remote_get(cpu, CMMU_CSSP, 0);
    printf("icache valmask 0x%x\n", (unsigned)valmask);
    printf("icache tag ports 0x%x 0x%x 0x%x 0x%x\n",
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP0, 0),
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP1, 0),
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP2, 0),
    (unsigned)cmmu_remote_get(cpu, CMMU_CTP3, 0));

#if 0
    if ( (pa12 == cmmu_remote_get(cpu, CMMU_CTP0, 0) &&
	(valmask & 0x3000 != 0x3000)) ||
	(pa12 == cmmu_remote_get(cpu, CMMU_CTP1, 0) &&
	(valmask & 0xb000 != 0xb000)) ||
	(pa12 == cmmu_remote_get(cpu, CMMU_CTP2, 0) &&
	(valmask & 0x30000 != 0x30000)) ||
	(pa12 == cmmu_remote_get(cpu, CMMU_CTP3, 0) &&
	(valmask & 0xb0000 != 0xb0000)) )
    {
	db_printf("line held in icache\n");
    }
#endif
}

static void luna_db_freeze_cpu(addr,have_addr)
int addr;
boolean_t have_addr;
{
    if (!have_addr)
    {
	db_printf("freeze requires cpu number as argument\n");
	return;
    }

    if (addr >= 0 && addr < NCPUS && machine_slot[addr].is_cpu)
	if (db_frozen[addr])
	    db_printf("cpu %d already frozen\n",addr);
	else
	{
	    db_printf("freezing cpu %d\n",addr);
	    db_frozen[addr] = 1;
	}
    else
	db_printf("cpu number %d invalid\n",addr);
}

static void luna_db_thaw_cpu(addr,have_addr)
int addr;
boolean_t have_addr;
{
    if (!have_addr)
    {
	db_printf("thaw requires cpu number as argument\n");
	return;
    }

    if (addr >= 0 && addr < NCPUS && machine_slot[addr].is_cpu)
    if (!db_frozen[addr])
	db_printf("cpu %d already thawed\n",addr);
    else
    {
	db_printf("thawing cpu %d\n",addr);
	db_frozen[addr] = 0;
    }
    else
	db_printf("cpu number %d invalid\n",addr);
}

static void luna_db_status_cpu()
{
    int i;

    for (i=0; i<NCPUS; i++)
	if (machine_slot[i].is_cpu)
	    db_printf("cpu %d: %s ",i,db_frozen[i] ? "frozen" : "thawed");
	else
	    db_printf("cpu %d: missing",i);
    db_printf("\n\n");
    return;
}


static int freeze_stack[10];
static int freeze_depth = 0;

static void luna_db_hold_down()
{
    extern int db_run_mode;
    int cpu = cpu_number();
    int i;

    freeze_stack[freeze_depth] = 0;
    if (!(db_run_mode==STEP_NONE || db_run_mode == STEP_CONTINUE))
    {
	/* freeze the other cpus */
	for (i=0; i<NCPUS; i++)
	    if (i != cpu)
	    {
		freeze_stack[freeze_depth] |= db_frozen[i] << i;
		db_frozen[i] = 1;
	    }
	    else
	    {
		freeze_stack[freeze_depth] |= db_frozen[i] << i;
		db_frozen[i] = 0;
	    }
	freeze_depth++;
	NOISY(luna_db_str1("(Because in step mode is nonzero, freezing all cpus except %d)\n", cpu);)
    }
    else
    {
	/* store idempotently */
	for (i=0; i<NCPUS; i++)
	    freeze_stack[freeze_depth] |= db_frozen[i] << i;
	freeze_depth++;
    }
}

static void luna_db_release()
{
    int i;
    if (freeze_depth)
    {
	freeze_depth--;
	for (i=0; i<NCPUS; i++)
	    db_frozen[i] = (freeze_stack[freeze_depth] & 1<<i) ? 1:0;
    }
}

/*
 * control how much info the debugger prints about itself
 */

static void luna_db_noise(addr,have_addr)
int addr;
boolean_t have_addr;
{
    if (!have_addr)
    {
	/* if off make noisy; if noisy or very noisy turn off */
	if (db_noisy)
	{
	    db_printf("changing debugger status from %s to quiet\n",
		db_noisy == 1 ? "noisy" :
		db_noisy == 2 ? "very noisy" : "violent");
	    db_noisy = 0;
	}
	else
	{
	    db_printf("changing debugger status from quiet to noisy\n");
	    db_noisy = 1;
	}
    }
    else
	if (addr < 0 || addr > 3)
	    db_printf("invalid noise level to luna_db_noisy; should be 0, 1, 2, or 3\n");
	else
	{
	    db_noisy = addr;
	    db_printf("debugger noise level set to %s\n",
		db_noisy == 0 ? "quiet" :
		(db_noisy == 1 ? "noisy" :
		db_noisy==2 ? "very noisy" : "violent"));
	}
}

/*
 * Check access to that address in user space on the
 * current processor.
 */

unsigned int db_check_user_addr(va)
vm_offset_t va;
{
    vm_offset_t pa;
    return db_task_va_to_phys(va, current_thread()->task, &pa);
}

/*
 * Check if the two (task,va) pairs map to the same
 * address
 */
unsigned int db_phy_eq(t1,va1,t2,va2)
task_t t1, t2;
vm_offset_t va1, va2;
{
    vm_offset_t pa1, pa2;
    if (db_task_va_to_phys(va1,t1,&pa1) &&
	db_task_va_to_phys(va2,t2,&pa2) &&
	(pa1 == pa2))
	    return 1;
    return 0;
}

unsigned int db_check_access(task,va,size)
task_t task;
vm_offset_t va;
unsigned int size;
{
    vm_offset_t pa;

    while (size)
    {
	if (!db_task_va_to_phys(va,task,&pa))
	    return 0;
	size--;
	va++;
    }
    return 1;
}
