/*
 * Mach Operating System
 * Copyright (c) 1992, 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	db_machdep.h,v $
 * Revision 2.7  92/05/21  17:18:21  jfriedl
 * 	Appended 'U' to constants with high bit on.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.6  92/05/04  11:26:33  danner
 * 	Set DB_ACCESS_LEVEL to DB_ACCESS_ANY.
 * 	[92/04/18            danner]
 * 
 * Revision 2.5  92/04/01  10:54:37  rpd
 * 	Made suitable for inclusion in assembly files.
 * 	[92/03/03            danner]
 * 
 * Revision 2.4  92/02/18  18:00:52  elf
 * 	Changed BKPT_INST, ENTRY_ASM to be trap 144, which is excutable
 * 	 from user space.
 * 	[92/02/13            danner]
 * 
 * 	Several new macros to support new ddb.
 * 	[92/02/12            danner]
 * 	Updated include of trap.h
 * 	[92/01/30            danner]
 * 	Moved redefinitions of _longjmp and _setjmp under 2.5
 * 	 conditionals.
 * 	[92/01/19            danner]
 * 
 * Revision 2.3  91/08/24  12:07:20  af
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:49  danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:46:16  danner
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:49  danner]
 * 
 * Revision 2.3  91/04/10  16:05:47  mbj
 * 	Added setjmp, longjmp defined to make ddb use the kernel
 * 	 versions, not the user level versions.
 * 	[91/03/31            danner]
 * 
 * Revision 2.2  91/04/05  14:01:53  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * 	Created
 * 	[91/01/11            danner]
 * 
 */

/*
 * Machine-dependent defined for the new kernel debugger 
 */

#ifndef  LUNA_DB_MACHDEP_H_
#define  LUNA_DB_MACHDEP_H_ 1

#define BKPT_SIZE        (4)                /* number of bytes in bkpt inst. */
#define BKPT_INST       (0xF000D082U)             /* tb0, 0,r0, vector 132 */
#define BKPT_SET(inst)  (BKPT_INST)

/* Entry trap for the debugger - used for inline assembly breaks*/
#define ENTRY_ASM       "tb0	0, r0, 132"
#define DDB_ENTRY_TRAP_NO	132

#include <cpus.h>

#include <mach/machine/vm_types.h>
#include <mach/machine/vm_param.h>
#include <luna88k/thread.h>                /* for luna88k_saved_state */
#include <motorola/m88k/trap.h>                          /* for T_KDB_BREAK */

typedef vm_offset_t   db_addr_t;
typedef int           db_expr_t;
typedef struct m88100_saved_state db_regs_t;

extern db_regs_t      *db_cur_exc_frame[NCPUS];     /* register state for debugger */
extern int             ddb_master;

extern int db_noisy;
#define NOISY(x) if (db_noisy) x
#define NOISY2(x) if (db_noisy >= 2) x
#define NOISY3(x) if (db_noisy >= 3) x



/* get the registers from the master cpu */
#define  DDB_REGS   (db_cur_exc_frame[ddb_master])
/*
  the low two bits of sxip, snip, sfip have valid bits
  in them that need to masked to get the correct addresses */

/* This is an actual function due to the fact that the sxip 
   or snip could be nooped out due to a jmp or rte */
#define  PC_REGS(regs) (luna_pc(regs))

/* These versions are not constantly poing SPL */
#define  cnmaygetc()    db_getc()
#define  cngetc()       db_getc()
#define  cnputc(x)      db_putc(x)

/* breakpoint/watchpoint foo */

#define IS_BREAKPOINT_TRAP(type,code) (((type)==T_KDB_BREAK))
#define IS_WATCHPOINT_TRAP(type,code) ((0))

#define SOFTWARE_SSTEP 1                  /* software single steping */

/*
  Debugger can get to any address space 
*/

#define DB_ACCESS_LEVEL DB_ACCESS_ANY

#define DB_VALID_KERN_ADDR(addr) (!badaddr((void*)(addr), 1))
#define DB_VALID_ADDRESS(addr,user) \
  (user ? db_check_user_addr(addr) : DB_VALID_KERN_ADDR(addr))

#define DB_PHYS_EQ(task1,addr1,task2,addr2) \
  db_phy_eq(task1,addr1,task2,addr2)

#define DB_CHECK_ACCESS(addr,size,task) db_check_access(task,addr,size)

/* instruction type checking - others are implemented in db_sstep.c */

#define inst_trap_return(ins)            (ins == 0xf400fc00U) 


/* don't need to load symbols */
#define DB_SYMBOLS_PRELOADED 1

/* softcall vector for softint */
#define LUNA_DDB_SOFTCALL 3

#ifndef MACH_KERNEL
/* redefinitions for 2.5 */
#  define ipc_port_print port_print
#define _longjmp longjmp
#define _setjmp  setjmp
#endif

/* machine specific commands have been added to ddb */
#define DB_MACHINE_COMMANDS 1

#endif LUNA_DB_MACHDEP_H_
