 /*
 * Mach Operating System
 * Copyright (c) 1991, 1992 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	db_trace.c,v $
 * Revision 2.10  92/08/03  17:41:31  jfriedl
 * 	Header file updates. [danner]
 * 
 * Revision 2.9  92/05/21  17:18:35  jfriedl
 * 	gcc warning cleanup.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.8  92/05/04  11:26:39  danner
 * 	Added 'f' flag to the stack trace command. When f is specifed
 * 	 (t/f), the stack pointer of each call frame is printed. 
 * 	[92/04/27            danner]
 * 	Support for tracing code without symbols. More AI in the
 * 	 kernel....
 * 	[92/04/13            danner]
 * 	Simplify (and correct) the address space selection code for
 * 	recursive stack trace calls.
 * 	[92/04/10            danner]
 * 
 * Revision 2.7  92/04/01  10:54:41  rpd
 * 	Added t/t, t/tu functionality.
 * 	[92/03/24            danner]
 * 	Switch use of TASK_NULL to kernel_task in stack trace code.
 * 	[92/03/20            danner]
 * 
 * 	Changed last usages of db_printsym to db_task_printsym.
 * 	[92/03/20            danner]
 * 
 * Revision 2.6  92/03/03  15:38:21  rpd
 * 	Now traces user space. Made much less verbose.
 * 	[92/03/02            danner]
 * 
 * Revision 2.5  92/02/19  13:50:01  elf
 * 	Additional bug fixes.
 * 	[92/01/27            danner]
 * 
 * 	Modified exception frame identifing code to take advantage of the
 * 	 duplicate frame pointer generated in locore. Had sign backwards
 * 	      in store double test in argument recovery code.
 * 	[92/01/23            danner]
 * 	Added code to recover arguments that are saved on the stack.
 * 	[92/01/22            danner]
 * 	Db_stack_trace_cmd now happily walks over multiple exception
 * 	 frames on a stack. It will also attempt to find the user
 * 	 exception frame, which no longer resides on the stack.
 * 	[91/09/05            danner]
 * 
 * Revision 2.4  91/10/09  16:11:53  af
 * 	Added parens for initializers to db_regs[].
 * 	[91/10/07            af]
 * 
 * Revision 2.3  91/08/24  12:08:45  af
 * 	Conditionalized pid field on ifndef MACH_KERNEL. Modified stack
 * 	 trace command to recognize integers 0 - 3 as references to the
 * 	 exception frame of the specified cpu.
 * 	[91/08/08            danner]
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:28  danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:46:23  danner
 * 	Conditionalized pid field on ifndef MACH_KERNEL. Modified stack
 * 	 trace command to recognize integers 0 - 3 as references to the
 * 	 exception frame of the specified cpu.
 * 	[91/08/08            danner]
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:28  danner]
 * 
 * Revision 2.3  91/04/10  16:05:54  mbj
 * 	Added stack tracing code. Hairy.
 * 	[91/03/28            danner]
 * 
 * Revision 2.2  91/04/05  14:02:04  mbj
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * 	Created
 * 	[91/01/08            danner]
 * 
 */

/*
 * luna trace/register state interface for ddb.
 */

#include <kern/task.h>
#include <mach/boolean.h>
#include <machine/db_machdep.h>
#include <ddb/db_variables.h>
#include <ddb/db_sym.h>
#include <machine/setjmp.h>



extern jmp_buf_t *db_recover;

/* lifted from mips */

static boolean_t
db_setf_regs(vp, valuep, op)
	db_expr_t	*valuep;
	struct db_variable *vp;
	int		op;
{
	register int   *regp;

	regp = (int *) ((char *) DDB_REGS + (int) (vp->valuep));

	if (op == DB_VAR_GET) {
		*valuep = *regp;
	} else if (op == DB_VAR_SET) {
		*regp = *valuep;
	} else
		return FALSE;
	return TRUE;
}

#define N(s,x)  {s, (int *)&(((db_regs_t *) 0)->x), db_setf_regs}

struct db_variable db_regs[] = {
  N("r1",r[1]),
  N("r2",r[2]),
  N("r3",r[3]),
  N("r4",r[4]),
  N("r5",r[5]),
  N("r6",r[6]),
  N("r7",r[7]),
  N("r8",r[8]),
  N("r9",r[9]),
  N("r10",r[10]),
  N("r11",r[11]),
  N("r12",r[12]),
  N("r13",r[13]),
  N("r14",r[14]),
  N("r15",r[15]),
  N("r16",r[16]),
  N("r17",r[17]),
  N("r18",r[18]),
  N("r19",r[19]),
  N("r20",r[20]),
  N("r21",r[21]),
  N("r22",r[22]),
  N("r23",r[23]),
  N("r24",r[24]),
  N("r25",r[25]),
  N("r26",r[26]),
  N("r27",r[27]),
  N("r28",r[28]),
  N("r29",r[29]),
  N("r30",r[30]),
  N("r31",r[31]),
#ifndef MACH_KERNEL
  N("pid",pid),
#endif
  N("epsr",epsr),
  N("sxip",sxip),
  N("snip",snip),
  N("sfip",sfip),
  N("ssbr",ssbr),
  N("dmt0",dmt0),
  N("dmd0",dmd0),
  N("dma0",dma0),
  N("dmt1",dmt1),
  N("dmd1",dmd1),
  N("dma1",dma1),
  N("dmt2",dmt2),
  N("dmd2",dmd2),
  N("dma2",dma2),
  N("fpecr",fpecr),
  N("fphs1",fphs1),
  N("fpls1",fpls1),
  N("fphs2",fphs2),
  N("fpls2",fpls2),
  N("fppt",fppt),
  N("fprh",fprh),
  N("fprl",fprl),
  N("fpit",fpit),
  N("fpsr",fpsr),
  N("fpcr",fpcr),
  N("mask",mask),    /* interrupt mask */
  N("mode",mode),    /* interrupt mode */
  N("exvc",vector),  /* exception vector */
};

#undef N

struct db_variable *db_eregs = db_regs + sizeof(db_regs)/sizeof(db_regs[0]);

/*
  returns 
  	1 if this looks like a reasonable kernel exception frame.
        2 if this looks like a reasonable user exception frame (in
	the current task).
	0 if this looks like neither.
*/

int frame_is_sane(regs)
db_regs_t *regs;
{
  unsigned int epsr;
  /* 
    Heuristics:

    We check 
             r[0]=0, 
	     r[31] & 3 = 0 (stack is aligned)
	     sxip & 2  = 2 (pc is valid)
	     epsr & EFFFFFF5 = 0x800003f0 (kernel mode)
	     epsr & EFFFFFFF = 0x000003f0 (user mode)


	     We first check to make sure the whole frame is accessible.
  */

  /* Frame accessibility */
  if (badwordaddr(regs) || badwordaddr(&regs->mode))
    return 0;

  /* r[0], stack, sxip,  sanity */

  if ((regs->r[0] != 0) 
      || ((regs->r[31] & 3) != 0) 
      /* || ((regs->sxip & 2) != 2) */)
    return 0;
  
  /* epsr sanity */
  epsr = regs->epsr; 
  if ((epsr & 0xEFFFFFF5U) == 0x800003f0U) /* kernel mode */
    return 1;
  if ((epsr & 0xEFFFFFFFU) == 0x000003f0U) /* user mode */
    return 2;
  return 0;
}
    
static char *exception_cause(vector)
unsigned int vector;
{
  char *cause;

  switch  (vector)
    {
    case 0:
      cause = "unknown";
      break;
    case 1:
      cause = "Interrupt";
      break;
    case 2:
      cause = "Instruction Access Exception";
      break;
    case 3:
      cause = "Data Access Exception";
      break;
    case 4:
      cause = "Misaligned Access Exception";
      break;
    case 5:
      cause = "Unimplemented Opcode Exception";
      break;
    case 6:
      cause = "Privilege Violation";
      break;
    case 7:
      cause = "Bounds Check";
      break;
    case 8:
      cause = "Integer Divide Exception";
      break;
    case 9:
      cause = "Integer Overflow Exception";
      break;
    case 10:
      cause = "Error Exception";
      break;
    case 64:
      cause = "Ddb break";
      break;
    case 65:
      cause = "Ddb trace";
      break;
    case 66:
      cause = "Ddb trap";
      break;
    case 114:
      cause = "FPU precise";
      break;
    case 115:
      cause = "FPU imprecise";
      break;
    case 128:
      cause = "Syscall";
      break;
    case 129:
      cause = "Mach syscall";
      break;
    default:
      cause = "Unknown trap";
      break;
    }
  return cause;
}

/*
 * Read a word from address space tk at address addr.
 * Return 1 if was able to read, 0 otherwise.
 */
static unsigned int db_trace_get_val(addr,ptr,tk)
vm_offset_t addr;
unsigned int *ptr;
task_t tk;
{
  jmp_buf_t db_jmpbuf;
  jmp_buf_t *prev = db_recover;

  if (_setjmp(db_recover = &db_jmpbuf) != 0)
    {
      db_recover = prev;
      return 0;
    }
  else
    db_read_bytes((char*)addr, 4, (char*)ptr, tk);

  db_recover = prev;
  return 1;
}

static void db_print_args(args,args_mask)
unsigned int *args;
unsigned int args_mask;
{
  int i,j= -1;

  db_printf("(");
  for (i=0; i < 8; i++)
    if (args_mask & (1<<i))
      j = i;
  for (i=0; i <= j; i++)
    {
      if (args_mask & (1<<i))
	db_printf("0x%x",args[i]);
      else
	db_printf("?");
      if (i<j)
	db_printf(",");
    }
  db_printf(")");
  return;
}

/*
  stack trace - needs a pointer to a luna88k saved state. If none is given,
  assume the current cpu. If given a number between 0 and 3, it traces
  that cpus.

  If argument t is given, it traces that thread starting from its 
  saved pcb state.

  If argument f is given, the stack pointer of each call frame is
  printed.
*/

void
db_stack_trace_cmd(addr, have_addr, count, modif)
	db_regs_t	*addr;
	int		have_addr;
	db_expr_t	count;
	char		*modif;
{
  int trace_user = 0;
  int spec_thread = 0;
  int print_frames = 0;
  task_t tk;

  db_regs_t *regs;
  union {
    db_regs_t *d;
    thread_t t;
    unsigned int i;
  } a;

  a.d = addr;

  if (modif != ((char *) 0))
    for ( ; *modif != '\0'; modif++)
      switch (*modif)
	{
	case 'u':
	  trace_user = 1;
	  break;
	case 't':
	  spec_thread = 1;
	  break;
	case 'f':
	  print_frames = 1;
	default:
	  db_printf("unknown modified %c passed to trace command\n",
		    *modif);
	  break;
	}

  if (spec_thread && have_addr)
    {
      db_regs_t frame; /* a m88100_saved_state */
      struct m88100_pcb *p; 
      
      if (badwordaddr(a.t) || badwordaddr(&a.t->swap_func) ||
	  badwordaddr(&a.t->pcb->kernel_state))
	{
	  db_printf("bad thread pointer given to stack trace cmd\n");
	  return;
	}

      /* if the thread is blocked with a continuation, 
	 we can only do a user space trace */
      if ((a.t)->swap_func) /* blocked with a continuation */
	{
	  db_printf("thread 0x%x blocked with continuation; ",a.t);	
	  if (trace_user)
	    {
	      regs = &((a.t)->pcb->user_state);
	      db_printf("only tracing user space\n");
	      db_printf("---------------- %s [EF : 0x%x] -------------\n",
			exception_cause(regs->vector), regs);
	    }
	  else
	    {
	      db_printf("trace aborted\n");
	      return;
	    }
	}
      else
	{
	  /* translate from saved state into exception frame format */
	  p = &(a.t->pcb->kernel_state);
	  frame.epsr = 0x800003f0U;
	  frame.sxip = p->pcb_pc | 2;
	  frame.snip = 0;
	  frame.sfip = 0;
	  bzero((char*)frame.r,sizeof(int)*32);
	  /* snag r14 - r30 */
	  bcopy((void*)&(p->pcb_r14), &(frame.r[14]), sizeof(int)*(30-14+1));
	  /* an stack pointer */
	  frame.r[31] = p->pcb_sp; 
	  
	  regs = &frame;
	}
      tk = a.t->task;
    }
  else
    {
      if (current_thread())
	tk = current_thread()->task;
      else
	tk = TASK_NULL;
      
      if (!have_addr)
	regs = DDB_REGS;
      else
	if ((0 <= a.i) && (a.i <=  3))
	  regs = (db_regs_t *) db_cur_exc_frame[a.i];
	else
	  regs = (db_regs_t *) addr;
    }

  db_stack_trace_cmd2(regs, tk, trace_user, print_frames);
}
    

/* How much slop we expect in the stack trace */

#define FRAME_PLAY 8

static void db_stack_trace_cmd2(regs, task, trace_user,print_frames)
  db_regs_t *regs;
  task_t task;
  int trace_user; /* trace into user space? */
  int print_frames;
{
  unsigned int stack;
  unsigned int depth=1;
  unsigned int where;
  unsigned int ft;
  unsigned int args[8];
  unsigned int pair[2];
  unsigned int args_mask;
  task_t   tk;
  int i;

  if(!(ft=frame_is_sane(regs)))
    {
      db_printf("Register frame 0x%x is suspicous; skipping trace\n",
		regs);
      return;
    }

  /* if user space and no user space trace specified, puke */
  if (ft == 2 && !trace_user)
    return;

  tk = (ft == 2 ? task : kernel_task /* TASK_NULL */);

  /* fetch address */
  /* use sxip if valid, otherwise try snip or sfip */ 
  where = ((regs->sxip & 2) ? regs->sxip : 
	   ((regs->snip & 2) ? regs->snip : regs->sfip) ) & ~3;
  stack = regs->r[31];
  db_printf("stack base = 0x%x\n",stack);
  db_printf("(0) "); 
  db_task_printsym(where,DB_STGY_PROC,tk);
  /* see if this routine had a stack frame */
  if ((where=stack_decode(where,&stack,args,&args_mask,tk))==0)
    {
      where = regs->r[1];
      db_printf("(stackless)");
    }
  else
    {
      db_print_args(args,args_mask);
      if (print_frames)
	db_printf(" [frame 0x%x]",stack);
    }
  db_printf("\n");

  while(1)
      {
	db_printf("(%d) ",depth++);
	db_task_printsym(where,DB_STGY_PROC,tk);
	where=stack_decode(where,&stack,args,&args_mask,tk);
	db_print_args(args,args_mask);
	if (print_frames)
	  db_printf(" [frame 0x%x]",stack);
	db_printf("\n");
	if (where==0) break;
      }

  /* try to trace back over trap/exception */

  stack &= ~7; /* double word aligned */
  /* take last top of stack, and try to find an exception frame near it */
      
  i = FRAME_PLAY;
  /* db_printf("(searching for exception frame at 0x%x\n",stack); */
  while (i)
    {

      /* On the stack, a pointer to the exception frame is written
	 in two adjacent words. In the case of a fault from the kernel,
	 this should point to the frame right above them:

	Exception Frame Top
	..
	Exception Frame Bottom  <-- frame addr
	frame addr
	frame addr		<-- stack pointer 
	
	In the case of a fault from user mode, the top of stack
	will just have the address of the frame (off in the thread pcb)
	replicated twice.

	frame addr		<-- top of stack
	frame addr

	Here we are just looking for kernel exception frames. 
	*/
	

      if (badwordaddr((void*)stack) || badwordaddr((void*)(stack+4)))
	break;

      db_read_bytes((char*)stack, 2*sizeof(int), (char*)pair, kernel_task);
      /* 
	the pairs should match and equal stack+8
      */
      if ((pair[0] == pair[1]))
	{
	  if (pair[0] != stack+8)
	    {
	      /*
		if (!badwordaddr((void*)pair[0]) && (pair[0]!=0))
		db_printf("stack_trace:found pair 0x%x but != to stack+8\n",
		pair[0]);
		*/
	    }
	  else
	    if(frame_is_sane((db_regs_t*)pair[0]))
	      {
		db_regs_t *frame = (db_regs_t *) pair[0];
		char *cause = exception_cause(frame -> vector);
		
		db_printf("-------------- %s [EF: 0x%x] -------------\n",
			  cause, frame);
		db_stack_trace_cmd2(frame, task, trace_user,print_frames);
		return;
	      }
	  /*
	    else
	      db_printf("pair matched, but frame at 0x%x looks insane\n",
			stack+8);
  	  */
	}
      stack += 8;
      i--;
    }
	      
  /* if we go here, crawling back on the stack failed to find us
     a previous exception frame. Look for a user frame pointer 
     pointed to by a word 8 bytes off of the top of the stack 
     if the "u" option was specified. */

  if (trace_user)
    {
      db_regs_t *user;

      /* Make sure we are back on the right page */
      stack -= 4*FRAME_PLAY; 
      stack = stack & ~(KERNEL_STACK_SIZE-1); /* point to the bottom */
      stack += KERNEL_STACK_SIZE - 8;

      if (badwordaddr((void*)stack) || badwordaddr((void*)stack))
	  return;
 
      db_read_bytes((char*)stack, 2*sizeof(int), (char*)pair, kernel_task);
      if (pair[0] != pair[1])
	return;

      /* have a hit */
      user = *((db_regs_t **) stack);
/*
      db_printf("checking stack link 0x%x for user ef pointer (*link=0x%x)\n",
		stack, user);
*/
      
      if (frame_is_sane(user) == 2)
	{
	  /* db_printf("[found pointer to user EF at top of stack (0x%x)]\n",stack);*/
	  db_printf("---------------- %s [EF : 0x%x] -------------\n",
		    exception_cause(user->vector), user);
	  db_stack_trace_cmd2(user,task,1,print_frames);

	}
    }
  
  return;
}

/*
  Stack decode -
  	unsigned int addr;    program counter 
	unsigned int *stack; IN/OUT stack pointer 
	unsigned int *argp;  OUT: pointer to argument save area
	unsigned int *args_found; OUT: pointer to int to tell which args 
			where found.

  	given an address within a function and a stack pointer,
	try to find the function from which this one was called
	and the stack pointer for that function.

	In addition, we attempt to recover the arguments for this
	function (registers r2-r9). If we can find any on the stack,
	there values are recorded in argp (register 2 in argp[0],
	etc) and the corresponding bits are set in args_found. 
	
	The return value is zero (if we get confused) or 
	we determine that the return address has not yet
	been saved (early in the function prologue). Otherwise
	the return value is the address from which this function
	was called. 

	Note that even is zero is returned (the second case) the
	stack pointer can be adjusted.
	
	
*/


static int stack_decode(addr,stack,argp,args_found,tk)
unsigned int addr, *stack;
unsigned int *argp; /* pointer to argument save area */
unsigned int *args_found; /* pointer to bit mask of arguments found */
task_t tk; /* what address space */
{
  int off;
  db_sym_t proc =  db_search_task_symbol(addr,DB_STGY_PROC,&off,tk);
  char *names;
  unsigned int val;
  unsigned int instr;
  unsigned int of, fs; /* offset, frame_size */
  unsigned int st;
  unsigned int ret; /* instuction returning to */
  unsigned int i;

  if (off == addr) 
    proc = DB_SYM_NULL;

#if 0
  /* initialize to no arguments found */
  if (badwordaddr(args_found))
    {
      db_printf("bad args_found argument to stack_decode\n");
      return 0;
    }
#endif 
  *args_found = 0;
  
#if 0
  if (badwordaddr(argp) || badwordaddr(argp+7))
    {
      db_printf("bad args_found argument to stack_decode\n");
      return 0;
    }
#endif

  if (proc != DB_SYM_NULL)
    {
      db_symbol_values(proc,&names,&val);
      if (names == 0)
	return 0;
#ifdef DEBUG
      db_printf("name %s address 0x%x\n",names,val);
#endif
    }
  else
    { /* hmm - unable to find symbol. Search back the last 400
	 instructions looking for a function prolog.
      */
      for (val=addr, i=0; i < 400; i++, val -= 4)
	{
	  if (db_trace_get_val(val,&instr,tk)==0)
	    break;
	  if ((instr & 0xffff0000U) == 0x67ff0000U) /* subu r31, r31, IMM */
	    {
	      /* good - found first instruction */
	      if (db_trace_get_val(val+4,&instr,tk)==0)
		continue;
	      if ((instr & 0xffff0000U) == 0x243f0000U) /* st r1, r31, IMM16 */
		break; /* sucess */
	    }
	  /* if we come across a jmp.n r1 assume we have hit the
	     previous functions epilogue and stop our search */
	  if (instr == 0xf400c401U) /* jmp.n r1 */
	    break;
	}
    }

  /* val now has the address of the first address of the function */
  /* get the instruction - see if it a stack frame setup */

  if (db_trace_get_val(val,&instr,tk)==0)
    return 0;

#ifdef DEBUG
  db_printf("instruction at 0x%x (0x%08x)",val,instr);
  db_disasm(val,0);
  db_printf("\n");
#endif

  if ((instr & 0xffff0000U) == 0x67ff0000U) /* subu r31, r31, IMM */
    fs = instr & 0x0000ffffU;  /* size of frame */
  else
    return 0; 

  /* get the next instruction - see if it a return address store */
  if (db_trace_get_val(val+4,&instr,tk)==0)
    return 0;

#ifdef DEBUG
  db_printf("instruction at 0x%x (0x%08x)",val+4,instr);
  db_disasm(val+4,0);
  db_printf("\n");
#endif

  if ((instr & 0xffff0000U) == 0x243f0000U) /* st r1, r31, IMM16 */
    of = instr & 0x0000ffffU;
  else
    return 0;

#ifdef DEBUG
  db_printf("frame size = 0x%x, offset 0x%x\n",fs,of);
#endif

  /* check that both are word aligned */
  if ((of & 3) || (fs & 3))
    return 0;

  /* if we were stopped with no offset, return 0 - haven't gotten
     far enough to do damage */
  if (off == 0)
    return 0;

  /* if we were stopped after setting up the stack but before we
     stored our return address, just knock off this fractional 
     frame and return */

  if (off == 4)
    {
      *stack += fs;
      return 0;
    }

  /* look back into the stack, see if the address makes sense */
  st = *stack;
  if (db_trace_get_val(st+of,&ret,tk)==0)
    return 0;

#ifdef DEBUG
  db_printf("calling sane_jump with ret = 0x%x val = 0x%x\n",ret,val);
#endif
  if (!sane_jump(ret,val,tk)) 
    return 0;

  /* 
    Frame looks reasonable. 
    Lets see if we can find any function arguments.
    */

  /* currently we found */

  /* val is current pointing the to the entry of the function.
     We look forward through the instruction stream looking
     for stores of r2-r9 relative to r31 of the form

     st{.d} r<n>, r31, IMM16

     We stop when we see the first instruction that is not a 
     store immediate off r31 or we reach the instruction <addr>.
     We skip the first two instructions of the function 
     which have already been examined above.

     We also skip over frame pointer (R30) adjustments of the form

     addu	r30, r31, IMM16

  */

  for (i=8; val+i < addr; i += 4)
    {
      unsigned int doub, reg, offset;

      if (db_trace_get_val(val+i,&instr,tk)==0)
	return 0;

      /* skip addu r30, r31, IMM16 */ 
      if ((instr & 0xffff0000U) == 0x63df0000U)
	continue;

      if (instr & 0xf81f0000U != 0x201f0000U) /* st{.d} r<n>,r31, IMM16 */
	break;
      /* extract register being stored */
      reg = (instr & 0x03e00000U) >> 21;
      offset = instr & 0xffff;
      doub  = instr & 0x0c000000U;
      if (reg < 2 || reg > 9) /* a store, but not of an argument */
	continue;
      /* offset should be word aligned */
      if (offset & 0x3)
	break;
      if (db_trace_get_val(st+offset,&(argp[reg-2]),tk)==0)
	break;
      *args_found |= 1<<(reg-2);
      if (!doub && reg < 9) /* double word store */
	{
	  if (db_trace_get_val(st+offset+4,&(argp[reg-1]),tk)==0)
	    break;
	  *args_found |= 1<<(reg-1);
	}
    } 

  
  /* decrement stack, return ret */
  *stack += fs; /* stacks grow down - shrink up */
  return ret;
}

/* 
 * int sane_jump(ret,val)
 * See if the instruction (or two) before
 * ret could have been a subroutine call to val
 */
static int sane_jump(ret,val,tk)
     unsigned int ret;
     unsigned int val;
     task_t tk;
{
  unsigned int inst;
  unsigned int i;
  
  for (i=2; i>0; i--)
    {
      if (db_trace_get_val(ret-i*4,&inst,tk)==0)
	return 0;
#ifdef DEBUG
      db_printf("sj: instruction -%d ",i);
      db_disasm(ret-i*4,0);
      db_printf("\n");
#endif
      if (inst_call(inst))
	{
	  /* if i==1, better be nondelayed form */
	  if (i==1)
	    {
	      if ((inst & 0xf8000000U) == 0xc8000000U) /* bsr */
		{
		  if ((inst & 0x04000000U) != 0 )
		    {
#ifdef DEBUG
		      db_printf("delayed when undelayed desired\n");
#endif
		      continue;
		    }
		}
	      else /* jsr */
		if ((inst & 0x400) != 0)
		  {
#ifdef DEBUG
		    db_printf("delayed when undelayed desired\n");
#endif
		    continue;
		  }
	    }
	  else /* i==2 better be delayed form */
	    {
	      if ((inst & 0xf8000000U) == 0xc8000000U) /* bsr */
		{
		  if ((inst & 0x04000000U) == 0 )
		    {
#ifdef DEBUG
		      db_printf("undelayed when delayed desired\n");
#endif
		      continue;
		    }
		}
	      else /* jsr */
		if ((inst & 0x400) == 0)
		  continue;
	    }
	  /* if branch instruction, check if jump is to the correct place */
	  if ((inst & 0xf8000000U) == 0xc8000000U) /* bsr */
	    {
	      unsigned int disp = inst & ((1<<26)-1);
	      unsigned int off;
	      
	      if (disp & 1<<25)
		off = ret-i*4 + ((disp << 2) | (~0 << 26));
	      else
		off = ret-i*4 + (disp<<2);
	      if (off != val)
		{
#ifdef DEBUG
		  db_printf("-%d was bsr, target wrong (off 0x%x tar 0x%x\n"
			    ,i,off,val);
#endif DEBUG
		  continue;
		}
#ifdef DEBUG
	      db_printf("good sane jump\n");
#endif
	      return 1;
	    }
	  else /* jsr, assume to the right place */
	    return 1;
	} /* end if (inst_call()) */
    } /* end for */
  return 0;
}
