/* 
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon 
 * the rights to redistribute these changes.
 */

/*
 * HISTORY
 * $Log:	genassym.c,v $
 * Revision 2.6  92/05/21  17:18:42  jfriedl
 * 	     Cleaned up and de-linted.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.5  92/02/18  18:01:26  elf
 * 	Added DISP_MIN to correspond to disp_min field added in the
 * 	 emulation vector definition introducted in MK68.
 * 	[92/01/17            danner]
 * 
 * Revision 2.4  91/08/24  12:09:54  af
 * 	Updated to reflect new exception frame thread /
 * 	 luna88k_saved_state structure. Added sanity checking code to
 * 	 make sure the thread structure is an integral multiple of double
 * 	 words. 
 * 	[91/08/07            danner]
 * 
 * Revision 2.3.3.1  91/08/19  13:46:35  danner
 * 	Updated to reflect new exception frame thread /
 * 	 luna88k_saved_state structure. Added sanity checking code to
 * 	 make sure the thread structure is an integral multiple of double
 * 	 words. 
 * 	[91/08/07            danner]
 * 
 * Revision 2.3  91/07/11  11:00:46  danner
 * 	Copyright Fixes
 * 
 * Revision 2.2  91/07/09  23:17:22  danner
 * 	Added generation of ENTRY_ASM, the assembly instruction to
 * 	 enter the debugger.
 * 	[91/07/08            danner]
 * 
 * 	Added THREAD_SWAP_FUNC for switch_to_shutdown_context
 * 	[91/06/25            danner]
 * 
 */

#include <mach/machine/thread_status.h>
#include <motorola/m88k/m88100/pcb.h>
#include <kern/thread.h>
#include <kern/task.h>
#include <kern/syscall_emulation.h>
#include <vm/vm_map.h>
#include <luna88k/pmap.h>
#include <luna88k/thread.h>
#include <mach/message.h>
#include <luna88k/db_machdep.h>

int main()
{
	struct m88100_saved_state *ss = (struct m88100_saved_state *) 0;
	struct thread *	thread = (struct thread *) 0;
	struct task *	task = (struct task *) 0;
	pcb_t	        pcb = (pcb_t) 0;
	eml_dispatch_t	eml = (eml_dispatch_t) 0;

        printf("#ifndef __GENASSYM_INCLUDED\n");
        printf("#define __GENASSYM_INCLUDED 1\n\n");

	printf("#define\tTHREAD_KERNEL_STACK %d\n", 	&thread->kernel_stack);
	printf("#define\tTHREAD_SWAP_FUNC %d\n", 	&thread->swap_func);
	printf("#define\tTHREAD_PCB %d\n", 		&thread->pcb);
	printf("#define\tTHREAD_RECOVER %d\n", 		&thread->recover);
	printf("#define\tTHREAD_TASK %d\n", 		&thread->task);

	printf("#define\tTASK_EMUL %d\n",	&task->eml_dispatch);

	printf("#define\tDISP_COUNT %d\n",	&eml->disp_count);
	printf("#define\tDISP_MIN %d\n",	&eml->disp_min);
	printf("#define\tDISP_VECTOR %d\n",	&eml->disp_vector[0]);

	printf("#define\tPCB_USER  %d\n",        &(pcb->user_state));
	printf("#define\tPCB_KERNEL %d\n",        &(pcb->kernel_state));
	
	/* 
           Needed in the exception handler. These are in ifdef assembler
           Because they are duplications of constants found in other headers
           that can't be included by assembly routines. 
        */
	printf("#ifdef ASSEMBLER\n");
	printf("#define\tMACH_SEND_INTERRUPTED %d\n", MACH_SEND_INTERRUPTED);
        printf("#define\tENTRY_ASM\t%s\n",ENTRY_ASM);
	printf("#endif /* ASSEMBLER */\n");

	/* these were all originally hard coded in m88k.h */
	printf("#define\tEF_R0 %d\n",	((int) &(ss->r[0]))/sizeof(int));
	printf("#define\tEF_R31 %d\n",	((int) &(ss->r[31]))/sizeof(int));
	printf("#define\tEF_FPSR %d\n",	((int) &(ss->fpsr))/sizeof(int));
	printf("#define\tEF_FPCR %d\n",	((int) &(ss->fpcr))/sizeof(int));
	printf("#define\tEF_EPSR %d\n",	((int) &(ss->epsr))/sizeof(int));
	printf("#define\tEF_SXIP %d\n",	((int) &(ss->sxip))/sizeof(int));
	printf("#define\tEF_SFIP %d\n",	((int) &(ss->sfip))/sizeof(int));
	printf("#define\tEF_SNIP %d\n",	((int) &(ss->snip))/sizeof(int));
	printf("#define\tEF_SSBR %d\n",	((int) &(ss->ssbr))/sizeof(int));
	printf("#define\tEF_DMT0 %d\n",	((int) &(ss->dmt0))/sizeof(int));
	printf("#define\tEF_DMD0 %d\n",	((int) &(ss->dmd0))/sizeof(int));
	printf("#define\tEF_DMA0 %d\n",	((int) &(ss->dma0))/sizeof(int));
	printf("#define\tEF_DMT1 %d\n",	((int) &(ss->dmt1))/sizeof(int));
	printf("#define\tEF_DMD1 %d\n",	((int) &(ss->dmd1))/sizeof(int));
	printf("#define\tEF_DMA1 %d\n",	((int) &(ss->dma1))/sizeof(int));
	printf("#define\tEF_DMT2 %d\n",	((int) &(ss->dmt2))/sizeof(int));
	printf("#define\tEF_DMD2 %d\n",	((int) &(ss->dmd2))/sizeof(int));
	printf("#define\tEF_DMA2 %d\n",	((int) &(ss->dma2))/sizeof(int));
	printf("#define\tEF_FPECR %d\n",((int) &(ss->fpecr))/sizeof(int));
	printf("#define\tEF_FPHS1 %d\n",((int) &(ss->fphs1))/sizeof(int));
	printf("#define\tEF_FPLS1 %d\n",((int) &(ss->fpls1))/sizeof(int));
	printf("#define\tEF_FPHS2 %d\n",((int) &(ss->fphs2))/sizeof(int));
	printf("#define\tEF_FPLS2 %d\n",((int) &(ss->fpls2))/sizeof(int));
	printf("#define\tEF_FPPT %d\n",	((int) &(ss->fppt))/sizeof(int));
	printf("#define\tEF_FPRH %d\n",	((int) &(ss->fprh))/sizeof(int));
	printf("#define\tEF_FPRL %d\n",	((int) &(ss->fprl))/sizeof(int));
	printf("#define\tEF_FPIT %d\n",	((int) &(ss->fpit))/sizeof(int));
	printf("#define\tEF_VECTOR %d\n",((int) &(ss->vector))/sizeof(int));
	printf("#define\tEF_MASK %d\n",	((int) &(ss->mask))/sizeof(int));
	printf("#define\tEF_MODE %d\n",	((int) &(ss->mode))/sizeof(int));
        /* next: I know, I know */
	printf("#define\tEF_RET %d\n",	((int) &(ss->scratch1))/sizeof(int));  
	printf("#define\tEF_NREGS %d\n",	sizeof(*ss)/sizeof(int));
	/* make a sanity check */
	if (sizeof(*ss) & 7)
	{
	    /* 
	     * This contortion using write instead of fputs(stderr)
	     * is necessary because we can't include stdio.h in here.
	     */
	    static char buf[] = 
	      "Exception frame not a multiple of double words\n";
	    write(2 /* stderr */,buf,sizeof(buf));
	    exit(1);
	}
	printf("#define\tSIZEOF_EF %d\n",	sizeof(*ss));

        printf("\n#endif /* __GENASSYM_INCLUDED\n */");
	return 0;
}
