/*
 * JEFF_UART_DEBUG
 * if defined, makes available raw_putchar() and raw_putstr() to output
 * characters and strings to the serial line (NOT THE BITMAPED CONSOLE).
 */
#ifdef JEFF_DEBUG
# define JEFF_UART_DEBUG
#endif
/*
 * Mach Operating System
 * Copyright (c) 1991, 1992 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */

/* $RCSfile: locore.c,v $ --  asm boot routines
 *
 **********************************************************************
 * This file created by Omron Corporation.
 *
 * HISTORY
 * $Log:	locore.c,v $
 * Revision 2.7  92/05/22  09:21:29  jfriedl
 * 	took out 'cpu=1' at start of slave.
 * 
 * Revision 2.6  92/05/21  17:19:17  jfriedl
 * 	Added raw_putstr and raw_putchar.
 * 	[92/05/21  16:08:15  jfriedl]
 * 
 * Revision 2.5  92/02/19  14:03:41  elf
 * 	br should have been br.n. Pointed out by Dan Malek of HRI.
 * 	[92/02/10            danner]
 * 
 * 	Removed release_other_cpus. Moved call_rom_abort. Removed
 * 	      thread_bootstrap_return.
 * 	[92/01/30            danner]
 * 
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 	Removed esym calculation.
 * 	[92/01/19            danner]
 * 
 * 	Removed esym definition, as it is obsoleted by the
 * 	 move_bootstrap.
 * 	[92/01/19            danner]
 * 	The interrupt_stack for the master processor was not getting set;
 * 	 it needs to be set to _intstack once the master knows its cpu
 * 	 number.
 * 
 * 	Fixed problem with master processor interrupt stack.
 * 	[91/11/04            danner]
 * 	Changed thread_bootstrap to return to
 * 	 return_from_exception_handler instead of check_ast. 
 * 	      No longer sets need_ast[cpu#] = 1.
 * 	[91/09/05            danner]
 * 
 * Revision 2.4  91/08/28  20:22:12  jsb
 * 	As a hack to make things compile, added null definition
 * 	of thread_bootstrap_return.
 * 
 * Revision 2.3  91/08/24  12:10:50  af
 * 	New slave stack initialization.
 * 	[91/08/15            rvb]
 * 	Removed the unused scASM* and ocsum functions.
 * 	[91/08/06            danner]
 * 	Made save context switch to interupt stack.
 * 	[91/07/15            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:46:48  danner
 * 	New slave stack initialization.
 * 	[91/08/15            rvb]
 * 	Removed the unused scASM* and ocsum functions.
 * 	[91/08/06            danner]
 * 	Made save context switch to interupt stack.
 * 	[91/07/15            danner]
 * 
 * Revision 2.2  91/07/09  23:17:51  danner
 * 	Removed OLD_MACH2_5 code (where did it come from). Also
 * 	 switch_task_context, which is obsolete.
 * 	[91/07/08            danner]
 * 
 * 	Updated save_context to use new pcb definition.
 * 	[91/07/05            danner]
 * 	Added switch_to_shutdown_context
 * 	[91/06/25            danner]
 * 
 * 	Added intstack definition for boot.
 * 	[91/05/09            danner]
 * 
 * 	Removed sigcode (2.5 trampoline code)
 * 	Other 2.5isms ifdef 0 out.
 * 	[91/05/05            danner]
 * 
 * 	Removed include of sys/syscall.h
 * 	[91/05/05            danner]
 * 	Flush jkdb.
 * 	[91/04/01            rvb]
 * 	Removed a comma that violated motorola assembler syntax
 * 	[91/03/07            danner]
 * 
 * Revision 2.2  91/04/05  13:55:36  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * Revision 2.1  91/03/06  15:03:02  rvb
 * my changes to omron for X134
 *
 * Revision 2.1  90/09/14  02:55:52  jfriedl
 *  Omron Luna/88k MACH Release 1.00 [release name version_1_00]
 * 
 * Revision 2.0  90/08/28  13:05:38  jfriedl
 * about to redo exception handler
 * 
 * Revision 1.4  90/06/05  10:49:24  jfriedl
 * Brought RCS files from version 2.0 to here (version 2.5)
 *
 * Revision 1.2  90/04/16  15:42:06  jfriedl
 * added RCS header.
 *****************************************************************RCS**/

#define ASSEMBLER
#include "locore.h"
#include "motorola/m88k/m88100/m88100.h"
#include "motorola/m88k/trap.h"
#include "luna88k/asm.h"
#include "sys/version.h"	/* for INCLUDE_VERSION define */
#include "luna88k/board.h"		/* lots of stuff (OBIO_PIO*, SYSV_BASE, etc)*/
#include "luna88k/pio.h"		/* lots of stuff */
#include "mach/machine/vm_param.h"      /* INTSTACK_SIZE */

/***********************************************************************/

/*
 *  Arrange for the include file version number to appear directly in
 *  the namelist.
 */
global	_INCLUDE_VERSION
def	_INCLUDE_VERSION, INCLUDE_VERSION

/*
 * The memory looks like:
 *   0x00000 - 0x01000	 : exception vectors (512 vectors x 2 instructions)
 *   0x10000 - 0x20000	 : ROM work area (I don't know how much actually used).
 *   0x20000 == start    : branch to start_text
 *   0x20004 - start_text: some locks and temporary stack space.
 *	The locks:
 *  		master_processor_chosen
 *   		inter_processor_lock
 *	are described below.
 *
 *   All slave cpus now come up using allocated interrupt stacks.
 *   The masters comes up an uses intstack, a pre allocated interrupt stack.
 *
 ***********************************************************************/
	text

LABEL(start)
	/* hardware (ROM) jumps here (this address, not this label) on reset*/
	br _start_text

    _LABEL(initialized_cpu_lock)
        /* XMEM spin lock -- to count CPUs */
	word 0

    LABEL(_initialized_cpus)
        /* CPU counter to initialize */
	word 0

    _LABEL(master_processor_chosen)
	/* The first processor that XMEMs this becomes the master */
	word 0

    LABEL(_inter_processor_lock)
	/* XMEM spin lock -- controls access to master_processor_chosen */
	word 0

    _LABEL(vector_list)		  /* references memory BELOW this line */
	#include "exception_vectors.h"
	word	END_OF_VECTOR_LIST

    _LABEL(_msgsw)
	word 0 /* Bits here turn on/off debugging somewhere. */

#ifdef JEFF_UART_DEBUG
#   define RAW_INTERCHAR_PAUSE	4000	/* range 0x0000 .. 0xffff */
#   define PRINT_MSG2(msg, cpu)		 \
	or.u	r2, r0, hi16(msg)	&\
	or  	r2, r2, lo16(msg)	&\
        bsr.n	_raw_putstr		&\
	ld	r2, r2 [cpu]

#   define PRINT_MSG(msg)		 	\
	or.u	r2, r0, hi16(msg)	&\
        bsr.n	_raw_putstr		&\
	or  	r2, r2, lo16(msg)

    cpu0_poweron_message: string "[cpu x0 poweron]\n\r\0"
    cpu1_poweron_message: string "[cpu x1 poweron]\n\r\0"
    cpu2_poweron_message: string "[cpu x2 poweron]\n\r\0"
    cpu3_poweron_message: string "[cpu x3 poweron]\n\r\0"
    cpu_poweron_message:
	word cpu0_poweron_message
	word cpu1_poweron_message
	word cpu2_poweron_message
	word cpu3_poweron_message
    cpu0_have_ip_lock: string "[cpu x0 has IP lock]\n\r\0"
    cpu1_have_ip_lock: string "[cpu x1 has IP lock]\n\r\0"
    cpu2_have_ip_lock: string "[cpu x2 has IP lock]\n\r\0"
    cpu3_have_ip_lock: string "[cpu x3 has IP lock]\n\r\0"
    cpu_have_ip_lock:
	word cpu0_have_ip_lock
	word cpu1_have_ip_lock
	word cpu2_have_ip_lock
	word cpu3_have_ip_lock
    cpu0_is_master: string "[cpu x0 is master]\n\r\0"
    cpu1_is_master: string "[cpu x1 is master]\n\r\0"
    cpu2_is_master: string "[cpu x2 is master]\n\r\0"
    cpu3_is_master: string "[cpu x3 is master]\n\r\0"
    cpu_is_master:
	word cpu0_is_master
	word cpu1_is_master
	word cpu2_is_master
	word cpu3_is_master
    cpu0_is_slave: string "[cpu x0 is slave]\n\r\0"
    cpu1_is_slave: string "[cpu x1 is slave]\n\r\0"
    cpu2_is_slave: string "[cpu x2 is slave]\n\r\0"
    cpu3_is_slave: string "[cpu x3 is slave]\n\r\0"
    cpu_is_slave:
	word cpu0_is_slave
	word cpu1_is_slave
	word cpu2_is_slave
	word cpu3_is_slave

    real_mayput_addr: word 0
    real_mayput_return: word 0
    orig_mayput_psr: word 0
    pre_mayput:
	/*
	 * on entry,
	 *	r1 is return address
	 *	r2 is byte to output
	 * we can trash r3-r9 as we like.
	 */
	;; disable interrupts, save original PSR in r4
	ldcr	r4, PSR
	set	r3, r4, 1<PSR_INTERRUPT_DISABLE_BIT>
	stcr	r3, PSR
	FLUSH_PIPELINE

	;; try to grab lock
	or.u	r3, r0, hi16(raw_put_lock)
	or	r3, r3, lo16(raw_put_lock)
	xmem	r3, r3, r0	;; if returns 0, we have the lock.
	FLUSH_PIPELINE
	bcnd	eq0, r3, mayput__have_raw_put_lock

	;; put PSR back (maybe interrupts go back on)
	stcr	r4, PSR
	FLUSH_PIPELINE

	;; pause for a while before checking
     1:	or	r3, r0, RAW_INTERCHAR_PAUSE
     2:	bcnd.n	ne0, r3, 2b
        subu	r3, r3, 1

	;; take a peek
	or.u	r3, r0, hi16(raw_put_lock)
	ld	r3, r3, lo16(raw_put_lock)
	bcnd	eq0, r3, _raw_putchar ;; might be able to get it.. try
	br	1b

    mayput__have_raw_put_lock:
	;; save real return address.
	or.u r3, r0, hi16(real_mayput_return)
	st   r1, r3, lo16(real_mayput_return)
	;; save PSR
	or.u r3, r0, hi16(orig_mayput_psr)
	st   r4, r3, lo16(orig_mayput_psr)
	;; jump to real mayput
	or.u r3, r0, hi16(real_mayput_addr)
	ld   r3, r3, lo16(real_mayput_addr)
	jsr r3

	;; get back original return address
	or.u r3, r0, hi16(real_mayput_return)
	ld   r1, r3, lo16(real_mayput_return)
	;; get original psr
	or.u r3, r0, hi16(orig_mayput_psr)
	ld   r4, r3, lo16(orig_mayput_psr)

	;; all done. Release the lock, restore PSR, and return.
	or.u    r3, r0, hi16(raw_put_lock)
	st      r0, r3, lo16(raw_put_lock)
	FLUSH_PIPELINE
	stcr    r4, PSR
	jmp     r1
#else
#  define PRINT_MSG2(msg, cpu) /*empty*/
#  define PRINT_MSG(msg) /*empty*/
#endif /*ifdef JEFF_UART_DEBUG */

/**************************************************************************/
LABEL(_start_text)	/* This is the *real* start upon poweron or reset */
	/*
	 * CPU Initialization
	 *
	 * Every CPU starts from here..
	 * (well, from 'start' above, which just jumps here).
	 *
	 * I use r11 and r22 here 'cause they're easy to not
	 * get mixed up -- r10, for example, looks too similar
	 * to r0 when not being careful....
	 *
	 * Ensure that the PSR is as we like:
	 *	supervisor mode
	 *	big-endian byte ordering
	 *	concurrent operation allowed
	 *	carry bit clear (I don't think we really care about this)
	 *	FPU enabled
	 *	misaligned access raises an exception
	 *	interrupts disabled
	 *	shadow registers frozen
	 *
	 * The manual says not to disable interrupts and freeze shadowing
	 * at the same time because interupts are not actually disabled
	 * until after the next instruction. Well, if an interrupt
	 * occurs now, we're in deep   anyway, so I'm going to do
	 * the two together.
	 *
	 * Upon a reset (or poweron, I guess), the PSR indicates:
	 *   supervisor mode
	 *   interrupts, shadowing, FPU, missaligned exception: all disabled
	 *
	 * We'll just construct our own turning on what we want.
	 *
	 *	jfriedl@omron.co.jp
	 */
	stcr	r0, SSBR	/* clear this for later */
	stcr	r0, SR0		/* clear "current thread" */
	stcr	r0, SR1		/* clear the CPU Flags */
	stcr	r0, SR2		/* not strictly necessary, but looks clean */
	stcr	r0, SR3		/* not strictly necessary, but looks clean */

	set	r11, r0,  1<PSR_SUPERVISOR_MODE_BIT>
	set	r11, r11, 1<PSR_INTERRUPT_DISABLE_BIT>
	stcr	r11, PSR
	;; shadowing, FPU, misalgined access exception: all enabled now.

	/* set the vector base */
	#if SYSV_BASE == 0	/* SYSV is "System Vector", not "AT&T SysV"! */
	   stcr	r0, VBR
	#else
	   or.u	r11,  r0, hi16(SYSV_BASE)
    	   or	r11, r11, lo16(SYSV_BASE)
	   stcr	r11, VBR
	#endif

	/*
	 * First time to count how many CPU to attached
	 */
	or.u	r11, r0,  hi16(initialized_cpu_lock)
	or	r11, r11, lo16(initialized_cpu_lock)
    _LABEL(check_init_lock)
	FLUSH_PIPELINE
	or	r22, r0,  1
	xmem    r22, r11, r0	  /* If r22 gets 0, we have the lock.. */
	bcnd	eq0, r22, have_init_lock /* ..but if not, we must wait */

    _LABEL(wait_for_init_lock)
	/* just watch the lock until it looks clear */
	ld	r22, r11, r0
	bcnd	eq0, r22, check_init_lock
	br	wait_for_init_lock		/* looks clear -- try to grab */

    _LABEL(have_init_lock)
	FLUSH_PIPELINE
	or.u	r11, r0,  hi16(_initialized_cpus)
        ld	r22, r11, lo16(_initialized_cpus)
	add	r23, r22, 1
	st	r23, r11, lo16(_initialized_cpus)

#if  defined(JEFF_UART_DEBUG)
  or   r14, r0, r22
  #if 0
    PRINT_MSG2(cpu_poweron_message, r14)
  #endif
#endif

	or.u	r11, r0,  hi16(initialized_cpu_lock)
        st	r0,  r11, lo16(initialized_cpu_lock)

	/*
 	 * Now we vie with any other processors to see who's the master.
 	 * We first try to obtain a lock to see who's allowed
	 * to check/set the master lock.
	 */
	or.u	r11, r0,  hi16(_inter_processor_lock)
	or	r11, r11, lo16(_inter_processor_lock)
    _LABEL(check_ip_lock)
	FLUSH_PIPELINE
	or	r22, r0,  1
	xmem    r22, r11, r0	       /* If r22 gets 0, we have the lock.. */
	bcnd	eq0, r22, have_ip_lock /* ..but if not, we must wait */
    _LABEL(wait_for_ip_lock)
	/* just watch the lock until it looks clear */
	ld	r22, r11, r0
	bcnd	ne0, r22, wait_for_ip_lock
        /* since we can be here with caches off, add a few nops to 
	   keep the bus from getting overloaded */
        or      r2, r0, lo16(1000)
    _LABEL(ip_loop)
        subu    r2, r2, 1
        bcnd	eq0, r2, ip_loop
	br	check_ip_lock		/* looks clear -- try to grab */

    _LABEL(have_ip_lock)


#if 0
  PRINT_MSG2(cpu_have_ip_lock, r14)
#endif

	/* now try to grab the master_processor_chosen prize */
	FLUSH_PIPELINE
	or.u	r11, r0,  hi16(master_processor_chosen)
	or	r11, r11, lo16(master_processor_chosen)
	or	r22, r0,  1
	xmem	r22, r11, r0

	/*
	 * If r22 is not clear we're a slave,
	 * otherwise we're first and the master.
	 *
	 * Note that we haven't released the interprocessor lock....
	 * We'll do that when we're ready for another CPU to go.
	 * (if we're the master, we'll do that in master_start below.
	 *  if we're a slave, we'll do it in slave_start below).
	 */
	bcnd	ne0, r22, slave_start
	/* fall through to master start if that's appropriate */

/************************************************************************/
_LABEL(master_start)
#ifdef JEFF_UART_DEBUG
	/*
	 * Want all uses of mayput() to use our lock.
	 * Therefore, we install a wrapper routine that the system
	 * will think is mayput. It will get the lock, call the real
	 * mayput, and then release the lock and return.
	 *
	 * That wrapper routine is pre_mayput(), and we install
	 * it into the ROM vector table here.
	 *
	 * Note that address 0x1110 is where the system expects to
	 * find the address of the mayput routine.
	 */
	ld	r1, r0, 0x1110	; get address of original mayput
	or.u	r2, r0, hi16(real_mayput_addr)
	st	r1, r2, lo16(real_mayput_addr) ; stuff away
	or.u	r1, r0, hi16(pre_mayput)
	or	r1, r1, lo16(pre_mayput)
	st	r1, r0, 0x1110  ; stuff	  ; replace with ours.
#endif

        /* pick up the master stack */
	or.u	r31, r0,  hi16(_intstack_end)
	or	r31, r31, lo16(_intstack_end)
	clr	r31, r31, 3<0>	;; round down to 8-byte boundary

   PRINT_MSG2(cpu_is_master, r14)

	/*
	 * Want to make the call:
	 * 	vector_init(VBR, vector_list)
	 */
	or.u	r3, r0,  hi16(vector_list)
	or	r3, r3, lo16(vector_list)
	bsr.n	_vector_init
	ldcr	r2, VBR		
	
       /* PIO stuff */
        or      r10, r0, PIO0_MODED		/* initialize pio 0 */
        or.u    r11, r0, hi16(OBIO_PIO0)
        st.b    r10, r11, lo16(OBIO_PIO0)

	/* read dispswitch setting */
	ld.bu	r10, r11, lo16(OBIO_PIO0A)	/* dipsw-1 (from portA) */
	mak	r10, r10, 0<8>			/* shift left 8 bit */
	ld.bu	r12, r11, lo16(OBIO_PIO0B)	/* dipsw-2 (from portB) */
	or	r10, r10, r12

	or.u	r11,r0,hi16(_dipswitch)
	st.h	r10,r11,lo16(_dipswitch)

	or	r10,  r0, PIO1_MODEO		/* initialize pio1 */
	or.u	r11,  r0, hi16(OBIO_PIO1)
	st.b	r10, r11, lo16(OBIO_PIO1)
	or	r10,  r0, POWER_ON*2+1		/* port c bit 1 on */
	st.b	r10, r11, lo16(OBIO_PIO1)

	or	r10,  r0, hi16(SCSI_ADDR+SCSI_INTS)	/* clear scsi int */
	ld.b	r11, r10, lo16(SCSI_ADDR+SCSI_INTS)
	st.b	r11, r10, lo16(SCSI_ADDR+SCSI_INTS)

        /* write 0x41000000 to escape rom */
        or      r2,  r0,  hi16(0x41000000)
        st      r0,  r2,  lo16(0x41000000)
	/* make the call: luna_init() */
	bsr.n	_luna_init
	subu	r31, r31, 40
	addu	r31, r31, 40

        /* we now know our cpu number, so we
	   can set interrupt_stack[cpu_number()] = _intstack
        */
 	ldcr	r10, SR1
	mak     r10, r10, FLAG_CPU_FIELD_WIDTH<0>    /* r10 <-- CPU# */
	/* figure interrupt_stack[cpu_number()] */
	or.u	r11,  r0, hi16(_interrupt_stack)  
	or	r11, r11, lo16(_interrupt_stack)
        or.u	r12, r0,  hi16(_intstack)
        or	r12, r12,  lo16(_intstack)
	st	r12, r11 [r10]

	/* make the call: setup_main() */
        /* when we leave we haven't released the _inter_processor_lock 
	   so all the slaves are stuck spinning. They will be relased
	   when we call start_other_cpus in kern/startup.c */
	bsr.n	_setup_main   /* returns initial thread in r2 */
	subu	r31, r31, 40

       /* never return */
       
/*****************************************************************************/
/* I don't know why this stuff is here -- why not in some C file? -- jfriedl */

	data
LABEL(_intstack)        zero    INTSTACK_SIZE /* initial interrupt stack */
LABEL(_intstack_end)                          /* top of stack */
LABEL(_slavestack)      zero    INTSTACK_SIZE /* initial find my cpu number stack */   
LABEL(_slavestack_end)
LABEL(_cpu)		word	0
LABEL(_dipswitch)	word	0
LABEL(_boothowto)	word	0
LABEL(_bootdev) 	word	0


	text

/***********************************************************************/
/*
 *	slave CPU starts here
 */
_LABEL(slave_start)
        /* while holding the inter_processor_lock, the slave cpu
	   can find use the slavestack to call slave_pre_main and
	   determine its cpu number. After that, however, it should
	   switch over to the interrupt stack associated with its
           cpu */

	/* r31 <-- slavestack */
	or.u	r31,  r0, hi16(_slavestack_end)
        or      r31, r31, lo16(_slavestack_end)
	clr	r31, r31, 3<0>	;; round down to 8-byte boundary
  PRINT_MSG2(cpu_is_slave, r14)
#if 0
I THINK THIS IS WRONG -- jfriedl
           /* just guess a cpu number for the debugger */
           bsr.n   _set_cpu_number   ;;
           or	r2,  r0, 0x2
#endif
        bsr.n    _slave_pre_main ;; set cpu number
        subu    r31, r31, 48    ;; allocate frame

	bsr	_get_slave_stack
	addu	r31, r2, INTSTACK_SIZE

        /*      SR1 now contains our cpu number. We can now release
		the inter_processor_lock, as we are done with the 
		slavestack.  We also have an interrupt stack  */

	or.u	r10,  r0, hi16(_inter_processor_lock)
	st	 r0, r10, lo16(_inter_processor_lock)	

	br.n	_slave_main     ;; does not return
	subu	r31, r31, 40    /* allocate frame */


/*
 *	thread_bootstrap:
 *
 *	Bootstrap a new thread using the thread state that has been
 *	placed on the stack.  Our fp has been set up for us, we only need
 *	to fix up a few things in the saved frame, then get into
 *	usermode.
 */
ENTRY(thread_bootstrap)
        /* 
	  Here r31 should point to the place on our stack which
	  contains a pointer to our exception frame.
	*/	
#if MACH_KDB
        ENTRY_ASM
#endif
	br      return_from_exception_handler

/*
 *	save_context
 */
ENTRY(save_context)
	subu	r31,r31,40		/* allocate stack for r1 and args */
	st	r1,r31,36		/* save return address */
	bsr	_spl			/* get the current interrupt mask */
	ld	r1,r31,36		/* recover return address */
	addu	r31,r31,40		/* put stack pointer back */
	ldcr	r10,SR0			/* r10 <- current_thread() */
	ld	r10,r10,THREAD_PCB	/* r10 <- pcb */
#if (PCB_KERNEL!=0)
     	addu	r10, r10, PCB_KERNEL    /* point to kernel save region */
#endif
	st	r1,r10,0	/* do setjmp */ /* save return address */
	st	r14,r10,4
	st	r15,r10,2*4
	st	r16,r10,3*4
	st	r17,r10,4*4
	st	r18,r10,5*4
	st	r19,r10,6*4
	st	r20,r10,7*4
	st	r21,r10,8*4
	st	r22,r10,9*4
	st	r23,r10,10*4
	st	r24,r10,11*4
	st	r25,r10,12*4
        /* In principle, registers 26-29 are never manipulated in the
           kernel. Maybe we can skip saving them? */
	st	r26,r10,13*4
	st	r27,r10,14*4
	st	r28,r10,15*4
	st	r29,r10,16*4
	st	r30,r10,17*4		/* save frame pointer */
	st	r31,r10,18*4		/* save stack pointer */
	st	r2,r10,19*4		/* save interrupt mask */
       /* we need to switch to the interrupt stack here -
	  extract our cpu number */
 	ldcr	r10, SR1
	mak     r10, r10, FLAG_CPU_FIELD_WIDTH<0>    /* r10 <-- CPU# */
	/* figure interrupt_stack[cpu_number()] */
	or.u	r11,  r0, hi16(_interrupt_stack)  
	or	r11, r11, lo16(_interrupt_stack)
	ld	r31, r11 [r10]
        addu    r31, r31, INTSTACK_SIZE             /* end of stack */
	clr	r31, r31, 3<0>	;; round down to 8-byte boundary
	jmp.n	r1
	or	r2,r0,r0


;------------------------------------------------------------------------

#ifndef JEFF_UART_DEBUG
    /* If not defined, just make them no-ops in case someone else calls them */
    LABEL(_raw_putstr)
    LABEL(_raw_putchar)
	jmp r1
#else
##
## Very low-level raw output to the serial line *ONLY*.
## (nothing goes to the bitmapped console).
##
## Only registers used are r1, r2, r3, and r4 (no stack used).
## This routine is very safe and robust and can be called at the most
## delecate of times.  About the only thing one must watch out for is
## if the cache state between processors is inconsistant, which happens
## at times during the boot.
##
## void raw_putstr(char *string)
##	Output the null-terminated string to the serial line.
##	The print lock is held througout, so concurrent calls won't have
##	their output intermixed.
##
## void raw_putstr(char *c)
##	Output the character to the serial line.
##
## _raw_putstr:
##   ENTER:
##	r1 = return address
##	r2 = string pointer
##   EXIT:
##	r2, r3, r4 trashed.
##	All other registers unchanged.
##
## _raw_putchar:
##   ENTER:
##	r1 = return address
##	r2 = data byte
##   EXIT:
##	r3, r4 trashed.
##	All other regesters unmodified..
##

    raw_put_lock: word 0
    raw_putstr_retval: word 0

LABEL(_raw_putstr)
	;; access the entire string to make sure it's all resident.
	or	r3, r2, r0
     1: ld.bu	r4, r3, 0
	bcnd.n	ne0, r4, 1b
	addu	r3, r3, 1
	
	FLUSH_PIPELINE
	set	r1, r1, 1<0> ;; use the low bit of R1 to mark 'putstr'.
	/* fall-through */

LABEL(_raw_putchar)
	;; disable interrupts, save original PSR in r4
	ldcr	r4, PSR
	set	r3, r4, 1<PSR_INTERRUPT_DISABLE_BIT>
	stcr	r3, PSR
	FLUSH_PIPELINE

	;; try to grab lock
	or.u	r3, r0, hi16(raw_put_lock)
	or	r3, r3, lo16(raw_put_lock)
	xmem	r3, r3, r0	;; if returns 0, we have the lock.
	FLUSH_PIPELINE
	bcnd	eq0, r3, have_raw_put_lock

	;; put PSR back (maybe interrupts go back on)
	stcr	r4, PSR
	FLUSH_PIPELINE

	;; pause for a while before checking
     1:	or	r3, r0, RAW_INTERCHAR_PAUSE
     2:	bcnd.n	ne0, r3, 2b
        subu	r3, r3, 1

	;; take a peek
	or.u	r3, r0, hi16(raw_put_lock)
	ld	r3, r3, lo16(raw_put_lock)
	bcnd	eq0, r3, _raw_putchar ;; might be able to get it.. try
	br	1b

   have_raw_put_lock:
	;; split now depending upon putchar and putstr
	bb1	0, r1, do_raw_putstr		;; check mark

	;; access the serial line control port
	or	r3, r0, hi16(0x51000004)
	ld.b	r3, r0, hi16(0x51000004)

	;; now do *(unsigned char *)0x51000000 = r2 to send char
	or.u	r3, r0, hi16(0x51000000)
	st.b	r2, r3, lo16(0x51000000)

	;; pause a bit before we shove the character.
       	or	r3, r0, RAW_INTERCHAR_PAUSE
     3: bcnd.n	ne0, r3, 3b
	subu	r3, r3, 1

   done_with_put:
        ;; all done. Release the lock, restore PSR, and return.
	or.u	r3, r0, hi16(raw_put_lock)
	st	r0, r3, lo16(raw_put_lock)
	FLUSH_PIPELINE
	stcr	r4, PSR
	jmp	r1

   do_raw_putstr:
	clr	r1, r1, 1<0> ;; remove mark.
	;; stuff return value away so we can use R1
	or.u	r3, r0, hi16(raw_putstr_retval)
	st	r1, r3, lo16(raw_putstr_retval)

     4: ld.bu	r1, r2, r0  ;; get byte to print.
	bcnd	eq0, r1, 6f ;; done if zero

       	or	r3, r0, RAW_INTERCHAR_PAUSE
     5: bcnd.n	ne0, r3, 5b
	subu	r3, r3, 1

	;; pause a bit before we shove the character.
	;; access the serial line control port
	or	r3, r0, hi16(0x51000004)
	ld.b	r3, r0, hi16(0x51000004)

	;; now do *(unsigned char *)0x51000000 = r1
	or.u	r3, r0, hi16(0x51000000)
	st.b	r1, r3, lo16(0x51000000)

	;; now increment r2 (string pointer)
	addu	r2, r2, 1
	br	4b	;; continue with loop
	
     6: or.u	r3, r0, hi16(raw_putstr_retval) ;; reclaim R1, return.
	ld	r1, r3, lo16(raw_putstr_retval)
	br done_with_put

#endif

;------------------------------------------------------------------------
/*
 * unsigned measure_pause(volatile int *flag)
 *
 * Count cycles executed until *flag becomes nonzero.
 * Return the number of cycles counted.
 */
ENTRY(measure_pause)
	/* R2 is pointer to flag */
	def GRANULAIRTY, 10000

	or	r3, r0, 1	/* r3 is my counter, this is the first */

    measure_pause_outer_loop:
	or	r4, r0, GRANULAIRTY

    measure_pause_inner_loop:
	/*
	 * Execute a tight loop of a known number of cycles.
	 * This assumes, of course, that the instruction cache is on.
	 * This loop takes two cycles per iteration.
	 */
	bcnd.n	ne0, r4, measure_pause_inner_loop
	subu	r4, r4, 1


	/*
	 * Now add the number of cycles done above (plus the overhead
	 * of the outer loop) to the total count.
	 * Also, check the *flag and exit the outer loop if it's non-zero.
	 *
	 * The overhead is really unknown because it's not known how
	 * the memory system will tread the access to *flag, so we just
	 * take a guess.
	 */
	ld	r4, r2, r0			/* get the flag */
	addu	r3, r3, (GRANULAIRTY * 2 + 10)	/* account for the cost */
	bcnd	eq0, r4, measure_pause_outer_loop /* continue or exit the loop*/

	jmp.n	r1
	or	r2, r3, r0	;/* pass count back */

/*
 * void delay_in_microseconds(int count)
 *
 * The processor loops (busy waits) for the given number of microseconds:
 * Thus, delay_in_microseconds(1000000) will delay for one second.
 *
 * REGISTER USAGE:
 *  IN  r1   - return address
 *  IN  r2   - (signed int) number of microseconds
 *      r3   - (float) number of microseconds
 *      r4/5 - (double) number of cycles per microsecond
 *      r6   - (float) number of cycles to delay
 *	r7   - (signed) number of cycles to delay
 */
ENTRY(delay_in_microseconds)
ENTRY(delay)
	flt.ss	 r3, r2	    /* convert microseconds from signed int to float */
	or.u	 r4, r0, hi16(_cycles_per_microsecond)
	ld.d	 r4, r4, lo16(_cycles_per_microsecond)
	fmul.ssd r6, r3, r4 /* convert microseconds to cycles */
	int.ss   r7, r6	    /* convert cycles from float to signed int */
	subu	 r7, r7, 25 /* subtract for overhead of above instruction */

	/* now loop for the given number of cycles */
    pause_loop:
	bcnd.n	gt0, r7, pause_loop
	subu	r7, r7, 2	/* two cycles per iteration */

	jmp	r1 /* return */

/*
 * void switch_to_shutdown_context(thread_t thread,
 *				   void (*routine)(processor_t),
 *				   processor_t processor)
 *
 * saves the kernel context of the thread,
 * switches to the interrupt stack,
 * continues the thread (with thread_dispatch),
 * then runs routine on the interrupt stack.
 *
 */
  
ENTRY(switch_to_shutdown_context)
/* call save_context to save the thread state */
     	subu 	r31, r31, 40
	or      r25, r3, r0    /* save arguments */
	or	r24, r4, r0     
     	bsr.n 	_save_context
	st	r1, r31, 36
     	addu 	r31, r31, 40
	ldcr	r10, SR0			/* r10 <- current_thread() */
	st      r31, r10, THREAD_KERNEL_STACK   /* save stack pointer */      
	st	r0,  r10, THREAD_SWAP_FUNC      /* null continuation */
        ldcr	r11, SR1
	mak	r11, r11, FLAG_CPU_FIELD_WIDTH<0> /* r1 = cpu # */
	or	r12, r12, lo16(_interrupt_stack)
	ld	r31, r12 [r11]
        addu    r31, r31, INTSTACK_SIZE             /* end of stack */
	clr	r31, r31, 3<0>	;; round down to 8-byte boundary
        /* save the thread; switched to the interrupt stack; now call thread
	   dispatch to get rid of this thread */
	or	r2, r10, r0      
	bsr.n	_thread_dispatch
     	subu 	r31, r31, 40
  	/* call the continuation routine */
  	jsr.n	r25
	or	r2, r24, r0
        /* panic if here */
        or.u	r2, r0, hi16(1f)
	bsr.n	_panic
	or	r2, r2, lo16(1f)
1:
  	string "switch_to_shutdown_context"
