/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	locore_c_routines.c,v $
 * Revision 2.5  92/05/21  17:19:42  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4  92/02/18  18:01:59  elf
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 
 * Revision 2.3  91/08/24  12:12:04  af
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:05  danner
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2  91/07/09  23:18:07  danner
 * 	Initial Checkin
 * 	[91/06/26  12:23:24  danner]
 * 
 * Revision 2.2  91/04/05  14:03:18  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/* $RCSfile: locore_c_routines.c,v $
 *
 **********************************************************************
 * This file created by Omron Corporation, 1990.
 *
 * $Locker:  $
 * $Revision: 2.5 $
 * $Author: jfriedl $
 * $Date: 92/05/21 17:19:42 $
 * $State: Exp $
 * 
 * HISTORY
 * 18-May-92  Jeffrey Friedl (jfriedl) at Carnegie-Mellon University
 *	Cleanup to quiet gcc warnings.
 *	Added prototypes.
 * $Log:	locore_c_routines.c,v $
 * Revision 2.5  92/05/21  17:19:42  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4.1.1  92/05/21  16:09:19  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4  92/02/18  18:01:59  elf
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 
 * Revision 2.3.1.1  92/02/18  16:33:22  danner
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 
 * Revision 2.3  91/08/24  12:12:04  af
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:05  danner
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.2.1  91/08/13  16:17:02  rvb
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.1.1  91/08/09  18:30:41  danner
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2  91/07/09  23:18:07  danner
 * 	Initial Checkin
 * 	[91/06/26  12:23:24  danner]
 * 
 * Revision 2.1.1.1  91/06/26  12:23:24  danner
 * 	Initial Checkin
 * 
 * Revision 2.2  91/04/05  14:03:18  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * Revision 2.1.1.1  91/04/02  18:33:15  rvb
 * 	Flush jkdb.
 * 	[91/04/01            rvb]
 * 
 * Revision 2.1  91/03/06  15:02:59  rvb
 * my changes to omron for X134
 * 
 * Revision 2.1  90/09/14  02:56:05  jfriedl
 *  Omron Luna/88k MACH Release 1.00 [release name version_1_00]
 * 
 * Revision 2.0  90/08/28  13:05:56  jfriedl
 * about to redo exception handler
 * 
 * Revision 1.8  90/08/21  11:54:32  jfriedl
 * ripped out the interrupt stack stuff.
 * 
 *    Revision 1.6  90/06/12  13:49:46  jfriedl
 *    Added DEBUG_MSG printf-type routine.  Debugged for data-access emulation
 *    loop.  Also allowed all exceptions to set the valid bit of DMT0 so
 *    that it's checked just before return.  This is so that a user program
 *    can set it when servicing a bus error and have it redone.
 *    
 *    Revision 1.5  90/06/05  10:49:31  jfriedl
 *    Brought RCS files from version 2.0 to here (version 2.5).
 *    Also added DEBUG_MSG() to print strings to the terminal
 *    without a lot of overhead.
 *    
 *    Revision 1.4  90/04/18  15:06:19  jfriedl
 *    after debugging new processor. Now has kluges for new processors.
 *    
 *    Revision 1.1.1.1  90/04/12  22:28:36  jfriedl
 *    Jeff's locore, initial check-in to RCS.
 *****************************************************************RCS**/

#include "locore.h"
#include "luna88k/asm.h"
#include "motorola/m88k/trap.h"      /* for T_ defines */
#include "motorola/m88k/m88100/m88100.h"
#include "kern/thread.h"
#include "kern/lock.h"

#ifdef WANT_PROTOTYPES
/* NO [raw_printf vsnprintf vector_init] */

 /* luna88k/locore/locore.c */
 extern void raw_putstr(char*);

 /* luna88k/locore/locore_asm_routines.c */
 extern unsigned int do_load_byte(unsigned,boolean_t,boolean_t);
 extern unsigned int do_load_half(unsigned,boolean_t,boolean_t);
 extern unsigned int do_load_word(unsigned,boolean_t,boolean_t);
 extern unsigned int do_xmem_byte(unsigned,unsigned,boolean_t,boolean_t);
 extern unsigned int do_xmem_word(unsigned,unsigned,boolean_t,boolean_t);
 extern void do_store_byte(unsigned,unsigned,boolean_t,boolean_t);
 extern void do_store_half(unsigned,unsigned,boolean_t,boolean_t);
 extern void do_store_word(unsigned,unsigned,boolean_t,boolean_t);
 extern void memcpy(void*,void*,unsigned int);

 /* forwards */
 static int show_num(unsigned char*,unsigned int,int,int,int,int,int);
 void data_access_emulation(long unsigned int*);
#endif /* ifdef WANT_PROTOTYPES */


/*
 *  data access emulation for M88100 exceptions
 */
#define DMT_BYTE	1
#define DMT_HALF	2
#define DMT_WORD	4

static struct
{
    unsigned char 	offset;
    unsigned char 	size;
} dmt_en_info[16] =
{
    {0, 0}, {3, DMT_BYTE}, {2, DMT_BYTE}, {2, DMT_HALF},
    {1, DMT_BYTE}, {0, 0}, {0, 0}, {0, 0},
    {0, DMT_BYTE}, {0, 0}, {0, 0}, {0, 0},
    {0, DMT_HALF}, {0, 0}, {0, 0}, {0, DMT_WORD}
};

#if EH_DEBUG
   int data_access_emulation_debug = 0;
   static char *bytes[] =
   {
	    "____", "___x", "__x_", "__xx",
	    "_x__", "_x_x", "_xx_", "_xxx",
	    "x___", "x__x", "x_x_", "x_xx",
	    "xx__", "xx_x", "xxx_", "xxxx",
   };
#  define DAE_DEBUG(stuff) { if (data_access_emulation_debug) { stuff ;}}
#else
#  define DAE_DEBUG(stuff)
#endif

#ifdef LOOP_DEBUG
  int loop_debug = 0;
#endif

#ifdef LOOP_DEBUG
# define LEVELS 10
  int loop_level[4];
  int loop_value[LEVELS][4];
  int loop_fault_location[LEVELS][4];
#endif

#include "luna88k/locore/asm_utils.h"

void data_access_emulation(eframe)
unsigned long eframe[EF_NREGS];
{
    register int x;
    register struct dmt_reg *dmtx;
    register unsigned  dmax, dmdx;
    register unsigned  v, reg;
#   ifdef LOOP_DEBUG
       int cpu;
#   endif

    if (!(eframe[EF_DMT0] & DMT_VALID))
	return;

#ifdef LOOP_DEBUG
        cpu = cpu_number();
#endif
#if defined(LOOP_DEBUG)
	if (loop_debug)
	{
		DEBUG_MSG("cpu %d level %d DAE: [%s:%x] [%s:%x] [%s:%x]\n",
		    cpu, loop_level[cpu], 
		    eframe[EF_DMT0] & DMT_VALID ? "#0" : "_",
			eframe[EF_DMA0],
		    eframe[EF_DMT1] & DMT_VALID ? "#1" : "_",
			eframe[EF_DMA1],
		    eframe[EF_DMT2] & DMT_VALID ? "#2" : "_",
			eframe[EF_DMA2]);
	}
#endif
    DAE_DEBUG
    (
	printf("Data access emulation. Valid pipelines: %s %s %s.\n",
	    eframe[EF_DMT0] & DMT_VALID ? "#0" : "_",
	    eframe[EF_DMT1] & DMT_VALID ? "#1" : "_",
	    eframe[EF_DMT2] & DMT_VALID ? "#2" : "_");
    )

#   ifdef LOOP_DEBUG
        if ((loop_level[cpu] > 3) &&
	    (loop_value[cpu][loop_level[cpu]] == eframe[EF_DMA0]) &&
	    (loop_value[cpu][loop_level[cpu]-1] == eframe[EF_DMA0]))
        {
	    DEBUG_MSG("cpu %d: DATA ACCESS LOOP! From %x [fp=%x], SXIP=%x\n",
		    cpu, return_addr, current_sp, eframe[EF_SXIP]);

	    for (x = 0; x < loop_level[cpu]; x++)
	    	DEBUG_MSG("loop_value[%d][cpu%d] is %08x., location is %08x.\n",
		    x, cpu, loop_value[x][cpu], loop_fault_location[x][cpu]);

	    for (x = 0; x < 3; x++) {
		    dmtx = (struct dmt_reg *)&eframe[EF_DMT0+x*3];

		    if (!dmtx->dmt_valid)
			    continue;

		    dmdx = eframe[EF_DMD0+x*3];
		    dmax = eframe[EF_DMA0+x*3];

		    DEBUG_MSG("DMT%d: %08x, DMD%d: %08x, DMA%d: %08x\n",
			    x, eframe[EF_DMT0+x*3],
			    x, eframe[EF_DMD0+x*3],
			    x, eframe[EF_DMA0+x*3]);

		    if (dmtx->dmt_write)
			    DEBUG_MSG("%d: write %x to addr %x(%c)",
				x, dmdx, dmax, dmtx->dmt_das ? 's' : 'u');
		    else
			    DEBUG_MSG("%d: read from addr %x(%c) to r%d, %s, ",
				 x, dmax, dmtx->dmt_das ? 's' : 'u', dmtx->dmt_dreg,
				 dmtx->dmt_signed ? "signed" : "unsigned");

		    DEBUG_MSG(" as [%s]: %s, %s\n",
			bytes[dmtx->dmt_en],
			dmtx->dmt_doub1 ? "double": "not double",
			dmtx->dmt_lockbar ? "xmem": "not xmem");
	    }
	    eframe[EF_DMT0] = 0;
	    asm("jeffmark3: global jeffmark3");
	    loop_level[cpu] = 0;
	    return;
	}
	x = loop_level[cpu]++;
	loop_value[x][cpu] = eframe[EF_DMA0];
	loop_fault_location[x][cpu] = eframe[EF_SXIP];
#   endif /* if LOOP_DEBUG */

    for (x = 0; x < 3; x++)
    {
	dmtx = (struct dmt_reg *)&eframe[EF_DMT0+x*3];

	if (!dmtx->dmt_valid)
		continue;

	dmdx = eframe[EF_DMD0+x*3];
	dmax = eframe[EF_DMA0+x*3];

	DAE_DEBUG
	(
	    printf("DMT%d: %08x, DMD%d: %08x, DMA%d: %08x\n",
			x, eframe[EF_DMT0+x*3],
			x, eframe[EF_DMD0+x*3],
			x, eframe[EF_DMA0+x*3]);

	    if (dmtx->dmt_write)
		printf("%d: write %x to addr %x(%c)",
		    x, dmdx, dmax, dmtx->dmt_das ? 's' : 'u');
	    else
		printf("%d: read from addr %x(%c) to r%d, %s, ",
		     x, dmax, dmtx->dmt_das ? 's' : 'u', dmtx->dmt_dreg,
		     dmtx->dmt_signed ? "signed" : "unsigned");

	    printf(" as [%s]: %s, %s\n",
		    bytes[dmtx->dmt_en],
		    dmtx->dmt_doub1 ? "double": "not double",
		    dmtx->dmt_lockbar ? "xmem": "not xmem");
	)

	dmax += dmt_en_info[dmtx->dmt_en].offset;
	reg = dmtx->dmt_dreg;

	if ( ! dmtx->dmt_lockbar)
	{
	    /* the fault is not during an XMEM */

	    if (x == 2 && dmtx->dmt_doub1)
	    {
		/* pipeline 2 (earliest stage) for a double */

		if (dmtx->dmt_write)
		{
		    /* STORE DOUBLE WILL BE RE-INITIATED BY rte */
		}
		else
		{
		    /* EMULATE ld.d INSTRUCTION */
		    if (dmtx->dmt_bo)
		    {
			/* LITTLE ENDIAN */
			v = do_load_word(dmax, dmtx->dmt_das, 1);
			if (reg != 31)
			    eframe[EF_R0 + reg + 1] = v;
			v = do_load_word(dmax ^ 4, dmtx->dmt_das, 1);
			if (reg != 0)
			    eframe[EF_R0 + reg] = v;
		    }
		    else
		    {
			/* BIG ENDIAN */
			v = do_load_word(dmax, dmtx->dmt_das, 0);
			if (reg != 0)
			    eframe[EF_R0 + reg] = v;
			v = do_load_word(dmax ^ 4, dmtx->dmt_das, 0);
			if (reg != 31)
			    eframe[EF_R0 + reg + 1] = v;
		    }
		}
	    }
	    else	 /* not pipeline #2 with a double */
	    {
		if (dmtx->dmt_write) switch (dmt_en_info[dmtx->dmt_en].size)
		{
		  case DMT_BYTE:
		    DAE_DEBUG(printf("byte %x -> [%x(%c)]\n",
			dmdx & 0xff, dmax, dmtx->dmt_das ? 's' : 'u'))
		    do_store_byte(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		    break;
		  case DMT_HALF:
		    DAE_DEBUG(printf("half %x -> [%x(%c)]\n",
			dmdx & 0xffff, dmax, dmtx->dmt_das ? 's' : 'u'))
		    do_store_half(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		    break;
		  case DMT_WORD:
		    DAE_DEBUG(printf("word %x -> [%x(%c)]\n", dmdx, dmax,
			dmtx->dmt_das ? 's' : 'u'))
		    do_store_word(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		    break;
		}
		else	/* else it's a read */
		{
		    switch (dmt_en_info[dmtx->dmt_en].size)
		    {
		      case DMT_BYTE:
			v = do_load_byte(dmax, dmtx->dmt_das, dmtx->dmt_bo);
			if (!dmtx->dmt_signed)
			    v &= 0x000000ff;
			break;
		      case DMT_HALF:
			v = do_load_half(dmax, dmtx->dmt_das, dmtx->dmt_bo);
			if (!dmtx->dmt_signed)
			    v &= 0x0000ffff;
			break;
		      case DMT_WORD:
		      default: /* 'default' just to shut up lint */
			v = do_load_word(dmax, dmtx->dmt_das, dmtx->dmt_bo);
			break;
		    }
	   	    if (reg == 0) {
			DAE_DEBUG(printf("no write to r0 done\n"));
		    }
		    else
		    {
			DAE_DEBUG(printf("r%d <- %x(%s)\n",
				reg, v, dmtx->dmt_das))
			eframe[EF_R0 + reg] = v;
		    }
		}
	    }
	}
	else /* if lockbar is set... it's part of an XMEM */
	{
	    /*
	     * According to Motorola's "General Information",
	     * the dmt_doub1 bit is never set in this case, as it should be.
	     * They call this "general information" - I call it a f*cking bug!
	     *
	     * Anyway, if lockbar is set (as it is if we're here) and if
	     * the write is not set, then it's the same as if doub1
	     * was set...
	     */
	    if ( ! dmtx->dmt_write)
	    {
		if (x != 2)
		{
		    /* RERUN xmem WITH DMD(x+1) */
		    x++;
		    dmdx = eframe[EF_DMD0 + x*3];
		}
		else
		{
		    /* RERUN xmem WITH DMD2 */
		}

		if (dmt_en_info[dmtx->dmt_en].size == DMT_WORD)
		    v = do_xmem_word(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		else
		    v = do_xmem_byte(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		eframe[EF_R0 + reg] = v;
	    }
	    else
	    {
		if (x == 0)
		{
		    asm("xmem_dae: global xmem_dae");
		    eframe[EF_R0 + reg] = dmdx;
		    eframe[EF_SFIP] = eframe[EF_SNIP];
		    eframe[EF_SNIP] = eframe[EF_SXIP];
		    eframe[EF_SXIP] = 0;
		    /* xmem RERUN ON rte */
		    eframe[EF_DMT0] = 0;
#		    ifdef LOOP_DEBUG
			if (--loop_level[cpu] < 0)
				loop_level[cpu] = 0;

			
#		    endif
		    return;
		}
	    }
	}
    }
    eframe[EF_DMT0] = 0;

#   ifdef LOOP_DEBUG
	--loop_level[cpu];
	if (loop_level[cpu] < 0)
	    loop_level[cpu] = 0;
#   endif
}



/*
 ***********************************************************************
 ***********************************************************************
 */
#define SIGSYS_MAX      501
#define SIGTRAP_MAX     511

#define EMPTY_BR	0xC0000000U      /* empty "br" instruction */
#define NO_OP 		0xf4005800U      /* "or r0, r0, r0" */

#define BRANCH(FROM, TO) (EMPTY_BR | ((unsigned)(TO) - (unsigned)(FROM)) >> 2)

#define SET_VECTOR(NUM, to, VALUE)  \
	vector[NUM].word_one = NO_OP; 	\
	vector[NUM].word_two = BRANCH(&vector[NUM].word_two, (VALUE));

typedef struct
{
	unsigned word_one, word_two;
} m88k_exception_vector_area;

/*
 * vector_init(vector, vector_init_list)
 *
 * This routine sets up the m88k vector table for the running processor.
 * It is called with a very little stack, and interrupts disabled,
 * so don't call any other functions!
 */
void
vector_init(vector, vector_init_list)
register m88k_exception_vector_area vector[];
register unsigned vector_init_list[];
{
	register unsigned num;
	register unsigned vec;
	extern void sigsys(), sigtrap(), stepbpt(), userbpt();

	for (num = 0; vec = vector_init_list[num], vec != END_OF_VECTOR_LIST;
	      num++)
	{
		if (vec == PREDEFINED_BY_ROM)
			continue;

		SET_VECTOR(num, to, vec);
	}

	while (num <= SIGSYS_MAX)
	{
		SET_VECTOR(num, to, sigsys);
		num++;
	}

	while (num <= SIGTRAP_MAX)
	{
		SET_VECTOR(num, to, sigtrap);
		num++;
	}

	SET_VECTOR(504, to, stepbpt);
	SET_VECTOR(511, to, userbpt);
}

/******* INCLUSION OF 2.5 vsnprintf.c starts here ****************/

/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/* 
 * HISTORY
 * $Log:	locore_c_routines.c,v $
 * Revision 2.5  92/05/21  17:19:42  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4.1.1  92/05/21  16:09:19  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.3  91/05/07  11:14:00  rvb
 * 	Omron changes April 1991 via J. Friedl
 * 	[91/05/07            rvb]
 * 
 *
 * 90-Sep-24 - jfriedl@omron.co.jp
 * Added the CHOP length to %s.
 *
 * Revision 2.2  91/04/05  14:03:23  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * Revision 2.1  91/03/06  15:03:10  rvb
 * my changes to omron for X134
 * 
 * Revision 2.1  90/09/14  02:56:10  jfriedl
 *  Omron Luna/88k MACH Release 1.00 [release name version_1_00]
 *
 *****************************************************************RCS***
 *
 *	VSNprintf()  ---
 *		v is for Vector (takes a vector of printf arguments)
 *		s is for String (prints to a given buffer)
 *		n is as in strNcpy ... gives maximum length of the buffer.
 *
 *	I got the idea for the N part of this (making this vsnprintf
 *	rather than vsprintf) from some code from CMU I saw once).
 *
 *	Written for the MACH kernel for use in debugging routines.
 *	Jeffrey Friedl (jfriedl@omron.co.jp)
 *	Spring 1990
 *
 * My [jfriedl@nff.ncl.omron.co.jp] vsnprintf() routine for kernel debugging.
 *
 * It use rather little stack (i.e. no internal buffer), and so is good for
 * debugging things like kernel stack overflows, etc (which is one reason I
 * wrote it).  It does all it's work in the given buffer itself.
 *
 *	recognizes:
 *		%%		%
 *		%d		decimal int
 *		%u		unsigned int
 *		%x		hex (unsigned) int
 *		%o		octal (unsigned) int
 *		%b		binary (unsigned) int
 *		%c		int as a character
 *		%C		"cat -v"-like character.
 *		%s		string (char *)
 *		%f		float (sort of).
 *
 *
 * Between '%' and '[duxobcsf]' you may have a value (or a '*' in which case
 * the next argument is used) to specify a LENGTH value.
 *
 * Examples using sprintf(), etc., assumes that they use this vsnprintf()
 * as their base routine (makes the examples easier).
 *
 * You may also have a period followed by a value (or '*') to specify the
 * maximum length for numbers and strings. For example,
 *		sprintf(buffer, "%*.*s", len, len, string);
 * prints exactly <len> characters each time, regardless of what's in <string>.
 *
 * Note:  If numbers get chopped off because of a too-short length argument,
 * 	  or because of running out of room in the buffer, they will NOT come
 *	  out as expected....
 *
 * For %[duxobs], the LENGTH is the minimum length used.  If the number has
 * fewer characters (for %d, '-' counts as a character if it is present) than
 * LENGTH, it is pre-padded with spaces (with '0' if the first digit of LENGTH
 * is '0'). [to get padding with '0' and use '*', use something like "%0*x"].
 *
 *
 *	Example: the following are the same
 *		sprintf(buffer, "%05b", value);
 *		sprintf(buffer, "%0*b", 5, value);
 *
 * For %f, LENGTH concerns only the integral part. The decimal part is always
 * '.' plus three digits.  It is very simple... the following are the same
 * because the first just does (internaly) the 2nd.
 *		printf("%*f", len, value)
 *
 *		printf("%*d.%03d", len, (int)value,
 *			(int)((value - (int)value) * 1000.0));
 *
 * LENGTH is ignored for %C and %c.
 *
 * If vsnprintf() is called with the buffer area as NULL, it just returns
 * the number of characters that would be output [WITHOUT the trailing null].
 * The maximum length given, however, DOES include the trailing null.
 * Thus,
 *		snprintf(buffer, 5, "%s", "abcdefghijk")
 * would fill buffer as 'abcd\0' and return 4.
 */

/* name of the routine */
#ifndef VSNPRINTF_FCN_NAME
#define VSNPRINTF_FCN_NAME vsnprintf
#endif

/* to have it translate \n to \n\r on ouput */
#ifndef TRANSLATE_NL_TO_NLCR
# define TRANSLATE_NL_TO_NLCR	 1
#endif

/* to use a local (but less efficient) ctype rather than libc one */
#ifndef USE_LOCAL_CTYPE
#define USE_LOCAL_CTYPE 1
#endif


#ifdef __STDC__
#  include <stdarg.h>
#else
#  include <varargs.h>
#endif

#if USE_LOCAL_CTYPE
# define isprint(c)	(c >= ' ' && c <= '~')
# define isdigit(c)	(c >= '0' && c <= '9')
#else
# include </usr/include/ctype.h>
#endif


#define SIGNED 1
#define UNSIGNED 0

/*
 * Insert2buffer(character)
 *
 * Macro that works with the variables 'max_len', 'out' and 'buffer'
 * in whatever scope it's used.  It works like:
 *
 * 	If there is room to send a new character
 *		Decrement the "space left" value.
 *		If there is a buffer to whoich to put it.
 *			put the character in the buffer, increment the ptr
 *		else
 *			increment the pointer.
 */
#define insert2buffer(character)		\
    (						\
	max_len == 0 ?				\
	    (void)0				\
	:					\
	(					\
	    (max_len--, buffer) ?		\
		(void)(*out++ = (character))	\
	    :					\
		(void)out++			\
	)					\
    )						\


/*
 * show_num(buffer, value, min_len, max_len, base, zerofill, sign)
 *
 * Fill the textual representation of VALUE (as base BASE)
 * to the buffer starting at OUT.
 *
 * A minimum of MIN_LEN characters will be output.  The number is treated
 * as signed if SIGN is non-zero.  If extra characters need to be pre-printed
 * to bring the length up to the minimum, ' ' will be used unles ZEROFILL
 * is non-zero in which case '0' is used.
 *
 * A maximum of MAX_LEN characters will be output. MAX_LEN overrides MIN_LEN.
 *
 * The number of characters actually "printed" is returned.
 */
#ifdef __STDC__
static int show_num(
    unsigned char *buffer,
    unsigned value,
    int min_len,
    int max_len,
    int base,
    int zerofill,
    int sign)
#else
static int show_num(buffer, value, min_len, max_len, base, zerofill, sign)
    unsigned char *buffer;
    unsigned value;
    int min_len;
    int max_len;
    int base;
    int zerofill;
    int sign;
#endif
{
	unsigned char *out;
	char *strcpy();
	int count;

	out = buffer;

	/*
	 * If we're to look at the sign, take care of all that stuff
	 * now.  Leave SIGN set to mean we must output a leading '-' if
	 * it is true later.
 	 */
	if (sign)
	{
		int signed_value = value;
		if (signed_value >= 0)
			sign = 0;
		else
		{
			min_len--;
			value = -signed_value;
		}
	}

	/* main loop to output digits */
	do
	{
		unsigned char C;

		if (C = '0' + (value % base), C > '9')
			C += 'a' - '9' - 1;

		/*
		 * The following is to attempt to make chopped-off numbers
		 * come out correctly (chopped off at the end).
		 */
		if (max_len == 0 && out != buffer)
		{
		    #ifndef bcopy
	               extern void bcopy();
		    #endif
		    /* I hope bcopy gets this overlapping move correct! */
		    out--;
		    max_len++;
		    bcopy(buffer+1, buffer, out-buffer);
		}


		insert2buffer(C);

		value /= base;
		min_len--;
	} while (value);

	while (min_len-- > 0)
		insert2buffer(zerofill ? '0' : ' ');

	if (sign)
		insert2buffer('-');

	count = out - buffer;

	if (buffer)
	{
		*out = '\0';

		/*
		 * Characters are now there, but backwards. Put right.
		 */
		while (buffer < --out)
		{
			unsigned char C;
			C = *buffer;
			*buffer++ = *out;
			*out = C;
		}
	}
	return(count);
}

#ifdef __STDC__
int VSNPRINTF_FCN_NAME(
    unsigned char *buffer,
    unsigned int max_len,
    unsigned char const *format,
    va_list ap)
#else
int VSNPRINTF_FCN_NAME(buffer, max_len, format, ap)
    unsigned char *buffer;
    unsigned int max_len;
    unsigned char *format;
    va_list ap;
#endif
{
    unsigned char *out = buffer;/* current "output" pointer */
    unsigned char C;		/* current character from the format string */
    unsigned zerofill;		/* if true, pad with '0' (rather than ' ') */
    int len;			/* minimum length as given in format string */
    int max_arg_len;		/* if nonzero, the max output from a %s */

    #define NEXT_ARG(type)	(va_arg(ap, type))

    if (max_len-- == 0)	/* subtract for the '\0' we always append */
	return 0;

    while (max_len != 0  &&  (C = *format++))
    {
	if (C != '%')
	{
		/* easy case.. just "print" the character */
		insert2buffer(C);

		#if TRANSLATE_NL_TO_NLCR
		    if (C == '\n')
			    insert2buffer('\r');
		#endif

		continue;
	}

	zerofill = (*format=='0'); /* pad with zero if 1st length char is '0'*/
	max_arg_len = max_len;

    /* at this point, C is '%', *format is the next character after */
	/* grab the length, if it's there */
	for (len = 0; C = *format, isdigit(C); format++)
		len = len * 10 + (C - '0');

    /* at this point, C and *format are the first character after any length */
	/*
	 * Maybe we read a length, maybe not.
	 * See if there's a '*' there to override
	 */
	if (C == '*')
	{
		len = NEXT_ARG(int);
		/* we don't yet, support negative lengths */
		if (len < 0)
			len = -len;

		C = *++format;
	}

    /* at this point, C and *format are the first character after any '*' */
	/* try to read a max string length */
	if (C == '.')
	{
		++format;	/* not beyond the '.' */

		/* now either a value or '*' */
		if ('*' == *format)
		{
			max_arg_len = NEXT_ARG(unsigned int);
			++format;	/* now beyond the '*' */
		}
		else
		{
			for (max_arg_len = 0; C = *format, isdigit(C); format++)
				max_arg_len = max_arg_len * 10 + (C - '0');
		}

		/* even if they want a lot, they still need to have the room */
		if (max_arg_len > max_len)
			max_arg_len = max_len;
	}

	/* should be at the format character now */
	C = *format++;

	switch(C)
	{
	    case '%':
		insert2buffer('%');
		break;

	    default:	/* huh? */
		insert2buffer('%');
		insert2buffer('?');
		insert2buffer(C);
		break;

	    case 'c':	/* character */
	    {
		C =NEXT_ARG(unsigned int);
		insert2buffer(C);
		#if TRANSLATE_NL_TO_NLCR
		    if (C == '\n')
			    insert2buffer('\r');
		#endif
		break;
	    }

	    case 'C':	/* CAT-V like character */
	    {
		unsigned char value = NEXT_ARG(unsigned int);

		if (isprint(value))
			insert2buffer(value);
		else if (value >= ('A' - 'A' + 1) && value <= ('Z' - 'A' + 1))
		{
			insert2buffer('^');
			insert2buffer(value + 'A' - 1);
		}
		else
		{
			int chars;
			insert2buffer('\\');
			chars = show_num(buffer ? out : 0, value,
				3, max_arg_len, 8, 1, UNSIGNED);
			out += chars;
			max_len -= chars;
		}
		break;
	    }

	    case 's':	/* string */
	    {
		unsigned char *ptr = NEXT_ARG(unsigned char *);
		while ((C = *ptr) && max_arg_len-- > 0)
		{
			insert2buffer(C);
			#if TRANSLATE_NL_TO_NLCR
			    if (C == '\n')
				    insert2buffer('\r');
			#endif
			ptr++;
			len--;
		}
		while (len-- > 0)
			insert2buffer(' ');
		break;
	    }

	    case 'x':	/* hex */
	    {
		int chars = show_num(buffer ? out : 0, NEXT_ARG(unsigned int),
			len, max_arg_len, 16, zerofill, UNSIGNED);
		out += chars;
		max_len -= chars;
		break;
	    }
	    case 'd':	/* decimal (signed) */
	    {
		int chars = show_num(buffer ? out : 0, NEXT_ARG(int),
			len, max_arg_len, 10, zerofill, SIGNED);
		out += chars;
		max_len -= chars;
		break;
	    }
	    case 'u':	/* unsigned decimal */
	    {
		int chars = show_num(buffer ? out : 0, NEXT_ARG(unsigned int),
			len, max_arg_len, 10, zerofill, UNSIGNED);
		out += chars;
		max_len -= chars;
		break;
	    }
	    case 'o':  /* octal */
	    {
		int chars = show_num(buffer ? out : 0, NEXT_ARG(unsigned int),
			len, max_arg_len, 8, zerofill, UNSIGNED);
		out += chars;
		max_len -= chars;
		break;
	    }
	    case 'b':	/* binary */
	    {
		int chars = show_num(buffer ? out : 0, NEXT_ARG(unsigned int),
			len, max_arg_len, 2, zerofill, UNSIGNED);
		out += chars;
		max_len -= chars;
		break;
	    }
	    case 'F':	/* "float" */
	    case 'E':
	    case 'G':
	    case 'f':
	    case 'e':
	    case 'g':
	    {
		int chars;
		double value = NEXT_ARG(double);
		chars = show_num(buffer ? out : 0,
			(int)value, len, max_arg_len, 10, 0, SIGNED);
		out += chars;
		max_len -= chars;
		max_arg_len -= chars;

		value -= (int)value;
		insert2buffer('.');
		chars = show_num(buffer ? out : 0, (int)(value * 1000.0),
			3, max_arg_len, 10, 1, UNSIGNED);
		out += chars;
		max_len -= chars;
		break;
	    }
	}
    }
    if (buffer)
	*out = '\0';
    return(out - buffer);
}

/*
 * Raw_printf() is a very low-level print routine.
 * It uses a small buffer on the stack to so a sprintf of the data,
 * and then calls raw_putstr.
 *
 * Raw_putstr is defined in locore.c and, if JEFF_UART_DEBUG is turned
 * on, it sends the string to the serial tty line.
 * Otherwise, nothing is done and the whole call to raw_printf is just
 * about a No-op.
 */
#ifdef __STDC__
void
raw_printf(const char *fmt, ...)
{
	va_list ap;
	unsigned char buffer[100];

	va_start(ap, fmt);
	(void)vsnprintf(buffer, sizeof(buffer), (unsigned char *)fmt, ap);
	va_end(ap);
	raw_putstr(buffer);
}

#else /* else not STDC */

void
raw_printf(fmt, va_alist)
unsigned char *fmt;
va_dcl
{
	va_list ap;
	unsigned char buffer[100];

	va_start(ap);
	(void)vsnprintf(buffer, sizeof(buffer), (unsigned char *)fmt, ap);
	va_end(ap);
	raw_putstr(buffer);
}
#endif	/* end elst not STDC */


#ifdef TEST
#define LEN 100

int main()
{
    char buffer[LEN];
    extern int write();
    int i, ret;
    extern void printf();
    extern int end;

printf("debugger symbol table: 0x%04x -- 0x%04x\n", 0x1670c8, 0x17c9b4);

#define DO_TEST(SNPRINTF_ARGS, EXPECTED_RETURN, EXPECTED_RESULT) 	\
	for (i = 0; i < LEN; i++)					\
	    buffer[i] = 'X';						\
								    	\
	ret = snprintf SNPRINTF_ARGS;					\
	if (ret != EXPECTED_RETURN)					\
	    printf("test on line %d returns %d (expected %d)\n",	\
		    __LINE__, ret, EXPECTED_RETURN);			\
								    	\
	if (strcmp(EXPECTED_RESULT, buffer) != 0)			\
	    printf("test on line %d: wanted \"%s\", got \"%s\".\n",	\
		    __LINE__, EXPECTED_RESULT, buffer);

DO_TEST(
  (buffer, LEN, "hi, (%x) [%d] [%o] %%%b", 0x12345, 12345, 012345, 12345),
  43,
  "hi, (12345) [12345] [12345] %11000000111001");

DO_TEST(
  (buffer, LEN, "hi, (%.2x) [%.3d] [%.4o] %%%.5b",0x12345, 12345,012345, 12345),
  28,
  "hi, (12) [123] [1234] %11000");

    return 0;
}
#endif
