/*
 * Mach Operating System
 * Copyright (c) 1992,1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	locore_c_routines.c,v $
 * Revision 2.2  92/08/03  18:14:48  jfriedl
 * 	Moved from locore directory.
 * 	[92/07/24            jfriedl]
 * 
 * Revision 2.4.2.1  92/05/27  14:40:05  danner
 * 	Merged with revision 2.5, minor corrections.
 * 
 * Revision 2.5  92/05/21  17:19:42  jfriedl
 * 	Added a light-weight vsnprintf for debugging purposes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4  92/02/18  18:01:59  elf
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 
 * Revision 2.3  91/08/24  12:12:04  af
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:05  danner
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2  91/07/09  23:18:07  danner
 * 	Initial Checkin
 * 	[91/06/26  12:23:24  danner]
 * 
 * Revision 2.2  91/04/05  14:03:18  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/* $RCSfile: locore_c_routines.c,v $
 *
 **********************************************************************
 * This file created by Omron Corporation, 1990.
 *
 * $Locker:  $
 * $Revision: 2.2 $
 * $Author: jfriedl $
 * $Date: 92/08/03 18:14:48 $
 * $State: Exp $
 * 
 * HISTORY
 * 18-May-92  Jeffrey Friedl (jfriedl) at Carnegie-Mellon University
 *	Cleanup to quiet gcc warnings.
 *	Added prototypes.
 *	Added raw_printf support.
 * $Log:	locore_c_routines.c,v $
 * Revision 2.2  92/08/03  18:14:48  jfriedl
 * 	Moved from locore directory.
 * 	[92/07/24            jfriedl]
 * 
 * Revision 2.1.2.1  92/08/03  14:57:04  jfriedl
 * 	Moved from locore directory.
 * 	[92/07/24            jfriedl]
 * 
 * Revision 2.1.1.1  92/07/18  00:23:33  jfriedl
 * 	17-July-92  Jeffrey Friedl (jfriedl) at Carnegie Mellon University
 * 		Moved from locore/
 * 
 * Revision 2.4.2.1  92/05/27  14:40:05  danner
 * 	Merged with revision 2.5, minor corrections.
 * 
 * Revision 2.5  92/05/21  17:19:42  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4.1.1  92/05/21  16:09:19  jfriedl
 * 	     Added a light-weight vsnprintf for debugging purposes.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4  92/02/18  18:01:59  elf
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 
 * Revision 2.3.1.1  92/02/18  16:33:22  danner
 * 	Updated includes.
 * 	[92/01/30            danner]
 * 
 * Revision 2.3  91/08/24  12:12:04  af
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:05  danner
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.2.1  91/08/13  16:17:02  rvb
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2.1.1  91/08/09  18:30:41  danner
 * 	Removed everything except data access emulation and vector init.
 * 	 The rest of the functions are defunct in 3.0
 * 	[91/08/06            danner]
 * 
 * Revision 2.2  91/07/09  23:18:07  danner
 * 	Initial Checkin
 * 	[91/06/26  12:23:24  danner]
 * 
 * Revision 2.1.1.1  91/06/26  12:23:24  danner
 * 	Initial Checkin
 * 
 * Revision 2.2  91/04/05  14:03:18  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * Revision 2.1.1.1  91/04/02  18:33:15  rvb
 * 	Flush jkdb.
 * 	[91/04/01            rvb]
 * 
 * Revision 2.1  91/03/06  15:02:59  rvb
 * my changes to omron for X134
 * 
 * Revision 2.1  90/09/14  02:56:05  jfriedl
 *  Omron Luna/88k MACH Release 1.00 [release name version_1_00]
 * 
 * Revision 2.0  90/08/28  13:05:56  jfriedl
 * about to redo exception handler
 * 
 * Revision 1.8  90/08/21  11:54:32  jfriedl
 * ripped out the interrupt stack stuff.
 * 
 *    Revision 1.6  90/06/12  13:49:46  jfriedl
 *    Added raw_printf printf-type routine.  Debugged for data-access emulation
 *    loop.  Also allowed all exceptions to set the valid bit of DMT0 so
 *    that it's checked just before return.  This is so that a user program
 *    can set it when servicing a bus error and have it redone.
 *    
 *    Revision 1.5  90/06/05  10:49:31  jfriedl
 *    Brought RCS files from version 2.0 to here (version 2.5).
 *    Also added raw_printf() to print strings to the terminal
 *    without a lot of overhead.
 *    
 *    Revision 1.4  90/04/18  15:06:19  jfriedl
 *    after debugging new processor. Now has kluges for new processors.
 *    
 *    Revision 1.1.1.1  90/04/12  22:28:36  jfriedl
 *    Jeff's locore, initial check-in to RCS.
 *****************************************************************RCS**/

#include "locore.h"
#include "luna88k/asm.h"
#include "motorola/m88k/trap.h"      /* for T_ defines */
#include "motorola/m88k/m88100/m88100.h"
#include "kern/thread.h"
#include "kern/lock.h"


#if defined(RAW_PRINTF) && RAW_PRINTF
#define DEBUG_MSG raw_printf
#else
# if MACH_KDB
#  define DEBUG_MSG db_printf
# else
#  define DEBUG_MSG printf
# endif
#endif

/*
 *  data access emulation for M88100 exceptions
 */
#define DMT_BYTE	1
#define DMT_HALF	2
#define DMT_WORD	4

static struct
{
    unsigned char 	offset;
    unsigned char 	size;
} dmt_en_info[16] =
{
    {0, 0}, {3, DMT_BYTE}, {2, DMT_BYTE}, {2, DMT_HALF},
    {1, DMT_BYTE}, {0, 0}, {0, 0}, {0, 0},
    {0, DMT_BYTE}, {0, 0}, {0, 0}, {0, 0},
    {0, DMT_HALF}, {0, 0}, {0, 0}, {0, DMT_WORD}
};

#if EH_DEBUG
   int data_access_emulation_debug = 0;
   static char *bytes[] =
   {
	    "____", "___x", "__x_", "__xx",
	    "_x__", "_x_x", "_xx_", "_xxx",
	    "x___", "x__x", "x_x_", "x_xx",
	    "xx__", "xx_x", "xxx_", "xxxx",
   };
#  define DAE_DEBUG(stuff) { if (data_access_emulation_debug) { stuff ;}}
#else
#  define DAE_DEBUG(stuff)
#endif

#ifdef LOOP_DEBUG
  int loop_debug = 0;
#endif

#ifdef LOOP_DEBUG
# define LEVELS 10
  int loop_level[4];
  int loop_value[LEVELS][4];
  int loop_fault_location[LEVELS][4];
#endif


void data_access_emulation(eframe)
unsigned long eframe[EF_NREGS];
{
    register int x;
    register struct dmt_reg *dmtx;
    register unsigned  dmax, dmdx;
    register unsigned  v, reg;
#   ifdef LOOP_DEBUG
       int cpu;
#   endif

    if (!(eframe[EF_DMT0] & DMT_VALID))
	return;

#ifdef LOOP_DEBUG
        cpu = cpu_number();
#endif
#if defined(LOOP_DEBUG)
	if (loop_debug)
	{
		DEBUG_MSG("cpu %d level %d DAE: [%s:%x] [%s:%x] [%s:%x]\n",
		    cpu, loop_level[cpu], 
		    eframe[EF_DMT0] & DMT_VALID ? "#0" : "_",
			eframe[EF_DMA0],
		    eframe[EF_DMT1] & DMT_VALID ? "#1" : "_",
			eframe[EF_DMA1],
		    eframe[EF_DMT2] & DMT_VALID ? "#2" : "_",
			eframe[EF_DMA2]);
	}
#endif
    DAE_DEBUG
    (
	DEBUG_MSG("Data access emulation. Valid pipelines: %s %s %s.\n",
	    eframe[EF_DMT0] & DMT_VALID ? "#0" : "_",
	    eframe[EF_DMT1] & DMT_VALID ? "#1" : "_",
	    eframe[EF_DMT2] & DMT_VALID ? "#2" : "_");
    )

#   ifdef LOOP_DEBUG
        if ((loop_level[cpu] > 3) &&
	    (loop_value[cpu][loop_level[cpu]] == eframe[EF_DMA0]) &&
	    (loop_value[cpu][loop_level[cpu]-1] == eframe[EF_DMA0]))
        {
	    DEBUG_MSG("cpu %d: DATA ACCESS LOOP! From %x [fp=%x], SXIP=%x\n",
		    cpu, return_addr, current_sp, eframe[EF_SXIP]);

	    for (x = 0; x < loop_level[cpu]; x++)
	    	DEBUG_MSG("loop_value[%d][cpu%d] is %08x., location is %08x.\n",
		    x, cpu, loop_value[x][cpu], loop_fault_location[x][cpu]);

	    for (x = 0; x < 3; x++) {
		    dmtx = (struct dmt_reg *)&eframe[EF_DMT0+x*3];

		    if (!dmtx->dmt_valid)
			    continue;

		    dmdx = eframe[EF_DMD0+x*3];
		    dmax = eframe[EF_DMA0+x*3];

		    DEBUG_MSG("DMT%d: %08x, DMD%d: %08x, DMA%d: %08x\n",
			    x, eframe[EF_DMT0+x*3],
			    x, eframe[EF_DMD0+x*3],
			    x, eframe[EF_DMA0+x*3]);

		    if (dmtx->dmt_write)
			    DEBUG_MSG("%d: write %x to addr %x(%c)",
				x, dmdx, dmax, dmtx->dmt_das ? 's' : 'u');
		    else
			    DEBUG_MSG("%d: read from addr %x(%c) to r%d, %s, ",
				 x, dmax, dmtx->dmt_das ? 's' : 'u', dmtx->dmt_dreg,
				 dmtx->dmt_signed ? "signed" : "unsigned");

		    DEBUG_MSG(" as [%s]: %s, %s\n",
			bytes[dmtx->dmt_en],
			dmtx->dmt_doub1 ? "double": "not double",
			dmtx->dmt_lockbar ? "xmem": "not xmem");
	    }
	    eframe[EF_DMT0] = 0;
	    asm("jeffmark3: global jeffmark3");
	    loop_level[cpu] = 0;
	    return;
	}
	x = loop_level[cpu]++;
	loop_value[x][cpu] = eframe[EF_DMA0];
	loop_fault_location[x][cpu] = eframe[EF_SXIP];
#   endif /* if LOOP_DEBUG */

    for (x = 0; x < 3; x++)
    {
	dmtx = (struct dmt_reg *)&eframe[EF_DMT0+x*3];

	if (!dmtx->dmt_valid)
		continue;

	dmdx = eframe[EF_DMD0+x*3];
	dmax = eframe[EF_DMA0+x*3];

	DAE_DEBUG
	(
	    DEBUG_MSG("DMT%d: %08x, DMD%d: %08x, DMA%d: %08x\n",
			x, eframe[EF_DMT0+x*3],
			x, eframe[EF_DMD0+x*3],
			x, eframe[EF_DMA0+x*3]);

	    if (dmtx->dmt_write)
		DEBUG_MSG("%d: write %x to addr %x(%c)",
		    x, dmdx, dmax, dmtx->dmt_das ? 's' : 'u');
	    else
		DEBUG_MSG("%d: read from addr %x(%c) to r%d, %s, ",
		     x, dmax, dmtx->dmt_das ? 's' : 'u', dmtx->dmt_dreg,
		     dmtx->dmt_signed ? "signed" : "unsigned");

	    DEBUG_MSG(" as [%s]: %s, %s\n",
		    bytes[dmtx->dmt_en],
		    dmtx->dmt_doub1 ? "double": "not double",
		    dmtx->dmt_lockbar ? "xmem": "not xmem");
	)

	dmax += dmt_en_info[dmtx->dmt_en].offset;
	reg = dmtx->dmt_dreg;

	if ( ! dmtx->dmt_lockbar)
	{
	    /* the fault is not during an XMEM */

	    if (x == 2 && dmtx->dmt_doub1)
	    {
		/* pipeline 2 (earliest stage) for a double */

		if (dmtx->dmt_write)
		{
		    /* STORE DOUBLE WILL BE RE-INITIATED BY rte */
		}
		else
		{
		    /* EMULATE ld.d INSTRUCTION */
		    if (dmtx->dmt_bo)
		    {
			/* LITTLE ENDIAN */
			v = do_load_word(dmax, dmtx->dmt_das, 1);
			if (reg != 31)
			    eframe[EF_R0 + reg + 1] = v;
			v = do_load_word(dmax ^ 4, dmtx->dmt_das, 1);
			if (reg != 0)
			    eframe[EF_R0 + reg] = v;
		    }
		    else
		    {
			/* BIG ENDIAN */
			v = do_load_word(dmax, dmtx->dmt_das, 0);
			if (reg != 0)
			    eframe[EF_R0 + reg] = v;
			v = do_load_word(dmax ^ 4, dmtx->dmt_das, 0);
			if (reg != 31)
			    eframe[EF_R0 + reg + 1] = v;
		    }
		}
	    }
	    else	 /* not pipeline #2 with a double */
	    {
		if (dmtx->dmt_write) switch (dmt_en_info[dmtx->dmt_en].size)
		{
		  case DMT_BYTE:
		    DAE_DEBUG(DEBUG_MSG("byte %x -> [%x(%c)]\n",
			dmdx & 0xff, dmax, dmtx->dmt_das ? 's' : 'u'))
		    do_store_byte(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		    break;
		  case DMT_HALF:
		    DAE_DEBUG(DEBUG_MSG("half %x -> [%x(%c)]\n",
			dmdx & 0xffff, dmax, dmtx->dmt_das ? 's' : 'u'))
		    do_store_half(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		    break;
		  case DMT_WORD:
		    DAE_DEBUG(DEBUG_MSG("word %x -> [%x(%c)]\n", dmdx, dmax,
			dmtx->dmt_das ? 's' : 'u'))
		    do_store_word(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		    break;
		}
		else	/* else it's a read */
		{
		    switch (dmt_en_info[dmtx->dmt_en].size)
		    {
		      case DMT_BYTE:
			v = do_load_byte(dmax, dmtx->dmt_das, dmtx->dmt_bo);
			if (!dmtx->dmt_signed)
			    v &= 0x000000ff;
			break;
		      case DMT_HALF:
			v = do_load_half(dmax, dmtx->dmt_das, dmtx->dmt_bo);
			if (!dmtx->dmt_signed)
			    v &= 0x0000ffff;
			break;
		      case DMT_WORD:
		      default: /* 'default' just to shut up lint */
			v = do_load_word(dmax, dmtx->dmt_das, dmtx->dmt_bo);
			break;
		    }
	   	    if (reg == 0) {
			DAE_DEBUG(DEBUG_MSG("no write to r0 done\n"));
		    }
		    else
		    {
			DAE_DEBUG(DEBUG_MSG("r%d <- %x\n", reg, v));
			eframe[EF_R0 + reg] = v;
		    }
		}
	    }
	}
	else /* if lockbar is set... it's part of an XMEM */
	{
	    /*
	     * According to Motorola's "General Information",
	     * the dmt_doub1 bit is never set in this case, as it should be.
	     * They call this "general information" - I call it a f*cking bug!
	     *
	     * Anyway, if lockbar is set (as it is if we're here) and if
	     * the write is not set, then it's the same as if doub1
	     * was set...
	     */
	    if ( ! dmtx->dmt_write)
	    {
		if (x != 2)
		{
		    /* RERUN xmem WITH DMD(x+1) */
		    x++;
		    dmdx = eframe[EF_DMD0 + x*3];
		}
		else
		{
		    /* RERUN xmem WITH DMD2 */
		}

		if (dmt_en_info[dmtx->dmt_en].size == DMT_WORD)
		    v = do_xmem_word(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		else
		    v = do_xmem_byte(dmax, dmdx, dmtx->dmt_das, dmtx->dmt_bo);
		eframe[EF_R0 + reg] = v;
	    }
	    else
	    {
		if (x == 0)
		{
		    asm("xmem_dae: global xmem_dae");
		    eframe[EF_R0 + reg] = dmdx;
		    eframe[EF_SFIP] = eframe[EF_SNIP];
		    eframe[EF_SNIP] = eframe[EF_SXIP];
		    eframe[EF_SXIP] = 0;
		    /* xmem RERUN ON rte */
		    eframe[EF_DMT0] = 0;
#		    ifdef LOOP_DEBUG
			if (--loop_level[cpu] < 0)
				loop_level[cpu] = 0;

			
#		    endif
		    return;
		}
	    }
	}
    }
    eframe[EF_DMT0] = 0;

#   ifdef LOOP_DEBUG
	--loop_level[cpu];
	if (loop_level[cpu] < 0)
	    loop_level[cpu] = 0;
#   endif
}

/*
 ***********************************************************************
 ***********************************************************************
 */
#define SIGSYS_MAX      501
#define SIGTRAP_MAX     511

#define EMPTY_BR	0xC0000000U      /* empty "br" instruction */
#define NO_OP 		0xf4005800U      /* "or r0, r0, r0" */

typedef struct
{
	unsigned word_one,
		 word_two;
} m88k_exception_vector_area;

#define BRANCH(FROM, TO) (EMPTY_BR | ((unsigned)(TO) - (unsigned)(FROM)) >> 2)

#define SET_VECTOR(NUM, to, VALUE) {                                       \
	unsigned _NUM = (NUM);                                             \
	unsigned _VALUE = (VALUE);                                         \
	vector[_NUM].word_one = NO_OP; 	                                   \
	vector[_NUM].word_two = BRANCH(&vector[_NUM].word_two, _VALUE);    \
}


/*
 * vector_init(vector, vector_init_list)
 *
 * This routine sets up the m88k vector table for the running processor.
 * It is called with a very little stack, and interrupts disabled,
 * so don't call any other functions!
 */
void
vector_init(vector, vector_init_list)
register m88k_exception_vector_area vector[];
register unsigned vector_init_list[];
{
    register unsigned num;
    register unsigned vec;
    extern void sigsys(), sigtrap(), stepbpt(), userbpt();

    for (num = 0; (vec = vector_init_list[num]) != END_OF_VECTOR_LIST; num++)
    {
	if (vec != PREDEFINED_BY_ROM)
	    SET_VECTOR(num, to, vec);
    }

    while (num <= SIGSYS_MAX)
	SET_VECTOR(num++, to, sigsys);

    while (num <= SIGTRAP_MAX)
	SET_VECTOR(num++, to, sigtrap);

    SET_VECTOR(504, to, stepbpt);
    SET_VECTOR(511, to, userbpt);
}
