/*
 *
 * Mach Operating System
 * Copyright (c) 1991, 1992 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	luna_init.c,v $
 * Revision 2.7  92/05/21  17:20:03  jfriedl
 * 	     Cleaned up and de-linted.
 * 	Added prototypes.
 * 	Removed include of <jkdb.h>.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.6  92/05/04  11:27:11  danner
 * 	Correct version declaration.
 * 	[92/04/13  10:04:08  danner]
 * 
 * 	Print version.
 * 	[92/04/10  20:15:12  danner]
 * 
 * Revision 2.5  92/03/03  15:38:36  rpd
 * 	Fixed irritating palette initialization code.
 * 	[92/03/02            danner]
 * 
 * Revision 2.4  92/02/18  18:02:14  elf
 * 	Fixed type of end, conditionalized db_pallete configuration on
 * 	 diptest.
 * 	[92/02/11            danner]
 * 	Changed types of edata and end to char. Without this, only 1/4
 * 	 of the bss gets zeroed.
 * 	[92/01/28            danner]
 * 
 * 	Made first_addr and last_addr initialized so they will not be in
 * 	 the BSS. They are refered to before move bootstrap
 * 	[92/01/28            danner]
 * 	luna88k/cpu.h -> luna88k/cpu_number.h
 * 	[92/01/21            danner]
 * 
 * 	Removed include of luna88k/m88k.h
 * 	[92/01/20            danner]
 * 
 * 	Adapted to use move_bootstrap.
 * 	[92/01/19            danner]
 * 
 * Revision 2.3  91/08/24  12:12:25  af
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:31  danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:09  danner
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:31  danner]
 * 
 * Revision 2.2  91/07/09  23:18:25  danner
 * 	Added avail_next.
 * 	[91/06/25            danner]
 * 
 * 	Moved kdb_init here from machdep.
 * 	[91/05/13            danner]
 * 
 * Revision 2.2  91/04/05  14:03:33  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 *	File:	luna88k/luna_init.c
 *	Basic initialization for LUNA.
 */

#include <mach_kdb.h>
#include <ddb/nlist.h>
#include <sys/types.h>
#include <luna88k/scr.h>
#include <luna88k/cpu_number.h>
#include <luna88k/board.h>
#include <luna88k/machparam.h>
#include <luna88k/pmap.h>
#include <luna88kdev/ramdisk.h>
#include <luna88k/pio.h> /* for diptest */

#include <mach/machine.h>
#include <mach/vm_param.h>
#include <mach/vm_prot.h>
#include <vm/vm_page.h>
#include <sys/reboot.h>

int boothowto;	/* read in kern/bootstrap */

#ifdef WANT_PROTOTYPES

 /* kern/bootstrap.c */
 extern vm_offset_t move_bootstrap(void);

 /* kern/printf.c */
 extern void printf(char*,...);

 /* luna88k/db_interface.c */
 extern void kdb_init(void);
 extern void luna_db_set_palette(unsigned int,unsigned int,unsigned int,unsigned int,unsigned int,unsigned int);

 /* luna88k/locore/eh.c */
 extern boolean_t badwordaddr(void*);

 /* luna88k/locore/locore_asm_routines.c */
 extern void blkclr(void*,unsigned int);
 extern void set_cpu_number(unsigned);

 /* luna88k/pmap.c */
 extern void pmap_bootstrap(vm_offset_t,vm_offset_t*,vm_offset_t*,vm_offset_t*,vm_offset_t*);
 extern void pmap_set_kernel_mmu(void);

 /* motorola/m88k/m88100/cmmu.c */
 extern unsigned int cmmu_cpu_number(void);
 extern void cmmu_init(void);
 extern void cmmu_parity_enable(void);

 /* vm/vm_resident.c */
 extern void vm_set_page_size(void);

 /* forwards */
 static void size_memory(void);
 void luna_init(void);
#endif /* ifdef WANT_PROTOTYPES */


#if	MACH_LOAD
extern int	loadpt;
#else	MACH_LOAD
#define	loadpt	0
#endif	MACH_LOAD

#define SYSBASE		SYSV_BASE
#define PHYS(addr)	((unsigned)(addr) - SYSV_BASE + loadpt)
#define roundup(value, stride) (((unsigned)(value) + (stride) - 1) & ~((stride)-1))

vm_size_t	mem_size;
vm_size_t	rawmem_size;
vm_offset_t	first_addr = 0;
vm_offset_t	last_addr = 0;

vm_offset_t	avail_start, avail_next, avail_end;
vm_offset_t	virtual_avail, virtual_end;

/*
 * Figure out how much real memory is available.
 * Start looking from the megabyte after the end of the kernel data,
 * until we find non-memory.
 */
static void size_memory()
{
    volatile unsigned int *look;
    unsigned int *max;
    extern char end;
#   define PATTERN   0x5a5a5a5a
#   define STRIDE    (1024*1024)

    /*
     * count it up.
     */
    max = (void*)MAXPHYSMEM;
    for (look = (void*)roundup(&end, STRIDE); look < max; look += STRIDE/sizeof(*look))
    {
	unsigned save;

	/* if can't access, we've reached the end */
	if (badwordaddr((void*)look))
		break;

	/*
	 * If we write a value, we expect to read the same value back.
	 * We'll do this twice, the 2nd time with the opposite bit
	 * pattern from the first, to make sure we check all bits.
	 */
	save = *look;
	if (*look = PATTERN, *look != PATTERN)
		break;
	if (*look = ~PATTERN, *look != ~PATTERN)
		break;
	*look = save;
    }
    rawmem_size = (vm_size_t)look;
    mem_size = (vm_size_t)((unsigned)look - (unsigned)loadpt);

    /* first addr is now set by move_bootstrap, which is also
       smart enough to preserve the symbol table. */

    last_addr = trunc_page(look);
}

void luna_init()
{
    extern int clock_enabled[];
    extern char version[];

    extern char edata, end;
    extern vm_offset_t kernel_segment_map;

    /* these must be set before move_bootstrap is called,
       else it might align things to the wrong page size */

    page_size = 8192;
    vm_set_page_size();

    first_addr = move_bootstrap();
    first_addr = round_page(first_addr);

    /* zero the kernel bss */

    blkclr((caddr_t) PHYS(&edata),
	   (unsigned) ((&end - &edata) - sizeof(int)));

    /*
     *	We won't really know our cpu number till after
     *	we configure... but we still need to have a valid
     *	cpu number for routines that use it.  Fortunately,
     *	set_cpu_number will do the right thing for us here
     *	because cpu_number doesn't know we have multi-port
     *	memory yet.
     */
    cmmu_init();
    set_cpu_number(master_cpu = cmmu_cpu_number());

    /* get rid of that horrible black text on a white background */
    luna_db_set_palette(15,15,15, 0, 0, 0);

#if    MACH_KDB
    /* so debugger won't wake slaves up until they are ready */
    {
	int i;
	for (i = 0; i < NCPUS; i++)
	    machine_slot[i].is_cpu = 0;
	machine_slot[master_cpu].is_cpu = 1;
    }
    /* initialize kernel debugger */
    kdb_init();
#endif MACH_KDB

    size_memory();
    cmmu_parity_enable();

    /*
     *	Initialize kernel physical map, mapping the
     *	region from loadpt to avail_start.
     *	Kernel virtual address starts at VM_KERNEL_MIN_ADDRESS.
     */

    avail_start = first_addr;
    avail_end = last_addr;
    printf("%s",version);
    printf("LUNA boot: memory from 0x%x to 0x%x\n", avail_start, avail_end);
    pmap_bootstrap(loadpt, &avail_start, &avail_end, &virtual_avail,
			    &virtual_end);
    /*
     * MMU enable. Kernel code is the same address both V and R
     */
    pmap_set_kernel_mmu();

    /*
     * Read in kern/bootstrap....
     * Always turn on KDB if compiled in.
     * Boot to singleuser only if dip#0 [leftmost] is on [down].
     */
    boothowto = RB_KDB | (diptest(0) ? RB_SINGLE : 0);
}
