/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	m88k.h,v $
 * Revision 2.2  91/07/09  23:18:28  danner
 * 	Removed all the EF_* defines - they are now created by genassym.
 * 	[91/07/04            danner]
 * 	Initial Checkin
 * 	[91/06/26  12:25:58  danner]
 * 
 * Revision 2.2  91/04/05  14:03:38  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

#ifndef	_M88K_DEF_READ_
#define	_M88K_DEF_READ_

#ifndef LOCORE
#include "sys/types.h"
#endif

/*
 *	88100 RISC definitions
 */

/* DMT0, DMT1, DMT2 */
#define DMT_BO		0x00008000	/* Byte-Ordering */
#define DMT_DAS		0x00004000	/* Data Access Space */
#define DMT_DOUB1	0x00002000	/* Double Word */
#define DMT_LOCKBAR	0x00001000	/* Bud Lock */
#define DMT_DREG	0x00000F80	/* Destination Registers 5bits */
#define DMT_SIGNED	0x00000040	/* Sign-Extended Bit */
#define DMT_EN		0x0000003C	/* Byte Enable Bit */
#define DMT_WRITE	0x00000002	/* Read/Write Transaction Bit */
#define	DMT_VALID	0x00000001	/* Valid Transaction Bit */

#ifndef	LOCORE
struct dmt_reg {
    u_int :16,
	  dmt_bo:1,
          dmt_das:1,
	  dmt_doub1:1,
	  dmt_lockbar:1,
	  dmt_dreg:5,
	  dmt_signed:1,
	  dmt_en:4,
	  dmt_write:1,
	  dmt_valid:1;
};
#endif !LOCORE

/*
 * exception vector number
 */
#define EVN_RESET	0	/* reset exception */
#define EVN_INT		1	/* intrrupt exception */
#define EVN_INS		2	/* instruction access exception */
#define EVN_DATA	3	/* data access exception */
#define ENV_MISALIGN	4	/* misaligned access exception */
#define EVN_UNIMP	5	/* unimplemented opcode exception */
#define EVN_PRIV	6	/* privilege violation exception */
#define EVN_BOUND	7	/* bounds check violation exception */
#define EVN_DIVIDE	8	/* illegal integer divide exception */
#define EVN_OVERFLOW	9	/* integer overflow exception */
#define EVN_ERROR	10	/* error exception */
#define	EVN_BREAK	64	/* break point (luna specific) */
#define	EVN_TRACE	65	/* trace (luna specific) */
#define	EVN_KDB		66	/* enter to KDB (luna specific) */
#define	EVN_UNKNOWN	67	/* undefined exception (luna specific) */
#define EVN_ROM_BREAK	68	/* ROM break point */
#define EVN_ROM_TRACE	69	/* ROM trace */
#define EVN_ROM		70	/* ROM enter at initialization */
#define EVN_ROM_ABORT	71      /* ROM enter on ABORT interrupt */
#define EVN_FLOAT_P	114	/* floating-point precise exception */
#define EVN_FLOAT_I	115	/* floating-point imprecise exception */
#define EVN_USER	128	/* user defined exception */
#define EVN_SYSCALL	128	/* BSD system call */
#define EVN_MACH	129	/* MACH system call */
#define EVN_SIGRET	448	/* sigret BCS system call */
#define EVN_SYSLOCAL	449	/* sys_local BCS system call */
#define	EVN_BCS_SYSCALL	450	/* BCS system call */
#define EVN_SIGTRAP0	504	/* SIGTRAP */
#define EVN_SIGTRAP1	505	/* SIGTRAP */
#define EVN_SIGTRAP2	506	/* SIGTRAP */
#define EVN_SIGTRAP3	507	/* SIGTRAP */
#define EVN_SIGTRAP4	508	/* SIGTRAP */
#define EVN_SIGTRAP5	509	/* SIGTRAP */
#define EVN_SIGTRAP6	510	/* SIGTRAP */
#define EVN_SIGTRAP7	511	/* SIGTRAP */
#define EVN_MAX		512	/* maximum exception number <= 512 */

/*
 *	88200 CMMU definitions
 */

#define CMMU_IDR	0x000	/* CMMU id register */
#define CMMU_SCR	0x004	/* system command register */
#define CMMU_SSR	0x008	/* system status register */
#define CMMU_SAR	0x00C	/* system address register */
#define CMMU_SCTR	0x104	/* system control register */
#define CMMU_PFSR	0x108	/* P bus fault status register */
#define CMMU_PFAR	0x10C	/* P bus fault address register */
#define CMMU_SAPR	0x200	/* supervisor area pointer register */
#define CMMU_UAPR	0x204	/* user area pointer register */
#define CMMU_BWP0	0x400	/* block ATC writer port 0 */
#define CMMU_BWP1	0x404	/* block ATC writer port 1 */
#define CMMU_BWP2	0x408	/* block ATC writer port 2 */
#define CMMU_BWP3	0x40C	/* block ATC writer port 3 */
#define CMMU_BWP4	0x410	/* block ATC writer port 4 */
#define CMMU_BWP5	0x414	/* block ATC writer port 5 */
#define CMMU_BWP6	0x418	/* block ATC writer port 6 */
#define CMMU_BWP7	0x41C	/* block ATC writer port 7 */
#define CMMU_CDP0	0x800	/* cache data port 0 */
#define CMMU_CDP1	0x804	/* cache data port 1 */
#define CMMU_CDP2	0x808	/* cache data port 2 */
#define CMMU_CDP3	0x80C	/* cache data port 3 */
#define CMMU_CTP0	0x840	/* cache tag port 0 */
#define CMMU_CTP1	0x844	/* cache tag port 1 */
#define CMMU_CTP2	0x848	/* cache tag port 2 */
#define CMMU_CTP3	0x84C	/* cache tag port 3 */
#define CMMU_CSSP	0x880	/* cache set status register */

/* 88204 CMMU definitions  */
#define CMMU_CSSP0	0x880	/* cache set status register */
#define CMMU_CSSP1	0x890	/* cache set status register */
#define CMMU_CSSP2	0x8A0	/* cache set status register */
#define CMMU_CSSP3	0x8B0	/* cache set status register */

/* CMMU systerm commands */
#define CMMU_FLUSH_USER_LINE		0x30	/* flush PATC */
#define CMMU_FLUSH_USER_PAGE		0x31
#define CMMU_FLUSH_USER_SEGMENT		0x32
#define CMMU_FLUSH_USER_ALL		0x33
#define CMMU_FLUSH_SUPER_LINE		0x34
#define CMMU_FLUSH_SUPER_PAGE		0x35
#define CMMU_FLUSH_SUPER_SEGMENT	0x36
#define CMMU_FLUSH_SUPER_ALL		0x37
#define CMMU_PROBE_USER			0x20	/* probe user address */
#define CMMU_PROBE_SUPER		0x24	/* probe supervisor address */
#define CMMU_FLUSH_CACHE_INV_LINE	0x14	/* data cache invalidate */
#define CMMU_FLUSH_CACHE_INV_PAGE	0x15
#define CMMU_FLUSH_CACHE_INV_SEGMENT	0x16
#define CMMU_FLUSH_CACHE_INV_ALL	0x17
#define CMMU_FLUSH_CACHE_CB_LINE	0x18	/* data cache copyback */
#define CMMU_FLUSH_CACHE_CB_PAGE	0x19
#define CMMU_FLUSH_CACHE_CB_SEGMENT	0x1A
#define CMMU_FLUSH_CACHE_CB_ALL		0x1B
#define CMMU_FLUSH_CACHE_CBI_LINE	0x1C	/* copyback and invalidate */
#define CMMU_FLUSH_CACHE_CBI_PAGE	0x1D
#define CMMU_FLUSH_CACHE_CBI_SEGMENT	0x1E
#define CMMU_FLUSH_CACHE_CBI_ALL	0x1F

/* CMMU system control command */
#define CMMU_SCTR_PE	0x00008000	/* parity enable */
#define CMMU_SCTR_SE	0x00004000	/* snoop enable */
#define CMMU_SCTR_PR	0x00002000	/* priority arbitration */

/* CMMU P bus fault status */
#define CMMU_PFSR_SUCCESS	0	/* no fault */
#define CMMU_PFSR_BERROR	3	/* bus error */
#define CMMU_PFSR_SFAULT	4	/* segment fault */
#define CMMU_PFSR_PFAULT	5	/* page fault */
#define CMMU_PFSR_SUPER		6	/* supervisor violation */
#define CMMU_PFSR_WRITE		7	/* writer violation */

/* Area Description */
#define AREA_D_WT	0x00000200	/* write through */
#define AREA_D_G	0x00000080	/* global */
#define AREA_D_CI	0x00000040	/* cache inhibit */
#define AREA_D_TE	0x00000001	/* translation enable */

/* Segment Description */
#define SEG_D_WT	0x00000200	/* write through */
#define SEG_D_SP	0x00000100	/* supervisor protection */
#define SEG_D_G		0x00000080	/* global */
#define SEG_D_CI	0x00000040	/* cache inhibit */
#define SEG_D_WP	0x00000004	/* write protect */
#define SEG_D_V		0x00000001	/* valid */

#ifndef	LOCORE
/*
 * This file defines the data structures for the mmu.
 * One major data structure, the page descriptor, is not defined here
 * but rather in pte.h as struct pte.
 */

struct area_d {			/* area descriptor */
u_int	ad_addr:20,		/* segment table base address */
	:2,
	ad_wt:1,		/* write through */
	:1,
	ad_g:1,			/* global */
	ad_ci:1,		/* cache inhibit */
	:5,
	ad_te:1;		/* translation enable */
};

struct segment_d {		/* segment descriptor */
u_int	sd_addr:20,		/* page table base address */
	:2,
	sd_wt:1,		/* write through */ 
	sd_sp:1,		/* supervisor protection */
	sd_g:1,			/* global */
	sd_ci:1,		/* cache inhibit */
	:3,
	sd_wp:1,		/* write protect */
	:1,
	sd_v:1;			/* valid */
};

typedef	struct segment_d segment_d_t;

struct pfsr {			/* P bus fault status register */
u_int	:13,
	pfsr_fc:3,		/* falut code */
	:16;
};

struct batc {			/* block address translation register */
u_int	batc_lba:13,		/* logical block address */
	batc_pba:13,		/* physical block address */
	batc_s:1,		/* supervisor */
	batc_wt:4,		/* write through */
	batc_g:1,		/* global */
	batc_ci:1,		/* cache inhibit */
	batc_wp:1,		/* write protect */
	batc_v:1;		/* valid */
};
#endif	LOCORE

/*
 *	Flags for cmmu_flush_tlb
 */
#define FLUSH_KERNEL	1
#define FLUSH_USER	0

#define	FLUSH_ALL	(-1)

#define NBPW    sizeof(int) /* bytes per word */
#define NBSG    4*1024*1024 /* segment size */
#define NBPG    4*1024      /* bytes per mmu page */

#endif	_M88K_DEF_READ_
