/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	pmap.h,v $
 * Revision 2.7  92/05/21  17:20:48  jfriedl
 * 	Made batc entries in struct pmap unions to allow access to them
 * 	as unsigned entities.
 * 	[92/05/18            jfriedl]
 * 
 * Revision 2.6  92/03/31  15:18:06  rpd
 * 	Removed pmap_remove_attributes.
 * 	[92/03/25            rpd]
 * 
 * Revision 2.5  92/02/18  18:02:39  elf
 * 	Added CACHE_INHIBIT definition, removed the M88K_* defines.
 * 	Removed segmap definition. 
 * 	Merged up.
 * 
 * 	Commented out broken assert 
 * 	[92/01/12            danner]
 * 
 * Revision 2.4  92/01/14  16:45:10  rpd
 * 	Removed pmap_list_resident_pages.
 * 	[91/12/31            rpd]
 * 
 * Revision 2.3  91/08/24  12:15:28  af
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:48:59  danner]
 * 	Volatilized
 * 	[91/07/19            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:37  danner
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:48:59  danner]
 * 
 * 	Volatilized
 * 	[91/07/19            danner]
 * 
 * Revision 2.2  91/07/09  23:19:14  danner
 * 	Added noops for pmap_list_resident_pages and
 * 	 pmap_remove_attributes.
 * 	[91/05/09            danner]
 * 
 * Revision 2.3  91/04/05  14:04:26  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

#define bcopy(a,b,c) memcpy(b,a,c)

/*
 *	Copyright (C) 1985, Avadis Tevanian, Jr., Michael Wayne Young
 *
 *	Machine-dependent structures for the physical map module.
 */

#ifndef	_PMAP_MACHINE_
#define _PMAP_MACHINE_	1

#ifndef	ASSEMBLER

#include <mach/machine/vm_param.h>
#include <mach/vm_statistics.h>

#ifdef	KERNEL
#include <kern/lock.h>
#include <luna88k/pcb.h>
#endif	KERNEL

#define CACHE_INHIBIT 0x40  

/*
 *	MC88200 Page Descriptor
 */

struct pt_entry {	/* naming for VAX simulation */
    unsigned int
      pfn:20,		/* physical frame address */
      :2,		/* reserved for future extension */
      wt:1,		/* is writethrough ? */
      sp:1,		/* is supervisor-protected */
      g:1,		/* is global ? */
      ci:1,		/* is cache inhibited ? */
      :1,		/* reserved for future extension */
      modify:1,		/* is modified ? */
      u:1,		/* is used (sccessed) ? */
      wp:1,		/* is write-protected ? */
      wired:1,
      valid:1;		/* is valid ? */
};

typedef struct pt_entry	pt_entry_t;
#define	PT_ENTRY_NULL	((pt_entry_t *) 0)

struct BATC {
    unsigned int
      lba:13,		/* logical block address */
      pba:13,		/* physical block address */
      s:1,		/* is supervisor-mode ? */
      wt:1,		/* is writethrough ? */
      g:1,		/* is global ? */
      ci:1,		/* is cache inhibited ? */
      wp:1,		/* is write-protected ? */
      v:1;		/* is valid ? */
};

typedef struct BATC batc_t;
#endif	ASSEMBLER

/*
 *	Misc. useful macros.
 */
#define bit_set(n,p)	(*(p) |= (1 << (n)))
#define bit_clear(n,p)	(*(p) &= ~(1 << (n)))

#define i_bit_set(n,p,l)	(simple_lock(l), bit_set(n, p), simple_unlock(l))
#define i_bit_clear(n,p,l)	(simple_lock(l), bit_clear(n, p), simple_unlock(l))

#ifndef	ASSEMBLER
#ifdef	KERNEL
#include <kern/macro_help.h>

typedef	volatile long	cpu_set;	/* set of CPUs - must be <= 32 */

struct pmap {
	pt_entry_t	*p0br;		/* P0 base register */
	long		p0lr;		/* P0 length register */
	pt_entry_t	*p1br;		/* P1 base register */
	long		p1lr;		/* P1 length register */
	volatile int	ref_count;	/* reference count */
	decl_simple_lock_data(, lock)	/* lock on map */
	struct pmap_statistics	stats;	/* map statistics */
	cpu_set		cpus_using;	/* bitmap of cpus using pmap */
	cpu_set		seg_changed;	/* bitmap of cpu must update segment */
#define	NBATC	8		/* Number of block address translation cache */
#if 1
	union batc_access
	{
		batc_t	 cooked;
	        unsigned raw;
	} i_batc[NBATC], d_batc[NBATC];
#else
	batc_t		i_batc[NBATC];  /* instruction block address	     */
					/* translation cache		     */
	batc_t		d_batc[NBATC];	/* data block address translation    */
#endif
};

typedef struct pmap	*pmap_t;

#define PMAP_NULL	((pmap_t) 0)

/*
 *	List of cpus that are actively using mapped memory.  Any
 *	pmap update operation must wait for all cpus in this list.
 *	Update operations must still be queued to cpus not in this
 *	list.
 */
cpu_set			cpus_active;
simple_lock_data_t	cpus_active_lock;
/*
 *	List of cpus that are idle, but still operating, and will want
 *	to see any kernel pmap updates when they become active.
 */
cpu_set			cpus_idle;
simple_lock_data_t 	cpus_idle_lock;
simple_lock_data_t	cpus_using_lock;

/*
 *	Quick test for pmap update requests.
 */
volatile boolean_t	cpu_update_needed[NCPUS];

/*
 *	External declarations for PMAP_ACTIVATE.
 */

void		process_pmap_updates();
void		pmap_update_interrupt();
extern	pmap_t	kernel_pmap;


/*
 *	For multiple CPUS, PMAP_ACTIVATE and PMAP_DEACTIVATE must manage
 *	fields to control TLB invalidation on other CPUS.
 */
/*
 *	Routine:	PMAP_ACTIVATE (,pmap_activate)
 *	Function:
 *		Bind the given physical map to the given processor
 */

#define PMAP_ACTIVATE(pmap, th, my_cpu)					\
{									\
    register pmap_t   tpmap = (pmap);					\
    /*									\
     *	Let pmap updates proceed while we wait for this pmap.		\
     */									\
    i_bit_clear((my_cpu), &cpus_active, &cpus_active_lock);		\
    /*									\
     *	Lock the pmap to put this cpu in its active set. 		\
     *	Wait for updates here.						\
     */									\
    simple_lock(&tpmap->lock);						\
	                                                                \
    if (tpmap == kernel_pmap){						\
	/*								\
	 *  invalidate requests for the kernel pmap. 			\
	 *	Don't load the map registers.				\
	 */								\
	if (cpu_update_needed[(my_cpu)])				\
	    process_pmap_updates(tpmap);				\
    }									\
    /*									\
     *	Mark that this cpu is using the pmap.				\
     *  tpmap is already locked, so i_bit_set is not used , but bit_set \
     */									\
    i_bit_set((my_cpu), &tpmap->cpus_using, &cpus_using_lock);		\
    /*									\
     *	Mark this cpu active - IPL will be lowered by			\
     *	load_context().							\
     */									\
	                                                                \
    i_bit_set((my_cpu), &cpus_active, &cpus_active_lock);		\
	                                                                \
    simple_unlock(&tpmap->lock);					\
}

#define PMAP_DEACTIVATE(pmap, thread, my_cpu)	{			\
        pmap_t _pm_ = (pmap);						\
        unsigned int _mc_ = (my_cpu);					\
	/*								\
	 *	Mark pmap no longer in use by this cpu even if		\
	 *	pmap is locked against updates.				\
	 */								\
	i_bit_clear((_mc_), &(_pm_)->cpus_using, &cpus_using_lock);	\
}

#define	MARK_CPU_IDLE(my_cpu)	{					\
	/*								\
	 *	Mark this cpu idle, and remove it from the active set,	\
	 *	since it is not actively using any pmap.  Signal_cpus	\
	 *	will notice that it is idle, and avoid signaling it,	\
	 *	but will queue the update request for when the cpu	\
	 *	becomes active.						\
	 */								\
	int	_s_ = splvm();						\
	i_bit_set((my_cpu), &cpus_idle, &cpus_idle_lock);		\
	/* i_bit_clear((my_cpu), &cpus_active, &cpus_active_lock); */	\
	splx(_s_);							\
}

#define	MARK_CPU_ACTIVE(my_cpu)						\
{									\
	int	_s_ = splvm();						\
	/*								\
	 *	If a kernel_pmap update was requested while this cpu	\
	 *	was idle, process it as if we got the interrupt. 	\
	 *	Before doing so, remove this cpu from the idle set. 	\
	 *	Since we do not grab any pmap locks while we flush	\
	 *	our TLB, another cpu may start an update operation	\
	 *	before we finish.  Removing this cpu from the idle	\
	 *	set assures that we will receive another update		\
	 *	interrupt if this happens.				\
	 */								\
	i_bit_clear((my_cpu), &cpus_idle, &cpus_idle_lock);		\
									\
	if (cpu_update_needed[(my_cpu)])				\
	 pmap_update_interrupt(my_cpu);					\
									\
	/*								\
	 *	Mark that this cpu is now active.			\
	 */								\
									\
	/* assert(cpus_active & (my_cpu));	*/			\
	/* i_bit_set((my_cpu), &cpus_active, &cpus_active_lock);*/	\
									\
	splx(_s_);							\
}

#define PMAP_PCB_INITIALIZE(pcb)

/*
 *	PMAP_CONTEXT: update any thread-specific hardware context that
 *		is managed by pmap module.
 */

#define PMAP_CONTEXT(pmap, thread)

#define FLUSH_TLB_PER_PAGE(addr)	(cmmu_flush_tlb(FLUSH_KERNEL, addr, M88K_PGBYTES))
#define FLUSH_TLB_ALL()			{cmmu_flush_tlb(FLUSH_KERNEL, 0, FLUSH_ALL);\
					 cmmu_flush_tlb(FLUSH_USER,   0, FLUSH_ALL);}


#define pmap_resident_count(pmap)	((pmap)->stats.resident_count)
#define	pmap_phys_address(frame)	((vm_offset_t) (m88k_ptob(frame)))
#define pmap_phys_to_frame(phys)	((int) (m88k_btop(phys)))
#define pmap_attribute(pmap,addr,siz,attr,val) 0

#endif	KERNEL
#endif	ASSEMBLER

#define zero_page(dst)		(bzero((char *)(dst) + VEQR_ADDR, M88K_PGBYTES))

#define	copy_page(src,dst)	(bcopy((char *)(src) + VEQR_ADDR, \
				       (char *)(dst) + VEQR_ADDR, M88K_PGBYTES))
#define	copy_to_phys(src,dst,cnt)     (bcopy((char *)(src), \
					     (char *)(dst) + VEQR_ADDR, (cnt)))
#define copy_from_phys(src,dst,cnt)   (bcopy((char *)(src) + VEQR_ADDR, \
					     (char *)(dst), (cnt)))
#define	putmemc(dst,byte)	(*((char *)(dst) + VEQR_ADDR) = byte)

#define	getmemc(src)	     	(*((char *)(src) + VEQR_ADDR))

#endif	_PMAP_MACHINE_
