/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	pte.h,v $
 * Revision 2.3  92/05/21  17:20:53  jfriedl
 * 	Appended 'U' to constants that would otherwise be signed.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.2  91/07/09  23:19:20  danner
 * 	Initial Checkin
 * 	[91/06/26  12:31:16  danner]
 * 
 * Revision 2.2  91/04/05  14:04:37  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 *	@(#)pte.h	7.1 (Berkeley) 6/5/86
 */

/*
 * M88000 page table entry
 *
 * There are two major kinds of pte's: those which have ever existed (and are
 * thus either now in core or on the swap device), and those which have
 * never existed, but which will be filled on demand at first reference.
 * There is a structure describing each.  There is also an ancillary
 * structure used in page clustering.
 */

#ifndef LOCORE
struct pte
{
unsigned int	pg_pfnum:20,		/* core page frame number or 0 */
		:2,
		pg_wt:1,		/* write through */
		pg_sp:1,		/* supervisor protection */
		pg_g:1,			/* global */
		pg_ci:1,		/* cache inhibit */
		:1,
		pg_m:1,			/* modified */
		pg_u:1,			/* used */
		pg_wp:1,		/* write protect */
		:1,
		pg_v:1;			/* valid bit */
};
struct hpte
{
unsigned int	pg_pfnum:20,
		:3,
		pg_high:9;		/* special for clustering */
};
struct fpte
{
unsigned int	pg_blkno:24,		/* file system block number */
		pg_fileno:1,		/* file mapped from or TEXT or ZERO */
		pg_fod:1,		/* is fill on demand (=1) */
		:1,
		pg_prot:4,
		pg_v:1;
};
#endif

#define	PG_V		0x00000001U
#define PG_WP		0x00000004U
#define PG_U		0x00000008U
#define PG_M		0x00000010U
#define PG_CI		0x00000040U
#define PG_G		0x00000080U
#define PG_SP		0x00000100U
#define PG_WT		0x00000200U
#define	PG_PFNUM	0xfffff000U
#define PG_SHIFT	12

#define	PG_NOACC	0
#define	PG_KW		0x00000181U
#define	PG_KR		0x00000185U
#define	PG_UW		0x00000081U
#ifdef not_perfect
#define	PG_URKW		0x00000081U
#endif not_perfect
#define	PG_URKR		0x00000085U

#ifndef	LOCORE
#ifdef	KERNEL

/* utilities defined in locore.s */
extern	struct pte *Sysmap;
			/* one of many names for the kernel page table */
#endif
#endif
