/*
 * Mach Operating System
 * Copyright (c) 1991, 1992 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	trap.c,v $
 * Revision 2.9  92/05/21  17:21:21  jfriedl
 * 	Added prototypes.
 * 	[92/05/18            jfriedl]
 * 
 * Revision 2.8  92/05/05  10:00:38  danner
 * 	Add continuations for user page faults.
 * 	[92/05/04            danner]
 * 
 * Revision 2.7  92/05/04  11:27:24  danner
 * 	More cleanup.
 * 	[92/05/03            danner]
 * 	debug
 * 	[92/04/09  19:04:41  danner]
 * 
 * Revision 2.6  92/04/01  10:56:02  rpd
 * 	Drop thread argument to exception.
 * 	[92/03/24            danner]
 * 	Was treating user breakpoint instructions like entry breakpoints,
 * 	 causing user space single stepping to act like noops.
 * 	[92/03/19            danner]
 * 	Removed syscall, mach_syscall and other 2.5 cruft.
 * 	[92/03/03            danner]
 * 
 * Revision 2.5  92/03/03  15:38:41  rpd
 * 	conditionalized gimmeabreak in trap path.
 * 	[92/03/02            danner]
 * 
 * Revision 2.4  92/02/19  13:59:09  elf
 * 	Added code for ddb traps from userspace.
 * 	[92/02/18            danner]
 * 	Updated include of trap.h
 * 	[92/01/30            danner]
 * 	updated includes.
 * 	[92/01/20            danner]
 * 	Removed T_ASTFLT case from trap, as ast_taken() is called
 * 	 directly from assembly in 3.0 instead.
 * 	[91/11/03            danner]
 * 	Support for user level ddb traps. 
 * 	[91/09/19  13:51:09  danner]
 * 
 * Revision 2.3  91/08/24  12:16:58  af
 * 	Replaced use of backtrace with db_stack_trace_cmd.
 * 	[91/08/06            danner]
 * 	Added missing argument to ddb_error_trap.
 * 	[91/07/14            danner]
 * 
 * Revision 2.2.3.1  91/08/19  13:47:52  danner
 * 	Replaced use of backtrace with db_stack_trace_cmd.
 * 	[91/08/06            danner]
 * 	Added missing argument to ddb_error_trap.
 * 	[91/07/14            danner]
 * 
 * Revision 2.2  91/07/09  23:20:02  danner
 * 	Removed frame is bogus warning that depended on scribbling magic
 * 	 constants into frames.
 * 	[91/07/08            danner]
 * 	Initial Checkin
 * 	[91/06/26  12:32:46  danner]
 * 
 * Revision 2.3  91/04/05  13:56:30  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * 		Flush jkdb
 * 
 * 	Support for ddb instead of kdb.
 * 	[91/03/09            danner]
 * 
 */

/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 *	@(#)trap.c	7.1 (Berkeley) 6/5/86
 */
 
#include <mach_kdb.h>
#include <sys/types.h>
#include <motorola/m88k/m88100/psl.h>
#include <motorola/m88k/m88100/m88100.h>
#include <motorola/m88k/trap.h>
#include <luna88k/pte.h>
#include <luna88k/locore/locore.h>
#include <kern/xpr.h>
#include <kern/thread.h>
#include <mach/thread_status.h>
#include <kern/task.h>
#include <vm/vm_kern.h>
#include <vm/vm_map.h>
#include <mach/vm_param.h>
#include <mach/kern_return.h>
#include <kern/sched_prim.h>
#include <kern/syscall_sw.h>
#include <mach/exception.h>
#include <luna88k/db_machdep.h>

#ifdef WANT_PROTOTYPES
 /* ddb/db_output.c */
 extern void db_printf(char*,...);

 /* kern/debug.c */
 extern void panic(char*,...);

 /* kern/exception.c */
 extern void exception(int,int,int);

 /* kern/printf.c */
 extern void printf(char*,...);

 /* kern/xpr.c */
 extern void xpr(char*,int,int,int,int,int);

 /* luna88k/db_interface.c */
 extern int ddb_break_trap(int,db_regs_t*);
 extern int ddb_entry_trap(int,db_regs_t*);
 extern void ddb_error_trap(char*,db_regs_t*);
 extern void gimmeabreak(void);

 /* luna88k/db_trace.c */
 extern void db_stack_trace_cmd(db_regs_t*,int,db_expr_t,char*);

 /* luna88k/locore/locore_asm_routines.c */
 extern unsigned int db_disable_interrupt(void);
 extern void db_enable_interrupt(void);

 /* luna88k/locore/locore_c_routines.c */
 extern void data_access_emulation(long unsigned int*);

 /* luna88k/machdep.c */
 extern unsigned int db_splhigh(void);
 extern unsigned int db_splx(unsigned int);

 /* vm/vm_fault.c */
 extern kern_return_t vm_fault(vm_map_t,vm_offset_t,vm_prot_t,boolean_t,boolean_t,void(*)());

 /* forwards */
 static void user_page_fault_continue(kern_return_t);
 void pre_trap(int,unsigned int*);
 void trap(int,int*);
#endif /* ifdef WANT_PROTOTYPES */


#define USER 040

/* extract user/supervisor mode bit of processor status word */

char	*trap_type[] = {
	"Trap type 0",
	"Priviledge violation",
	"Trap type 2",
	"Instruction access",
	"Data access",
	"Misaligned access",
	"Unimplemented opcode",
	"Bounds check violation",
	"Illegal divide",
	"Integer overflow",
	"Error exception",
	"Floating point precise",
	"Floating point imprecise",
	"Software trap",
	"Kernel debugger entry",
	"Kernel debugger break",
	"Kernel debugger trace",
	"Unknown",
	"Reserved vector 1",
	"Reserved vector 2"
	"Reserved vector 3"
	"Reserved vector 4"
};

#if MACH_KDB
#define DEBUG_MSG db_printf
#else
#define DEBUG_MSG printf
#endif

/*
 *  Need this accessible to the kernel debugger.
 */
int	TRAP_TYPES =	(sizeof trap_type / sizeof trap_type[0]);

#ifdef TRAPTRACE
int traptrace = 0;
#endif TRAPTRACE

/*
  typically entered with interrupts turned off 
*/

/*ARGSUSED*/
void
pre_trap(type, regs)
int type;
unsigned int *regs;
{
#if    MACH_KDB

  if (!USERMODE(regs[EF_EPSR]))
    {
      switch (type)
	{
	case T_KDB_BREAK:
	  {
	    int s = db_splhigh();
	    db_enable_interrupt(); /* turn interrupts on */
	    ddb_break_trap(T_KDB_BREAK,(db_regs_t*)regs);
	    db_disable_interrupt(); /* shut them back off */
	    db_splx(s);
	    return;
	  }
	case T_KDB_ENTRY:
	  {
	    int s = db_splhigh();
	    db_enable_interrupt(); /* turn interrupts on */
	    ddb_entry_trap(T_KDB_ENTRY,(db_regs_t*)regs);
	    db_disable_interrupt(); /* shut them back off */
	    db_splx(s);
	    return;
	  }
	case T_ERRORFLT:
	case T_ILLFLT:
	  {
	    int s = db_splhigh();
	    db_enable_interrupt(); /* turn interrupts on */
	    ddb_error_trap(type == T_ILLFLT ? "unimplemented opcode" :
			   "error fault", (db_regs_t*)regs);
	    db_disable_interrupt(); /* shut them back off */
	    db_splx(s);
	    return;
	  }
	case T_INSTFLT:
	case T_DATAFLT:
	  /* watch point support code should go here */
	  break;
	}
    }
  else
    {
      switch (type)
	{
	case T_KDB_ENTRY:
	  {
	    int s = db_splhigh();
	    db_enable_interrupt(); /* turn interrupts on */
	    ddb_entry_trap(T_KDB_ENTRY,(db_regs_t*)regs);
	    db_disable_interrupt(); /* shut them back off */
	    db_splx(s);
	    return;
	  }
	case T_KDB_BREAK:
	  {
	    int s = db_splhigh();
	    db_enable_interrupt(); /* turn interrupts on */
	    ddb_break_trap(T_KDB_BREAK,(db_regs_t*)regs);
	    db_disable_interrupt(); /* shut them back off */
	    db_splx(s);
	    return;
	  }
	}
    }
#else
  if (USERMODE(regs[EF_EPSR]) && (type == T_KDB_BREAK || 
				   type == T_KDB_ENTRY))
    type = T_UNKNOWNFLT;
#endif  
  
  /* hand off to normal trap routine */
  trap(type,regs);

#if  1
  if (regs[EF_SNIP] < 0x20 || regs[EF_SFIP] < 0x20)
    gimmeabreak();
#endif
}

/*
 *	Mach trap handler.  Called from locore when a processor trap
 *	occurs.
 */

int dae_trap = 0;

static void user_page_fault_continue(kr)
	kern_return_t kr;
{
  unsigned int *locore = (unsigned int *)&(current_thread()->pcb->user_state);
  if (kr == KERN_SUCCESS) /* all went well */
    {
      /* call data access emulation to handle the fault */
      data_access_emulation((unsigned *)&(current_thread()->pcb->user_state));
      locore[EF_DMT0] = 0;
      if (locore[EF_VECTOR] == 2) /* instruction fault */
	{
	  /* back up SXIP, SNIP clearing the the Error bit */
	  locore[EF_SFIP] = locore[EF_SNIP] & ~(FIP_E);
	  locore[EF_SNIP] = locore[EF_SXIP] & ~(NIP_E);
	}
      thread_exception_return();
    }

  /*
   * since we're going to send the BAD ACCESS signal,
   * clear DMT0 so that later we don't do anything in
   * data_access_emulation...
   *      jfriedl@omron   90-Mar-12
   */
  locore[EF_DMT0] = 0;
  
  exception(EXC_BAD_ACCESS,kr,0); 

}

/*ARGSUSED*/
void
trap(type, regs)
	int	type;
	int	*regs;
{
	int	code;
	int	mycpu = cpu_number();
	vm_map_t	map;
	kern_return_t	result;

	int	exc_type, exc_code, exc_subcode;

	if (!USERMODE(regs[EF_EPSR])) {
	    /*
	     *	Trap in system mode.  Only page-faults are valid,
	     *	and then only in special circumstances.
	     */

	    switch (type) {

		case T_INSTFLT:
		case T_DATAFLT:
		    code = (type == T_INSTFLT) 
		             ? regs[EF_SXIP] & XIP_ADDR	
			     : regs[EF_DMA0];

		    if (current_thread() == THREAD_NULL ||
			 (type == T_INSTFLT || (regs[EF_DMT0] & DMT_DAS)))
		      /* && (!current_thread()->task->kernel_vm_space)) */ 
			map = kernel_map;
		    else
			map = current_thread()->task->map;

		    XPR(XPR_VM_FAULT,
			("vm_fault(%x,%x) regs=%x pc=%x th=%x\n",
				    map, code, regs, regs[EF_SXIP],
			            current_thread()));
		    result = vm_fault(map,
				trunc_page((vm_offset_t)code),
				(((type == T_INSTFLT 
				  || !(regs[EF_DMT0] & DMT_WRITE)))
				  && !(regs[EF_DMT0] & DMT_LOCKBAR))
				    ? VM_PROT_READ
				    : VM_PROT_READ|VM_PROT_WRITE,
				FALSE, FALSE, (void  (*)()) 0);


		    if ((result != KERN_SUCCESS) &&
			/* ... and if it's a kernel problem .... */
			(regs[EF_DMT0] & DMT_DAS)
#if VM_MIN_KERNEL_ADDRESS
			&& ((vm_offset_t) code >= VM_MIN_KERNEL_ADDRESS)
#endif
		    )
		    {
			        XPR(XPR_VM_FAULT,
				    ("vm_fault(k,%x) lcr=%x pc=%x(%s) th=%x\n",
					       code, regs,
					       regs[EF_SXIP],
					       USERMODE(regs[EF_EPSR])?"u":"s",
				               current_thread()));
		    		result = vm_fault(kernel_map,
					trunc_page((vm_offset_t)code),
					(((type == T_INSTFLT 
					  || !(regs[EF_DMT0] & DMT_WRITE)))
					  && !(regs[EF_DMT0] & DMT_LOCKBAR))
					    ? VM_PROT_READ
					    : VM_PROT_READ|VM_PROT_WRITE,
				            FALSE, FALSE, (void  (*)()) 0);
		    }
		    if (result == KERN_SUCCESS) {
			return;
		    }

		    {
			extern int	ALLOW_FAULT_START,
					ALLOW_FAULT_END,
					FAULT_ERROR;

			vm_offset_t	recover;
			int pc = regs[EF_SXIP] & XIP_ADDR;
			recover = current_thread()->recover;
			if (recover == (vm_offset_t)0) {
			    if ((vm_offset_t)pc >
			        (vm_offset_t)&ALLOW_FAULT_START
			     && (vm_offset_t)pc <
				(vm_offset_t)&ALLOW_FAULT_END) {

				recover = (vm_offset_t) &FAULT_ERROR;
			    }
			}
			if (recover != (vm_offset_t)0) {
			    regs[EF_SNIP] = ((int)recover + 0) | FIP_V;
			    regs[EF_SFIP] = ((int)recover + 4) | FIP_V;
			    regs[EF_SXIP] = 0;
			    regs[EF_DMT0] = 0;                     
			    return;
			}
		    }
		    if (!(type == T_INSTFLT || regs[EF_DMT0] & DMT_DAS))
                      goto    user_fault;

		    /* Unanticipated page-fault errors in kernel
		       should not happen. */
		    /* fall through */

		default:

		if (regs[EF_VECTOR] == 3)
		{
		   DEBUG_MSG("(CPU %d) kernel bus error " 
			"@ 0x%08x accessing 0x%08x\n",
			mycpu, regs[EF_SXIP], regs[EF_DMA0]);
		}
		else if (regs[EF_VECTOR] == 4)
		{
		    DEBUG_MSG("(CPU %d) kernel misalgined "
			"access exception @ 0x%08x\n", mycpu, regs[EF_SXIP]);
		}
		else
		{
		    DEBUG_MSG("(CPU %d) trap type %d, vector=%d,",
			mycpu, type, regs[EF_VECTOR]);
#if MACH_KDB
		    ddb_error_trap("bad news in kernel\n", (db_regs_t*)regs);
#endif
		    DEBUG_MSG(" sxip=%x, dma0=%x, dmt0=%x\n",
			 regs[EF_SXIP], regs[EF_DMA0], regs[EF_DMT0]);
		}

#if MACH_KDB
		    /* backtrace */
		    db_stack_trace_cmd((db_regs_t*)regs,1,0,NULL);
		    DEBUG_MSG("Entering kernel debugger....\n");
		    DEBUG_MSG("Warning: noncontinuable error");
		    ddb_error_trap("page fault in kernel mode",(db_regs_t*)regs);
#else
		DEBUG_MSG("\n --- CPU %d HALTED --\n", mycpu);

		spl7();
		enable_interrupt();
		while(1)
		    ;
#endif /* end else not MACH_KDB */
		  }
	 }/* if not usermode */
      
	/*
	 *	Trap from user mode.
	 */

    user_fault:
	type |= USER;

	exc_code = 0;
	exc_subcode = 0;

	switch (type) {
	    case T_RESADFLT|USER:	/* never happens on MC88000 */
	    case T_RESOPFLT|USER:	/* never happens on MC88000 */
	    case T_ERRORFLT|USER:	/* cannot resume. */
		printf("{%d}trap type %d, vector = %d, sxip = %x, dma0 = %x\n",
		       mycpu,type, regs[EF_VECTOR], regs[EF_SXIP], regs[EF_DMA0]);
		panic(trap_type[type]);
		break;

	    case T_PRIVINFLT|USER:
		exc_type = EXC_BAD_INSTRUCTION;
		exc_code = EXC_M88K_PRIV;
		break;

	    case T_INSTFLT|USER:
	    case T_DATAFLT|USER:
		code = (type == (T_INSTFLT|USER)) 
		          ? regs[EF_SXIP] & XIP_ADDR
			  : regs[EF_DMA0];
		map = current_thread()->task->map;



		result = vm_fault(map,
				trunc_page((vm_offset_t)code),
				(type == (T_INSTFLT | USER)
				  || !(regs[EF_DMT0] & (DMT_WRITE|DMT_LOCKBAR)))
				    ? VM_PROT_READ
				    : VM_PROT_READ|VM_PROT_WRITE,
				FALSE, FALSE, user_page_fault_continue);
		/* NOTREACHED */
		break;

	    case T_MISALGNFLT|USER:
		exc_type = EXC_BAD_ACCESS;
		exc_code = EXC_M88K_MISALIGN;
		break;

	    case T_ILLFLT|USER:
		exc_type = EXC_BAD_INSTRUCTION;
		exc_code = EXC_M88K_UNIMP;
		break;

	    case T_BNDFLT|USER:
		exc_type = EXC_SOFTWARE;
		exc_code = EXC_M88K_BOUND;
		break;

	    case T_ZERODIV|USER:
		exc_type = EXC_ARITHMETIC;
		exc_code = EXC_M88K_DIVIDE;
		break;

	    case T_OVFFLT|USER:
		exc_type = EXC_ARITHMETIC;
		exc_code = EXC_M88K_OVERFLOW;
		break;

	    case T_FPEPFLT|USER:
		exc_type = EXC_ARITHMETIC;
		exc_code = EXC_M88K_FLOAT_P;
		break;

	    case T_FPEIFLT|USER:
		exc_type = EXC_ARITHMETIC;
		exc_code = EXC_M88K_FLOAT_I;
		break;

	    case T_SIGTRAP|USER:
		exc_type = EXC_BREAKPOINT;
		exc_code = regs[EF_VECTOR];
		break;

	    /*
	     * This trap is used by the kernel to support single-step
	     * debugging (although any user could generate this trap
	     * which should probably be handled differently.  When a
	     * process is continued by a debugger with the PT_STEP
	     * function of ptrace (single step), the kernel inserts
	     * one or two breakpoints in the user process so that only
	     * one instruction (or two in the case of a delayed branch)
	     * is executed.  When this breakpoint is hit, we get the
	     * T_STEPBPT trap.
	     */

	    case T_STEPBPT|USER:

		exc_type = EXC_BREAKPOINT;
		exc_code = regs[EF_VECTOR];

                regs[EF_SFIP] = regs[EF_SNIP];    /* set up next FIP */
                regs[EF_SNIP] = regs[EF_SXIP];    /* set up next NIP */

		break;

	    /*
  	     * This trap is meant to be used by debuggers to implement
	     * breakpoint debugging.  When we get this trap, we just
	     * return a signal which gets caught by the debugger.
	     */

	    case T_USERBPT|USER:

		exc_type = EXC_BREAKPOINT;
		exc_code = regs[EF_VECTOR];

                regs[EF_SFIP] = regs[EF_SNIP];    /* set up the next FIP */
                regs[EF_SNIP] = regs[EF_SXIP];    /* set up the next NIP */

		break;


	    case T_SIGSYS|USER:
		exc_type = EXC_SOFTWARE;
		exc_code = EXC_M88K_RESTRAP;
		exc_subcode = regs[EF_VECTOR];
		break;

	    default:
#if	MACH_KDB
		ddb_error_trap("unknown trap from user mode",(db_regs_t*)regs);
#endif	MACH_KDB
		printf("trap type %d, vector=%d, sxip=%x, dma0=%x from user\n",
			   type, regs[EF_VECTOR], regs[EF_SXIP], regs[EF_DMA0]);
		type &= ~USER;
		if ((unsigned)type < TRAP_TYPES)
			panic(trap_type[type]);
		panic("trap");
	      } 
      
	exception(exc_type, exc_code, exc_subcode);
      
    /* out: */
	/* check signals, other things? */
	return;
}
