/*
 * Mach Operating System
 * Copyright (c) 1992, 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	cons.c,v $
 * Revision 2.6  92/08/03  17:45:32  jfriedl
 * 	Added flexability to support /dev/ttya as the console.
 * 	[92/07/24            jfriedl]
 * 
 * Revision 2.4.2.1  92/05/27  14:45:52  danner
 * 	updated headers.
 * 	[92/05/17            danner]
 * 
 * Revision 2.5  92/05/21  17:21:46  jfriedl
 * 	Cleanup to quiet gcc warnings.
 * 	Made CHAR and DEV_T args to be UNSIGNED (ANSI/std problems).
 * 	Added prototypes.
 * 	[92/05/18            jfriedl]
 * 
 * Revision 2.4  92/05/04  11:27:32  danner
 * 	Added SPLX protection in cnputc. Possibly gratuitous.
 * 	[92/05/03            danner]
 * 
 * Revision 2.3  92/02/19  14:06:23  elf
 * 	Added hack to allow a keyboard character to cause a debugger
 * 	 entry. Set the integer ddb_trap_char to the ascii code of the
 * 	 character. This character will never make it past the debugger.
 * 	[91/09/17            danner]
 * 
 * Revision 2.2  91/07/09  23:20:42  danner
 * 	Initial Checkin
 * 	[91/06/26  11:42:24  danner]
 * 
 * Revision 2.2  91/04/05  14:07:01  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * LUNA-88K console driver
 *	For output character, console driver uses ROM mayput routine.
 *	And this driver supports console switch function.
 */
#include <mach_kdb.h>

#define uart_tty cons_tty
#if MACH_KERNEL
#include <sys/types.h>
#include <kern/time_out.h>
#include <device/conf.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/buf.h>		/* for struct uio (!) */
#else	MACH_KERNEL
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/file.h>
#endif MACH_KERNEL
#include <m88k/board.h>
#include <luna88kdev/kbms.h>
#include <luna88kdev/cons.h>
#include <luna88k/rom1ep.h>
#include <luna88k/softint.h> /* 4proto */


struct  tty     cons_tty[3];
int     cons_cnt = 3;

struct tty *curtty  = &uart_tty[CONSOLE];       /* current tty struct */

extern void ttrstrt();

/*ARGSUSED*/
int cnopen(dev, flag, ior)
	unsigned dev;
	int	flag;
	io_req_t ior;
{
    return utopen(dev,flag,ior);
}

/*ARGSUSED*/
int cnclose(dev)
unsigned dev;
{
    int	s = spltty();
    struct tty *tp;
    int ret;

    tp = &uart_tty[minor(dev)];
    simple_lock(&tp->t_lock);
    ret = utclose(dev);
    ttyclose(&uart_tty[minor(dev)]);
    simple_unlock(&tp->t_lock);
    splx(s);
    return ret;
}

/*ARGSUSED*/
int cnwrite(dev, uio)
unsigned dev;
struct uio *uio;
{
  return utwrite(dev, uio);
}

/*ARGSUSED*/
int cnread(dev, uio)
unsigned dev;
struct uio *uio;
{
  struct tty *tp;

  tp = &uart_tty[minor(dev)];
  tp->t_state |= TS_CARR_ON;
  return utread(dev,uio);
    }

io_return_t cngetstat(dev, flavor, data, count)
	unsigned	dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* OUT */
{
  return tty_get_status(&uart_tty[minor(dev)], flavor, data, count);
}

io_return_t cnsetstat(dev, flavor, data, count)
	unsigned	dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	count;
{
  return tty_set_status(&uart_tty[minor(dev)], flavor, data, count);
}

static void cnxint_cons(c)
int c;
{
    cnxint(&cons_tty[CONSOLE], c);
    }

static void cnxint_term(c)
int c;
{
    cnxint(&cons_tty[TERM], c);
    }

static void cnxint_kbd(c)
int c;
{
    cnxint(&cons_tty[KBD], c);
}

void cnstart(tp)
register struct tty *tp;
{
    register int s,c;
    
    s = spl6();
    if (tp->t_state & (TS_TIMEOUT|TS_BUSY|TS_TTSTOP)) 
	goto out;
    if (tp->t_outq.c_cc <= TTLOWAT(tp))
      tt_write_wakeup(tp);

    if (tp->t_outq.c_cc == 0)
	goto out;
    c = getc(&tp->t_outq);
    if ((tp->t_flags&(RAW|LITOUT|PASS8)) || c <= 0177) {
	if (tp == &cons_tty[CONSOLE])
	    softcall((func_t)cnxint_cons, (caddr_t)(c&0xff), cpu_number());
	else if (tp == &cons_tty[TERM])
	    softcall((func_t)cnxint_term, (caddr_t)(c&0xff), cpu_number());
	else
	    softcall((func_t)cnxint_kbd, (caddr_t)(c&0xff), cpu_number());
    }
    else {
	timeout((int(*)())ttrstrt, (caddr_t)tp, (c&0xff));
	tp->t_state |= TS_TIMEOUT;
	goto out;
    }
    tp->t_state |= TS_BUSY;
 out:
    splx(s);
}

static void cnxint(tp,c)
register struct tty *tp;
unsigned c;
{
    register s;
    
    s = spl6();
    if (c != 0 && mayput(c) == -1) {
	if (tp == &cons_tty[CONSOLE])
	    softcall((func_t)cnxint_cons, (caddr_t)c, cpu_number());
	else if (tp == &cons_tty[TERM])
	    softcall((func_t)cnxint_term, (caddr_t)c, cpu_number());
	else
	    softcall((func_t)cnxint_kbd, (caddr_t)c, cpu_number());
    }
    else {
	tp->t_state &= ~TS_FLUSH;
	if (tp->t_state & TS_BUSY) {
	    tp->t_state &= ~TS_BUSY;
	    if (tp->t_line)
		(*linesw[tp->t_line].l_start)(tp);
	    else
		cnstart(tp);
	}
    }
    splx(s);
}

void cnputc(c)
register int c;
{
  if (c)
    {
      register unsigned int s = SPLX();
      while (mayput(c & 0xff) == -1);
      if (c == '\n')
	cnputc('\r');
      splx(s);
    }
}

#if MACH_KDB
int ddb_trap_char = -2;
#endif

/*
 * Get a character from console.
 */
int cngetc()
{
    register int s,c;
    
    s = SPLX();
    c = mayget();
#if MACH_KDB
    if ( c == ddb_trap_char)
      {
	c = -1;
	gimmeabreak();
      }
#endif
    splx(s);
    return c;
}

int cnportdeath(dev, port)
	unsigned	dev;
	mach_port_t	port;
{
	return (tty_portdeath(&uart_tty[minor(dev)], (ipc_port_t)port));
}

/*ARGSUSED*/
int cnmmap(dev, off, prot)
	unsigned dev;
	off_t off;
	int prot;
{
  return -1;
}

/* moved from tty.c in 2.5 */

ushort g_sjis[SJISGAS];

void j_init(tp)
register struct tty *tp;
{
	register j;
	static gset = 0;

	if (!gset) {		/* sift JIS gaiji address table set */
		int i;
		unsigned char c1, c2;

		for (i = 0, c1 = SJISG1S, c2 = SJISG2S; i < SJISGAS; i++) {
			g_sjis[i] = (c1 << 8) + c2;
			c2++;
			if (c2 == 0x7f)
				c2++;
			if (c2 == SJISG2E + 1) {
				c2 = SJISG2S;
				c1++;
			}
	
			if (c1 > SJISG1E)
				break;
		}
	gset = 1;
	}

	tp->t_jt.j_level = 0;
	tp->t_jt.j_flg = 0;
	tp->t_jt.j_ecode = 0;
	tp->t_jt.j_jstl = 0;
	tp->t_jt.j_jendl = 0;
	tp->t_jt.j_gdspl = 0;
	tp->t_jt.j_gcsa[0] = 0;
	tp->t_jt.j_gcsa[1] = 0;
	tp->t_jt.j_firstcode = 0;
	tp->t_jt.outputp = 0;
	tp->t_jt.firstbyte = 0;
	tp->t_jt.kesccount = 0;
	tp->t_jt.inputconv = 0;
	tp->t_jt.eschit = 0;
	tp->t_jt.kesctime = 0;
	tp->t_jt.inputaddr[0] = 0;
	tp->t_jt.inputaddr[1] = 0;
	tp->t_jt.kana = 0;
	for (j = 0; j < EMAX; j++) {
		tp->t_jt.j_jst[j] = 0;
		tp->t_jt.j_jend[j] = 0;
		tp->t_jt.j_gdsp[j] = 0;
	}
}
