/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	cons.c,v $
 * Revision 2.5  92/05/21  17:21:46  jfriedl
 * 	Cleanup to quiet gcc warnings.
 * 	Made CHAR and DEV_T args to be UNSIGNED (ANSI/std problems).
 * 	Added prototypes.
 * 	[92/05/18            jfriedl]
 * 
 * Revision 2.4  92/05/04  11:27:32  danner
 * 	Added SPLX protection in cnputc. Possibly gratuitous.
 * 	[92/05/03            danner]
 * 
 * Revision 2.3  92/02/19  14:06:23  elf
 * 	Added hack to allow a keyboard character to cause a debugger
 * 	 entry. Set the integer ddb_trap_char to the ascii code of the
 * 	 character. This character will never make it past the debugger.
 * 	[91/09/17            danner]
 * 
 * Revision 2.2  91/07/09  23:20:42  danner
 * 	Initial Checkin
 * 	[91/06/26  11:42:24  danner]
 * 
 * Revision 2.2  91/04/05  14:07:01  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * LUNA-88K console driver
 *	For output character, console driver uses ROM mayput routine.
 *	And this driver supports console switch function.
 */
#include <mach_kdb.h>

#define uart_tty cons_tty
#if MACH_KERNEL
#include <sys/types.h>
#include <kern/time_out.h>
#include <device/conf.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/buf.h>		/* for struct uio (!) */
#else	MACH_KERNEL
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/file.h>
#endif MACH_KERNEL
#include <machine/board.h>
#include <luna88kdev/kbms.h>
#include <luna88kdev/cons.h>
#include <machine/rom1ep.h>
#include <luna88k/softint.h> /* 4proto */

#ifdef WANT_PROTOTYPES
 /* device/chario.c */
 extern boolean_t tty_portdeath(struct tty*,ipc_port_t);
 extern io_return_t tty_get_status(struct tty*,int,int*,unsigned int*);
 extern io_return_t tty_set_status(struct tty*,int,int*,unsigned int);
 extern void ttrstrt(struct tty*);
 extern void tty_queue_completion(queue_t);
 extern void ttyclose(struct tty*);

 /* device/cirbuf.c */
 extern int getc(struct cirbuf*);

 /* kern/mach_clock.c */
 extern void timeout(int(*)(),char*,int);

 /* luna88k/db_interface.c */
 extern void gimmeabreak(void);

 /* luna88k/locore/locore_asm_routines.c */
 extern void simple_lock(simple_lock_t);
 extern void simple_unlock(simple_lock_t);

 /* luna88k/machdep.c */
 extern unsigned int spl6(void);
 extern unsigned int spltty(void);
 extern unsigned int splx(unsigned int);

 /* luna88k/softint.c */
 extern void softcall(func_t,caddr_t,int);

 /* luna88kdev/uart.c */
 extern int utclose(unsigned int);
 extern int utopen(unsigned int,int,io_req_t);
 extern int utread(unsigned int,io_req_t);
 extern int utwrite(unsigned int,io_req_t);

 /* forwards */
 int cnclose(unsigned int);
 int cngetc(void);
 int cnmmap(unsigned int,off_t,int);
 int cnopen(unsigned int,int,io_req_t);
 int cnportdeath(unsigned int,mach_port_t);
 int cnread(unsigned int,struct io_req*);
 int cnwrite(unsigned int,struct io_req*);
 io_return_t cngetstat(unsigned int,int,int*,unsigned int*);
 io_return_t cnsetstat(unsigned int,int,int*,unsigned int);
 static void cnxint(struct tty*,unsigned int);
 static void cnxint_cons(int);
 static void cnxint_kbd(int);
 void cnputc(int);
 void cnstart(struct tty*);
 void j_init(struct tty*);
#endif /* ifdef WANT_PROTOTYPES */

struct  tty     cons_tty[3];
int     cons_cnt = 3;

#ifndef MACH_KERNEL
volatile struct conssw conssw = {
    utopen, utclose, utread, utwrite, utioctl, CONSOLE, 0, 0, 0, 0, 0
    };
#endif

struct tty *curtty  = &uart_tty[CONSOLE];       /* current tty struct */

extern void ttrstrt();

#ifdef unused
void cninit()
{
#ifndef MACH_KERNEL
    extern console_no;
    conssw.dev = makedev(console_no, CONSOLE);
#endif 
}
#endif

/*ARGSUSED*/
#ifdef MACH_KERNEL
int cnopen(dev, flag, ior)
	unsigned dev;
	int	flag;
	io_req_t ior;
#else	MACH_KERNEL
int cnopen(dev, flag)
unsigned dev;
int flag;
#endif
{
#ifndef MACH_KERNEL
    int s;
    
    switch(minor(dev)) {
    case KBD:
	if (conssw.kbopen)            /* dual open */
	    return(ENXIO);
	s = SPLX();
	conssw.kbopen = ON;
	curtty = &uart_tty[KBD];
	splx(s);
    case TERM:
	return(utopen(dev, flag));
    case CONSOLE:
	return ((*conssw.c_open)(conssw.dev, flag));
    default:
	return(ENXIO);
    }
#else
    return utopen(dev,flag,ior);
#endif 
}

/*ARGSUSED*/
int cnclose(dev)
unsigned dev;
{
#ifndef MACH_KERNEL
    int s;
    int ret;
    
    switch(minor(dev)) {
    case KBD:
	if (conssw.slpf)
	    wakeup((caddr_t) &conssw);
	else
	    ret = utclose(dev);
	s = SPLX();
	conssw.kbopen = OFF;
	curtty = &uart_tty[CONSOLE];
	flgmode = ENCODED;
	splx(s);
	return(ret);
    case TERM:
	return(utclose(dev));
    case CONSOLE:
	return((*conssw.c_close)(conssw.dev));
    default:
	return(ENXIO);
    }
#else
    int	s = spltty();
    struct tty *tp;
    int ret;

    tp = &uart_tty[CONSOLE];
    simple_lock(&tp->t_lock);
    ret = utclose(dev);
    ttyclose(&uart_tty[CONSOLE]);
    simple_unlock(&tp->t_lock);
    splx(s);
    return ret;
#endif
}

/*ARGSUSED*/
int cnwrite(dev, uio)
unsigned dev;
struct uio *uio;
{
#ifndef MACH_KERNEL
  if (minor(dev) == CONSOLE) {
    return((*conssw.c_write)(conssw.dev, uio));
  }
  else {
    return(utwrite(dev, uio));
  }
#else
  return utwrite(dev, uio);
#endif<
}

/*ARGSUSED*/
int cnread(dev, uio)
unsigned dev;
struct uio *uio;
{
#ifdef MACH_KERNEL
  struct tty *tp;

  tp = &uart_tty[CONSOLE];
  tp->t_state |= TS_CARR_ON;
  return utread(dev,uio);
  /* return((*linesw[uart_tty[CONSOLE].t_line].l_read)(tp, uio)); */
#else
    if (minor(dev) > CONSOLE) {
	return(utread(dev, uio));
    }
    if (conssw.c_read == utread) {
	while (conssw.kbopen) {
	    conssw.slpf = ON;
	    sleep( (caddr_t) &conssw, TTIPRI);
        }
    }
    return((*conssw.c_read)(dev, uio));
#endif
}

#ifdef MACH_KERNEL

io_return_t cngetstat(dev, flavor, data, count)
	unsigned	dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* OUT */
{
  return tty_get_status(&uart_tty[CONSOLE], flavor, data, count);
}

io_return_t cnsetstat(dev, flavor, data, count)
	unsigned	dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	count;
{
  return tty_set_status(&uart_tty[CONSOLE], flavor, data, count);
}

#else
/*ARGSUSED*/
void cnioctl(dev, cmd, addr, flag)
unsigned dev;
caddr_t addr;
int cmd, flag;
{
    if (minor(dev) == CONSOLE) {
        return((*conssw.c_ioctl)(conssw.dev, cmd, addr, flag));
    }
    else {
	return(utioctl(dev, cmd, addr, flag));
    }
}
#endif

static void cnxint_kbd(), cnxint_cons();

void cnstart(tp)
register struct tty *tp;
{
    register int s,c;
    
    s = spl6();
    if (tp->t_state & (TS_TIMEOUT|TS_BUSY|TS_TTSTOP)) 
	goto out;
    if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
#ifdef MACH_KERNEL
      tt_write_wakeup(tp);
#else
	if (tp->t_state&TS_ASLEEP) {
	    tp->t_state &= ~TS_ASLEEP;
	    wakeup((caddr_t)&tp->t_outq);
	}
	if (tp->t_wsel) {
	    selwakeup(tp->t_wsel, tp->t_state & TS_WCOLL);
	    tp->t_wsel = 0;
	    tp->t_state &= ~TS_WCOLL;
	}
#endif
    }
    if (tp->t_outq.c_cc == 0)
	goto out;
    c = getc(&tp->t_outq);
    if ((tp->t_flags&(RAW|LITOUT|PASS8)) || c <= 0177) {
	if (tp == &cons_tty[CONSOLE])
	    softcall((func_t)cnxint_cons, (caddr_t)(c&0xff), cpu_number());
	else
	    softcall((func_t)cnxint_kbd, (caddr_t)(c&0xff), cpu_number());
    }
    else {
	timeout((int(*)())ttrstrt, (caddr_t)tp, (c&0xff));
	tp->t_state |= TS_TIMEOUT;
	goto out;
    }
    tp->t_state |= TS_BUSY;
 out:
    splx(s);
}

static void cnxint(tp,c)
register struct tty *tp;
unsigned c;
{
    register s;
    
    s = spl6();
    if (c != 0 && mayput(c) == -1) {
	if (tp == &cons_tty[CONSOLE])
	    softcall((func_t)cnxint_cons, (caddr_t)c, cpu_number());
	else
	    softcall((func_t)cnxint_kbd, (caddr_t)c, cpu_number());
    }
    else {
	tp->t_state &= ~TS_FLUSH;
	if (tp->t_state & TS_BUSY) {
	    tp->t_state &= ~TS_BUSY;
	    if (tp->t_line)
		(*linesw[tp->t_line].l_start)(tp);
	    else
		cnstart(tp);
	}
    }
    splx(s);
}

static void cnxint_cons(c)
int c;
{
    cnxint(&cons_tty[CONSOLE],c);
}

static void cnxint_kbd(c)
int c;
{
    cnxint(&cons_tty[KBD],c);
}


#ifndef MACH_KERNEL
/*
 * Print a character on console.
 * Attempts to save and restore device
 * status.
 */
void cnputc(c)
register int c;
{
    extern ptsopen();
    
    if (conssw.c_open != ptsopen)	/* CONSOLE is console or uart */
	ocnputc(c);
    else{				/* CONSOLE is pty */
    
	register s = spltty();
	register struct tty  *tp;
	extern struct tty pt_tty[];
	
	tp = &pt_tty[minor(conssw.dev)];
	if (tp && (tp->t_state & (TS_CARR_ON | TS_ISOPEN)) ==
	    (TS_CARR_ON | TS_ISOPEN)) {
	    if (c == '\n')
		(void) ttyoutput('\r', tp);
	    (void) ttyoutput(c, tp);
	    ttstart(tp);
	}
	splx(s);
    }
}
#else
void cnputc(c)
register int c;
{
  if (c)
    {
      register unsigned int s = SPLX();
      while (mayput(c & 0xff) == -1);
      if (c == '\n')
	cnputc('\r');
      splx(s);
    }
}
#endif MACH_KERNEL

#if MACH_KDB
int ddb_trap_char = -2;
#endif

/*
 * Get a character from console.
 */
int cngetc()
{
    register int s,c;
    
    s = SPLX();
    c = mayget();
#if MACH_KDB
    if ( c == ddb_trap_char)
      {
	c = -1;
	gimmeabreak();
      }
#endif
    splx(s);
    return c;
}

#ifndef MACH_KERNEL
cnstop(tp, flag)
register struct tty *tp;
int flag;
{
    utstop(tp, flag);
}

conssw_reset(dev,mode)
unsigned dev;
int mode;
{
    extern struct tty pt_tty[];
    extern ptsopen(), console_no;
    
    if (conssw.c_open != ptsopen || minor(conssw.dev) != minor(dev))
	return (EINVAL);
    
    if ((mode & RESETMODE) == TIOCRESET ||
	conssw.previousdev == -1 ||
	!(pt_tty[minor(conssw.previousdev)].t_state & TS_CARR_ON)) {
	conssw.c_open = utopen;
	conssw.c_close = utclose;
	conssw.c_read = utread;
	conssw.c_write = utwrite;
	conssw.c_ioctl = utioctl;
	conssw.dev = makedev(console_no,CONSOLE);
	conssw.mode = TIOCNORMAL; 
    } else {
	conssw.dev = conssw.previousdev;
	conssw.mode = conssw.previousmode;
    }
    conssw.previousdev = -1;
    return (0);
}

int conssw_set(dev,mode)
unsigned dev;
int mode;
{
    extern struct tty pt_tty[];
    extern ptsopen(), ptsclose(), ptswrite(), ptyioctl();
    
    if ((conssw.mode & SETMODE) == TIOCFORCE) {
	return (EACCES);
    }
    
    if (conssw.c_open == ptsopen) {
	conssw.previousdev = conssw.dev;
	conssw.previousmode = conssw.mode;
	ttyflush(&pt_tty[minor(conssw.previousdev)], FREAD|FWRITE);
    } 
    conssw.dev = dev;
    conssw.mode = mode;
    if (conssw.c_open != ptsopen) {
	conssw.c_open = ptsopen;
	conssw.c_close = ptsclose;
	conssw.c_write = ptswrite;
	conssw.c_ioctl = ptyioctl;
    }
    return(0);
}

#endif

#ifdef	MACH_KERNEL
int cnportdeath(dev, port)
	unsigned	dev;
	mach_port_t	port;
{
	return (tty_portdeath(&uart_tty[CONSOLE], (ipc_port_t)port));
}

/*ARGSUSED*/
int cnmmap(dev, off, prot)
	unsigned dev;
	off_t off;
	int prot;
{
  return -1;
}
#endif /* MACH_KERNEL*/

#ifdef MACH_KERNEL

/* moved from tty.c in 2.5h */

ushort g_sjis[SJISGAS];

void j_init(tp)
register struct tty *tp;
{
	register j;
	static gset = 0;

	if (!gset) {		/* sift JIS gaiji address table set */
		int i;
		unsigned char c1, c2;

		for (i = 0, c1 = SJISG1S, c2 = SJISG2S; i < SJISGAS; i++) {
			g_sjis[i] = (c1 << 8) + c2;
			c2++;
			if (c2 == 0x7f)
				c2++;
			if (c2 == SJISG2E + 1) {
				c2 = SJISG2S;
				c1++;
			}
	
			if (c1 > SJISG1E)
				break;
		}
	gset = 1;
	}

	tp->t_jt.j_level = 0;
	tp->t_jt.j_flg = 0;
	tp->t_jt.j_ecode = 0;
	tp->t_jt.j_jstl = 0;
	tp->t_jt.j_jendl = 0;
	tp->t_jt.j_gdspl = 0;
	tp->t_jt.j_gcsa[0] = 0;
	tp->t_jt.j_gcsa[1] = 0;
	tp->t_jt.j_firstcode = 0;
	tp->t_jt.outputp = 0;
	tp->t_jt.firstbyte = 0;
	tp->t_jt.kesccount = 0;
	tp->t_jt.inputconv = 0;
	tp->t_jt.eschit = 0;
	tp->t_jt.kesctime = 0;
	tp->t_jt.inputaddr[0] = 0;
	tp->t_jt.inputaddr[1] = 0;
	tp->t_jt.kana = 0;
	for (j = 0; j < EMAX; j++) {
		tp->t_jt.j_jst[j] = 0;
		tp->t_jt.j_jend[j] = 0;
		tp->t_jt.j_gdsp[j] = 0;
	}
}
#endif
