/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	kbmain.c,v $
 * Revision 2.4  92/05/21  17:21:56  jfriedl
 * 	     Changes to quite gcc warnings.
 * 	Changed UNSIGNED CHAR args to fcns to UNSIGNED.
 * 	Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.3  92/02/18  18:02:58  elf
 * 	Made include of sys/types.h unconditional
 * 	[92/01/21            danner]
 * 
 * Revision 2.2  91/07/09  23:20:56  danner
 * 	Initial Checkin
 * 	[91/06/26  11:47:02  danner]
 * 
 * Revision 2.2  91/04/05  14:08:35  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * Keyboard and mouse driver
 */

#include <sys/types.h>
#ifdef MACH_KERNEL
#include <kern/time_out.h>
#include <device/conf.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/buf.h>		/* for struct uio (!) */
#include <device/errno.h>
#include <luna88kdev/compat_25.h>
#else
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/tty.h>
#include <sys/errno.h>
#include <sys/file.h>
#endif
#include <luna88kdev/kbms.h>
#include <luna88kdev/deftbl.t>
#include <luna88kdev/table.h>
#include <luna88kdev/codetbl.t>
#include <luna88kdev/keymap.t>
#include <luna88kdev/esctbl.t>

#ifdef WANT_PROTOTYPES

 /* luna88k/locore/locore_asm_routines.c */
 extern void memcpy(void*,void*,unsigned int);

 /* luna88k/machdep.c */
 extern unsigned int spltty(void);
 extern unsigned int splx(unsigned int);

 /* luna88kdev/uart.c */
 extern void bellcall(int);

 /* forwards */
 int kbencode(unsigned int,char*);
 int kbrepeat(char*);
 static int chrept(int);
 static int scheck(unsigned int);
 static int semienc(unsigned int,char*);
 static int sencode(unsigned int,char*);
 static int sinput(unsigned int);
 static int specal(unsigned int);
 static int srawout(unsigned int,char*);
 static void sflgset(void);
 void kbsetinit(unsigned int);
#endif /* ifdef WANT_PROTOTYPES */

#define SPLX spltty
#define SPLC spltty
static int clickfreq, kbclickfrag;
extern struct tty cons_tty[];
int kbreset;

int kbencode(data,buf)
register unsigned data;
char *buf;
{
    int  rtc;

    if(rtc=specal(data) != 0)
	return rtc;

    switch(flgmode) {
        case KB_RAW: 
            return srawout(data,buf);
        case ENCODED: 
 	    rtc = sencode(data, buf);
 	    if (kbclickfrag && rtc != KBNON && rtc != KBRESET) 
 	        bellcall(clickfreq); /* NEWKB */
	    return rtc;
        case SEMIENCODED: 
            return semienc(data,buf);
        default:
            return KBNON;
    }
}

static int srawout(data,buf)
register unsigned data;
register char *buf;
{
    *buf = data;
    reptflag = OFF;
    return 1;
}

static int sencode(data,buf)
unsigned data;
register char *buf; 
{
    unsigned short dd;
    register i,j;
    short mode;

    lastcnt = cntbuf;

    if( !scheck(data))
        return KBNON;
    if( !cntbuf || cntbuf>2){
	reptflag = OFF;
        return KBNON;
    }
    data = bufcode[cntbuf-1];
    dd = *(tbpint+data);

    if(bufout[cntbuf-1]  && !(dd & REPEATB)){
        reptflag = OFF;
	return KBNON;
    }
    if(dd & OK){
        if(dd & ESCLINK){
            mode = *(tblmode+(dd & 0x000f));

            for(i=0 ; i<ESCBYTE ; i=i+ESCSIZE){
                if(*(tblescp+i) == data){
                    for(i=i+1 ; i<ESCBYTE ; i=i+ESCSIZE){
                        if(mode == *(tblescp+i)){
                            for(i=i+1,j=0 ; *(tblescp+i)!=ENDCODE ; i++,j++){
                                buf[j] = *(tblescp+i);
                                bufrep[j] = *(tblescp+i);
                            }
                            bufout[cntbuf-1] = 1;
                            if(dd & REPEATB )
                                return chrept(j);
                            reptflag = OFF;
                            return j;
                        }
                        if(*(tblescp+i) == 0xff)
                            return KBNON; 
                    }
                    break; 
                }
                if(*(tblescp+i) == 0xff)
                    return KBNON; 
            }
            return KBNON;
        }else{
            buf[0] = (char)dd;
            bufout[cntbuf-1] = 1;
            if(dd & REPEATB){
                bufrep[0] = (char)dd;
                return chrept(1);
            }
            reptflag = OFF;
            return 1;
        }
    }
    return KBNON;
}

static int semienc(data,buf)
register unsigned data;
register char *buf;
{
    unsigned char dd;
    if (kbclickfrag && !(data & 0x80) && (data == 0x9 || 0x10 <= data)) 
 	bellcall(clickfreq); /* NEWKB */
    buf[0] = data;
    lastcnt = cntbuf;
    if( ! scheck(data)){
        reptflag = OFF;
        return 1;
    }
    if( !cntbuf || cntbuf>2){
	reptflag = OFF;
        return 1;
    }
    data = bufcode[cntbuf-1];
    dd = !*(keymap+data);
    if(bufout[cntbuf-1]  && !dd ){
        reptflag = OFF;
        return 1;
    }
    bufout[cntbuf-1] = 1;
    if( dd ){
        bufrep[0] = data;
        chrept(1);
        return 1;
    }
    reptflag = OFF; 
    return 1;
}

static int specal(data)
register unsigned data;
{
    switch(data){
        case 0x0f:{
            flgctrl |= 0x8000;         /* special on */
            return 0;
        }
        case 0x8f:{
            flgctrl &= 0x7fff;         /* special off */
            return 0;
        }
        case 0x73:{                    /* reset */
            if((kbreset == ENABLE) &&  (flgctrl & 0x8000)){
                kbsetinit(1);
                return KBRESET; 
            }
            return 0;
        }
        case 0x72:{                    /* break */
            if((flgctrl & 0x8000)&&(flgmode == ENCODED ))
                return KBBREAK;
            return 0;
        }
        default  :
            break;
    }
    return 0;
}

static int scheck(data)
register unsigned data;
{
    switch(data){
        case 0x0f:            /* special on */
            flgctrl |= 0x8000;
            sflgset();
            return 1;
        case 0x8f:            /* special off */
            flgctrl &= 0x7fff;
            sflgset();
            return 1;
        case 0x0a:            /* ctrl on */ 
            flgctrl |= 0x4000;
            sflgset();
            return 1;
        case 0x8a:            /* ctrl off */        
            flgctrl &= 0xbfff;
            sflgset();
            return 1;
        case 0x0b:            /* kana on */ 
            flgctrl |= 0x2000;
            sflgset();
            return 1;
        case 0x8b:            /* kana off */ 
            flgctrl &= 0xdfff;
            sflgset();
            return 1;
        case 0x0e:            /* caps lock on */
            flgctrl |= 0x1000;
            sflgset();
            return 1;
        case 0x8e:            /* caps lock off */
            flgctrl &= 0xefff;
            sflgset();
            return 1;
        case 0x0c:            /* shift on */
            flgctrl |= 0x0800;
            sflgset();
            return 1;
        case 0x8c:            /* shift off */
            flgctrl &= 0xf7ff; 
            sflgset();
            return 1;
        case 0x0d:            /* shift on */
            flgctrl |= 0x0400;
            sflgset();
            return 1;
        case 0x8d:            /* shift off */
            flgctrl &= 0xfbff; 
            sflgset();
            return 1;
	default:
	    return sinput(data);
    }
}

static void sflgset()
{
    if(flgctrl == 0x0000){
        tbpint = tblnoml;     /* normal encoded table */
        return;
    }
    if(flgctrl & 0x8000){
        tbpint = tblspcl;     /* special encoded table */
        return;
    }
    if(flgctrl & 0x4000){
        tbpint = tblctrl;     /* control encoded table */
        return;
    }
    if(flgctrl & 0x2000){
        if(flgctrl & 0x0c00){ tbpint = tblknsf; }/* kana+shift */
        else                { tbpint = tblkana; }/* kana table */
        return;
    }
    if(flgctrl & 0x1000){
        if(flgctrl & 0x0c00){ tbpint = tblcpsf; }/* caps+shift */
        else                { tbpint = tblcaps; }/* caps lock  */ 
        return;
    }
    if(flgctrl & 0x0c00){
        tbpint = tblsift;     /* shift encoded table */
        return;
    }
}

static int sinput(data)
register unsigned data;
{
    unsigned char dd;
    register i,j;

    if(data & 0x80){
        dd = data & 0x7f;
        for(i=0 ; i != cntbuf ; i++){
            if(bufcode[i] == dd){
                for(j=i ; j != BUFMAX ; j++){
                    bufcode[j] = bufcode[j+1];
                    bufout[j] = bufout[j+1];
                }
                if(i == (--cntbuf)){ return 3; }
                else               { return 4; }
            }
        }
    }else{
        if(cntbuf != BUFMAX){
            bufcode[cntbuf++] = data;
            return 2;
        }
    }
    return 0;
}

static int chrept(cnt)
register cnt;
{
    cntrep = cnt;
    repwait = cntwait;
    repintvl = 1;
    switch(lastcnt){
        case 0:{ reptflag = ON; return cnt; }
        case 1:{ if(cntbuf == 1){ reptflag = ON; return KBNON; }
                 else           { reptflag = OFF; return cnt; } 
        }
        case 2:{ if(cntbuf == 1){ reptflag = ON; return KBNON; }
                 else           { reptflag = OFF; return KBNON; } 
        }
        case 3:{ reptflag = OFF; return cnt; }
        default:{ reptflag = OFF; return KBNON; }
    }
}

void kbsetinit(flag)
unsigned flag;  /*  =0:initial  ;  =1:reset  */
{
    int s;
    /* encoded table initial */
    s = SPLC();
    /* special encoded table */
    memcpy((char*)tblspcl, (char*)t_specl, CODTBSZ*2);

    /* control encoded table */
    memcpy((char*)tblctrl, (char*)t_cntrl, CODTBSZ*2);

    /* kana    encoded table */
    memcpy((char*)tblkana, (char*)t_kana , CODTBSZ*2);

    /* capsloc encoded table */
    memcpy((char*)tblcaps, (char*)t_cploc, CODTBSZ*2);

    /* shift   encoded table */
    memcpy((char*)tblsift, (char*)t_shift, CODTBSZ*2);

    /* kana + shift encoded table */
    memcpy((char*)tblknsf, (char*)t_kansf, CODTBSZ*2);

    /* capslock + shift encoded table */
    memcpy((char*)tblcpsf, (char*)t_capsf, CODTBSZ*2);

    /* normal  encoded table */
    memcpy((char*)tblnoml, (char*)t_nomal, CODTBSZ*2);

    /* semi-encoded table */
    memcpy((char*)keymap,  (char*)t_keymap ,CODTBSZ);

    /* mode table */
    memcpy((char*)tblmode, (char*)t_mode , MODTBSZ*2);

    /* escape table */
    memcpy((char*)tblescp, (char*)t_escap, ESCBYTE);

    /* flag initial */

    if(!flag)
        flgctrl = OFF;
    else
	flgctrl &= 0x3000;

    flgmode = ENCODED;
    reptflag = OFF;
    reptsw = ON;

    /* counter initial */

    cntbuf = 0;
    cntwait = REPCNT;
    cntintvl = RPINTVL;

    /* buffer initial */

    bufcode[0] = 0;
    bufcode[1] = 0;
    bufcode[2] = 0;
    bufcode[3] = 0;
    bufcode[4] = 0;

    bufout[0] = 0;
    bufout[1] = 0;
    bufout[2] = 0;
    bufout[3] = 0;
    bufout[4] = 0;

    /* pointer initial */

    tbpint = tblnoml;      

    kbclickfrag = 0;
    kbreset = ENABLE;

    splx(s);
}

int kbrepeat(buf)
char	*buf;
{
    register j;

    if( !reptsw )
	return KBNON;
    
    if( !reptflag )
	return KBNON;
    
    if( --repwait )
	return KBNON;
    
    ++repwait;
    
    if( --repintvl )
	return KBNON;
    
    for(j=0;j<cntrep;j++)
	buf[j] = bufrep[j];

    repintvl = cntintvl;
    if (kbclickfrag) 
	bellcall(clickfreq); /* NEWKB */
    return cntrep;
}

int	bztime = 0;

#ifdef UNUSED /* HERE TO END OF FILE CURRENTLY UNUSED */

static int subyte(addr, byte)
unsigned addr;
unsigned byte;
{
    unsigned char byte_byte = byte;
    return copyout(&byte_byte, addr, sizeof(char)) == 0 ? 0 : -1;
}

static int fubyte(addr)
unsigned addr;
{
    unsigned char byte;

    if (copyin(addr, &byte, sizeof(char)))
	return -1;
    else
	return byte;
}

static int bsetm(arg)
int arg;
{
    int s;
    s = SPLX();
    switch(arg){
    case KB_RAW:
	flgmode = KB_RAW;
	break;
    case ENCODED:
	kbsetinit(1);	/* NAKA */
	/*          flgmode = ENCODED;	*/
	break;
    case SEMIENCODED:
	flgmode = SEMIENCODED;
	break;
    default:
	splx(s);
	return EINVAL;
    }
    splx(s);
    return 0;
}

static int bsetktm(kemode)
struct keyescmode *kemode;
{
    if((kemode->keytype < 0) || (kemode->keytype > 15))
        return EINVAL;

#if 0 /* this will always fail, because escmode is a char */
    if((kemode->escmode < 0) || (kemode->escmode > 255))
        return EINVAL;
#endif

    tblmode[kemode->keytype] = kemode->escmode;
    return 0;
}


static int bgetet(kesget)
struct kbesget *kesget;
{
    register unsigned short work1;

    if(kesget->keycode & 0xff80)
        return EINVAL;

    switch(kesget->tblno){
    case 0:
	work1 = tblspcl[kesget->keycode];
	break;
    case 1:
	work1 = tblctrl[kesget->keycode];
	break;
    case 2:
	work1 = tblknsf[kesget->keycode];
	break;
    case 3:
	work1 = tblkana[kesget->keycode];
	break;
    case 4:
	work1 = tblcpsf[kesget->keycode];
	break;
    case 5:
	work1 = tblcaps[kesget->keycode];
	break;
    case 6:
	work1 = tblsift[kesget->keycode];
	break;
    case 7:
	work1 = tblnoml[kesget->keycode];
	break;
    default:
	return EINVAL;
    }

    kesget->kbentry.flags = work1 >> 8;
    kesget->kbentry.code = (char)work1;
    
    return 0;
}

static int bgeteta(kesgeta)
struct kbesgeta *kesgeta;
{

    register unsigned short *tbpnts;

    switch(kesgeta->tblno){
    case 0:
	tbpnts = tblspcl;
	break;
    case 1:
	tbpnts = tblctrl;
	break;
    case 2:
	tbpnts = tblknsf;
	break;
    case 3:
	tbpnts = tblkana;
	break;
    case 4:
	tbpnts = tblcpsf;
	break;
    case 5:
	tbpnts = tblcaps;
	break;
    case 6:
	tbpnts = tblsift;
	break;
    case 7:
	tbpnts = tblnoml;
	break;
    default:
	return EINVAL;
    }

    if(copyout((char*)tbpnts,(unsigned)kesgeta->kbtblp,CODTBSZ*2))
        return EFAULT;
           
    return 0;
}

static int bgetsm(kssget)
struct kbssget *kssget;
{
    if(kssget->keycode & 0xff80)
        return EINVAL;
    
    kssget->kbsrept = !keymap[kssget->keycode];

    return 0;
}

static int bgetsma(kssgeta)
struct kbssgeta *kssgeta;
{
    register char *keymapp;
    register char *tbptr = kssgeta->kbsreptp;

    for(keymapp = (char *)&keymap[0];
	keymapp < (char *)&keymap[CODTBSZ];keymapp ++, tbptr ++)
	if(subyte((unsigned)tbptr, *keymapp) < 0)
	    return (EFAULT);
    
    return 0;
}

static int bsetet(kesget)
struct kbesget *kesget;
{
    register unsigned short work1, work2;

    if(kesget->keycode & 0xff80)
        return EINVAL;

    work1 = (short)kesget->kbentry.flags & 0x00ff;
    work1<<=8;
    work2 = (short)kesget->kbentry.code & 0x00ff;
    work1 |= work2;

    switch(kesget->tblno){
    case 0:
	tblspcl[kesget->keycode] = work1;
	break;
    case 1:
	tblctrl[kesget->keycode] = work1;
	break;
    case 2:
	tblknsf[kesget->keycode] = work1;
	break;
    case 3:
	tblkana[kesget->keycode] = work1;
	break;
    case 4:
	tblcpsf[kesget->keycode] = work1;
	break;
    case 5:
	tblcaps[kesget->keycode] = work1;
	break;
    case 6:
	tblsift[kesget->keycode] = work1;
	break;
    case 7:
	tblnoml[kesget->keycode] = work1;
	break;
    default:
        return EINVAL;
    }
    return 0;
}

static int bseteta(kesgeta)
struct kbesgeta *kesgeta;
{
    register unsigned short *tbpnts;

    switch(kesgeta->tblno) {
    case 0:
	tbpnts = tblspcl;
	break;
    case 1:
	tbpnts = tblctrl;
	break;
    case 2:
	tbpnts = tblknsf;
	break;
    case 3:
	tbpnts = tblkana;
	break;
    case 4:
	tbpnts = tblcpsf;
	break;
    case 5:
	tbpnts = tblcaps;
	break;
    case 6:
	tbpnts = tblsift;
	break;
    case 7:
	tbpnts = tblnoml;
	break;
    default:
        return EINVAL;
    }
    
    if(copyin((unsigned)kesgeta->kbtblp,(char*)tbpnts,CODTBSZ*2))
        return EFAULT;
          
    return 0;
}

static int bsetst(kssget)
struct kbssget *kssget;
{
    if(kssget->keycode & 0xff80)
        return EINVAL;

    keymap[kssget->keycode] = (kssget->kbsrept)?0:t_keymap[kssget->keycode];

    return 0;
}

static int bsetsta(kssgeta)
struct kbssgeta *kssgeta;
{
    register char tmp;
    register char *keymapp;
    register char *tbptr = kssgeta->kbsreptp;

    for(keymapp = (char *)&keymap[0];
	keymapp < (char *)&keymap[CODTBSZ];keymapp ++, tbptr ++){
	if((tmp =fubyte((unsigned)tbptr)) < 0)
	    return EFAULT;
	else
	    *keymapp = tmp?0:t_keymap[(int)keymapp -(int)keymap];
    }

    return 0;
}

static int bgetesc(eentry)
struct escentry *eentry;
{
    int pint;
    register i;
    register unsigned char *p;

    if(eentry->keycode > 0x00ff)
        return EINVAL;

    if(eentry->escmode > 0x00ff)
        return EINVAL;

       if(seresc(eentry->keycode,eentry->escmode,&pint))
        return EINVAL;

    p = &tblescp[pint];

    for(i=0 ; i<ESCSIZE-2 ; i++){
        if(subyte((unsigned)eentry->esctext++,*p) == -1)
	    return EFAULT;
        if(*p++ == 0)
            break;
    }
    return 0;
}

static int bsetesc(eentry)
struct escentry *eentry;
{
    int pint;
    register i,c;
    register unsigned char *p;

    if(eentry->keycode > 0x00ff)
        return EINVAL;
    if(eentry->escmode > 0x00ff)
        return EINVAL;
    if(seresc(eentry->keycode,eentry->escmode,&pint))
        return EINVAL;
    p = &tblescp[pint];
    for(i=0 ; i<ESCSIZE-2 ; i++) {
        if((c=fubyte((unsigned)eentry->esctext++)) == -1)
	    return EFAULT;
        if((*p++=c) == 0)
            break;
    }
    return 0;
}

static int breptim(krepeat)
struct kbrepeat *krepeat;
{
    if( cntwait=krepeat->repstart )
        reptsw = ON;
    else
        reptsw = OFF;
    if( (cntintvl=krepeat->repinterval ) < RPINTVL)
        cntintvl = RPINTVL;
    return 0;
}

static int kbrstenable(arg)
int arg;
{
    int s;

    s = SPLC();

    switch (arg) {
    case RSTENABLE :
	kbreset = ENABLE; 
	break;
    case RSTDISABLE :
	kbreset = DISABLE; 
	break;
    default:
	splx(s);
	return -1;
    }
    splx(s);
    return 0;
}

static int seresc(code,mode,pint)
register unsigned code;  
register unsigned mode;  
int  *pint;
{
    register i;
    register unsigned char *tbl;

    tbl = tblescp;

    for(i=0 ; i<ESCBYTE ; i=i+ESCSIZE)
        if(code == *(tbl+i)){
            for(i=i+1 ; i<ESCBYTE ; i=i+ESCSIZE)
                if(mode == *(tbl+i)){
                    *pint = i+1;
                    return 0;
                }
            return -1;
        }
    return -1;
}

static int bbuzzer(click)
register click;
{
    register s, i;

    s = SPLC();
    i = bztime;
    if(click==0) 
	bztime = 0;
    else {
	bztime += click;
	if(i==0)
	    bzscan();
    }
    splx(s);
    return 0;
}

static int settbl(buf)
short *buf[];
{
    register int s;

    s = SPLC();
    if(buf[0] != 0)
	copyin((unsigned)buf[0],(char *)tblspcl,CODTBSZ*2);
    else
        /* special encoded table */
	memcpy((char*)tblspcl,(char*)t_specl,CODTBSZ*2);

    if(buf[1] != 0)
	copyin((unsigned)buf[1],(char *)tblctrl,CODTBSZ*2);
    else
	/* control encoded table */
	memcpy((char*)tblctrl,(char*)t_cntrl,CODTBSZ*2);

    if(buf[2] != 0)
	copyin((unsigned)buf[2],(char *)tblkana,CODTBSZ*2);
    else
	/* kana    encoded table */
	memcpy((char*)tblkana,(char*)t_kana ,CODTBSZ*2);

    if(buf[3] != 0)
	copyin((unsigned)buf[3],(char *)tblcaps,CODTBSZ*2);
    else
	/* capsloc encoded table */
	memcpy((char*)tblcaps,(char*)t_cploc,CODTBSZ*2);

    if(buf[4] != 0)
	copyin((unsigned)(buf[4] ? buf[4] : t_shift),(char *)tblsift,CODTBSZ*2);
    else
	/* shift   encoded table */
	memcpy((char*)tblsift,(char*)t_shift,CODTBSZ*2);

    if(buf[5] != 0)
	copyin((unsigned)(buf[5] ? buf[5] : t_kansf),(char *)tblknsf,CODTBSZ*2);
    else
	/* kana + shift encoded table */
	memcpy((char*)tblknsf,(char*)t_kansf,CODTBSZ*2);

    if(buf[6] != 0)
	copyin((unsigned)(buf[6] ? buf[6] : t_capsf),(char *)tblcpsf,CODTBSZ*2);
    else
	/* capslock + shift encoded table */
	memcpy((char*)tblcpsf,(char*)t_capsf,CODTBSZ*2);

    if(buf[7] != 0)
	copyin((unsigned)(buf[7] ? buf[7] : t_nomal),(char *)tblnoml,CODTBSZ*2);
    else
	/* normal  encoded table */
	memcpy((char*)tblnoml,(char*)t_nomal,CODTBSZ*2);

    splx(s);
    return 0;
}

static int exchbs()
{
    register int s;
    s = SPLC();
    tblspcl[DELKEY] = t_specl[BSKEY];
    tblspcl[BSKEY] = t_specl[DELKEY];
    tblctrl[DELKEY] = t_cntrl[BSKEY];
    tblctrl[BSKEY] = t_cntrl[DELKEY];
    tblkana[DELKEY] = t_kana[BSKEY];
    tblkana[BSKEY] = t_kana[DELKEY];
    tblcaps[DELKEY] = t_cploc[BSKEY];
    tblcaps[BSKEY] = t_cploc[DELKEY];
    tblsift[DELKEY] = t_shift[BSKEY];
    tblsift[BSKEY] = t_shift[DELKEY];
    tblknsf[DELKEY] = t_kansf[BSKEY];
    tblknsf[BSKEY] = t_kansf[DELKEY];
    tblcpsf[DELKEY] = t_capsf[BSKEY];
    tblcpsf[BSKEY] = t_capsf[DELKEY];
    tblnoml[DELKEY] = t_nomal[BSKEY];
    tblnoml[BSKEY] = t_nomal[DELKEY];
    splx(s);
    return 0;
}

static int chmap(args)
unsigned args;
{
    register int s;

    s = SPLC();
    copyin(args, keymap, CODTBSZ);
    splx(s);
    return 0;
}

static int getmap(args)
unsigned args;
{
    register s;

    s = SPLC();
    copyout(keymap, args, CODTBSZ);
    splx(s);
    return 0;
}

static int setkeymap(arg)
struct km_ent *arg;
{
    register int s;

    if(arg->ent & 0xff80)
	return EINVAL;

    s = SPLC();
    keymap[arg->ent] = arg->value;
    splx(s);
    return 0;
}

static int getkeymap(arg)
struct km_ent *arg;
{
    register int s;

    if(arg->ent & 0xff80)
	return EINVAL;

    s = SPLC();
    arg->value = keymap[arg->ent];
    splx(s);
    return 0;
}

/* KBD click
	freq =  0 : OFF
		1 : 6000Hz
		2 : 3000Hz
		3 : 1500Hz
		4 : 1000Hz
 		5 :  600Hz
 		6 :  300Hz
 		7 :  150Hz
 		8 :  100Hz
*/

static void kbclick(freq)
int freq;
{
    if (freq <= 0 || 8 < freq) 
	kbclickfrag = 0;
    else {
	kbclickfrag = 1;
	clickfreq = 0x40 + freq - 1;
    }
}

int kbxioctl(tp,cmd,arg,mode)
int cmd, mode;
register struct tty *tp;
caddr_t *arg;
{
    switch(cmd){
    case KBSETTBL:
	return settbl((short int**)arg);
    case KBEXCHBS:
	return exchbs();
    case KBCHMAP:
	return chmap((unsigned)*arg);
    case KBGETMAP:
	return getmap((unsigned)*arg);
    case KBSETKME:
	return setkeymap((struct km_ent*)arg);
    case KBGETKME:
	return getkeymap((struct km_ent*)arg);
    case KBBUZZER:
	return bbuzzer((unsigned)*arg);
    case KBSETM:
	return bsetm((unsigned)*arg);
    case KBRSTM:
	kbsetinit(1);
	return 0;
    case KBSETKTM:
	return bsetktm((struct keyescmode*)arg);
    case KBEGETT:
	return bgetet((struct kbesget*)arg);
    case KBEGETTA:
	return bgeteta((struct kbesgeta*)arg);
    case KBSGETT:
	return bgetsm((struct kbssget*)arg);
    case KBSGETTA:
	return bgetsma((struct kbssgeta*)arg);
    case KBESETT:
	return bsetet((struct kbesget*)arg);
    case KBESETTA:
	return bseteta((struct kbesgeta*)arg);
    case KBSSETT:
	return bsetst((struct kbssget*)arg);
    case KBSSETTA:
	return bsetsta((struct kbssgeta*)arg);
    case KBGETESC:
	return bgetesc((struct escentry*)arg);
    case KBSETESC:
	return bsetesc((struct escentry*)arg);
    case KBREPTIM:
	return breptim((struct kbrepeat *)arg);
    case KBFLSH:
#ifdef MACH_KERNEL
	ttyflush(tp,D_READ| D_WRITE);
#else
	ttyflush(tp,(FREAD|FWRITE));
#endif
	return 0 ;
    case KBRSTENABLE:
	return kbrstenable((int)*arg);
    case KBCLICK :
	kbclick((int)*arg);
	return 0;
    case KBBELL : /* NEWKB */
    {
	unsigned int bzflag;
	bzflag = (int)*arg;
	bzflag |= 0x40;
	bellcall(bzflag);
	return 0;
    }
    default:
	return EINVAL;
    }
}
#endif UNUSED
