/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	msmain.c,v $
 * Revision 2.5  92/08/03  17:46:29  jfriedl
 * 	removed silly prototypes
 * 	[92/08/02            jfriedl]
 * 
 * Revision 2.4  92/05/21  17:22:02  jfriedl
 * 	Cleanup to quiet gcc warnings.
 * 	Changed CHAR and DEV_T args of fcns to UNSIGNED (ANSI/std problems)
 * 	[92/05/18            jfriedl]
 * 
 * Revision 2.3  92/02/18  18:03:01  elf
 * 	Made include of sys/types.h unconditional.
 * 	[92/01/21            danner]
 * 
 * Revision 2.2  91/07/09  23:21:08  danner
 * 	Initial Checkin
 * 	[91/06/26  11:46:02  danner]
 * 
 * Revision 2.2  91/04/05  14:09:27  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * Mouse driver
 */
#include <sys/types.h>

#ifdef MACH_KERNEL
#include <kern/time_out.h>
#include <device/conf.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/buf.h>		/* for struct uio (!) */
#include <device/errno.h>
#include <luna88kdev/compat_25.h>
#include <sys/time.h>
#else	MACH_KERNEL
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/systm.h>
#endif
#include <machine/machparam.h>
#include <luna88kdev/uart.h>
#include <luna88kdev/kbms.h>



/*
 * Mouse driver
 */
#define TIMEMAX       -1
#define MOUSE_BUFFER_SIZE       100
#define TSAMPL        3	   /* how often (in clock ticks) to sample the mouse */
#define IdleScan      100  /* how often (in clock ticks) to check to see if the
			      mouse is now open */

#define MICKYVALUE    200
#define MICKYBASE     100
#define MICKYMAX      1000

#define MIN_MOUSE_X      0
#define MIN_MOUSE_Y      0
#define MAX_MOUSE_X      1279
#define MAX_MOUSE_Y      1023

static short   ms_flag;	/* chosen from below */
#define MS_OPEN		0x1
#define MS_DWANTED	0x4
#define	MS_RCOLL	0x8 	/* for select */

/*****************************************************************************/
/*****************************************************************************/

/*
 * Standard circular buffer of MOUSE_BUFFER_SIZE elements of 'msdata'.
 * Elements go in at 'next_free' and are taken out at 'last_read'.
 *
 * If (next_free == last_read), there are no elements.
 *
 * Access is controlled by 'lock'.
 *
 * This whole thing is in a structure simply to keep it logically together.
 *
 */
static struct
{
    simple_lock_data_t lock;		/* Must have lock to access below. */
    struct msdata *next_free,		/* Elements inserted here.. */
		  *last_read,		/* ..and taken from here */
 		   data[MOUSE_BUFFER_SIZE];	/* The Elements */
} mouse_data_buffer;

/*
 * Returns the number of elements currently available in the mouse_data_buffer.
 * ASSUMES THAT THE LOCK IS HELD.
 */	
#define mouse_buffer_count()						\
    (mouse_data_buffer.next_free >= mouse_data_buffer.last_read ?	\
     mouse_data_buffer.next_free -  mouse_data_buffer.last_read :	\
     MOUSE_BUFFER_SIZE - (mouse_data_buffer.last_read - mouse_data_buffer.next_free))

/*
 * Returns true if the buffer is empty, false if there is data available.
 * ASSUMES THAT THE LOCK IS HELD.
 */
#define mouse_buffer_empty() \
    (mouse_data_buffer.next_free == mouse_data_buffer.last_read)
	
/*
 * Used to increment next_free and last_read (of mouse_data_buffer).
 * ASSUMES THAT THE LOCK IS HELD.
 */
#define bump_up(ptr)						\
{								\
	if (((ptr)+=1) == &mouse_data_buffer.data[MOUSE_BUFFER_SIZE])	\
		(ptr) = mouse_data_buffer.data;			\
}

#define data_buffer_lock()	simple_lock(&mouse_data_buffer.lock)
#define data_buffer_unlock()	simple_unlock(&mouse_data_buffer.lock)

/*****************************************************************************/
/*****************************************************************************/

static  char button;    /* button jyotai(old)  */
static  char indata[5]; /* input data buffer   */
static short   indcount;       /* counter for indata  */

static short	nx_current;	/* new current zahyo x */
static short	ny_current;	/* new current zahyo y */
static short	ox_current;	/* old current zahyo x */
static short	oy_current;	/* old current zahyo y */

static short	sox_current;	/* last msscan() nx_current (for delta) */
static short	soy_current;	/* last msscan() ny_current (for delta) */

static short 	limit_min_x;	/* gentei zahyo start x */
static short	limit_min_y;	/* gentei zahyo start y */ 
static short 	limit_max_x;	/* gentei zahyo end   x */
static short	limit_max_y;	/* gentei zahyo end   y */ 

static short	b_mickey;	/* mickey base   */
static short	v_mickey;	/* mickey value  */

short	flgmode;	       /* mode flag      */
static short   datamode;       /* data mode flag */

static short   t_sampl;        /* sample timing  */

static unsigned long o_time;   /* old time       */

#ifndef MACH_KERNEL
/* aditional for select */
#include <sys/proc.h>
#endif

#ifdef unused
static struct	thread	*ms_selr;
#endif
#define SPLC spltty
extern	struct uart_device volatile *uart_addr[];

/* logical to physical line number */
#define ldevtop(dev)	(minor(dev)==2?0:1)

/* forwards */
static void msputpacket();
static int msgetpacket();
static void msscan();
static void mssetinit();

static int delay_flag = 0; 

#ifdef MACH_KERNEL

extern time_value_t *mtime;
#define lbolt ((mtime) ? mtime->seconds : 0)

#endif

int msopen(dev, flag)
unsigned dev;
int flag;
{
    unsigned s;
    extern	kbms_init;

    dev = minor(dev);
    if (dev != MOUSE || (ms_flag&MS_OPEN)) 
	return(ENXIO);
    s = SPLC();
    if((ms_flag&MS_OPEN)==0) {
        static	short	ms_scan = 0;

	ms_flag |= MS_OPEN;
	uart_addr[ldevtop(KBD)]->data = 0x60;	/* enable mouse */
/*
 * IDEA
 * do the msscan() each time, and get rid of the idle scan.
 *
 * IDEA
 * allow only one open of the mouse -- no need to allow competing processes.
 */
	if (ms_scan == 0) {
	    msscan();
	    ms_scan = 1;
	}
	if(kbms_init==0) {
	    extern void msparam();
	    msparam();
	    kbms_init = 1;
	}
	mssetinit();
    }
    indata[3] = indata[4] = 0;
    splx(s);
    return(0);
}

int msclose(dev)
unsigned dev;
{
    if (dev != MOUSE || (ms_flag&MS_OPEN))
	return(ENXIO);

    uart_addr[ldevtop(KBD)]->data = 0x20;  /* disable mouse */
    ms_flag &= ~MS_OPEN;
    return(0);
}

static int	wk_mickey = 0;

#define abs(x)  ((x>0)?x:-x)

int
msreceive(data)
unsigned data;
{
    int    tmp,wk;

    if(indcount)
    {
	indata[3-indcount] = data;

        if (((--indcount)==0)&&(ms_flag & MS_OPEN))
	{
	    /* if the buttons have changed state, output that */
            if(button != (indata[0] & 0x07))
	    {
		struct msdata mouse;
                button = indata[0] & 0x07;

		if(button & 0x01)
		    mouse.event.button.R_button = 0;
		else
		    mouse.event.button.R_button = 1;

		if(button & 0x02)
		    mouse.event.button.M_button = 0;
		else
		    mouse.event.button.M_button = 1;

		if(button & 0x04)
		    mouse.event.button.L_button = 0;
		else
		    mouse.event.button.L_button = 1;

		mouse.time = (short)lbolt;
		mouse.type = BUTONEVNT;
		msputpacket(&mouse);
            }

            if(indata[1])
	    {
                ox_current = nx_current;

                if(indata[1] < 0)
                    wk = -5;
                else
                    wk = 5;

	        wk_mickey =  10 * v_mickey / b_mickey;

                wk = (short)indata[1] * wk_mickey + wk;
                tmp = wk / 10 + ox_current; 
                if(abs(tmp) > 32767)
                    nx_current = ((int)tmp>0)?32767:-32768;
                else
                    nx_current = tmp;
            }

            if(indata[2])
	    {
                oy_current = ny_current;

                indata[2] = -indata[2];
                if(indata[2] < 0)
                    wk = -5;
                else
                    wk = 5;

	        wk_mickey = 10 * v_mickey / b_mickey;
                wk = (short)indata[2] * wk_mickey + wk;
		
                tmp = wk / 10 + oy_current; 
                if(abs(tmp) > 32767)
                    ny_current = ((int)tmp>0)?32767:-32768;
                else
                    ny_current = tmp;
            }

            if((flgmode == 0) || (flgmode == 2))
		shcheck();
            else if(flgmode == 4)
                slimit();
	}
    }
    else if((data > 0x7f) && (data < 0x88))
    {
        indata[0] = data;
	indcount = 2;
    } else 
        return(0);
    return(1);
}

static void slimit()
{
    int x_x,y_y;
    int bx,i;
    short cross[4][2];

    if(nx_current == ox_current){
        ssamex();
        return;
    }
    if(ny_current == oy_current){
        ssamey();
        return;
    }
    y_y = oy_current - ny_current;
    x_x = ox_current - nx_current;
    bx = ny_current * 10  - y_y * nx_current * 10 / x_x;
    i = 0;
    if(((ny_current < limit_min_y) && (oy_current >= limit_min_y)) ||
       ((ny_current >= limit_min_y) && (oy_current < limit_min_y))){
        cross[i][0] = (x_x * limit_min_y * 10 - x_x * bx ) / y_y / 10;
        cross[i++][1] = limit_min_y;
    }
    if(((ny_current > limit_max_y) && (oy_current <= limit_max_y)) ||
       ((ny_current <= limit_max_y) && (oy_current > limit_max_y))){
        cross[i][0] = (x_x * limit_max_y * 10 - x_x * bx ) / y_y / 10;
        cross[i++][1] = limit_max_y;
    }
    if(((nx_current < limit_min_x) && (ox_current >= limit_min_x)) ||
       ((nx_current >= limit_min_x) && (ox_current < limit_min_x))){
        cross[i][0] = limit_min_x;
        cross[i++][1] = (y_y * limit_min_x * 10 / x_x + bx) / 10;
    }
    if(((nx_current > limit_max_x) && (ox_current <= limit_max_x)) ||
       ((nx_current <= limit_max_x) && (ox_current > limit_max_x))){
        cross[i][0] = limit_max_x;
        cross[i++][1] = (y_y * limit_max_x * 10 / x_x + bx) / 10;
    }
    if(!i)
        return;
    if(i==1){
        sputlimit(cross[0][0],cross[0][1]);
        return;
    }
    if(ox_current <= limit_min_x){
	if(cross[0][0] < cross[1][0]){
            sputlimit(cross[0][0],cross[0][1]);
            sputlimit(cross[1][0],cross[1][1]);
        } else {
            sputlimit(cross[1][0],cross[1][1]);
            sputlimit(cross[0][0],cross[0][1]);
        }
        return;
    }
    if(ox_current >= limit_max_x){
	if(cross[0][0] > cross[1][0]){
            sputlimit(cross[0][0],cross[0][1]);
            sputlimit(cross[1][0],cross[1][1]);
        } else {
            sputlimit(cross[1][0],cross[1][1]);
            sputlimit(cross[0][0],cross[0][1]);
        }
        return;
    }
    if(oy_current <= limit_min_y){
	if(cross[0][1] < cross[1][1]){
            sputlimit(cross[0][0],cross[0][1]);
            sputlimit(cross[1][0],cross[1][1]);
        } else {
            sputlimit(cross[1][0],cross[1][1]);
            sputlimit(cross[0][0],cross[0][1]);
        }
        return;
    }
    if(oy_current >= limit_max_y){
	if(cross[0][1] > cross[1][1]){
            sputlimit(cross[0][0],cross[0][1]);
            sputlimit(cross[1][0],cross[1][1]);
        } else {
            sputlimit(cross[1][0],cross[1][1]);
            sputlimit(cross[0][0],cross[0][1]);
        }
        return;
    }
}

static void ssamex()
{
    if(oy_current <= limit_min_y){
        if(((ny_current < limit_min_y) && (oy_current >= limit_min_y)) ||
           ((ny_current >= limit_min_y) && (oy_current < limit_min_y)))
            sputlimit(nx_current,limit_min_y);
        if(((ny_current > limit_max_y) && (oy_current <= limit_max_y)) ||
           ((ny_current <= limit_max_y) && (oy_current > limit_max_y)))
            sputlimit(nx_current,limit_max_y);

    } else {
        if(((ny_current > limit_max_y) && (oy_current <= limit_max_y)) ||
           ((ny_current <= limit_max_y) && (oy_current > limit_max_y)))
            sputlimit(nx_current,limit_max_y);
        if(((ny_current < limit_min_y) && (oy_current >= limit_min_y)) ||
           ((ny_current >= limit_min_y) && (oy_current < limit_min_y)))
            sputlimit(nx_current,limit_min_y);
    }
}


static void ssamey()
{
    if(ox_current <= limit_min_x){
        if(((nx_current < limit_min_x) && (ox_current >= limit_min_x)) ||
           ((nx_current >= limit_min_x) && (ox_current < limit_min_x)))
            sputlimit(limit_min_x,ny_current);
        if(((nx_current > limit_max_x) && (ox_current <= limit_max_x)) ||
           ((nx_current <= limit_max_x) && (ox_current > limit_max_x)))
            sputlimit(limit_max_x,ny_current);
    } else {
        if(((nx_current > limit_max_x) && (ox_current <= limit_max_x)) ||
           ((nx_current <= limit_max_x) && (ox_current > limit_max_x)))
            sputlimit(limit_max_x,ny_current);
        if(((nx_current < limit_min_x) && (ox_current >= limit_min_x)) ||
           ((nx_current >= limit_min_x) && (ox_current < limit_min_x)))
            sputlimit(limit_min_x,ny_current);
    }
}


static void sputlimit(x,y)
short x,y;
{
    struct msdata mouse;

    if (x < limit_min_x || limit_max_x < x ||
        y < limit_min_y || limit_max_y < y )
	    return;

    mouse.time = (short)lbolt;
    mouse.type = LIMITEVNT;
    mouse.event.coordinate.X_coord = x;
    mouse.event.coordinate.Y_coord = y;
    mouse.event.coordinate.X_speed = 0;
    mouse.event.coordinate.Y_speed = 0;

    msputpacket(&mouse);
}


static void mssetinit()
{
    datamode = 0;
    t_sampl = TSAMPL;

    limit_min_x = MIN_MOUSE_X;
    limit_min_y = MIN_MOUSE_Y;
    limit_max_x = MAX_MOUSE_X;
    limit_max_y = MAX_MOUSE_Y;

    v_mickey = MICKYVALUE / 2;
    b_mickey = MICKYBASE;

    nx_current = 0;
    ny_current = 0;
    indcount  = 0;

    msflush();
}

static void msflush()
{
    unsigned s;
    
    s = SPLC();

    data_buffer_lock();
    mouse_data_buffer.next_free =
	mouse_data_buffer.last_read = mouse_data_buffer.data;
    data_buffer_unlock();

    sox_current = nx_current;
    soy_current = ny_current;
    ox_current = nx_current;
    oy_current = ny_current;

    o_time = lbolt;

    splx(s);
}

static void msscan()
{
    struct msdata mouse;
    short  delta_x,delta_y,delta_t;

    if (ms_flag&MS_OPEN)
    {
        if ((nx_current != sox_current) || (ny_current != soy_current))
	{
            if(lbolt < o_time) 
                delta_t = (short)(TIMEMAX - (lbolt - o_time) + 1);
            else
                delta_t = (short)(lbolt - o_time);
            delta_x = nx_current - sox_current;
            delta_y = ny_current - soy_current;

            mouse.time = (short)lbolt;
            mouse.type = COORDEVNT;

            if(delta_t){
                mouse.event.coordinate.X_speed = delta_x / delta_t;
                mouse.event.coordinate.Y_speed = delta_y / delta_t;
            } else {
                mouse.event.coordinate.X_speed = delta_x;
                mouse.event.coordinate.Y_speed = delta_y;
            }
            if(datamode){
                mouse.event.coordinate.X_coord = delta_x;
                mouse.event.coordinate.Y_coord = delta_y;
            } else {
                mouse.event.coordinate.X_coord = nx_current;
                mouse.event.coordinate.Y_coord = ny_current;
            }
            msputpacket(&mouse);
            sox_current = nx_current;
            soy_current = ny_current;
        }
        o_time = lbolt;
        timeout((int(*)())msscan,0,t_sampl);
    } else 
        timeout((int(*)())msscan,0,IdleScan);
}

int msioctl(dev, cmd, arg, mode)
unsigned dev;
caddr_t arg;
int cmd, mode;
{
    switch(cmd)
    {
      case MSDGET:
      {
	struct msdget *m = (struct msdget *)arg;
	m->retval = msgetpacket(&m->msdata, mode & FNDELAY, m->mode, 1, 0);
	return (m->retval == -1 ? EFAULT : 0);
      }

      case MSDGETA:
      {
	struct msdgeta *m = (struct msdgeta *)arg;
	m->retval = msgetpacket(m->msdatap, mode & FNDELAY, m->mode, m->count, 1);
	return (m->retval == -1 ? EFAULT : 0);
      }

    case MSSMPLCLK:
	if(*(int*)arg < 1)
	    t_sampl = 1;
	else
	    t_sampl = *(int*)arg;
	msflush();
	return(0);

    case MSSETMICK:
	if(*(int*)arg > MICKYMAX)
	    v_mickey = MICKYMAX;
	else if(!*(int*)arg)
	    v_mickey = MICKYVALUE / 2;
	else
	    v_mickey = *(int*)arg;
	
	msflush();
	return(0);
    case MSSETLIMT:
	return smslimit((struct mssetlimt*)arg);
    case MSSETCRNT:
	return smscrnt((struct mssetcrnt*)arg);
    case MSSETMODE:
	if(*(int*)arg == 1)
	    datamode = 1;	/* if true, mouse X and Y are deltas */
	else
	    datamode = 0;	/* normal */
	
	msflush();
	return 0;
    case MSSETCURS:
	return 0;
    case MSDFLUSH:
	msflush();
	return 0;
    case FIONBIO:
	if(*(int*)arg)
	    delay_flag = 1;
	else
	    delay_flag = 0;
	return(0);
    default:
	return(EINVAL);
    }
}

static int smslimit(mssetlimt)
struct mssetlimt *mssetlimt;
{
    switch(mssetlimt->mode){
    case 0:
	limit_min_x = MIN_MOUSE_X;
	limit_min_y = MIN_MOUSE_Y;
	limit_max_x = MAX_MOUSE_X;
	limit_max_y = MAX_MOUSE_Y;
	
	flgmode = 0;
	shcheck();
	msflush();
	return(0);
    case 1:
	flgmode = 1;
	msflush();
	return(0);
    case 2:
	if(mssetlimt->X_scoord > mssetlimt->X_ecoord)
	    return(EINVAL);

	if(mssetlimt->Y_scoord > mssetlimt->Y_ecoord)
	    return(EINVAL);

	limit_min_x = mssetlimt->X_scoord;
	limit_max_x = mssetlimt->X_ecoord;

	limit_min_y = mssetlimt->Y_scoord;
	limit_max_y = mssetlimt->Y_ecoord;

	flgmode = 2;
	shcheck();
	msflush();
	return(0);

    case 4:
	if(mssetlimt->X_scoord > mssetlimt->X_ecoord)
	    return(EINVAL);
   
	if(mssetlimt->Y_scoord > mssetlimt->Y_ecoord)
	    return(EINVAL);

	limit_min_x = mssetlimt->X_scoord;
	limit_max_x = mssetlimt->X_ecoord;

	limit_min_y = mssetlimt->Y_scoord;
	limit_max_y = mssetlimt->Y_ecoord;

	flgmode = 4;
	return(0);
    default:
	return(EINVAL);
    }
}

static int smscrnt(mssetcrnt)
struct mssetcrnt *mssetcrnt;
{

    nx_current = mssetcrnt->X_coord;
    ny_current = mssetcrnt->Y_coord;

    if((flgmode == 0) || (flgmode == 2))
        shcheck();
    return(0);
}

static void shcheck()
{
    if(nx_current < limit_min_x)
        nx_current = limit_min_x;
    else if(nx_current > limit_max_x)
        nx_current = limit_max_x;

    if(ny_current < limit_min_y)
        ny_current = limit_min_y;
    else if(ny_current > limit_max_y)
        ny_current = limit_max_y;
}

/*
 * Put the given element into the mouse buffer.
 */
static void msputpacket(element)
struct msdata *element;
{
    unsigned s;

    s = SPLC();
    data_buffer_lock();
    *mouse_data_buffer.next_free = *element;
    bump_up(mouse_data_buffer.next_free);

    if (mouse_buffer_empty())
    {
	/*
	 * If we've just added an element, but it looks like the buffer
	 * is empty, we must have just come full circle and filled up the
	 * buffer.
 	 *
	 * We'll throw away the oldest element by pretending it was just read.
	 */
	bump_up(mouse_data_buffer.last_read);
    }
    data_buffer_unlock();

    if (ms_flag & MS_DWANTED)
    {
        ms_flag &= ~MS_DWANTED;
        wakeup((caddr_t)&ms_flag);
    }

#ifndef MACH_KERNEL /* XXX */
    /* additional for select */

    if(ms_selr){
	selwakeup(ms_selr, ms_flag & MS_RCOLL);
	ms_flag &= ~MS_RCOLL;
	ms_selr = 0;
    }
#endif
    splx(s);
}

/*
 * Returns the number of packets read.
 */
static int msgetpacket(addr, nodelay, flag, count, copyout_to_addr)
struct msdata *addr;
unsigned flag, count, copyout_to_addr, nodelay;
{
    int s, i, element_count;

    s = SPLC();
    data_buffer_lock();

    while (mouse_buffer_empty())
    {
	data_buffer_unlock();

        if(nodelay | delay_flag)
	{
            splx(s);
            return(0);
	}

	ms_flag |= MS_DWANTED;
	while(ms_flag & MS_DWANTED)
	    sleep((caddr_t)&ms_flag, TTIPRI);
        data_buffer_lock();
    }

    /* find out what's there */
    element_count = mouse_buffer_count();

    /* and make sure we try to read no more than that */
    if (element_count < count)
	count = element_count;

    /*
     * Apparently, flag==2 means to take the 'count' LATEST elements
     * and flush when we're done, rather than taking the next count
     * elements in chronolocial order.
     */
    if(flag == 2)
    {
	int skip = element_count - count;

	while (skip--)
	    bump_up(mouse_data_buffer.last_read);
    }

    i = count;	/* want to save 'count' as the return value */

    while (i--)
    {
	if (!copyout_to_addr)
	{
	    *addr = *mouse_data_buffer.last_read;
	}
	else
	{
	    if (copyout((char*)mouse_data_buffer.last_read,
		(unsigned)addr, sizeof(struct msdata)))
	    {
		    data_buffer_unlock();
		    splx(s);
		    return -1;
	    }
	}
	bump_up(mouse_data_buffer.last_read);
	addr++;
    }
    data_buffer_unlock();

    if (flag == 2)
	msflush();	/* data buffer must be unlocked when this is called */

    splx(s);
    return(count);
}

#ifndef MACH_KERNEL
/*
 * CURSOR
 */
#define	CURSINIT	01
#define	CURSON		010

/* additional for select */
int msselect(dev, rw)
unsigned dev;
int	rw;
{
    int	s;
    s =  SPLC();
    switch(rw)
    {
    case FREAD:
	if(!mouse_buffer_empty()) {	/* ??? what's this ??? */
	    splx(s);
	    return(1);
	}
	if(ms_selr && (ms_selr->wait_event == (int)&selwait))
	    ms_flag |= MS_RCOLL;
	else
	    ms_selr = current_thread();
	break;
    case FWRITE:
	splx(s);
	return(1);
    }
    splx(s);
    return(0);
}

#endif

#ifdef DEBUG
static void show(m, msg)
struct msdata *m;
char *msg;
{
    int count = m - mouse_data_buffer.data;

    switch(m->type)
    {
	default:
	    printf("%s %d: unknown %d\n", msg, count, m->type);
	    break;

	case BUTONEVNT:
	    printf("%s: %d: button: %d %d %d\n", msg, count,
		(int)m->event.button.L_button,
		(int)m->event.button.M_button,
		(int)m->event.button.R_button);
	    break;
    	case LIMITEVNT:
	    printf("%s %d: limit: %d %d %d %d\n", msg, count,
		(int)m->event.coordinate.X_coord,
		(int)m->event.coordinate.Y_coord,
		(int)m->event.coordinate.X_speed,
		(int)m->event.coordinate.Y_speed);
	    break;
	case COORDEVNT:
	    printf("%s: %d: coord: <%d %d> %d %d\n", msg, count,
                (int)m->event.coordinate.X_coord,
                (int)m->event.coordinate.Y_coord,
                (int)m->event.coordinate.X_speed,
                (int)m->event.coordinate.Y_speed);
	    break;
    }
}
#endif
