/*
 * Mach Operating System
 * Copyright (c) 1992, 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	uart.c,v $
 * Revision 2.6  92/05/21  17:22:09  jfriedl
 * 	Cleanup to quiet gcc warnings.
 * 	Changed UNSIGNED CHAR and DEV_T args to UNSIGNED (ANSI/std problems)
 * 	Removed unused flag argument from utclose.
 * 	Added prototypes.
 * 	[92/05/18            jfriedl]
 * 	Appended 'U' to constants that would otherwise be signed.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.5  92/05/04  11:27:37  danner
 * 	Added initialization of t_stop field (from jfriedl). 
 * 	Updated copyright.
 * 	[92/04/30            danner]
 * 
 * Revision 2.4  92/04/01  10:56:06  rpd
 * 	Added a missing ttyinput call.
 * 	[92/03/16  17:22:02  danner]
 * 
 * Revision 2.3  92/02/19  14:07:21  elf
 * 	Includes fixup.
 * 	[92/01/20            danner]
 * 	Added ddb trap via the funky two over from the BS key (right next
 * 	 to the DEL) key.
 * 	[91/09/17            danner]
 * 
 * Revision 2.2  91/07/09  23:21:19  danner
 * 	Initial Checkin
 * 	[91/06/26  11:46:18  danner]
 * 
 * Revision 1.1  91/04/21  16:22:26  jfriedl
 * Initial revision
 * 
 * Revision 2.2  91/04/05  14:10:51  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * 	cpu-board serial i/o driver
 */

/* For configuration. TTY struct must be cons_tty, */
/* Becouse enty name is cons_open ...		   */

#define uart_tty	cons_tty

#include <mach_kdb.h>

#include <sys/types.h>

#ifdef MACH_KERNEL
#include <kern/time_out.h>
#include <device/conf.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/buf.h>		/* for struct uio (!) */
#include <device/errno.h>
#include <luna88kdev/compat_25.h>
#else	MACH_KERNEL
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/tty.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/syslog.h>
#endif

#include <sys/ioctl.h>
#include <machine/board.h>

#include <luna88kdev/kbms.h>
#include <luna88kdev/cons.h>
#include <luna88kdev/uart.h>

/* OMRON extended flag (set in t_OMRON) */
#define WCLOSE          02
#define LCLOSE          04
#define HZ 60

#define TTY_NODELAY

struct uart_device volatile *uart_addr[] = {
    (struct uart_device volatile *)(OBIO_SIO),
    (struct uart_device volatile *)(OBIO_SIO+8)
    };

#define UART_COUNT (sizeof(uart_addr)/sizeof(uart_addr[0]))

/*
 * Apparently can handle only 2400, 4800, and 9600 baud
 */
unsigned char	uart_speeds[16] = {
    0,0,0,0,0,0,0,0,0,0,0,0xc0,0x80,0x40,0,0
    };

static	uart_init= 0;
static	char uart_brk[UART_COUNT];
static	char lsiprm[UART_COUNT][NWR];

#ifdef WANT_PROTOTYPES
 /* device/chario.c */
 extern int ttymodem(struct tty*,int);
 extern io_return_t char_open(int,struct tty*,dev_mode_t,io_req_t);
 extern io_return_t char_read(struct tty*,io_req_t);
 extern io_return_t char_write(struct tty*,io_req_t);
 extern void ttrstrt(struct tty*);
 extern void ttstart(struct tty*);
 extern void tty_queue_completion(queue_t);
 extern void ttychars(struct tty*);
 extern void ttyclose(struct tty*);
 extern void ttyinput(unsigned int,struct tty*);

 /* device/cirbuf.c */
 extern int getc(struct cirbuf*);

 /* kern/debug.c */
 extern void panic(char*,...);

 /* kern/mach_clock.c */
 extern void timeout(int(*)(),char*,int);

 /* luna88k/db_interface.c */
 extern void gimmeabreak(void);

 /* luna88k/locore/locore.c */
 extern void delay_in_microseconds(unsigned);

 /* luna88k/machdep.c */
 extern unsigned int spl7(void);
 extern unsigned int spltty(void);
 extern unsigned int splx(unsigned int);

 /* luna88kdev/cons.c */
 extern void cnstart(struct tty*);
 extern void j_init(struct tty*);

 /* luna88kdev/kbmain.c */
 extern int kbencode(unsigned int,char*);
 extern int kbrepeat(char*);
 extern void kbsetinit(unsigned int);

 /* luna88kdev/msmain.c */
 extern int msreceive(unsigned int);

 /* forwards */
 int utclose(unsigned int);
 int utopen(unsigned int,int,io_req_t);
 int utread(unsigned int,io_req_t);
 int utwrite(unsigned int,io_req_t);
 static int utmodem(unsigned int,int);
 static int utstart(struct tty*);
 static void kbreptscan(void);
 static void utlsi(int,unsigned int,unsigned int,int);
 static void utparam(unsigned int);
 static void utrint(int);
 static void utstop(struct tty*,int);
 static void utxint(int);
 void bellcall(int);
 void bzscan(void);
 void msparam(void);
 void utinit(void);
 void utintr(void);
#endif /* ifdef WANT_PROTOTYPES */

extern void ttrstrt();
extern void cnstart();

int	kbms_init = 0;		

#define	IFLAGS (TOSTOP|PASS8|EVENP|ODDP|ECHO|XTABS|CRMOD)

/* logical to physical line number */
#define ldevtop(dev)	(minor(dev)==2?0:1)

static void utstop(); /* forward */
static int utstart();

int utopen(dev, flag, ior)
unsigned dev;
int flag;
#ifdef MACH_KERNEL
io_req_t ior;
#else
int ior; /* unused */
#endif
{
    struct tty *tp;
    unsigned s;

    if (minor(dev) < 0 || (minor(dev) > TERM))
	return (ENXIO);
    tp = &uart_tty[minor(dev)];

#ifdef MACH_KERNEL
    {
	tp->t_stop = (int (*)())utstop;
    }
#endif

    s = SPLX();
    if (minor(dev) != TERM) {
	tp->t_oproc = (int (*)())cnstart;
    } else {
	tp->t_oproc = utstart;
    }
    if ((tp->t_state&TS_ISOPEN) == 0) {
	ttychars(tp);
	tp->t_flags = IFLAGS;
	tp->t_ospeed = tp->t_ispeed = B9600;
	j_init(tp);     		/* Japanese  tty.c j_init() */
	if(minor(dev) != TERM) {	/* BMcons      */
	    if ( kbms_init == 0 ) {	
		utparam(dev);
		kbms_init = 1;
	    }			
	    kbsetinit(0);       /* keyboard initialize    */
	} else {
	    utparam(dev);
	}
    } 
#ifndef MACH_KERNEL
    else if((tp->t_state&TS_XCLUDE) && 
	      ((u.u_uid!=0) ||
	       (u.u_procp->p_flag & SLOGIN))) 
      {
	splx(s);
	return (EBUSY);
      }
#endif MACH_KERNEL

    if(utmodem(dev, ON)) 
#ifdef MACH_KERNEL
      ttymodem(tp, 1);
#else    
    (*linesw[tp->t_line].l_modem)(tp, 1);
#endif MACH_KERNEL



#ifndef MACH_KERNEL
#ifdef TTY_NODELAY
    if (!(flag&FNDELAY)) {
#endif TTY_NODELAY
	while ((tp->t_state&TS_CARR_ON)==0) {
	    tp->t_state |= TS_WOPEN;
	    return (EBUSY);

	    sleep((caddr_t)&tp->t_rawq, TTIPRI);
	    if ((tp->t_state&TS_XCLUDE) &&
		(u.u_procp->p_flag&SLOGIN)) {
		splx(s);
		return (EBUSY);
	    }
	}
#ifdef TTY_NODELAY
    }
#endif TTY_NODELAY
#endif MACH_KERNEL

    splx(s);
#ifdef MACH_KERNEL
    return char_open(dev, tp, flag, ior);
#else
    return((*linesw[tp->t_line].l_open)(dev, tp));
#endif 
}

int utclose(dev)
unsigned dev;
{
    struct tty *tp;
    unsigned  s;
    extern kbreset;

    tp = &uart_tty[minor(dev)];
#ifndef MACH_KERNEL
    (*linesw[tp->t_line].l_close)(tp);
#endif
#ifndef MACH_KERNEL
    if (tp->t_jt.j_flg & CONVTOEXT) /* wait for KOUT transmit */
	delay(HZ/2);
#endif
    if (tp->t_state&TS_HUPCLS && !(tp->t_state&TS_ISOPEN))
	utmodem(dev, OFF);
    s = SPLX();
    if (minor(dev) == KBD)
	kbreset = ENABLE;
    ttyclose(tp);
    splx(s);
    return 0;
}

#ifdef MACH_KERNEL
int utread(dev, ior)
unsigned dev;
io_req_t ior;
#else
int utread(dev, uio)
unsigned dev;
struct uio *uio;
#endif
{
    struct tty *tp;

    tp = &uart_tty[minor(dev)];
#ifdef MACH_KERNEL
    return char_read(tp, ior);
#else
    return((*linesw[tp->t_line].l_read)(tp, uio));
#endif
}

#ifdef MACH_KERNEL
int utwrite(dev,ior)
unsigned dev;
io_req_t ior;
#else
int utwrite(dev,uio)
unsigned dev;
struct uio *uio;
#endif
{
    struct tty *tp;
    
    tp = &uart_tty[minor(dev)];
#ifdef MACH_KERNEL
    return char_write(tp, ior);
#else
    return((*linesw[tp->t_line].l_write)(tp, uio));
#endif
}

#ifndef MACH_KERNEL
int utioctl(dev, cmd, addr, flag)
unsigned dev;
caddr_t addr;
int cmd, flag;
{
    struct tty *tp = &uart_tty[minor(dev)];
    unsigned s, error = 0, flgmodesv;

    if (minor(dev) != TERM ) { 
	flgmodesv = flgmode;
	if((error = kbioctl(tp,cmd,addr,flag)) > 0)
	    return(error);
	if (flgmodesv == ENCODED && flgmode != ENCODED ){
	    s = SPLX();
	    tp->t_flags &= ~(ECHO|NOFLSH|
			     (CRTBS|CRTERA|CRTKIL)|LCASE|CRMOD);
	    tp->t_flags |= (RAW|DECCTQ|TANDEM);
	    splx(s);
	}
	return(error);
    } else {
	if ((error=(*linesw[tp->t_line].l_ioctl)(tp,cmd,addr,flag))>=0)
	    return (error);
	if ((error=ttioctl(tp, cmd, addr, flag)) >= 0) {
	    if (cmd==TIOCSETP||cmd==TIOCSETN||cmd==TIOCLSET||
		cmd == TIOCLBIS || cmd == TIOCLBIC)
		utparam(dev);
    /* WARNING-
     * I just noticed that in some cases (cmd is an unknown value),
     * ttioctl() returns -1.  In other parts of the kernel, this
     * is translated to ENOTTY, while yet in other parts, it's
     * not checked.  In some of those parts, it _SHOULD_ be checked,
     * and so I'm going around all the files looking for those that
     * should be checked.
     *
     * I can't tell if this is one of those, but hopefully someone who
     * knows what this code does will see this and check....
     */
	    return (error);
	}
	switch (cmd) {
	case TIOCSDTR:
	    utmodem(dev, ON);
	    break;
	case TIOCCDTR:
	    utmodem(dev, OFF);
	    break;
	case TIOCSBRK:
	default:
	    return(ENOTTY);
	}
	return(error);
    }
}
#endif MACH_KERNEL /* ifndef */

static void utparam(dev)
unsigned dev;
{
    struct tty *tp;
    char data;
    int line;
    unsigned s;
    char wr3,wr5, wr4 = 0;	
    static short kbrep_scan = 0;
    
    tp = &uart_tty[minor(dev)];
    line = ldevtop(dev);
    /* hang up line */
    if (tp->t_ispeed == 0) {
	utmodem(dev, OFF);
	return;
    }
    if ((minor(dev) != TERM)&&(kbrep_scan == 0)) {
	kbreptscan();
	kbrep_scan++;
    }
    if (minor(dev) != TERM)
	tp->t_ispeed = tp->t_ospeed = B9600;
    s = SPLX();
    lsiprm[line][WR4] = 0;
    lsiprm[line][WR3] = 0;
    lsiprm[line][WR5] = 0;
    if (!(data = uart_speeds[(unsigned char)tp->t_ispeed]))
	data = uart_speeds[B9600];
    if (tp->t_ispeed == B134) {			/* CS6 */
	wr3 = BITS6;
	wr5 = BITS6>>1;
    } else if (tp->t_flags & (RAW|LITOUT|PASS8)) {	/* CS8 */
	wr3 = BITS8;
	wr5 = BITS8>>1;
    } else {					/* CS7 */
	wr3 = BITS7;
	wr5 = BITS7>>1;
    }
    if (((tp->t_flags&ODDP)&&(tp->t_flags&EVENP)) ||
	(!(tp->t_flags&ODDP)&&!(tp->t_flags&EVENP))) {
	wr4 = 0;
    } else if (tp->t_flags&ODDP) {
	wr4 = PENABLE;
    } else {
	wr4 = (PENABLE|EVENPA);
    }
    if (tp->t_ospeed == B110 || tp->t_omron & TS_TWOSB) {
	wr4 |= TWOSB;
    } else {
	wr4 |= ONESB;
    }
    wr4 |= data;
    wr5 |= RTS;
    utlsi(line,WR4, wr4, ON );
    utlsi(line,WR3,(RXEN|wr3),ON);
    utlsi(line,WR5,(TXEN|wr5),ON);
    lsiprm[line][WR1] = 0;
    utlsi(line,WR1,RXIE|TXIE,ON); 
    splx(s);
}

/*
 * Called upon a level 5 interrupt from luna88k/machdep.c
 */
void utintr()
{
    struct uart_device volatile *uartaddr;
    int line;

    if (!uart_init){
	uartaddr = uart_addr[0];
	uartaddr->cmd = RSTINT;
	return;
    }

    for(line=0 ; line < UART_COUNT; line++)
    {
        uartaddr = uart_addr[line];
	if (uartaddr->status&RXRDY && !(uartaddr->status&RCVBRK))
	    utrint(line);
	if (uartaddr->status&TXRDY)
	    utxint(line);
	if(uartaddr->status&RCVBRK)
	    uart_brk[line]++;
	uartaddr->cmd = RSTESIN;
    }
}

char encodebuf[32];

static void utrint(line)
int line;
{
    struct tty *tp;
    struct uart_device volatile *uartaddr;
    unsigned char data;
    int count, i;
    unsigned short c;
    
    uartaddr = uart_addr[line];
    if (line == 0)
	tp = &uart_tty[TERM];
    else
    {
	if (line == KBD
#ifndef MACH_KERNEL
	    && !conssw.kbopen
#endif
	    )
	    tp = &uart_tty[CONSOLE];
	else
	    tp = curtty;

	if (tp != &uart_tty[CONSOLE] && tp != &uart_tty[KBD])
	    panic("utrint");
    }

    data = uartaddr->data;

    if ((tp->t_state&(TS_ISOPEN|TS_WOPEN)) == 0)
	return;

#ifndef MACH_KERNEL /* XXX */

    if((tp->t_flags &  (NOHANG
#ifdef  NOMDM
			| NOMDM
#endif  NOMDM
			| MDMBUF)) || (uartaddr->status&CD)) {
	if ((tp->t_state&TS_CARR_ON) == 0) {
	    wakeup(&tp->t_canq);
	    tp->t_state |= TS_CARR_ON;
	}
    } else {
	if (tp->t_state&TS_CARR_ON) {
	    gsignal(tp->t_pgrp, SIGHUP);
	    tp->t_state &= ~TS_CARR_ON;
	    ttyflush(tp, (FREAD|FWRITE));
	}
	return;
    }

#endif MACH_KERNEL /* XXX */

    uartaddr->cmd = WR1;
    c = uartaddr->status;
    if((c&RXOVRUN)|(c&PE))
	uartaddr->cmd = ERRST;

    if(uart_brk[line]) {
	uart_brk[line] = 0;
	c = (FRMER <<7);
#ifdef MACH_KERNEL
	data = tp->t_breakc; /* XXX was 0 */
	ttyinput(data, tp);
#else
	data = tp->t_intrc;
	(*linesw[tp->t_line].l_rint)(data, tp);
#endif
    } else { /* not break charcter */
	if (line == 0) { /* TTYA  */
	    if (tp->t_ispeed == B134)	/* CS6 */
		data &= 077;
	    else if (tp->t_flags&(RAW|LITOUT|PASS8))/*CS8*/
		;
	    else
		data &= 0177;		/* CS7 */
#if NBK > 0 || Nbk > 0
	    if(tp->t_line == NETLDISC) {
		data &= 0177;
		BKINPUT(data,tp);
	    } else {
#endif
#ifndef MACH_KERNEL
		if(tp->t_state&TS_TTSTOP)
		{
		    if ((tp->t_flags&DECCTQ)&& (data == tp->t_startc))
			tp->t_state &= ~TS_TTSTOP;
		}
#endif 

#ifdef MACH_KERNEL
		ttyinput(data, tp);
#else
		(*linesw[tp->t_line].l_rint)(data, tp);
#endif

#ifndef MACH_KERNEL /* XXX */
		if((tp->t_state&TS_TTSTOP) == 0)
		{
		    if ((tp->t_flags&DECCTQ)&& (data == tp->t_stopc))
			tp->t_state |= TS_TTSTOP;
		}
#endif

#if NBK > 0 || Nbk > 0
	    }
#endif
	} else { /* CONSOLE */

#ifdef MACH_KDB
	  if (data == 0x19)
	    {
	      gimmeabreak();
	      return;
	    }
#endif
	    if(msreceive(data))
		return;
	    if(!(tp->t_state&(TS_ISOPEN|TS_WOPEN)))
		return;
	    count = kbencode(data, encodebuf);
	    if(count == KBNON || count == KBRESET ||
	       tp->t_state & TS_TBLOCK)
		return;
	    for (i = 0; i < count; i++) {
		data = encodebuf[i];
#if NBK > 0 || Nbk > 0
		if(tp->t_line == NETLDISC) {
		    data &= 0177;
		    BKINPUT(data,tp);
		} else  {
#endif

#ifdef MACH_KERNEL
		  ttyinput(data,tp);
#else
		  (*linesw[tp->t_line].l_rint)(data, tp);
#endif

#if NBK > 0 || Nbk > 0
		}
#endif
	    }
	} /* CONSOLE */
    } /* not break charcter */
}

static void utxint(line)
int line;
{
    struct tty *tp;
    struct uart_device volatile *uartaddr;

    uartaddr = uart_addr[line];
    if (line == 0) 
	tp = &uart_tty[TERM];
    else 
	goto out;
    if((tp->t_state & (TS_ISOPEN|TS_WOPEN))==0) 
	goto out;

    if (tp->t_state & TS_FLUSH) 
	tp->t_state &= ~TS_FLUSH;
    if (tp->t_state & TS_BUSY) {
	tp->t_state &= ~TS_BUSY;
	if (tp->t_line) 
#ifdef MACH_KERNEL
	    ttstart(tp);
#else
	    (*linesw[tp->t_line].l_start)(tp);
#endif
	else 
	    utstart(tp);
	if (tp->t_state & TS_BUSY)
	    return;
    }
out:
    uartaddr->cmd = RSTPEDI;
    return;
}

static void utstop(tp, flag)
struct tty *tp;
int flag;
{
    unsigned s;

    s = SPLX();
    if (tp->t_state & TS_BUSY) 
	if ((tp->t_state&TS_TTSTOP)==0) 
	    tp->t_state |= TS_FLUSH;
    splx(s);
}

static int utstart(tp)
struct tty *tp;
{
    struct uart_device volatile *uartaddr;
    int c, x;

    uartaddr = uart_addr[ldevtop(TERM)];
    x = SPLX();
    if (tp->t_state & (TS_TIMEOUT|TS_BUSY|TS_TTSTOP))
	goto out;
    if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
#ifdef MACH_KERNEL
	    tt_write_wakeup(tp);
#else
	if (tp->t_state&TS_ASLEEP) {
	    tp->t_state &= ~TS_ASLEEP;
	    wakeup((caddr_t)&tp->t_outq);
	}
	if (tp->t_wsel) {
	    selwakeup(tp->t_wsel, tp->t_state & TS_WCOLL);
	    tp->t_wsel = 0;
	    tp->t_state &= ~TS_WCOLL;
	}
#endif
    }
    if (tp->t_outq.c_cc == 0) 
	goto out;
    tp->t_state |= TS_BUSY;
    c = getc(&tp->t_outq);
    if (tp->t_flags & (RAW|LITOUT|PASS8)) {
	uartaddr->data = c&0xff;
    } else if (c <= 0177) {
	uartaddr->data = c&0xff;
    } else {
	tp->t_state &= ~TS_BUSY;
	timeout((int(*)())ttrstrt, (caddr_t) tp, c&0xff);
	tp->t_state |= TS_TIMEOUT;
    }
 out:
    splx(x);
    return(1);
}

static int utmodem(dev, flag)
unsigned dev;
int flag;
{
    struct uart_device volatile *uartaddr;
    int line;

    line = ldevtop(dev);
    uartaddr = uart_addr[line];
    if (flag == OFF) 
	utlsi(line,WR5,~(DTR|RTS),OFF);
    else 
	utlsi(line,WR5,(DTR|RTS),ON); 
    return(uartaddr->status & CD);
}

void utinit()
{
    int line;
    int wr;
    struct uart_device volatile *uartaddr = (struct uart_device *)&wr;
    unsigned s;
    extern clev_tty;

#if 1
    /* the following delay approximately equals what might have been there */
    delay_in_microseconds(3200);	/* 0.003200 seconds */
#else
    for(s=0;s<20000;s++);
#endif
    (*(int *)uartaddr)++;
    s = spl7();
    if (uart_init == 0){
	for (line = 0; line < UART_COUNT; line++) {
	    uartaddr = uart_addr[line];	
	    for (wr = 0 ; wr < NWR ; wr++ )
		lsiprm[line][wr] = 0; 
	    uartaddr->cmd = CHANRST;              
	}
	uart_init = ON;
    }
    utlsi(0,WR2,WR2DATA,ON);
    utlsi(1,WR2,0 ,OFF);
    for (line = 0; line < UART_COUNT; line++) {
	utlsi(line,WR4,ONESB|WR4DATA ,ON);
	utlsi(line,WR3,BITS8|RXEN ,ON);
	utlsi(line,WR5,(BITS8>>1)|TXEN|RTS ,ON);
    }
    uartaddr = uart_addr[0];
    while ((uartaddr->status&INTPEND) == INTPEND)
	uartaddr->cmd = RSTINT;
    uartaddr->cmd = RSTPEDI;
    uartaddr = uart_addr[1];
    uartaddr->cmd = RSTPEDI;
#ifndef MACH_KERNEL /* XXX */
    clev_tty = 6;
#endif
    splx(s);
}

static void utlsi(line, wr, data, onoff)
int line, onoff;
unsigned wr, data; 
{
    char *prm;
    struct uart_device volatile *uartaddr;
    unsigned s;

    s = SPLX();
    prm = &lsiprm[line][wr];
    uartaddr = uart_addr[line];
    if (onoff==ON)
	*prm |= data;
    else
	*prm &= data;
    uartaddr->cmd = wr;
    uartaddr->cmd = *prm;
    uartaddr->cmd = ERRST;
    splx(s);
}

#ifndef MACH_KERNEL /* Mmmm, don't seem to use it at all */
static int kbioctl(tp, cmd, arg, mode)
struct tty *tp;
caddr_t arg;
int cmd, mode;
{
    int error;
    unsigned newflags;

#ifndef MACH_KERNEL
    unsigned s;

    switch(cmd) {
    case TIOCSETD:
    case TIOCSETP:
    case TIOCSETN:
    case TIOCFLUSH:
    case TIOCSETC:
    case TIOCSLTC:
    case TIOCSPGRP:
    case TIOCLBIS:
    case TIOCLBIC:
    case TIOCLSET:
    case TIOCSTI:
    case TIOCSWINSZ:
	while (tp->t_line == NTTYDISC &&
	       u.u_procp->p_pgrp != tp->t_pgrp && tp == u.u_ttyp &&
	       (u.u_procp->p_flag&SVFORK) == 0 &&
	       (u.u_procp->p_flag&SVFORK) == 0 &&
	       !(u.u_procp->p_sigignore & sigmask(SIGTTOU)) &&
	       !(u.u_procp->p_sigmask & sigmask(SIGTTOU))) {
	    gsignal(u.u_procp->p_pgrp, SIGTTOU);
	    sleep((caddr_t)&lbolt, TTOPRI);
	}
	break;
    }
#endif

    if (((cmd & IOCTYPE) >> 8) =='K') 
	return(kbxioctl(tp, cmd, arg, mode));

    newflags = tp->t_flags;
    switch(cmd)
    {
#ifndef MACH_KERNEL
      case TIOCLSET:
	newflags &= 0xffff;
	/*FALLTHROUGH*/

      case TIOCLBIS:
	newflags |= (*(unsigned *)arg << 16) | (PASS8|ODDP|EVENP);
	tp->t_omron |= TWOSB;
	tp->t_flags = newflags;
#ifdef  NOMDM
	if ((tp->t_flags&NOMDM) && (tp->t_state&TS_CARR_ON) == 0)
	    (void)(*linesw[tp->t_line].l_modem)(tp, 1);
#endif  NOMDM
	return 0;

      case TIOCLBIC:
	newflags &= ~(*(unsigned *)arg << 16);
	newflags |= (PASS8|ODDP|EVENP);
	tp->t_omron |= TWOSB;
	tp->t_flags = newflags;
	return 0;

    case TIOCSETP:
    case TIOCSETN: {
	struct sgttyb *sg = (struct sgttyb *)arg;
	
	tp->t_erase = sg->sg_erase;
	tp->t_kill = sg->sg_kill;
	tp->t_ispeed = sg->sg_ispeed;
	tp->t_ospeed = sg->sg_ospeed;
	newflags = (tp->t_flags&0xffff0000U) | (sg->sg_flags&0xffff);
	s = SPLX();
	if (tp->t_flags&RAW || newflags&RAW || cmd == TIOCSETP) {
	    ttywait(tp);
	    ttyflush(tp, (FREAD|FWRITE));
	} else if ((tp->t_flags&CBREAK) != (newflags&CBREAK)) {
	    if (newflags&CBREAK) {
		struct clist tq;
		catq(&tp->t_rawq, &tp->t_canq);
		tq = tp->t_rawq;
		tp->t_rawq = tp->t_canq;
		tp->t_canq = tq;
	    } else {
		tp->t_flags |= PENDIN;
		newflags |= PENDIN;
		ttwakeup(tp);
	    }
	}
	newflags |= (PASS8|ODDP|EVENP);
	tp->t_omron |= TWOSB;
	tp->t_flags = newflags;
	if (tp->t_flags&RAW) {
	    tp->t_state &= ~TS_TTSTOP;
	    ttstart(tp);
	}
	splx(s);
	return(0);
    }
#endif MACH_KERNEL
    default:
#ifndef MACH_KERNEL
	error = ttioctl(tp, cmd, arg, mode);
	if (error < 0)
#endif
	    error = ENOTTY;	/* jfriedl@omron 90-Mar-12 */
	return(error);
    }
    return(error);
}
#endif /*MACH_KERNEL*/

char	repeatbuf[32];

static void kbreptscan()
{
    struct tty *tp;
    unsigned short c;
    unsigned char ctmp = 0;
    int count, i, j;
    
    tp = curtty;
    
    if (!(tp->t_state&(TS_ISOPEN|TS_WOPEN))) {
	timeout((int(*)())kbreptscan,0,KBIDLESCAN);
	return;
    }
    c = 0;
    count = kbrepeat(repeatbuf);
    if(count == KBNON || count == KBRESET || tp->t_state & TS_TBLOCK) {
	timeout((int(*)())kbreptscan,0,KBREPTSCAN);
	return;
    }
    if((count > 0) && (tp->t_flags & DECCTQ)) {
	for(i=0;i<count;) {
	    c = repeatbuf[i] & 0177;
	    if ((ctmp == CSTART) || (ctmp == CSTOP)) {
		ctmp = c;
		count--;
		for(j=0;j<count;j++)
		    repeatbuf[j] = repeatbuf[j+1];
		continue;
	    } else 
		ctmp = c;
	    i++;
	}
    } else if(count == KBBREAK) 
	c = (BSD_FRERROR|BSD_PERROR|BSD_OVERRUN);
    
    if (tp->t_flags&DECCTQ) {
	if(tp->t_state & TS_TTSTOP ) {
	    if (((ctmp == CSTART) || (ctmp == CSTOP)) &&
		(count == KBNON)) {
		timeout((int(*)())kbreptscan,0,KBREPTSCAN);
		return;
	    }
	}
    }
    if (c & (BSD_FRERROR|BSD_PERROR|BSD_OVERRUN)) {
	timeout((int(*)())kbreptscan,0,KBREPTSCAN);
	return;
    }
    i = count;
    while (count) {
	c = repeatbuf[i-(count--)];
#ifdef MACH_KERNEL
	ttyinput(c,tp);
#else
	(*linesw[tp->t_line].l_rint)(c, tp);
#endif
    }
    timeout((int(*)())kbreptscan,0,KBREPTSCAN);
}

void msparam()
{
    int speed;

    lsiprm[ldevtop(KBD)][WR4] = 0;
    lsiprm[ldevtop(KBD)][WR3] = 0;
    lsiprm[ldevtop(KBD)][WR5] = 0;
    speed = uart_speeds[B9600];
    utlsi(ldevtop(KBD),WR4,(speed|ONESB|WR4DATA),ON);
    utlsi(ldevtop(KBD),WR3,(RXEN|BITS8),ON);
    utlsi(ldevtop(KBD),WR5,(TXEN|BITS8>>1),ON);
    lsiprm[ldevtop(KBD)][WR1] = 0;
    utlsi(ldevtop(KBD),WR1,RXIE|TXIE,ON);
}

void bzscan()
{
    extern bztime;
    struct uart_device volatile *uartaddr;

    if((--bztime)<=0)
	return;
    uartaddr = uart_addr[ldevtop(KBD)];
    uartaddr->data = 0x49;
    timeout((int(*)())bzscan, 0, 1);
}

void bellcall(flag) 
int flag;
{
    struct uart_device volatile *uartaddr;

    uartaddr = uart_addr[ldevtop(KBD)];
    uartaddr->data = flag;
}


#ifdef MACH_KERNEL
#ifdef unused
int utportdeath(dev, port)
unsigned	dev;
mach_port_t	port;
{
  struct tty *tp = &uart_tty[minor(dev)];
  return (tty_portdeath(tp, (ipc_port_t)port));
}
#endif /* unused */

#ifdef unused
io_return_t utgetstat(dev, flavor, data, count)
unsigned	dev;
int		flavor;
int		*data;		/* pointer to OUT array */
unsigned int	*count;		/* out */
{
	io_return_t	result = D_SUCCESS;
	unsigned	unit = minor(dev);

	switch (flavor) {
	default:
		result = tty_get_status(&uart_tty[unit], flavor, data, count);
		break;
	}
	return (result);
}
#endif /*unused*/

#ifdef unused
io_return_t utsetstat(dev, flavor, data, count)
unsigned	dev;
int		flavor;
int *		data;
unsigned int	count;
{
	io_return_t	result = D_SUCCESS;
	unsigned 	unit = minor(dev);

	switch (flavor) {
	default:
		result = tty_set_status(&uart_tty[unit], flavor, data, count);
		if (result == D_SUCCESS && flavor == TTY_STATUS)
			utparam(unit);
		return (result);
	}
	return (D_SUCCESS);
}
#endif /*unused*/

#endif
