/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	uart.h,v $
 * Revision 2.2  91/07/09  23:21:24  danner
 * 	Initial Checkin
 * 	[91/06/26  11:46:09  danner]
 * 
 * Revision 2.2  91/04/05  14:10:57  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 */

/*
 * Uart (on the CPU board) driver header
 */

#define UARTSCAN  	2	/* clock ticks for character scan */
#define UARTMSCAN	20	/* scan loops for modem scan */

#define KBIDLESCAN	HZ	/* SX8870 */
#define KBREPTSCAN	1	/* SX8870 */

#define UART_CHAR 01		/* character arrived flag (uart_work) */
#define UART_ACT  02		/* character in buffer flag (uart_work) */

extern lbolt;
extern struct conssw conssw;

#define RXRDY   0x01		/* Rx CHAR. Available (RR0) */
#define INTPEND 0x02		/* INT Pending (CH.A) (RR0) */
#define TXRDY   0x04		/* Tx Buffer Empty (RR0) */
#define CD	0x08		/* Data Carrier Detect (RR0) */
#define CTS	0x20		/* CTS (RR0) */
#define TXUDRUN	0x40		/* Tx Underrun/EOM (RR0) */
#define RCVBRK	0x80		/* Break/Abort (RR0) */

#define ALLSENT 0x01		/* All Sent (RR1) */
#define PE	0x10		/* Parity Error (RR1) */
#define RXOVRUN 0x20		/* Rx Overrun Error (RR1) */
#define FRMER	0x40		/* CRC/Framing Error (RR1) */

#define RSTINT  0x38		/* End of Interrupt (WR0) */
#define RSTESIN 0x10		/* Reset External/Status Interrupt (WR0) */
#define CHANRST 0x18		/* Channel Reset (WR0) */
#define RSTPEDI 0x28		/* Reset Tx Interrput/DMA Pending (WR0) */
#define ERRST	0x30		/* Error Reset (WR0) */
#define EOMRST	0xc0		/* Tx Underrun/EOM (WR0) */
#define ESIE    0x01		/* External Interrupt Enable (WR1) */
#define TXIE	0x02		/* Tx Interrupt/DMA Enable (WR1) */
#define VECMODE 0x04		/* Status Affects Vector (CH.B) (WR1) */
#define RXIE	0x10		/* Rx Interrupt - All Char or Specilal condition (WR1) */
#define RXID	0xe7		/* Rx Int/DMA disable (WR1) */
#define RR2     2
#define INTMODE 0x03		/* INT/DMA Mode 0,1 (WR2A) */
#define CHMSK   0x04		/* Priority Select (WR2A) */
#define RXAVL   0x02		/* Receive Character Available (WR2B) */
#define SPECIAL 0x03		/* Special Receive Condition (WR2B) */
#define TXEMPT  0x00		/* Transmit Buffer Empty (WR2B) */
#define WR1DATA 0x1e
/*#define WR2DATA 0x10*/
#define WR2DATA 0x14
#define WR4DATA 0x40
#define BITS0	0xc0
#define BITS5	0x00		/* Rx 5 Bits/Character (WR3) */
#define BITS7	0x40		/* Rx 7 Bits/Character (WR3) */
#define BITS6   0x80		/* Rx 6 Bits/Character (WR3) */
#define BITS8	0xc0		/* Rx 8 Bits/Character (WR3) */
#define AUTOEN  0x20		/* Auto Enables (WR3) */
#define RXEN    0x01		/* Rx Enable (WR3) */
#define ONESB	0x04		/* 1 Stop Bit/Character (WR4) */
#define TWOSB   0x0c		/* 2 Stop Bit/Character (WR4) */
#define PENABLE 0x01		/* Parity Enable (WR4) */
#define EVENPA	0x02		/* Parity Even/Odd (1=Even) (WR4) */
#define DTR     0x80		/* DTR (WR5) */
#define SENDBRK	0x10		/* Send Break (WR5) */
#define TXEN	0x08		/* Tx Enable (WR5) */
#define RTS	0x02		/* RTS (WR5) */
#define NOTXINT 0
#define TXINT	1
#define CHRSET	1
#define NOCHR	0

struct uart_device {
    unsigned char data;
    unsigned char revl;
    unsigned char rev3,rev4;
    unsigned char cmd;
    unsigned char rev2;
    unsigned char rev5,rev6;
};
#define WR0     0
#define WR1	1
#define WR2	2
#define WR3	3
#define WR4	4
#define WR5	5
#define NWR     6
#define	ON	1
#define	OFF	0
#define UARTA	0
#define status	cmd
#define MODESET 0xff00
#define OPMODE	0x0b22
#define CNTSET	0xff08
#define START	0xff20
#define CSMASK  060
#define BRK     1  

#define	QESC	0200	/* queue escape */
#define	HQEND	01	/* high queue end */

#define	TTIPRI	28
#define	TTOPRI	29

/* Hardware bits */
#define	BSD_DONE	0200
#define	BSD_NABLE	0100
#define	BSD_OVERRUN	040000
#define	BSD_FRERROR	020000
#define	BSD_PERROR	010000
