/*
 * HISTORY
 * $Log:	pmap.h,v $
 * Revision 2.2  92/08/03  17:49:02  jfriedl
 * 	Created from Horizon Research version.
 * 	[92/07/24            jfriedl]
 * 
 */
#ifndef _PMAP_MACHINE_
#define _PMAP_MACHINE_

/* use builtin memcpy in gcc 2.0 */
#if (__GNUC__ > 1)
#define bcopy(a,b,c) memcpy(b,a,c)
#endif

#include	<kern/lock.h>
#include	<machine/mmu.h>
#include	<mach/vm_statistics.h>

/*
 *	10/3/90 Fuzzy modfied
 *	Miss:	typedef sdt_entry_t -> typedef struct sdt_entry
 */
typedef struct sdt_entry *sdt_ptr_t;

/*
 * PMAP structure
 */

typedef struct pmap     *pmap_t;

/*
 *	10/3/90 Fuzzy modfied
 *	Miss:	typedef struct pmap -> struct pmap
 */
struct pmap {
    sdt_ptr_t		sdt_paddr;	/* physical pointer to sdt */
    sdt_ptr_t		sdt_vaddr;	/* virtual pointer to sdt */
    simple_lock_data_t	lock;		/* lock on map  */
    int			ref_count;	/* reference count */

    struct pmap_statistics stats;	/* pmap statistics */
    /*
     * cpus using of this pmap; NCPU must be <= 32
     */
    unsigned long      cpus_using;

#ifdef DBG
    pmap_t		next;
    pmap_t		prev;
#endif
};

#define PMAP_NULL ((pmap_t) 0)

extern	pmap_t	kernel_pmap;


/*
 * Macros to operate cpus_using field
 */

#define SETBIT_CPUSET(cpu_number, cpuset)	\
    (*(cpuset)) |= (1 << (cpu_number));


#define CLRBIT_CPUSET(cpu_number, cpuset)	\
    (*(cpuset)) &= ~(1 << (cpu_number));

/*
 * Macro PMAP_ACTIVATE and PMAP_DEACTIVATE have changed to simply call
 * function _pmap_activate and _pmap_deactivate for easy debugging.
 *		N. Sugai	3/11/91
 */

#define PMAP_ACTIVATE(pmap, th, my_cpu)	_pmap_activate(pmap, th, my_cpu)
#define PMAP_DEACTIVATE(pmap, th, my_cpu) _pmap_deactivate(pmap, th, my_cpu)

#define PMAP_CONTEXT(pmap, thread)

#define pmap_resident_count(pmap) ((pmap)->stats.resident_count)

/* Used in builtin/device_pager.c */
#define pmap_phys_address(frame)        ((vm_offset_t) (M88K_PTOB(frame)))

/* Used in kern/mach_timedev.c */
#define pmap_phys_to_frame(phys)        ((int) (M88K_BTOP(phys)))

/*
 * Since Our PCB has no infomation about the mapping,
 * we have nothing to do in PMAP_PCB_INITIALIZE.
 */
/* Used in machine/pcb.c */
#define PMAP_PCB_INITIALIZE(x)

/*
 * Modes used when calling pmap_cache_fulsh().
 */
#define	FLUSH_CACHE		0
#define	FLUSH_CODE_CACHE	1
#define	FLUSH_DATA_CACHE	2
#define	FLUSH_LOCAL_CACHE	3
#define	FLUSH_LOCAL_CODE_CACHE	4
#define	FLUSH_LOCAL_DATA_CACHE	5

#endif /* */

/* endif  _PMAP_MACHINE_ */
