.\" 
.\" Mach Operating System
.\" Copyright (c) 1991,1990 Carnegie Mellon University
.\" All Rights Reserved.
.\" 
.\" Permission to use, copy, modify and distribute this software and its
.\" documentation is hereby granted, provided that both the copyright
.\" notice and this permission notice appear in all copies of the
.\" software, derivative works or modified versions, and any portions
.\" thereof, and that both notices appear in supporting documentation.
.\" 
.\" CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
.\" CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
.\" ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
.\" 
.\" Carnegie Mellon requests users of this software to return to
.\" 
.\"  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
.\"  School of Computer Science
.\"  Carnegie Mellon University
.\"  Pittsburgh PA 15213-3890
.\" 
.\" any improvements or extensions that they make and grant Carnegie Mellon
.\" the rights to redistribute these changes.
.\" 
.\" 
.\" HISTORY
.\" $Log:	mach_port_extract_right.man,v $
.\" Revision 2.5  91/12/11  08:42:52  jsb
.\" 	Changed <mach/mach.h> to <mach.h>.
.\" 	[91/11/25  10:55:22  rpd]
.\" 
.\" Revision 2.4  91/05/14  17:06:16  mrt
.\" 	Correcting copyright
.\" 
.\" Revision 2.3  91/02/14  14:11:11  mrt
.\" 	Changed to new Mach copyright
.\" 	[91/02/12  18:11:23  mrt]
.\" 
.\" Revision 2.2  90/08/07  18:36:37  rpd
.\" 	Created.
.\" 
.TH mach_port_extract_right 2 4/13/87
.CM 4
.SH NAME
.nf
mach_port_extract_right \- extracts a port right from a task
.SH SYNOPSIS
.nf
.ft B
#include <mach.h>

kern_return_t
mach_port_extract_right(task, name, desired_type, right, acquired_type)
	task_t task;
	mach_port_t name;
	mach_msg_type_name_t desired_type;
	mach_port_t *right;			/* out */
	mach_msg_type_name_t *acquired_type;	/* out */
.fi
.ft P
.SH ARGUMENTS
.TP 12
.B
task
The task holding the port right.
.TP 12
.B
name
The task's name for the port right.
.TP 12
.B
desired_type
IPC type, specifying how the right should be extracted.
.TP 12
.B
right
The extracted right.
.TP 12
.B
acquired_type
The type of the extracted right.
.SH DESCRIPTION
\fBmach_port_extract_right\fR extracts a port right
from the target task and returns it to the caller
as if the task sent the right voluntarily, using \fBdesired_type\fR
as the value of \fBmsgt_name\fR.  See \fBmach_msg(2)\fR.

The returned value of \fBacquired_type\fR will be MACH_MSG_TYPE_PORT_SEND
if a send right is extracted, MACH_MSG_TYPE_PORT_RECEIVE
if a receive right is extracted, and MACH_MSG_TYPE_PORT_SEND_ONCE
if a send-once right is extracted.
.SH DIAGNOSTICS
.TP 25
KERN_SUCCESS
The call succeeded.
.TP 25
KERN_INVALID_TASK
\fBtask\fR was invalid.
.TP 25
KERN_INVALID_NAME
\fBname\fR did not denote a right.
.TP 25
KERN_INVALID_RIGHT
\fBname\fR denoted an invalid right.
.TP 25
KERN_INVALID_VALUE
\fBdesired_type\fR was invalid.
.PP
The \fBmach_port_extract_right\fR call is actually an RPC to \fBtask\fR,
normally a send right for a task port, but potentially any send right.
In addition to the normal diagnostic
return codes from the call's server (normally the kernel),
the call may return \fBmach_msg(2)\fR return codes.
.SH SEE ALSO
mach_port_insert_right(2), mach_msg(2).
