.\" 
.\" Mach Operating System
.\" Copyright (c) 1991,1990 Carnegie Mellon University
.\" All Rights Reserved.
.\" 
.\" Permission to use, copy, modify and distribute this software and its
.\" documentation is hereby granted, provided that both the copyright
.\" notice and this permission notice appear in all copies of the
.\" software, derivative works or modified versions, and any portions
.\" thereof, and that both notices appear in supporting documentation.
.\" 
.\" CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
.\" CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
.\" ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
.\" 
.\" Carnegie Mellon requests users of this software to return to
.\" 
.\"  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
.\"  School of Computer Science
.\"  Carnegie Mellon University
.\"  Pittsburgh PA 15213-3890
.\" 
.\" any improvements or extensions that they make and grant Carnegie Mellon
.\" the rights to redistribute these changes.
.\" 
.\" 
.\" HISTORY
.\" $Log:	memory_object_data_request.man,v $
.\" Revision 2.4  91/05/14  17:09:11  mrt
.\" 	Correcting copyright
.\" 
.\" Revision 2.3  91/02/14  14:12:50  mrt
.\" 	Changed to new Mach copyright
.\" 	[91/02/12  18:13:16  mrt]
.\" 
.\" Revision 2.2  90/08/07  18:40:06  rpd
.\" 	Created.
.\" 
.TH memory_object_data_request 2 12/22/89
.CM 4
.SH NAME
.nf
memory_object_data_request  \-  request for data from paging object
.SH SYNOPSIS
.nf
.ft B
#include <mach.h>

.nf
.ft B
kern_return_t memory_object_data_request(memory_object,memory_control,
			     offset, length, desired_access)
	memory_object_t	memory_object;
	memory_object_control_t
			memory_control;
	vm_offset_t	offset;
	vm_size_t	length;
	vm_prot_t	desired_access;


.fi
.ft P
.SH DESCRIPTION

.B memory_object_data_request
is a request for data from the specified
memory object, for at least the access specified.  The memory manager is
expected to return at least the specified data, with as much
access as it can allow, using 
.B memory_object_data_provided
.
If the memory manager is unable to provide the data (for example, because
of a hardware error), it may use the 
.B memory_object_data_error
call.
.B memory_object_data_unavailable
call may be used to tell the kernel
to supply zero-filled memory for this region.

.SH ARGUMENTS
.TP 15
.B
.B memory_object
The port that represents the memory object data, as 
supplied to the kernel in a 
.B vm_map
call.
.TP 15
.B
.B memory_control
The request port to which a response is 
requested.  [In the event that a memory object has been supplied 
to more than one the kernel that has made the request.]
.TP 15
.B
.B offset
The offset within a memory object to which this call refers.
This will be page aligned.
.TP 15
.B
.B length
The number of bytes of data, starting at 
.B offset
, to which this
call refers. This will be an integral number of memory object pages.
.TP 15
.B
.B desired_access
A protection value describing the memory access modes which
must be permitted on the specified cached data. One or more of:
.B VM_PROT_READ, VM_PROT_WRITE
or 
.B VM_PROT_EXECUTE
.

.SH DIAGNOSTICS
.TP 25
.B KERN_SUCCESS
Since this routine is called by the kernel, which does not
wait for a reply message, this value is ignored.


.SH SEE ALSO
.B memory_object_data_provided, memory_object_data_error, memory_object_data_unavailable

