/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	cmmu.c,v $
 * Revision 2.3  92/05/21  17:22:43  jfriedl
 * 	Cleanup to quiet gcc warnings.
 * 	Have cmmu_get_idr() and cmmu_get() actually return a value.
 * 	Add prototypes.
 * 	[92/05/18            jfriedl]
 * 
 * Revision 2.2  92/02/18  18:03:23  elf
 * 	Updated includes.
 * 	[92/02/18  16:37:40  danner]
 * 
 * 	Created for luna machine depedent source.
 * 	[92/02/06            danner]
 * 
 */

#include <mach_kdb.h>

#include <mach/machine/vm_types.h>
#include <sys/types.h>
#include <kern/lock.h>
#include <motorola/m88k/m88100/m882xx.h>
#include <motorola/m88k/cpus.h>
#include <machine/board.h> 
#include <machine/cpu_number.h>
#include <machine/pmap.h>

#ifdef WANT_PROTOTYPES
#ifdef luna88k    /*<nosave>*/

 /* luna88k/db_interface.c */
 extern void gimmeabreak(void);

 /* luna88k/locore/eh.c */
 extern boolean_t badwordaddr(void*);

 /* luna88k/locore/locore_asm_routines.c */
 extern void simple_lock(simple_lock_t);
 extern void simple_lock_init(simple_lock_t);
 extern void simple_unlock(simple_lock_t);

 /* luna88k/locore/locore_c_routines.c */
 extern void raw_printf(unsigned char*,...);

 /* luna88k/machdep.c */
 extern unsigned int splhigh(void);
 extern unsigned int splx(unsigned int);
#endif /*luna88k*/   /*<nosave>*/

 /* kern/debug.c */
 extern void panic(char*,...);

 /* kern/printf.c */
 extern void printf(char*,...);

 /* forwards */
 static unsigned int cmmu_get(unsigned int,unsigned int);
 static void cmmu_set(unsigned int,unsigned int,unsigned int);
 unsigned int cmmu_cpu_number(void);
 unsigned int cmmu_get_idr(unsigned int);
 unsigned int cmmu_remote_get(unsigned int,unsigned int,unsigned int);
 void cmmu_flush_cache(caddr_t,int);
 void cmmu_flush_data_cache(caddr_t,int);
 void cmmu_flush_remote_cpu_tlb(unsigned int,boolean_t,int,vm_offset_t);
 void cmmu_flush_tlb(unsigned int,caddr_t,int);
 void cmmu_init(void);
 void cmmu_parity_enable(void);
 void cmmu_remote_set(unsigned int,unsigned int,unsigned int,unsigned int);
 void cmmu_set_batc_entry(unsigned int,unsigned int,unsigned int);
 void cmmu_set_sapr(unsigned int);
 void cmmu_set_uapr(unsigned int);
#endif /* ifdef WANT_PROTOTYPES */


#define MAX_CMMUS	(MAX_CPUS*2)

#define SNOOP_ENABLE

struct cmmu_regs
{
        /* base + $000 */ volatile unsigned idr;
        /* base + $004 */ volatile unsigned scr;
        /* base + $008 */ volatile unsigned ssr;
        /* base + $00C */ volatile unsigned sar;
        /*             */ unsigned padding1[0x3D];
        /* base + $104 */ volatile unsigned sctr;
        /* base + $108 */ volatile unsigned pfSTATUSr;
        /* base + $10C */ volatile unsigned pfADDRr;
        /*             */ unsigned padding2[0x3C];
        /* base + $200 */ volatile unsigned sapr;
        /* base + $204 */ volatile unsigned uapr;
        /*             */ unsigned padding3[0x7E];
        /* base + $400 */ volatile unsigned bwp[8];
        /*             */ unsigned padding4[0xF8];
        /* base + $800 */ volatile unsigned cdp[4];
        /*             */ unsigned padding5[0x0C];
        /* base + $840 */ volatile unsigned ctp[4];
        /*             */ unsigned padding6[0x0C];
        /* base + $880 */ volatile unsigned cssp;

	/* The rest for the 88204 */
	#define cssp0 cssp
        /*             */ unsigned padding7[0x03];
        /* base + $890 */ volatile unsigned cssp1;
        /*             */ unsigned padding8[0x03];
        /* base + $8A0 */ volatile unsigned cssp2;
        /*             */ unsigned padding9[0x03];
        /* base + $8B0 */ volatile unsigned cssp3;
};

struct cmmu {
    struct cmmu_regs *cmmu_regs;	/* CMMU "base" area */
    unsigned char  cmmu_cpu;	/* cpu number it is attached to */
    unsigned char  cmmu_data;	/* flag for data/instruction cache */
    unsigned char  cmmu_alive;	/* alive or not */
} cmmu[MAX_CMMUS] = {
    {(void *)CMMU_I0, 0, 0, 0},
    {(void *)CMMU_D0, 0, 1, 0},
    {(void *)CMMU_I1, 1, 0, 0},
    {(void *)CMMU_D1, 1, 1, 0},
    {(void *)CMMU_I2, 2, 0, 0},
    {(void *)CMMU_D2, 2, 1, 0},
    {(void *)CMMU_I3, 3, 0, 0},
    {(void *)CMMU_D3, 3, 1, 0}};

struct cpu_cmmu {
    struct cmmu *inst_cmmu;
    struct cmmu *data_cmmu;
} cpu_cmmu[MAX_CPUS];

/* 
  This lock protects the cmmu SAR and SCR's; other ports 
  can be accessed without locking it 
*/

simple_lock_data_t cmmu_cpu_lock;

#define CMMU_LOCK     simple_lock(&cmmu_cpu_lock)
#define CMMU_UNLOCK simple_unlock(&cmmu_cpu_lock)

int	cpu_sets[MAX_CPUS];
int	ncpus = 0;

/*
 * Figure the initial setting of the SCTR (system control register)
 */
#ifdef SNOOP_ENABLE
# define SCTR_SNOOP 				CMMU_SCTR_SE
#else
# define SCTR_SNOOP 				0
#endif
#ifdef PRIORITY_ARBITRATION
# define SCTR_PA 				CMMU_SCTR_PR
#else
# define SCTR_PA 				0
#endif
#if 1 /* seems no need to have this be a variable */
 int cmmu_sctr = 	((SCTR_SNOOP)|(SCTR_PA)) ;
#else
# define cmmu_sctr	((SCTR_SNOOP)|(SCTR_PA))
#endif

#if !defined(MACH_KERNEL)    /* use with 2.5 */
# define NBPW    sizeof(int) /* bytes per word */
# define NBSG    4*1024*1024 /* segment size */
# define NBPG    4*1024      /* bytes per mmu page */
#endif

/*
 * CMMU initialization routine
 */
void cmmu_init()
{
    register unsigned cpu, s;
    union cpupid id;

    for (cpu = 0; cpu < MAX_CPUS; cpu++) {
	cpu_cmmu[cpu].inst_cmmu = 0;
	cpu_cmmu[cpu].data_cmmu = 0;
    }
    simple_lock_init(&cmmu_cpu_lock);

    for (cpu = 0; cpu < MAX_CMMUS; cpu++) {
	if (!badwordaddr(cmmu[cpu].cmmu_regs)) {
	    id.cpupid = cmmu[cpu].cmmu_regs->idr;
	    if (id.m88200.type != M88200 && id.m88200.type != M88204)
		continue;
	    cmmu[cpu].cmmu_alive = 1;
	    if (cmmu[cpu].cmmu_data)
		cpu_cmmu[cmmu[cpu].cmmu_cpu].data_cmmu = &cmmu[cpu];
	    else
		cpu_cmmu[cmmu[cpu].cmmu_cpu].inst_cmmu = &cmmu[cpu];

	    /*
	     * Reset cache data....
	     * as per M88200 Manual (2nd Ed.) section 3.11.
	     */
	    for (s = 0; s < 255; s++) {
		cmmu[cpu].cmmu_regs->sar = s << 4;
		cmmu[cpu].cmmu_regs->cssp = 0x3f0ff000; /* reset Set State */
	    }

	    /* 88204 has additional cache to clear */
	    if(id.m88200.type == M88204)
	    {
#	    if 0
		/*
		 * This is better, but I'd better check the manual to make
		 * sure it's kosher.
		 */
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp1 = 0x3f0ff000;
		    cmmu[cpu].cmmu_regs->cssp2 = 0x3f0ff000;
		    cmmu[cpu].cmmu_regs->cssp3 = 0x3f0ff000;
		}
#	    else
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp1 = 0x3f0ff000;
		}
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp2 = 0x3f0ff000;
		}
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp3 = 0x3f0ff000;
		}
#	    endif
	    }

	    cmmu[cpu].cmmu_regs->sctr = cmmu_sctr;

	    /*
	     * Turn on global, writethrough bits, turn off cache inhibit 
	     * for SAPR, UAPR to get coherency, write-through, until
	     * we come up virtual
	     */
	    cmmu[cpu].cmmu_regs->scr = CMMU_FLUSH_CACHE_INV_ALL;
	    cmmu[cpu].cmmu_regs->sapr |= AREA_D_G | AREA_D_WT;  
	}
    }

    /*
     * Now that we know which CMMUs are there, let's report on which
     * CPU/CMMU sets seem complete (hopefully all)
     */
    for (cpu = 0; cpu < MAX_CPUS; cpu++)
    {
	if (cpu_cmmu[cpu].inst_cmmu && cpu_cmmu[cpu].data_cmmu)
	{
	    cpu_sets[cpu] = 1;	/* this cpu installed */
	    ncpus++;		/* which makes one more installed */
	    id.cpupid = cmmu[cpu*(MAX_CMMUS/MAX_CPUS)].cmmu_regs->idr;
	    if(id.m88200.type == M88204)
		printf("CPU%d is attached with MC88204 CMMU\n", cpu);
	    else
		printf("CPU%d is attached with MC88200 CMMU\n", cpu);
	}
	else if (cpu_cmmu[cpu].inst_cmmu)
	{
		printf("CPU%d data CMMU is not working.\n", cpu);
		panic("cmmu-data");
	}
	else if (cpu_cmmu[cpu].data_cmmu)
	{
	    printf("CPU%d instruction CMMU is not working.\n", cpu);
	    panic("cmmu");
	}
	else
	{
	    /* no CMMU there? CPU must not be there at all... */
	    cpu_sets[cpu] = 0;
	}
    }
}

/*
 * enable parity
 */
void cmmu_parity_enable()
{
    register unsigned cpu;

    for (cpu = 0; cpu < MAX_CMMUS; cpu++) {
	if (cmmu[cpu].cmmu_alive) {
	    cmmu[cpu].cmmu_regs->sctr = cmmu_sctr
#ifdef	PARITY_ENABLE
		 |CMMU_SCTR_PE
#endif	PARITY_ENABLE
		;
	}
    }
}

/*
 * Find out the CPU number from accessing CMMU
 * better be at splhigh, or even better, with interrupts
 * disabled.
 */

unsigned cmmu_cpu_number()
{
  register unsigned cmmu_no;
  
  CMMU_LOCK;
  
  /* clear CMMU system address registers */
  for (cmmu_no = 0; cmmu_no < MAX_CMMUS; cmmu_no++) 
    {
      if (cmmu[cmmu_no].cmmu_alive > 1)
	continue;
      if (cmmu[cmmu_no].cmmu_alive && cmmu[cmmu_no].cmmu_data) 
	{
	  cmmu[cmmu_no].cmmu_regs->pfADDRr = 0;
	  cmmu[cmmu_no].cmmu_regs->pfSTATUSr = 0;
	}
    }
  
  /* access illegal address to create a fault */
  badwordaddr((void*)ILLADDRESS);
  
  /* check which CMMU reporting the fault */
  for (cmmu_no = 0; cmmu_no < MAX_CMMUS; cmmu_no++) 
    {
      if (cmmu[cmmu_no].cmmu_alive > 1)	
	continue;
      if (cmmu[cmmu_no].cmmu_alive && cmmu[cmmu_no].cmmu_data) 
	{
	  if (cmmu[cmmu_no].cmmu_regs->pfADDRr ||
	      (cmmu[cmmu_no].cmmu_regs->pfSTATUSr & 0x70000)) 
	    {
	      if (cmmu[cmmu_no].cmmu_regs->pfADDRr != ILLADDRESS)
		break;
	      cmmu[cmmu_no].cmmu_regs->pfADDRr = 0;
	      cmmu[cmmu_no].cmmu_regs->pfSTATUSr = 0;
	      cmmu[cmmu_no].cmmu_alive = 2;
	      CMMU_UNLOCK;
raw_printf("cmmu_cpu returns %d\n", cmmu[cmmu_no].cmmu_cpu);
	      return cmmu[cmmu_no].cmmu_cpu;
	    }
	}
    }
  
    gimmeabreak();
    panic("could not determine my cpu number");
}

/* 
 *  The caller of this should hold the cmmu_cpu_lock
 *  if modify the CMMU_SCR or CMMU_SAR.
 */

#if !MACH_KDB
static
#endif
void cmmu_remote_set(cpu, r, data, x)
unsigned cpu, r, data, x;
{
    struct cmmu *c = data ? cpu_cmmu[cpu].data_cmmu : cpu_cmmu[cpu].inst_cmmu;
    *(volatile unsigned *)((char*)c->cmmu_regs + r) = x;
}

static void cmmu_set(r, data, x)
unsigned r, data, x;
{
    cmmu_remote_set(cpu_number(), r, data, x);
}

/*
  cmmu_cpu_lock should be held when called if read
  the CMMU_SCR or CMMU_SAR.
*/

#if !MACH_KDB
static
#endif
unsigned cmmu_remote_get(cpu, r, data)
unsigned cpu, r, data;
{
    struct cmmu *c = data ? cpu_cmmu[cpu].data_cmmu : cpu_cmmu[cpu].inst_cmmu;
    return *(volatile unsigned *)((char*)c->cmmu_regs + r);
}

static unsigned cmmu_get(r, data)
unsigned r, data;
{
    return cmmu_remote_get(cpu_number(), r, data);
}

/* Needs no locking - read only registers */
unsigned cmmu_get_idr(data)
unsigned data;
{
    return cmmu_get(CMMU_IDR, data);
}

void cmmu_set_sapr(ap)
unsigned ap;
{
    extern int	cache_policy;

    if (cache_policy & CACHE_INHIBIT) {
	cmmu_set(CMMU_SAPR, 0, ap | AREA_D_CI);
	cmmu_set(CMMU_SAPR, 1, ap | AREA_D_CI);
    } else {
	cmmu_set(CMMU_SAPR, 0, ap);
	cmmu_set(CMMU_SAPR, 1, ap);
    }
}

void cmmu_set_uapr(ap)
unsigned ap;
{
    cmmu_set(CMMU_UAPR, 0, ap);
    cmmu_set(CMMU_UAPR, 1, ap);
}

/* set batc entry number entry_no to value in 
   the data or instruction cache depending on data */

void cmmu_set_batc_entry(entry_no, data, value)
     unsigned entry_no;
     unsigned data;   /* 1 = data, 0 = instruction */
     unsigned value;  /* the value to stuff into the batc */
{
    cmmu_set(CMMU_BWP0 + entry_no*sizeof(int), data, value);
}

/*
 *	flush my personal tlb TLB
 */
void cmmu_flush_tlb(kernel, vaddr, size)
unsigned kernel;
register caddr_t vaddr;
register int size;
{
    register s = splhigh();

    CMMU_LOCK;
    if (vaddr <= 0 || size > NBPG) 
      {
	cmmu_set(CMMU_SCR, 0,
		 kernel?CMMU_FLUSH_SUPER_ALL:CMMU_FLUSH_USER_ALL);
	cmmu_set(CMMU_SCR, 1,
		 kernel?CMMU_FLUSH_SUPER_ALL:CMMU_FLUSH_USER_ALL);
      }
    else /* a page or smaller */
      {
	cmmu_set(CMMU_SAR, 0, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 0,
		 kernel?CMMU_FLUSH_SUPER_PAGE:CMMU_FLUSH_USER_PAGE);
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,
		 kernel?CMMU_FLUSH_SUPER_PAGE:CMMU_FLUSH_USER_PAGE);
    }
    CMMU_UNLOCK;
    splx(s);
}

/*
 *	flush instruction cache
 */
void cmmu_flush_cache(vaddr, size)
register caddr_t vaddr;
register int size;
{
    register s = splhigh();

    CMMU_LOCK;

    if (size < 0 || size > NBSG) {
	cmmu_set(CMMU_SCR, 0,CMMU_FLUSH_CACHE_CBI_ALL);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_ALL);
    }
    else if (size <= 16) {
	cmmu_set(CMMU_SAR, 0, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 0,CMMU_FLUSH_CACHE_CBI_LINE);
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_LINE);
    }
    else if (size <= NBPG) {
	cmmu_set(CMMU_SAR, 0, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 0,CMMU_FLUSH_CACHE_CBI_PAGE);
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_PAGE);
    }
    else {
	cmmu_set(CMMU_SAR, 0, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 0,CMMU_FLUSH_CACHE_CBI_SEGMENT);
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_SEGMENT);
    }

    CMMU_UNLOCK;

    splx(s);
}

/* flush data cache */ 
void cmmu_flush_data_cache(vaddr, size)
register caddr_t vaddr; 
register int size; 
{ 
    register s = splhigh();

    CMMU_LOCK;

    if (size < 0 || size > NBSG) {
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_ALL);
    }
    else if (size <= 16) {
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_LINE);
    }
    else if (size <= NBPG) {
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_PAGE);
    }
    else {
	cmmu_set(CMMU_SAR, 1, (unsigned)vaddr);
	cmmu_set(CMMU_SCR, 1,CMMU_FLUSH_CACHE_CBI_SEGMENT);
    }
    
    CMMU_UNLOCK;
    splx(s);
}

/*
 *
 * TLB shootdown on remote cpu
 *
 */

void cmmu_flush_remote_cpu_tlb(cpu, kernel, addr, size)
     unsigned cpu;
     boolean_t kernel;
     int addr;	/* was vm_offset_t, but that's unsigned */
     vm_offset_t size;
{
  int s = splhigh();
  
  /* check that its a valid cpu */
  if (cpu >= NCPUS)
    panic("out of range cpu number passed to tlb shootdown\n");

  CMMU_LOCK; /* acquire the system lock */

  if (addr <= 0 || size > NBPG)
    {
      int code = kernel ? CMMU_FLUSH_SUPER_ALL : CMMU_FLUSH_USER_ALL;

      cpu_cmmu[cpu].data_cmmu->cmmu_regs->scr = code;
      cpu_cmmu[cpu].inst_cmmu->cmmu_regs->scr = code;
    }
  else
    { /* shooting down a page */
      int code = kernel ? CMMU_FLUSH_SUPER_PAGE : CMMU_FLUSH_USER_PAGE;
            
      /* set up addresses */
      cpu_cmmu[cpu].data_cmmu->cmmu_regs->sar = addr;
      cpu_cmmu[cpu].data_cmmu->cmmu_regs->scr = code;

      cpu_cmmu[cpu].inst_cmmu->cmmu_regs->sar = addr;
      cpu_cmmu[cpu].inst_cmmu->cmmu_regs->scr = code;
    }

  CMMU_UNLOCK;
  splx(s);
}
