/*
 * Mach Operating System
 * Copyright (c) 1992, 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	cmmu.c,v $
 * Revision 2.4  92/08/03  17:51:47  jfriedl
 * 	Cmmu_cpu_num fixes for 3.0 [jfriedl]
 * 	Picked up new version from H R. Extensive changes.[danner]
 * 
 *
 * Revision 2.2  92/02/18  18:03:23  elf
 * 	Updated includes.
 * 	[92/02/18  16:37:40  danner]
 * 
 * 	Created for luna machine depedent source.
 * 	[92/02/06            danner]
 * 
 */

#include <mach_kdb.h>

#include <mach/machine/vm_types.h>
#include <sys/types.h>
#include <kern/lock.h>
#include <motorola/m88k/m88100/m882xx.h>
#include <motorola/m88k/cpus.h>
#include <machine/board.h> 
#include <machine/cpu_number.h>
#include <machine/pmap.h>

#define MAX_CMMUS	(MAX_CPUS*2)

#define SNOOP_ENABLE

struct cmmu_regs
{
        /* base + $000 */ volatile unsigned idr;
        /* base + $004 */ volatile unsigned scr;
        /* base + $008 */ volatile unsigned ssr;
        /* base + $00C */ volatile unsigned sar;
        /*             */ unsigned padding1[0x3D];
        /* base + $104 */ volatile unsigned sctr;
        /* base + $108 */ volatile unsigned pfSTATUSr;
        /* base + $10C */ volatile unsigned pfADDRr;
        /*             */ unsigned padding2[0x3C];
        /* base + $200 */ volatile unsigned sapr;
        /* base + $204 */ volatile unsigned uapr;
        /*             */ unsigned padding3[0x7E];
        /* base + $400 */ volatile unsigned bwp[8];
        /*             */ unsigned padding4[0xF8];
        /* base + $800 */ volatile unsigned cdp[4];
        /*             */ unsigned padding5[0x0C];
        /* base + $840 */ volatile unsigned ctp[4];
        /*             */ unsigned padding6[0x0C];
        /* base + $880 */ volatile unsigned cssp;

	/* The rest for the 88204 */
	#define cssp0 cssp
        /*             */ unsigned padding7[0x03];
        /* base + $890 */ volatile unsigned cssp1;
        /*             */ unsigned padding8[0x03];
        /* base + $8A0 */ volatile unsigned cssp2;
        /*             */ unsigned padding9[0x03];
        /* base + $8B0 */ volatile unsigned cssp3;
};

struct cmmu {
    struct cmmu_regs *cmmu_regs;	/* CMMU "base" area */
    unsigned char  cmmu_cpu;	/* cpu number it is attached to */
    unsigned char  cmmu_data;	/* flag for data/instruction cache */
    unsigned char  cmmu_alive;
	#define CMMU_DEAD	0	/* This cmmu not there */
	#define CMMU_AVAILABLE	1	/* It's there, but which cpu's? */
	#define CMMU_MARRIED	2	/* Know which cpu it belongs to. */
} cmmu[MAX_CMMUS] = {
    {(void *)CMMU_I0, 0, 0, 0},
    {(void *)CMMU_D0, 0, 1, 0},
    {(void *)CMMU_I1, 1, 0, 0},
    {(void *)CMMU_D1, 1, 1, 0},
    {(void *)CMMU_I2, 2, 0, 0},
    {(void *)CMMU_D2, 2, 1, 0},
    {(void *)CMMU_I3, 3, 0, 0},
    {(void *)CMMU_D3, 3, 1, 0}};

struct cpu_cmmu {
    struct cmmu *inst_cmmu;
    struct cmmu *data_cmmu;
} cpu_cmmu[MAX_CPUS];

/* 
  This lock protects the cmmu SAR and SCR's; other ports 
  can be accessed without locking it 
*/

simple_lock_data_t cmmu_cpu_lock;

#define CMMU_LOCK simple_lock(&cmmu_cpu_lock)
#define CMMU_UNLOCK simple_unlock(&cmmu_cpu_lock)

int	cpu_sets[MAX_CPUS];
int	ncpus = 0;
int	cache_policy = 0;


#if !defined(MACH_KERNEL)    /* use with 2.5 */
# define NBPW    sizeof(int) /* bytes per word */
# define NBSG    4*1024*1024 /* segment size */
# define NBPG    4*1024      /* bytes per mmu page */
#endif

/*
 * CMMU initialization routine
 */
void cmmu_init()
{
    register unsigned cpu, s;
    union cpupid id;

    for (cpu = 0; cpu < MAX_CPUS; cpu++) {
	cpu_cmmu[cpu].inst_cmmu = 0;
	cpu_cmmu[cpu].data_cmmu = 0;
    }
    simple_lock_init(&cmmu_cpu_lock);

    for (cpu = 0; cpu < MAX_CMMUS; cpu++) {
#ifdef notdef
	if (!badaddr((caddr_t)cmmu[cpu].cmmu_regs, sizeof(int))) {
#else
	if (!wprobe((caddr_t)cmmu[cpu].cmmu_regs, -1)) {
#endif
	    id.cpupid = cmmu[cpu].cmmu_regs->idr;
	    if (id.m88200.type != M88200 && id.m88200.type != M88204)
		continue;
	    cmmu[cpu].cmmu_alive = CMMU_AVAILABLE;
	    if (cmmu[cpu].cmmu_data)
		cpu_cmmu[cmmu[cpu].cmmu_cpu].data_cmmu = &cmmu[cpu];
	    else
		cpu_cmmu[cmmu[cpu].cmmu_cpu].inst_cmmu = &cmmu[cpu];

	    /*
	     * Reset cache data....
	     * as per M88200 Manual (2nd Ed.) section 3.11.
	     */
	    for (s = 0; s < 255; s++) {
		cmmu[cpu].cmmu_regs->sar = s << 4;
		cmmu[cpu].cmmu_regs->cssp = 0x3f0ff000; /* reset Set State */
	    }

	    /* 88204 has additional cache to clear */
	    if(id.m88200.type == M88204)
	    {
#	    if 0
		/*
		 * This is better, but I'd better check the manual to make
		 * sure it's kosher.
		 */
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp1 = 0x3f0ff000;
		    cmmu[cpu].cmmu_regs->cssp2 = 0x3f0ff000;
		    cmmu[cpu].cmmu_regs->cssp3 = 0x3f0ff000;
		}
#	    else
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp1 = 0x3f0ff000;
		}
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp2 = 0x3f0ff000;
		}
		for (s = 0; s < 255; s++) {
		    cmmu[cpu].cmmu_regs->sar = s<<4;
		    cmmu[cpu].cmmu_regs->cssp3 = 0x3f0ff000;
		}
#	    endif
	    }

	    /*
	     * Only Data CMMUs get snoop enabled.  We don't enable snoop
	     * on Instruction CMMUs to minimize M-bus snoop traffic and
	     * internal contention.  The Instruction CMMUs get invalidated
	     * when a page is removed from a vm_map.
	     */

#if 0
#ifdef SNOOP_ENABLE
	    if (cmmu[cpu].cmmu_data)
		cmmu[cpu].cmmu_regs->sctr |= CMMU_SCTR_SE;
	    else
#endif
		cmmu[cpu].cmmu_regs->sctr = 0;
#endif
	    cmmu[cpu].cmmu_regs->sctr |= CMMU_SCTR_SE;

	    /*
	     * Enable some caching until virtual mode tells us different.
      	     * Data caches get writethrough, snooping (global) mode.
	     * The instruction cache is also mapped wt/snooping so 
	     * the kernel debugger can modify code. 
	     *
	     */
	    cmmu[cpu].cmmu_regs->scr = CMMU_FLUSH_CACHE_INV_ALL;
	    /* one thinks one could just use an assignment. But
	       on the luna, this causes problems. Why ?
	    */
	    cmmu[cpu].cmmu_regs->sapr |= AREA_D_G | AREA_D_WT;  
	}
    }

    /*
     * Now that we know which CMMUs are there, let's report on which
     * CPU/CMMU sets seem complete (hopefully all)
     */
    for (cpu = 0; cpu < MAX_CPUS; cpu++)
    {
	if (cpu_cmmu[cpu].inst_cmmu && cpu_cmmu[cpu].data_cmmu)
	{
	    cpu_sets[cpu] = 1;	/* this cpu installed */
	    ncpus++;		/* which makes one more installed */
	    id.cpupid = cmmu[cpu*(MAX_CMMUS/MAX_CPUS)].cmmu_regs->idr;

	    if(id.m88200.type == M88204)
		printf("CPU%d is attached with MC88204 CMMU\n", cpu);
	    else
		printf("CPU%d is attached with MC88200 CMMU\n", cpu);

	}
	else if (cpu_cmmu[cpu].inst_cmmu)
	{
		printf("CPU%d data CMMU is not working.\n", cpu);
		panic("cmmu-data");
	}
	else if (cpu_cmmu[cpu].data_cmmu)
	{
	    printf("CPU%d instruction CMMU is not working.\n", cpu);
	    panic("cmmu");
	}
	else
	{
	    /* no CMMU there? CPU must not be there at all... */
	    cpu_sets[cpu] = 0;
	}
    }
}

/*
 * enable parity
 */
void
cmmu_parity_enable()
{
#ifdef	PARITY_ENABLE
    register int cpu;

    for (cpu = 0; cpu < MAX_CMMUS; cpu++) {
	if (cmmu[cpu].cmmu_alive != CMMU_DEAD) {
	    cmmu[cpu].cmmu_regs->sctr |= CMMU_SCTR_PE;
	}
    }
#endif	PARITY_ENABLE
}

/*
 * Find out the CPU number from accessing CMMU
 * better be at splhigh, or even better, with interrupts
 * disabled.
 */

unsigned
cmmu_cpu_number()
{
    register unsigned cmmu_no;
    int i;

    CMMU_LOCK;
    for (i=0; i < 10; i++)
    {
	/* clear CMMU p-bus status registers */
	for (cmmu_no = 0; cmmu_no < MAX_CMMUS; cmmu_no++) 
	{
	    if (cmmu[cmmu_no].cmmu_alive == CMMU_AVAILABLE &&
	        cmmu[cmmu_no].cmmu_data) 
		    cmmu[cmmu_no].cmmu_regs->pfSTATUSr = 0;
	}

	/* access faulting address */
	badwordaddr((void *)ILLADDRESS);

	/* check which CMMU reporting the fault  */
	for (cmmu_no = 0; cmmu_no < MAX_CMMUS; cmmu_no++) 
	{
	    if (cmmu[cmmu_no].cmmu_alive == CMMU_AVAILABLE &&
	        cmmu[cmmu_no].cmmu_data  &&
		cmmu[cmmu_no].cmmu_regs->pfSTATUSr & 0x70000)
	    {
		if (cmmu[cmmu_no].cmmu_regs->pfSTATUSr & 0x70000)
		{
		    cmmu[cmmu_no].cmmu_regs->pfSTATUSr = 0; /* to be clean */
		    cmmu[cmmu_no].cmmu_alive = CMMU_MARRIED;
		    CMMU_UNLOCK;
		    return cmmu[cmmu_no].cmmu_cpu;
		}
	    }
	}
    }

    panic("could not determine my cpu number");
}

/**
 **	Funcitons that actually modify CMMU registers.
 **/
/* 
 *  The caller of this should hold the cmmu_cpu_lock
 *  if modify the CMMU_SCR or CMMU_SAR.
 */

#if !MACH_KDB
static
#endif
void
cmmu_remote_set(cpu, r, data, x)
unsigned cpu, r, data, x;
{

    struct cmmu *c = data ? cpu_cmmu[cpu].data_cmmu : cpu_cmmu[cpu].inst_cmmu;

#if MACH_KDB
    if (!cpu_sets[cpu])
      panic("cmmu_remote_set called on nonexistent cpu\n");
#endif
    *(volatile unsigned *)((char*)c->cmmu_regs + r) = x;
}

static void cmmu_set(r, data, x)
unsigned r, data, x;
{
  cmmu_remote_set(cpu_number(), r, data, x);
}

/*
  cmmu_cpu_lock should be held when called if read
  the CMMU_SCR or CMMU_SAR.
*/

#if !MACH_KDB
static
#endif
unsigned
cmmu_remote_get(cpu, r, data)
unsigned cpu, r, data;
{
    struct cmmu *c = data ? cpu_cmmu[cpu].data_cmmu : cpu_cmmu[cpu].inst_cmmu;

#if MACH_KDB
    if (!cpu_sets[cpu])
      panic("cmmu_remote_get called on nonexistent cpu\n");
#endif

    return *(volatile unsigned *)((char*)c->cmmu_regs + r);
}

static void cmmu_get(r, data)
unsigned r, data;
{
    cmmu_remote_get(cpu_number(), r, data);
}

/* Needs no locking - read only registers */
void cmmu_get_idr(data)
unsigned data;
{
  cmmu_get(CMMU_IDR, data);
}

void
cmmu_set_sapr(ap)
unsigned ap;
{
	cmmu_remote_set_sapr(cpu_number(), ap);
}

void
cmmu_remote_set_sapr(cpu, ap)
unsigned cpu;
unsigned ap;
{
    extern int	cache_policy;

#ifdef notdef
    if (cache_policy & CACHE_INHIBIT) {
#else
    if (cache_policy & CACHE_INH) {
#endif
	ap |= AREA_D_CI;
    }
    cmmu_remote_set(cpu, CMMU_SAPR, INST_CMMU, ap);
    cmmu_remote_set(cpu, CMMU_SAPR, DATA_CMMU, ap);
}

void
cmmu_set_uapr(ap)
unsigned ap;
{
    cmmu_set(CMMU_UAPR, INST_CMMU, ap);
    cmmu_set(CMMU_UAPR, DATA_CMMU, ap);
}

/* set batc entry number entry_no to value in 
   the data or instruction cache depending on data */

void
cmmu_set_batc_entry(cpu, entry_no, data, value)
     unsigned cpu;
     unsigned entry_no;
     unsigned data;   /* 1 = data, 0 = instruction */
     unsigned value;  /* the value to stuff into the batc */
{
  cmmu_remote_set(cpu, CMMU_BWP0 + entry_no*sizeof(int), data, value);
}

/**
 **	Functions that invalidate TLB entries.
 **/
/*
 *	flush my personal tlb
 */
void
cmmu_flush_tlb(kernel, vaddr, size)
unsigned kernel;
register caddr_t vaddr;
register int size;
{
	cmmu_flush_remote_tlb(cpu_number(), kernel, vaddr, size);
}

/*
 *	flush any tlb
 */
void
cmmu_flush_remote_tlb(cpu, kernel, vaddr, size)
unsigned cpu;
unsigned kernel;
register caddr_t vaddr;
register unsigned size;
{
    register s = splhigh();

    CMMU_LOCK;
    if (size > M88K_PGBYTES) 
      {
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,
		 kernel?CMMU_FLUSH_SUPER_ALL:CMMU_FLUSH_USER_ALL);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,
		 kernel?CMMU_FLUSH_SUPER_ALL:CMMU_FLUSH_USER_ALL);
      }
    else /* a page or smaller */
      {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, vaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,
		 kernel?CMMU_FLUSH_SUPER_PAGE:CMMU_FLUSH_USER_PAGE);
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, vaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,
		 kernel?CMMU_FLUSH_SUPER_PAGE:CMMU_FLUSH_USER_PAGE);
    }
    CMMU_UNLOCK;
    splx(s);
}

/**
 **	Functions that invalidate caches.
 **
 ** Cache invalidates require physical addresses.  Care must be exercised when
 ** using segment invalidates.  This implies that the starting physical address
 ** plus the segment length should be invalidated.  A typical mistake is to
 ** extract the first physical page of a segment from a virtual address, and
 ** then expecting to invalidate when the pages are not physically contiguous.
 **
 ** We don't push Instruction Caches prior to invalidate because they are not
 ** snooped and never modified (I guess it doesn't matter then which form
 ** of the command we use then).
 **/
/*
 *	flush both Instruction and Data caches
 */
void
cmmu_flush_cache(physaddr, size)
register caddr_t physaddr;
register int size;
{
	cmmu_flush_remote_cache(cpu_number(), physaddr, size);
}

/*
 *	flush both Instruction and Data caches
 */
void
cmmu_flush_remote_cache(cpu, physaddr, size)
register int	cpu;
register caddr_t physaddr;
register int size;
{
    register s = splhigh();

    CMMU_LOCK;

    if (size < 0 || size > NBSG ) {
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU, CMMU_FLUSH_CACHE_CBI_ALL);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU, CMMU_FLUSH_CACHE_CBI_ALL);
    }
    else if (size <= 16) {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,CMMU_FLUSH_CACHE_CBI_LINE);
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_LINE);
    }
    else if (size <= NBPG) {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,CMMU_FLUSH_CACHE_CBI_PAGE);
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_PAGE);
    }
    else {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,CMMU_FLUSH_CACHE_CBI_SEGMENT);
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_SEGMENT);
    }

    CMMU_UNLOCK;

    splx(s);
}

/*
 *	flush Instruction caches
 */
void
cmmu_flush_inst_cache(physaddr, size)
register caddr_t physaddr;
register int size;
{
	cmmu_flush_remote_inst_cache(cpu_number(), physaddr, size);
}

/*
 *	flush Instruction caches
 */
void
cmmu_flush_remote_inst_cache(cpu, physaddr, size)
register int	cpu;
register caddr_t physaddr;
register int size;
{
    register s = splhigh();

    CMMU_LOCK;

    if (size < 0 || size > NBSG) {
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU, CMMU_FLUSH_CACHE_INV_ALL);
    }
    else if (size <= 16) {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,CMMU_FLUSH_CACHE_INV_LINE);
    }
    else if (size <= NBPG) {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,CMMU_FLUSH_CACHE_INV_PAGE);
    }
    else {
	cmmu_remote_set(cpu, CMMU_SAR, INST_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, INST_CMMU,CMMU_FLUSH_CACHE_INV_SEGMENT);
    }

    CMMU_UNLOCK;

    splx(s);
}


/*
 * flush data cache
 */ 
void
cmmu_flush_data_cache(physaddr, size)
register caddr_t physaddr; 
register int size; 
{ 
	cmmu_flush_remote_data_cache(cpu_number(), physaddr, size);
}

void
cmmu_flush_remote_data_cache(cpu, physaddr, size)
register int	cpu;
register caddr_t physaddr; 
register int size; 
{ 
    register s = splhigh();

    CMMU_LOCK;

    if (size < 0 || size > NBSG) {
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_ALL);
    }
    else if (size <= 16) {
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_LINE);
    }
    else if (size <= NBPG) {
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_PAGE);
    }
    else {
	cmmu_remote_set(cpu, CMMU_SAR, DATA_CMMU, physaddr);
	cmmu_remote_set(cpu, CMMU_SCR, DATA_CMMU,CMMU_FLUSH_CACHE_CBI_SEGMENT);
    }
    
    CMMU_UNLOCK;
    splx(s);
}
