/*
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/* 
 * Motorola 88100 pcb definitions 
 *
 */
/*
 * HISTORY
 * $Log:	pcb.h,v $
 * Revision 2.3  92/03/03  14:22:28  rpd
 * 	Added dummy definition of syscall_emulation_sync.
 * 	[92/03/03            rpd]
 * 
 * Revision 2.2  92/02/18  18:03:49  elf
 * 	Disabled KEEP_STACKS.
 * 	[92/01/30            danner]
 * 	Created.
 * 	[92/01/20            danner]
 * 
 */
#ifndef _M88100_THREAD
#define _M88100_THREAD

#include <mach/kern_return.h>  

/* This structure is used to save the kernel state during a save_context
   or other handoff. It was the original pcb in the impure kernel */

/* This must always be an even number of words long */

struct m88100_pcb {
	unsigned int	pcb_pc;		/* address to return */
	unsigned int	pcb_r14;
	unsigned int	pcb_r15;
	unsigned int	pcb_r16;
	unsigned int	pcb_r17;
	unsigned int	pcb_r18;
	unsigned int	pcb_r19;
	unsigned int	pcb_r20;
	unsigned int	pcb_r21;
	unsigned int	pcb_r22;
	unsigned int	pcb_r23;
	unsigned int	pcb_r24;
	unsigned int	pcb_r25;
	unsigned int	pcb_r26;
	unsigned int	pcb_r27;
	unsigned int	pcb_r28;
	unsigned int	pcb_r29;
	unsigned int	pcb_r30;
	unsigned int	pcb_sp;		/* kernel stack pointer */
	unsigned int	pcb_mask;	/* to use in save_context() */
};


/*
 *	m88100_saved_state	this structure corresponds to the state
 *				of the user registers as saved on the
 *				stack upon kernel entry.  This structure
 *				is used internally only.  Since this
 *				structure may change from version to
 *				version, it is hidden from the user.
 */

/* This must always be an even number of words long */

struct m88100_saved_state {
    unsigned int r[32];
    unsigned int fpsr;
    unsigned int fpcr;
    unsigned int epsr;
    unsigned int sxip;
    unsigned int snip;
    unsigned int sfip;
    unsigned int ssbr;
    unsigned int dmt0;
    unsigned int dmd0;
    unsigned int dma0;
    unsigned int dmt1;
    unsigned int dmd1;
    unsigned int dma1;
    unsigned int dmt2;
    unsigned int dmd2;
    unsigned int dma2;
    unsigned int fpecr;
    unsigned int fphs1;
    unsigned int fpls1;
    unsigned int fphs2;
    unsigned int fpls2;
    unsigned int fppt;
    unsigned int fprh;
    unsigned int fprl;
    unsigned int fpit;
    unsigned int vector;   /* exception vector number */
    unsigned int mask;	   /* interrupt mask level */
    unsigned int mode;     /* interrupt mode */
    unsigned int scratch1; /* used by locore trap handling code */
    unsigned int pad;      /* to make an even length */
} ;

struct pcb
{
  struct m88100_pcb            kernel_state;
  struct m88100_saved_state    user_state;
};

typedef	struct pcb	*pcb_t;		/* exported */

				/* user's stack pointer is in the PCB,
				   or in the processor register if
				   the thread is active.
				 */

/*
 *	Location of saved user registers for thread.
 */
#define	USER_REGS(thread) \
	(((struct m88100_saved_state *)  (&((thread)->pcb->user_state))))


#define syscall_emulation_sync(task)	/* do nothing */

#endif  _M88100_THREAD
