#!/bin/sh
# 
# Mach Operating System
# Copyright (c) 1992 Carnegie Mellon University
# All Rights Reserved.
# 
# Permission to use, copy, modify and distribute this software and its
# documentation is hereby granted, provided that both the copyright
# notice and this permission notice appear in all copies of the
# software, derivative works or modified versions, and any portions
# thereof, and that both notices appear in supporting documentation.
# 
# CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
# CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
# ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
# 
# Carnegie Mellon requests users of this software to return to
# 
#  Software Distribution Coordinator   or   Software.Distribution@CS.CMU.EDU
#  School of Computer Science
#  Carnegie Mellon University
#  Pittsburgh PA 15213-3890
# 
# any improvements or extensions that they made and grant Carnegie Mellon the
# rights to redistribute these changes.
#
#
# HISTORY
# $Log:	m3build,v $
# Revision 2.4  92/04/14  01:35:32  mrt
# 	Added a help option.
# 	[92/04/14            mrt]
# 
# Revision 2.3  92/03/26  09:11:29  rpd
# 	Changed bootstrap to just bootstrap without forcing mk, ux, and user.
# 	[92/03/26            rpd]
# 
# 	Changed /afs/cs.cmu.edu/project/mach/mach3
# 	to /afs/cs.cmu.edu/project/mach3.
# 	[92/03/26            rpd]
# 
# Revision 2.2  92/03/08  14:41:20  rpd
# 	Created.
# 	[92/03/08            rpd]
# 

# ABSTRACT
#	run m3build in the top-level src directory.
#	This script builds and installs whatever parts of
#	the system are specified. It is most useful for bootstapping 
#	a system, where it ensures that the build-tools and libraries
#	are  built and installed in the correct order.
#	Run as m3build bootstrap [-mach3release <rel-dir>] 
#	where the rel-dir specifies the name of the release directory,
#	One or more of the targets bootstrap, mk, ux, and user may be
#	chosen.

# clean object/release directories first?
Clean=
# make extra configurations?
Extra=
# what do we build?
Bootstrap=
MK=
UX=
USER=

# system release area
Mach3Release=/afs/cs.cmu.edu/project/mach3/latest/release/@sys

make="make"
mflags="-c -k"
help=

if [ "$#" -eq 0 ]; then help=yes; fi;
while [ "$#" -gt 0 ]; do
	case "$1" in
	-mach3release ) shift; Mach3Release="$1"; shift;;
	-clean ) Clean=yes; shift;;
	-extra ) Extra=yes; shift;;
	bootstrap ) Bootstrap=yes; shift;;
	mk ) MK=yes; shift;;
	ux ) UX=yes; shift;;
	user ) USER=yes; shift;;
	* ) help=yes; break;;
	esac
done


if [ "$help" ]; then
 	echo "usage: m3build  bootstrap | mk | ux | user" 1>&2; 
		echo "     	      [-mach3release <release-dir>] [-clean] [-extra] " 1>&2; 
		echo "  bootstrap - build and release  build tools and libraries" 1>&2;
		echo "  mk - build the kernel" 1>&2;
		echo "  ux - build the Unix server" 1>&2;
		echo "  user - build the user tree" 1>&2;
		echo "  -mach3release <name of release directory> " 1>&2;
		echo "  -clean  - delete contents of release or object directory before building" 1>&2;
		echo "  -extra - build a complete set of kernel and/or Uxix server configurations" 1>&2;
		exit 1;
fi;
# reset paths:

# Ideally a bootstrapping build wouldn't need Mach3Release on the paths,
# but right now some things like gcc/cpp aren't built.

PATH=$Mach3Release/bin:/usr/cs/bin:/usr/ucb:/bin:/usr/bin; export PATH
CPATH=$Mach3Release/include:/usr/cs/include:/usr/include; export CPATH
LPATH=$Mach3Release/lib:/usr/cs/lib:/lib:/usr/lib; export LPATH
GCC_EXEC_PREFIX=$Mach3Release/lib/gcc-; export GCC_EXEC_PREFIX;

# what kind of machine is this?

case `/etc/machine` in
	SUN)	machine="sun3";;
	VAX)	machine="vax";;
	I386)	machine="i386";;
	PMAX)	machine="mips";;
	LUNA88K)machine="luna88k";;
	*)	echo "Unknown machine type"; exit 1;;
esac

# show what we do:

set -x

if [ "$Clean" ]; then
  date

  if [ "$Bootstrap" ]; then
    rm -rf ../release/@sys/*
  fi
  if [ "$MK" ]; then
    rm -rf ../obj/@sys/mk/*
  fi
  if [ "$UX" ]; then
    rm -rf ../obj/@sys/ux/*
  fi
  if [ "$USER" ]; then
    rm -rf ../obj/@sys/user/*
  fi
fi

if [ "$Bootstrap" ]; then

  # make top-level directories for installing into
  (cd user && date && $make EXIT_RULE=@ mkidirs)

  # on the luna, we need the ascpp script

  if [ $machine = luna88k ]; then
    (cd user/bin/ascpp && date && $make $mflags EXIT_RULE=@ install)
  fi

  # XXX These initial binaries are made with the 'wrong' crt0.o/includes.
  # (Whatever is lying around instead of ours.)
  # This isn't very important because they don't use Mach features.

  # first we make md; it doesn't need libcmucs

  (cd user/bin/md && date && $make $mflags DEP_TEMPLATE= EXIT_RULE=@ install)

  # now we install most header files
  # without wh, we have to handle subdirectories ourself

  (cd mk/include/$machine && date && $make $mflags preinstall)
  (cd mk/include && date && $make $mflags RECURSE_TEMPLATE=@ preinstall)
  (cd ux/include/$machine && date && $make $mflags preinstall)
  (cd ux/include && date && $make $mflags RECURSE_TEMPLATE=@ preinstall)

  # now we make libcmucs, wh, and xstrip

  (cd user/lib/libcmucs && date && $make $mflags install)
  (cd user/bin/wh && date && $make $mflags DEP_TEMPLATE= install)
  (cd user/bin/xstrip && date && $make $mflags install)

  # clean up some object areas, so binaries get remade
  # with proper crt0.o/includes

  if [ "$Clean" ]; then
    date

    rm -rf ../obj/@sys/user/bin/md
    rm -rf ../obj/@sys/user/bin/wh
    rm -rf ../obj/@sys/user/bin/xstrip
  fi
fi

# now we can make mk, ux, user

if [ "$MK" ]; then
  (cd mk && date && $make $mflags install)
fi

if [ "$UX" ]; then
  if [ "$machine" != luna88k ]; then
    (cd ux && date && $make $mflags install)
  fi
fi

if [ "$USER" ]; then
  (cd user && date && $make $mflags install)
fi

# extra configurations

if [ "$Extra" ]; then
  if [ "$MK" ]; then
    if [ $machine = i386 ]; then
      (cd mk/kernel && date && $make $mflags CONFIG=STD+SQT)
      (cd mk/kernel && date && $make $mflags CONFIG=STD+WS+TEST)
      (cd mk/kernel && date && $make $mflags CONFIG=STD+WS+NORMA+norma_ether)
    fi
    if [ $machine = vax ]; then
      (cd mk/kernel && date && $make $mflags CONFIG=STD+16+fixpri+VAX6bb+BI+de)
    fi
    if [ $machine = mips ]; then
      (cd mk/kernel && date && $make $mflags CONFIG=STD+ANY+TEST)
    fi
  fi

  if [ "$UX" ]; then
    if [ "$machine" != luna88k ]; then
      (cd ux/server && date && $make $mflags CONFIG=STDVICE+WS)
      (cd ux/server && date && $make $mflags CONFIG=STD+WS)
    fi
  fi
fi

date
exit 0
