/*
 * Copyright (c) 1991 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 *
 * Purpose:
 *	Contains the code that is resident in the DOS addressable
 *	memory that is used by the Emulator task.
 *
 * HISTORY: 
 * $Log:	bios_dosres.c,v $
 * Revision 2.2  91/12/06  15:28:06  grm
 * 	Initial version.
 * 	[91/12/06            grm]
 * 
 *
 */
#include "base.h"
#include "bios.h"

#define Adjptr(ptr)	{int x; x = (int)(ptr);x /=16;x++;x *=16; (ptr) = (u_char *)x;}

/*
 * Fake_mouse_driver (FMD) includes the header to the fake mouse interface
 * Int 0x33.  The interrupt vector points into this code.  Currently
 * it has a flaw.  The interrupt dies horribly if called instead of the
 * INT entrance. XXX 11/4/91 grm.
 */
u_long FAKE_MOUSE_DRIVER;
u_char fake_mouse_driver[] =  {
	0x2a, 0x2a, 0x2a, 0x20, 0x54, 0x68, 0x69, 0x73,	/* 0x00 */
	0x20, 0x69, 0x73, 0x20, 0x43, 0x6f, 0x70, 0x79, 
	0x72, 0x69, 0x67, 0x68, 0x74, 0x20, 0x31, 0x39,	/* 0x10 */
	0x38, 0x33, 0x20, 0x4d, 0x69, 0x63, 0x72, 0x6f,
	0x73, 0x6f, 0x66, 0x74, 0x20, 0x2a, 0x2a, 0x2a,	/* 0x20 */
	0xeb, 0x2f, 0xe9, 0xdf, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x30 */
	0x4c, 0x4f, 0x47, 0x49, 0x54, 0x45, 0x43, 0x48,
	0x20, 0x4d, 0x4f, 0x55, 0x53, 0x45, 0x20, 0x44,	/* 0x40 */
	0x52, 0x49, 0x56, 0x45, 0x52, 0x20, 0x56, 0x34,
	0x2e, 0x31, 0x30, 0x0a, 0x0d, 0x20, 0x43, 0x53,	/* 0x50 */
#ifdef NOVEMBER_91
	0x01, 0x46 };
#else
        0x01, 0xcd, 0xfb, 0xcf };
#endif NOVEMBER_91

/*
 *  VDISK Header used with the XMS support.
 */
u_long VDISK_HEADER;
u_char vdisk_header[] = {
	/* The VDISK header */
	0x00, 0x00, 0xe7, 0x19, 0x00, 0x08, 0xa9, 0x00,	/* 00 */
	0xd4, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x56, 0x44, 0x49, 0x53, 0x4b, 0x20, /* 10 */
	0x20, 0x56, 0x33, 0x2e, 0x33, 0x28, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 20 */
	0x00, 0x60, 0x86, 0x09, 0x00, 0x00, 0x11, 0xe0,
	/* Possibly empty space or used by VDISK header */
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,		/* 30 */
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,	
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,		/* 40 */
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,		/* 50 */
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,		/* 60 */
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,		/* 70 */
	0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0 };

/*
 *  Used by bios_mouse.c mon_space+0x80
 */
u_long MOUSE_HELPER_2;
u_char mouse_helper_2[] = {
	0x58, 0x5b, 0x59, 0x5a, 0x5f, 0x5e, 0x5d, 0x1f,
	0x07, 0xcf };

/*
 *  Mouse helper routine #1 (MACH_MOUSE_ADDRESS)
 */
u_long MOUSE_HELPER_1;
u_char mouse_helper_1[] = {
	0xcd, 0xfd, 0xcf };

/*
 *  Keyboard redirect ISR (Int 0x16)
 */
u_long INT16_ISR;
u_char int16_isr[] = {
	0xcd, 0xff, 0xcf };

/*
 *  XMS Driver (given as the address of XMS driver)
 */
u_long XMS_DRIVER;
u_char xms_driver[] = {
	0xeb, 0x04, 0x90, 0x90, 0x90, 0x90, 0xcd, 0xfc,
	0xcb };

/*
 *  Int 15h Func C0h Sys Config Info
 */
u_long SYS_CONFIG;
u_char sys_config[] = {
	0x08, 0x00, 0xfc, 0x81, 0x00, 0x20, 0x00, 0x00,
	0x00 };

/*
 *  Keyboard ISR
 */
u_long KEYISR;
u_char keyisr[] = {
	0xff, 0x00, 0xfa, 0x50, 0x53, 0x51, 0x52, 0x1e,	/* 0x00 */
	0x06, 0x57, 0x56, 0x55, 0x0e, 0x1f, 0xa1, 0x00,
	0x00, 0x0a, 0xe4, 0x74, 0x29, 0xbf, 0x40, 0x00,	/* 0x10 */
	0x8e, 0xc7, 0x06, 0x1f, 0x33, 0xff, 0x26, 0x8b,
	0x5d, 0x1c, 0x53, 0x43, 0x43, 0x83, 0xfb, 0x3e,	/* 0x20 */
	0x7c, 0x03, 0xbb, 0x1e, 0x00, 0x26, 0x3b, 0x5d,
	0x1a, 0x74, 0x38, 0x26, 0x89, 0x5d, 0x1c, 0x5f,	/* 0x30 */
	0x26, 0x89, 0x05, 0xeb, 0x2f, 0x90, 0x3c, 0xff,
	0x74, 0x2a, 0x50, 0x8a, 0xc8, 0xb8, 0x01, 0x00,	/* 0x40 */
	0xd3, 0xe0, 0x8b, 0xc8, 0xbf, 0x40, 0x00, 0x8e,
	0xc7, 0xbf, 0x17, 0x00, 0x58, 0x0a, 0xc0, 0x78,	/* 0x50 */
	0x0a, 0x8b, 0xc1, 0xf7, 0xd0, 0x26, 0x21, 0x05,
	0xeb, 0x0a, 0x90, 0x8b, 0xc1, 0x26, 0x09, 0x05,	/* 0x60 */
	0xeb, 0x02, 0x90, 0x5f, 0x5d, 0x5e, 0x5f, 0x07,
	0x1f, 0x5a, 0x59, 0x5b, 0x58, 0xfb, 0xcf };	/* 0x70 */

/*
 *  IRET for general use.
 */
u_long IRET_LOCATION;
u_char iret_location[] = {
	0xcf };

/*
 *  EMM Location for checking if EMM driver is present.
 *  This will be pointed at by Int 0x67 Vector.  The EMMXXXX0 must
 *  be at offset:0x0a, where the offset is the offset in vector 67.
 *  This means that the pointer to this location must be adjusted.
 */
u_long EMM_LOCATION;
u_char emm_location[] = {
	0xcf, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00,  'E',  'M',  'M',  'X',  'X',  'X',
	 'X',  '0' };

void
init_dosres()
{
	u_char * ptr;
	int adjust;

	/*
	 *  Copy information into the DOS Shared Memory Area
	 */

	ptr = (u_char *)mon_space;

	FAKE_MOUSE_DRIVER = (u_long) ptr;
	bcopy (fake_mouse_driver, ptr, sizeof(fake_mouse_driver));
	ptr += sizeof(fake_mouse_driver);

	VDISK_HEADER = (u_long) ptr;
	bcopy (vdisk_header, ptr, sizeof(vdisk_header));
	ptr += sizeof(vdisk_header);

	MOUSE_HELPER_2 = (u_long) ptr;
	bcopy (mouse_helper_2, ptr, sizeof(mouse_helper_2));
	ptr += sizeof(mouse_helper_2);

	MOUSE_HELPER_1 = (u_long) ptr;
	bcopy (mouse_helper_1, ptr, sizeof(mouse_helper_1));
	ptr += sizeof(mouse_helper_1);

	INT16_ISR = (u_long) ptr;
	bcopy (int16_isr, ptr, sizeof(int16_isr));
	ptr += sizeof(int16_isr);

	XMS_DRIVER = (u_long) ptr;
	bcopy (xms_driver, ptr, sizeof(xms_driver));
	ptr += sizeof(xms_driver);

	SYS_CONFIG = (u_long) ptr;
	bcopy (sys_config, ptr, sizeof(sys_config));
	ptr += sizeof(sys_config);

#ifndef MONKEY
	/*
	 *  Keyisr needs to be at a seg:0000
	 */
	adjust = (int)ptr;
	adjust /= 16;
	adjust += 1;
	adjust *= 16;
	ptr = (u_char *)adjust;

	KEYISR = (u_long) ptr;
	bcopy (keyisr, ptr, sizeof(keyisr));
	ptr += sizeof(keyisr);
#else
	KEYISR = IRET_LOCATION;
#endif MONKEY

	IRET_LOCATION = (u_long) ptr;
	bcopy (iret_location, ptr, sizeof(iret_location));
	ptr += sizeof(iret_location);

	Adjptr(ptr);

	EMM_LOCATION = (u_long) ptr;
	bcopy (emm_location, ptr, sizeof(emm_location));
	ptr += sizeof(emm_location);
}

